# -*- coding: utf-8 -*-
"""
Created on Mon Aug 22 15:24:49 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .multiplyForm import multiplyForm
from ..ind.plane import plane2D

"""
 * INTERVAL is a behavioral category for Sorts (see 'Sort.java').
 * An interval sort is a multiply-associated sort for which
 * ...
 * Therefore, the IntervalForm class extends on the MultiplyForm
 * class (see 'MultiplyForm.java') and implements the algebra methods
 * from the Form interface (see 'Form.java').
 *
 * ASSUMPTIONS: the following algebraic property is assumed on forms
 *   a + b = b <=> a < b
 * 
"""
class cyclical(multiplyForm):
    type = 'form-intervalC'
    
    def __init__(self, sort, *args, **kwargs):
        super(cyclical, self).__init__(sort)
    
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' _ ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return
        
    
    def insert(self, ind, start = 0):
        if not self.sort == ind.sort:
            raise ValueError('not of the same sort')
            return
        index = start
        thisInd = self.get(index)
        while thisInd is not None and thisInd.compareInterval(ind) < 1:
            if thisInd.compareInterval(ind) == 0:
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            else:
                index = index+1
                thisInd = self.get(index)
        self.add(index, ind)
        self.maximal = False
        return
        
        
    def oneStepForward(self, index):
        current = self.get(index)
        nextInd = self.get(index+1)
        if (nextInd is not None and current.touches(nextInd) and 
            nextInd.attribute == current.attribute and nextInd.end != current.start):
                nextInd.combine(current)
                self.remove(current)
        else:
            index += 1
        return index


    def combineIndividuals(self, ind1, ind2):
        if ind1 is ind2:
            return
        cmpI = ind2.compareInterval(ind1)
        if cmpI == 0:
            if ind2.hasAttribute() or ind1.hasAttribute():
                ind2.attribute.sum(ind1.attribute)                
            self.remove(ind1)
        elif ind2.touches(ind1):
            if (not ind2.hasAttribute() and not ind1.hasAttribute()) or ind2.attribute == ind1.attribute:
                self.remove(ind1)
                ind2.combine(ind1)
        elif ind2.disjoint(ind1):
            return
        elif ind2.contains(ind1):
            if (not ind2.hasAttribute() and not ind1.hasAttribute()) or ind2.attribute == ind1.attribute:
                self.remove(ind1)
            else:
                temp = ind2.complement(ind1)
                if ind2.hasAttribute() or ind1.hasAttribute():
                    ind1.attribute.sum(ind2.attribute)
                    
                self.remove(ind2)
                
                for i in temp:
                    self.insert(i)
                
        else:
            if (not ind2.hasAttribute() and not ind1.hasAttribute()) or ind2.attribute == ind1.attribute:
                self.remove(ind1)
                ind2.combine(ind1)
            else:
                temp1 = ind2.complement(ind1)[0]
                temp2 = ind1.complement(ind2)[0]
                
                tempCommon = ind2.common(ind1)
                        
                self.remove(ind1)
                self.remove(ind2)
                
                self.insert(temp1)
                self.insert(temp2)
                for com in tempCommon:
                    self.insert(com)
                    
            
    def splitArcs(self):
        if len(self.elements) == 0:
            return self
        if 'Arc' not in self.elements[0].type:
            return self
        i = 0
        while i < len(self):
            rev = self.get(i)
            l = rev.splitArc()
            if l is not None:
                self.remove(rev)
                self.insert(l[0])
                self.insert(l[1])
                i += 1
            i+= 1
                        
        return self
                    
        
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    def maximalize(self, ignoreMaximal = False):
        if self.maximal and ignoreMaximal == False:
            return
    

        self.splitArcs()
        
        #self.printE('ooo')

        if len(self.elements) == 0:
            return
        ind = self.first()
        
        index = 1
        nextInd = self.get(index)
        
        disorganized = []
        while nextInd is not None:
            #print('\n"""',ind.printVal(),'\n|||', nextInd.printVal(),'\n')
            if nextInd.isNil():
                self.remove(nextInd)
                nextInd = self.get(index)
                continue
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.get(index)
                continue
            cmp = ind.compareValue(nextInd)
            if cmp == 0:
                cmpI = ind.compareInterval(nextInd)
                if cmpI == 0:
                    if ind.hasAttribute() or nextInd.hasAttribute():
                        ind.attribute.sum(nextInd.attribute)                
                    self.remove(nextInd)
                    nextInd = self.get(index)
                elif ind.touches(nextInd):
                    if nextInd.end == ind.start:
                        ind = nextInd
                        index += 1
                        nextInd = self.get(index)
                        continue
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        ind.combine(nextInd)
                        nextInd = self.get(index)
                    else:
                        if cmpI == 1:
                            self.remove(nextInd)
                            disorganized.append(nextInd)
                            nextInd = self.get(index)
                        else:
                            ind = nextInd
                            index += 1
                            nextInd = self.get(index)
                elif ind.disjoint(nextInd):
                    if cmpI == 1:
                        self.remove(nextInd)
                        disorganized.append(nextInd)
                        nextInd = self.get(index)
                    else:
                        ind = nextInd
                        index += 1
                        nextInd = self.get(index)
                    
                elif ind.contains(nextInd):
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        nextInd = self.get(index)
                    else:
                        temp = ind.complement(nextInd)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            nextInd.attribute.sum(ind.attribute)
                            
                        self.remove(ind)
                        
                        ind = nextInd
                        
                        for i in temp:
                            self.insert(i)
                            
                        index += 1
                        nextInd = self.get(index)
                        
                elif nextInd.contains(ind):
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(ind)
                        ind = nextInd
                        nextInd = self.get(index)
                    else:
                        temp = nextInd.complement(ind)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            ind.attribute.sum(nextInd.attribute)
                        self.remove(nextInd)
                        if temp[0].compareInterval(ind) == -1:
                            ind = temp[0]
                        for i in temp:
                            self.insert(i)
                        nextInd = self.get(index)
                        
                else:
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        ind.combine(nextInd)
                        nextInd = self.get(index)
                    else:
                        temp1 = ind.complement(nextInd)[0]
                        temp2 = nextInd.complement(ind)[0]
                        
                        tempCommon = ind.common(nextInd)
                        self.remove(nextInd)
                        self.remove(ind)
                        
                        ind = temp1
                        self.insert(temp1)
                        self.insert(temp2)
                        for com in tempCommon:
                            com.attribute.sum(nextInd.attribute)
                            self.insert(com)
                        ind = self.get(index-1)
                        nextInd = self.get(index)

            elif cmp == 1:                
                self.remove(nextInd)
                disorganized.append(nextInd)
                nextInd = self.get(index)
            else:
                
                ind = nextInd
                index += 1
                nextInd = self.get(index)
                
        for ins in disorganized:
            self.insert(ins)
            
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()   
        
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
            
        self.maximal = True
        return
    
        
    
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        other.maximalize()
        
        for i in other.elements:
            self.insert(i)
                
        self.maximalize()
        return

    #####
    #    going through the two form and split any arc segment that goes over  
    #    the anchor point (last segment of a codescriptor) and do normal 
    #    difference then go through the list once again to join the segments together
    #####
    def difference(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        
        self.maximalize()
        other.maximalize()
            
        thisInd = self.first()       
        otherInd = other.first()
        
        thisIndex = 0
        otherIndex = 0
        
        while thisInd is not None and otherInd is not None:
            #print('\n"""',thisInd.printVal(),'\n|||', otherInd.printVal(),'\n')
            cmp = thisInd.compareValue(otherInd) 
            if cmp == 0:
                cmpI = thisInd.compareInterval(otherInd)
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if  cmpI == 1:                
                        otherIndex += 1
                        otherInd = other.get(otherIndex)
                    else:     
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                        
                elif cmpI == 0:
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        self.remove(thisInd)
                        thisInd = self.get(thisIndex)
                    otherIndex += 1
                    otherInd = other.get(otherIndex)
                    
                elif thisInd.contains(otherInd):
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                        
                        if thisInd.hasAttribute():
                            self.remove(thisInd)
                            thisInd = thisInd.common(otherInd)[0]
                            self.insert(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisIndex += 1
                            if thisInd == self.get(thisIndex):
                                thisIndex += 1
                                thisInd = self.get(thisIndex)
                        else:
                            self.remove(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisInd = self.get(thisIndex)
                    elif not thisInd.hasAttribute():
                        self.remove(thisInd)
                        for i in temp:
                            self.insert(i)
                        thisInd = self.get(thisIndex)
   
                elif otherInd.contains(thisInd):
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        self.remove(thisInd)
                        thisInd = self.get(thisIndex)
                    else:
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                elif cmpI == -1:   
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        tempCommon = thisInd.common(otherInd)
                        self.remove(thisInd)
                        for com in tempCommon:
                            self.insert(com)
                        self.insert(temp[0])
                        thisIndex += 1 + len(tempCommon)
                        thisInd = self.get(thisIndex)
                    else:
                        self.remove(thisInd)
                        self.insert(temp[0])
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                else:
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        tempCommon = thisInd.common(otherInd)
                        self.remove(thisInd)
                        for com in tempCommon:
                            self.insert(com)
                        thisIndex += len(tempCommon)
                    else:
                        self.remove(thisInd)
                    self.insert(temp[0])
                    thisInd = self.get(thisIndex)
            elif cmp == 1: 
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:            
                thisIndex += 1
                thisInd = self.get(thisIndex)
        
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
        return
    """



    
    #####
    #    going through the two forms and add segments with the same 
    #    codescriptors into two temp forms for taking difference, then add
    #    the result into a new result form
    #####
    def difference(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        
        self.maximalize()
        other.maximalize()
            
            
        index = 0 
        tempThis = self.sort.newForm()
        tempOther = self.sort.newForm()
        result = self.sort.newForm()

        while len(self) > 0:
            tempInd = self.elements.pop(0)
            tempThis.elements.append(tempInd)
            
            if len(self) != 0:
                comp = tempInd.compareValue(self.get(0))
                if comp == 0:
                    continue
                elif comp == 1:
                    raise ValueError('Form is not maximalized before difference operation')
            
            while index < len(other) and tempInd.compareValue(other.get(index)) == 1:
                index += 1
            while index < len(other) and tempInd.compareValue(other.get(index)) == 0:
                tempOther.elements.append(other.get(index))
                index += 1
            
            if len(tempOther) == 0:
                tempThis.move(result)
                continue
            
            thisInd = tempThis.first()       
            otherInd = tempOther.first()
            
            thisIndex = 0
            otherIndex = 0
            check = False

            while check is False:
                if thisInd is None or otherInd is None:
                    check = True
                    if len(tempThis) == 0:
                        break
                    if 'Arc' not in tempThis.elements[0].type:
                        break
                    crossThis = tempThis.get(-1).crossingComparator()
                    crossOther = tempOther.get(-1).crossingComparator()
                    if crossThis and crossOther:
                        c = tempThis.get(-1).compareEnd(tempOther.get(-1))
                        if c == 1:
                            thisInd = tempThis.get(-1)
                            otherInd = tempOther.get(0)
                        elif c == -1:
                            thisInd = tempThis.get(0)
                            otherInd = tempOther.get(-1)
                        else:
                            break
                    elif crossThis:
                        thisInd = tempThis.get(-1)
                        otherInd = tempOther.get(0)
                    elif crossOther:
                        thisInd = tempThis.get(0)
                        otherInd = tempOther.get(-1)
                    else:
                        break
                    
                cmpI = thisInd.compareInterval(otherInd)
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if  cmpI == 1:                
                        otherIndex += 1
                        otherInd = tempOther.get(otherIndex)
                    else:     
                        thisIndex += 1
                        thisInd = tempThis.get(thisIndex)
                        
                elif cmpI == 0:
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        tempThis.remove(thisInd)
                        thisInd = tempThis.get(thisIndex)
                    otherIndex += 1
                    otherInd = tempOther.get(otherIndex)
                    
                elif thisInd.contains(otherInd):
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                        if thisInd.hasAttribute():
                            tempThis.remove(thisInd)
                            thisInd = thisInd.common(otherInd)[0]
                            tempThis.insert(thisInd)
                            for i in temp:
                                tempThis.insert(i)
                            thisIndex += 1
                            if thisInd == tempThis.get(thisIndex):
                                thisIndex += 1
                                thisInd = tempThis.get(thisIndex)
                        else:
                            tempThis.remove(thisInd)
                            for i in temp:
                                tempThis.insert(i)
                            thisInd = tempThis.get(thisIndex)
                    elif not thisInd.hasAttribute():
                        tempThis.remove(thisInd)
                        for i in temp:
                            tempThis.insert(i)
                        thisInd = tempThis.get(thisIndex)
   
                elif otherInd.contains(thisInd):
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        tempThis.remove(thisInd)
                        thisInd = tempThis.get(thisIndex)
                    else:
                        thisIndex += 1
                        thisInd = tempThis.get(thisIndex)
                elif cmpI == -1:   
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                        
                    if thisInd.hasAttribute():
                        tempCommon = thisInd.common(otherInd)
                        tempThis.remove(thisInd)
                        for com in tempCommon:
                            tempThis.insert(com)
                        tempThis.insert(temp[0])
                        thisIndex += 1 + len(tempCommon)
                        thisInd = tempThis.get(thisIndex)
                    else:
                        tempThis.remove(thisInd)
                        tempThis.insert(temp[0])
                        thisIndex += 1
                        thisInd = tempThis.get(thisIndex)
                else:
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        tempCommon = thisInd.common(otherInd)
                        tempThis.remove(thisInd)
                        for com in tempCommon:
                            tempThis.insert(com)
                        thisIndex += len(tempCommon)
                    else:
                        tempThis.remove(thisInd)
                    tempThis.insert(temp[0])
                    thisInd = tempThis.get(thisIndex)
            tempThis.move(result, True)
            tempThis.clear()
            tempOther.clear()
        result.move(self, True)
        
        index = 0
        while index < len(self.elements):
            index = tempThis.oneStepForward(index)
        return
    """
        
    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    """
    def product(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        
        index = 0 
        tempThis = self.sort.newForm()
        tempOther = self.sort.newForm()
        result = self.sort.newForm()
        
        
        while len(self) > 0:
            tempInd = self.elements.pop(0)
            tempThis.elements.append(tempInd)
            
            if len(self) != 0:
                comp = tempInd.compareValue(self.get(0))
                if comp == 0:
                    continue
                elif comp == 1:
                    raise ValueError('Form is not maximalized before difference operation')
            
            while index < len(other) and tempInd.compareValue(other.get(index)) == 1:
                index += 1
            while index < len(other) and tempInd.compareValue(other.get(index)) == 0:
                tempOther.elements.append(other.get(index))
                index += 1
            
            if len(tempOther) == 0:
                tempThis.clear()
                continue
            
            thisInd = tempThis.first()       
            otherInd = tempOther.first()
            
            thisIndex = 0
            otherIndex = 0
            check = False
            
            while check is False:
                if thisInd is None or otherInd is None:
                    check = True
                    if len(tempThis) == 0:
                        break
                    if 'Arc' not in tempThis.elements[0].type:
                        break
                    crossThis = tempThis.get(-1).crossingComparator()
                    crossOther = tempOther.get(-1).crossingComparator()
                    if crossThis and crossOther:
                        c = tempThis.get(-1).compareEnd(tempOther.get(-1))
                        if c == 1:
                            thisInd = tempThis.get(-1)
                            otherInd = tempOther.get(0)
                        elif c == -1:
                            thisInd = tempThis.get(0)
                            otherInd = tempOther.get(-1)
                        else:
                            break
                    elif crossThis:
                        thisInd = tempThis.get(-1)
                        otherInd = tempOther.get(0)
                    elif crossOther:
                        thisInd = tempThis.get(0)
                        otherInd = tempOther.get(-1)
                    else:
                        break

                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherIndex += 1
                        otherInd = tempOther.get(otherIndex)
                    else:
                        thisIndex += 1
                        thisInd = tempThis.get(thisIndex)
                else:
                    if thisInd.hasAttribute() and otherInd.hasAttribute():
                        temp = thisInd.duplicateInd()
                        temp.attribute.product(otherInd.attribute)
                        if temp.hasAttribute():
                            tempCommon = temp.common(otherInd)
                            for com in tempCommon:
                                result.insert(com)
                    else:
                        tempCommon = thisInd.common(otherInd)
                        for com in tempCommon:
                            result.insert(com)
                            
                    #######
                    if thisInd.compareEnd(otherInd) < 1:
                        thisIndex += 1
                        thisInd = tempThis.get(thisIndex)
                    else:
                        otherIndex += 1
                        otherInd = tempOther.get(otherIndex)
                      
            tempThis.clear()
            tempOther.clear()
            
        result.move(self, True)
            
        return

    """
    def product(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        
        thisInd = self.first()
        nextInd = self.next(thisInd)
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherInd = other.next(otherInd)
                    else:
                        self.remove(thisInd)
                        thisInd = nextInd
                        nextInd = self.next(thisInd)
                else:
                    checkSame = False
                    if thisInd.hasAttribute() and otherInd.hasAttribute():
                        temp = thisInd.duplicateInd()
                        temp.attribute.product(otherInd.attribute)
                        if temp.hasAttribute():
                            tempCommon = temp.common(otherInd)
                            for com in tempCommon:
                                self.insert(com)
                    else:
                        tempCommon = thisInd.common(otherInd)
                        for com in tempCommon:
                            if com == thisInd:
                                checkSame = True
                            else:
                                self.insert(com)
                            
                    #######
                    if thisInd.compareEnd(otherInd) < 1:
                        if not checkSame:
                            self.remove(thisInd)
                        thisInd = nextInd
                        nextInd = self.next(thisInd)
                    else:
                        otherInd = other.next(otherInd)
  
            elif thisInd.compareValue(otherInd) == 1: 
                otherInd = other.next(otherInd)
            else:         
                toBeRemovedInd = thisInd
                thisInd = self.next(thisInd)
                self.remove(toBeRemovedInd)
                
        while thisInd is not None:   
            self.remove(thisInd)
            thisInd = nextInd
            nextInd = self.next(thisInd)
            
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
            
        return



    """
    symdifference of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.symdifference(formB)
      -->formA: a,       d
    """
    def symdifference(self,other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherInd = other.next(otherInd)
                    else:         
                        thisInd = self.next(thisInd)    
                elif thisInd.compareInterval(otherInd) == 0:
                    thisInd = self.next(thisInd)    
                    otherInd = other.next(otherInd)
                else:         
                    temp1 = thisInd.complement(otherInd)
                    temp2 = otherInd.complement(thisInd)
                    temp3 = otherInd
                    temp4 = thisInd
                    if thisInd.compareEnd(otherInd) >= 0:
                        otherInd = other.next(otherInd)
                    elif thisInd.compareEnd(otherInd) <= 0:
                        thisInd = self.next(thisInd)    
                    temp3.common(temp4)
                    temp4.common(temp3)
                    for i in temp1:
                        self.insert(i)
                    for i in temp2:
                        other.insert(i)
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)    
                        
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        self.insert(otherInd)                            
                        otherInd = other.next(otherInd)
                    else:                             
                        thisInd = self.next(thisInd)    
                elif thisInd.compareInterval(otherInd) == 0:
                    #if has attribute, calculate difference
                    if thisInd.hasAttribute() or otherInd.hasAttribute() :
                        thisInd.attribute.symdifference(otherInd.attribute)
                        if not thisInd.hasAttribute():
                            toBeRemoved = thisInd
                            thisInd = self.next(thisInd)    
                            self.remove(toBeRemoved)
                        else:
                            thisInd = self.next(thisInd)    
                        otherInd = other.next(otherInd)
                    else:
                        toBeRemoved = thisInd
                        thisInd = self.next(thisInd)    
                        self.remove(toBeRemoved)
                        otherInd = other.next(otherInd)
                
            elif thisInd.compareValue(otherInd) == 1:
                self.insert(otherInd)                            
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)    
        while otherInd is not None:
            self.insert(otherInd)                            
            otherInd = other.next(otherInd)
            
        ind = self.first()
        index = 1
        while ind is not None:
            ind = self.oneStepForward(ind, index)
            index += 1
        ind = other.first()
        index = 1
        while ind is not None:
            ind = other.oneStepForward(ind, index)
            index += 1

        return 

    """
    partition of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
         
        _common = formA.partition(formB)
      -->formA: a
         formB:          d
         common:   b, c
    """
    def partition(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
        

        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherInd = other.next(otherInd)
                    else:         
                        thisInd = self.next(thisInd)    
                elif thisInd.compareInterval(otherInd) == 0:
                    thisInd = self.next(thisInd)    
                    otherInd = other.next(otherInd)
                else:         
                    temp1 = thisInd.complement(otherInd)
                    temp2 = otherInd.complement(thisInd)
                    temp3 = otherInd
                    temp4 = thisInd
                    if thisInd.compareEnd(otherInd) >= 0:
                        otherInd = other.next(otherInd)
                    elif thisInd.compareEnd(otherInd) <= 0:
                        thisInd = self.next(thisInd)    
                    temp3.common(temp4)
                    temp4.common(temp3)
                    for i in temp1:
                        self.insert(i)
                    for i in temp2:
                        other.insert(i)
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)    
                
        
        common = interval(self.sort)
        
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:      
                        otherInd = other.next(otherInd)
                    else:                             
                        thisInd =self.next(thisInd)        
                elif thisInd.compareInterval(otherInd) == 0:
                    
                    attForm = otherInd.duplicateInd()
                    
                    if attForm.hasAttribute():
                        attForm.attribute.product(thisInd.attribute)
                        
                    if not (not attForm.hasAttribute() and (thisInd.hasAttribute() or otherInd.hasAttribute())):
                        common.insert(attForm)    
                    
                        if thisInd.hasAttribute():
                            thisInd.attribute.difference(attForm.attribute)
                        if otherInd.hasAttribute():
                            otherInd.attribute.difference(attForm.attribute)
                        
                        if not thisInd.hasAttribute():
                            toBeRemovedInd = thisInd
                            thisInd = self.next(thisInd)
                            self.remove(toBeRemovedInd)
                        else:
                            thisInd = self.next(thisInd)
                            
                        if not otherInd.hasAttribute():
                            toBeRemovedInd = otherInd
                            otherInd = other.next(otherInd)
                            other.remove(toBeRemovedInd)
                        else:
                            otherInd = other.next(otherInd)
                    else:
                        thisInd = self.next(thisInd)
                        otherInd = other.next(otherInd)
                        
            elif thisInd.compareValue(otherInd) == 1:      
                otherInd = other.next(otherInd)
            else:                             
                thisInd =self.next(thisInd)       
            
        ind = self.first()
        index = 1
        while ind is not None:
            ind = self.oneStepForward(ind, index)
            index += 1
        ind = other.first()
        index = 1
        while ind is not None:
            ind = other.oneStepForward(ind, index)
            index += 1
        return common
        
    """
    check if this form contains a certain individual
    """
    def containsInd(self, ind):
        com = None
        if self.sort == ind.sort:
            for i in self.elements:
                if i.containsInd(ind):
                    return True
                elif i.containsPoint(ind.start) and self.sort.type == 'sort-attribute':
                    if ind.hasAttribute():
                        if not ind.attribute.partOf(i.attribute):
                            continue
                    com = i.duplicateInd()
                elif com is not None:
                    if i.start == com.end and i.containsPoint(ind.end) and ind.attribute.partOf(i.attribute):
                        com.combine(i)
                        if com.contains(ind):
                            return True
                    elif i.start == com.end and ind.contains(i) and ind.attribute.partOf(i.attribute):
                        com.combine(i)
        return False
        
        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        self.maximalize()
        other.maximalize()
        thisInd = self.first()
        otherInd = other.first()
        for thisInd in self.elements:
            check = False
            com = None
            for otherInd in other.elements:
                if thisInd == otherInd or otherInd.contains(thisInd):
                    if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                        return False
                    check = True
                    break

                elif otherInd.containsPoint(thisInd.start):
                    if thisInd.hasAttribute():
                        if not thisInd.attribute.partOf(otherInd.attribute):
                            continue
                    com = otherInd.duplicateInd()
                elif com is not None:
                    if (otherInd.start == com.end and otherInd.containsPoint(thisInd.end) 
                    and thisInd.attribute.partOf(otherInd.attribute)):
                        com.combine(otherInd)
                        if com.contains(thisInd):
                            check = True
                            break
                    elif (otherInd.start == com.end and thisInd.contains(otherInd) 
                    and thisInd.attribute.partOf(otherInd.attribute)):
                        com.combine(otherInd)

            if check is False:
                return False
        return True