# -*- coding: utf-8 -*-
"""
Created on Tue Aug 30 14:30:41 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .multiplyForm import multiplyForm
"""
 * A meta-form denotes a form of a {DisjunctiveSort}.
 * It is composed of zero, one or more forms corresponding to the component
 * sorts of the disjunctive sort. The operational behavior of a meta-form is
 * defined by the individual behaviors of the element forms.
 * A meta-form is maximal if it contains at most one form
 * corresponding to each component sort and each element form is maximal.
 * 
 * The meta-form class extends on the {MultiplyForm} class and
 * specifies the operational behavior of meta-forms in terms of the behaviors
 * of its constituent element forms.
"""
class metaForm(multiplyForm):
    type = 'form-metaForm'
    
    def __init__(self, sort, *args, **kwargs):
        super(metaForm, self).__init__(sort)
        
    def __str__(self):
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            s += i.sort.name+': '+str(i)
        s += '}'
        return s

    def __eq__(self, other):
        if other is None or self.type != other.type or self.sort != other.sort or len(self.elements) != len(other.elements):
            return False
        
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
            
        lenSelf = 0
        for i in self:
            lenSelf += len(i)
        lenOther = 0
        for i in other:
            lenOther += len(i)
        if lenSelf != lenOther:
            return False
            
        for i in self.elements:
            check = False
            for j in other.elements:
                if i.type != j.type or i.sort != j.sort:
                    continue
                if i == j:
                    check = True
                    break
            if check == False:
                return False
        return True
        
    
    """
    print out the form for easier assessment
    structure:
      |disjunctive: r elements
        |form 1 sort name: n elements
         _ individual 1
         _ individual 2
             ...
         _ individual n
          |attribute1 sort name: m elements
           _ attribute1 1
               ...
          |attribute2 sort name: 1 elements
           _ attribute2
        |form 2 sort name: p elements
         _ individual 1
         _ individual 2
             ...
         _ individual p
        ...
        |form ??? sort name: ??? elements
            ...
    """
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ele in self.elements:
            ele.printE(string+'  ')
        return
        
    def printRhino(self, string = '', displacement = 0):
        for ele in self.elements:
            if ele.type in ('form-interval','form-discrete','form-coForm','form-areal'):
                ele.printRhino(string+'  ', displacement)
        return
                
    def accept(self, visitor, direction=-1):
        if visitor.visitEnter(self, direction) and not self.isEmpty():
            for element in self.elements:
                element.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    def splitElements(self):
        splits = []
        for i in self.elements:
            splits += i.splitElements()
        return splits
                
    """
    inserting an element into the form
    """
    def insert(self, element):
        if 'form' in element.type:
            if not self.sort.contains(element.sort):
                print('~~~~~~~')
                print("ERROR: Argument sorts do not adhere to containment relationship:")
                print(self.sort.name,"does not contain",element.sort.name)
                return
            self.maximal = False
            index = 0
            for i in self.elements:
                if i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                    if i.sort.name == element.sort.name:
                        i.merge(element)
                        return
                    elif element.sort.name < i.sort.name:
                        self.add(index, element)
                        return
                elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                    self.add(index, element)
                    return
                index = index + 1
            self.add(index, element)
        else:
            if not self.sort.contains(element.sort):
                print('!!!!!!!!!!')
                print(self.sort.name, element.sort.name)
                print("Argument sorts do not adhere to containment relationship")
                return
            self.maximal = False
            index = 0
            for i in self.elements:
                if i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                    if i.sort.name == element.sort.name:
                        i.insert(element)
                        return
                    elif element.sort.name < i.sort.name:
                        form = element.sort.newForm()
                        form.insert(element)
                        self.add(index, form)
                        return
                elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                    form = element.sort.newForm()
                    form.insert(element)
                    self.add(index, form)
                    return
                index = index + 1
            form = element.sort.newForm()
            form.insert(element)
            self.add(index, form)
        return
               
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    
    def maximalize(self):
        if self.maximal:
            return
    
        self.maximal = True
    
        for i in self.elements:
            i.maximalize()     
        i = self.first()
        disorganized = []
        index = 1
        j = self.get(index)
        while j is not None:
            if i.sort.baseSort().dimension() == j.sort.baseSort().dimension():
                if j.sort.name == i.sort.name:
                    i.sum(j)
                    self.remove(j)
                    j = self.get(index)
                elif i.sort.name > j.sort.name:
                    self.remove(j)
                    disorganized.append(j)
                    j = self.get(index)
                else:
                    index += 1
                    i = j
                    j = self.get(index)
            elif i.sort.baseSort().dimension() > j.sort.baseSort().dimension():
                self.remove(j)
                disorganized.append(j)
                j = self.get(index)
            else:
                index += 1
                i = j
                j = self.get(index)
        for i in disorganized:
            self.insert(i)
        return
        
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
        
        self.maximalize()
        other.maximalize()
        
        thisForm = self.first()
        otherForm = other.first()
        index = 0
        while thisForm is not None and otherForm is not None:   
            """
            if i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                if i.sort.name == element.sort.name:
                    i.merge(element)
                    return
                elif element.sort.name < i.sort.name:
                    self.add(index, element)
                    return
            elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                self.add(index, element)
                return
            """
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, sum
                    thisForm.sum(otherForm)
                    thisForm = self.next(thisForm)                            
                    index = index+1
                    otherForm = other.next(otherForm)
                elif thisForm.sort.name > otherForm.sort.name:
                    self.add(index, otherForm)                            
                    index = index+1
                    otherForm = other.next(otherForm)
                else:                          
                    index = index+1
                    thisForm =self.next(thisForm)       
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                self.add(index, otherForm)                            
                index = index+1
                otherForm = other.next(otherForm)
            else:
                index = index+1
                thisForm =self.next(thisForm)       
                 
        while otherForm is not None:
            self.add(index, otherForm)                            
            index = index+1
            otherForm = other.next(otherForm)
        return
        
    """
    difference of two forms
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.difference(formB)
      -->formA: a
    """
    def difference(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if self.sort != other.sort:
            print(self.sort, other.sort)
            raise ValueError('not of the same sort')
            return
        
        self.maximalize()
        other.maximalize()
            
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, calculate difference
                    thisForm.difference(otherForm)
                    if thisForm.isEmpty():
                        toBeRemoved = thisForm
                        thisForm = self.next(thisForm)
                        self.remove(toBeRemoved)
                    otherForm = other.next(otherForm)
                elif thisForm.sort.name > otherForm.sort.name:                            
                    otherForm = other.next(otherForm)
                else:                             
                    thisForm =self.next(thisForm)   
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherForm = other.next(otherForm)
            else:
                thisForm =self.next(thisForm)   
        return        
        
    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    def product(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, product
                    thisForm.product(otherForm)
                    if thisForm.isEmpty():
                        toBeRemoved = thisForm
                        thisForm = self.next(thisForm)
                        otherForm = other.next(otherForm)
                        self.remove(toBeRemoved)
                    else:
                        thisForm = self.next(thisForm)                            
                        otherForm = other.next(otherForm)
                elif thisForm.sort.name > otherForm.sort.name:                
                    otherForm = other.next(otherForm)
                else:                             
                    toBeRemoved = thisForm
                    thisForm = self.next(thisForm)
                    self.remove(toBeRemoved)
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherForm = other.next(otherForm)
            else:
                toBeRemoved = thisForm
                thisForm = self.next(thisForm)
                self.remove(toBeRemoved)
        while thisForm is not None:   
            toBeRemoved = thisForm
            thisForm = self.next(thisForm)
            self.remove(toBeRemoved)
        return

    """
    symdifference of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.symdifference(formB)
      -->formA: a,       d
    """
    def symdifference(self,other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisForm = self.first()
        otherForm = other.first()
        index = 0
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, calculate difference
                    thisForm.symdifference(otherForm)
                    if thisForm.isEmpty():
                        toBeRemoved = thisForm
                        thisForm = self.next(thisForm)
                        self.remove(toBeRemoved)
                    else:
                        thisForm = self.next(thisForm)
                    otherForm = other.next(otherForm)
                    
                elif thisForm.sort.name > otherForm.sort.name:
                    self.add(index, otherForm)                            
                    index = index+1
                    otherForm = other.next(otherForm)
                else:                             
                    index = index+1
                    thisForm =self.next(thisForm)       
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                self.add(index, otherForm)                            
                index = index+1
                otherForm = other.next(otherForm)
            else:
                index = index+1
                thisForm =self.next(thisForm)       
            
        while otherForm is not None:
            self.add(index, otherForm)                            
            index = index+1
            otherForm = other.next(otherForm)
        return 


    """
    partition of two forms (XOR)
    e.g. formA: a, b, c
         formB:    b, c, d
         
        _common = formA.partition(formB)
      -->formA: a
         formB:          d
         common:   b, c
    """
    def partition(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                       
        thisForm = self.first()
        otherForm = other.first()
        common = metaForm(self.sort)
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    common.insert(thisForm.partition(otherForm))
                    if thisForm.isEmpty():
                        toBeRemoved = thisForm
                        
                        thisForm = self.next(thisForm)
                        self.remove(toBeRemoved)
                    else:
                        thisForm = self.next(thisForm)
                    if otherForm.isEmpty():
                        toBeRemoved = otherForm
                        
                        otherForm = other.next(otherForm)
                        other.remove(toBeRemoved)
                    else:
                        otherForm = other.next(otherForm)
                        
                elif thisForm.sort.name > otherForm.sort.name:
                    otherForm = other.next(otherForm)
                else:                             
                    thisForm =self.next(thisForm)       
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherForm = other.next(otherForm)
            else:
                thisForm =self.next(thisForm)       
        return common
        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                
        for thisForm in self.elements:
            check = False
            for otherForm in other.elements:
                if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                    if thisForm.sort.name == otherForm.sort.name:
                        if thisForm.partOf(otherForm):
                            check = True
            if check == False:
                return False
        return True
        """
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.name == otherForm.sort.name:
                if not thisForm.partOf(otherForm):
                    return False
                thisForm = self.next(thisForm)
                otherForm = other.next(otherForm)
                
            if thisForm.sort.name > otherForm.sort.name:
                otherForm = other.next(otherForm)
            else:                             
                return False
        return True
        """
        
    def checkPartOf(self, other, flagDesc = True):
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            raise ValueError('not of the same sort')
            return
            
        self.maximalize()
        other.maximalize()
                
        for thisForm in self.elements:
            check = False
            for otherForm in other.elements:
                if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                    if thisForm.sort.name == otherForm.sort.name:
                        if thisForm.checkPartOf(otherForm, flagDesc):
                            check = True
            if check == False:
                return False
        return True
        
