from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import


import warnings
from ..visit.elementVisitor import elementVisitor
from .ruleAnalysis import ruleAnalysis
from .ruleApplication import ruleApplication
from ..map.similarityMap import similarityMap2D
from ..map.similarityMap3D import similarityMap3D
from ..map.parametricMap import parametricMap
from ..map.exactMap import exactMap
from ..transform.transformSet import transformSet
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..ind.description import description
from ..ind.descriptions.tree import DescriptionEvaluatingError

class detection(elementVisitor):
    applicationDict = {}
    
    def __init__(self, rule):
        self.rule = rule
        self.applicationDict[rule.name] = []
        ###############################################
        self.analysis= ruleAnalysis(self.rule.lhs)
        """
        if not self.isValid():
            warnings.warn("\nrule "+self.rule.name+" does not meet the minimum requirement for similarity transformation\n")
        """
        ###############################################
        
        self.form = None
        self.target = None
        self.uptrail = []
        self.downTrail = []
        self.success = False
        self.applications = []
        return
        
    def isValid(self):
        for i in self.analysis.maps:
            if not self.analysis.maps[i].isValid():
                return False
        return True
        
    def count(self):
        return len(self.applications)
        
        
    def getApplication(self, index = 0):
        if index >= len(self.applications):
            raise IndexError
        return self.applications[index]
        
    def matches(self):
        matches = []
        for i in self.applications:
            match = i.matchForm
            if len(matches) == 0 or match not in matches:
                matches.append(match)
        return matches
    
    def results(self, match):
        r = []
        for i in self.applications:
            if match is i.matchForm or match == i.matchForm:
                r.append(i.resultForm)
        return r
        
    #..................
    def detect(self, form):
        self.clear()
        ###########################
        self.form = form
        ###########################
        
        for i in self.applicationDict[self.rule.name]:
            if i[0] == form:
                self.applications = i[1]
                return i[1]
        

        self.analysis.unsetVisitedAll()
        fragments = self.analysis.fragments
        
        for i in fragments:
            self.target = i
            self.uptrail= []
            if not self.target.content.visited:
                visit = True
                self.uptrail.append(self.target)
                while(self.uptrail[-1].parent is not None):
                    self.uptrail.append(self.uptrail[-1].parent)
                    if self.uptrail[-1].content.visited:
                        visit = False
                if visit:
                    self.form.accept(self)
                    self.target.content.visited = True
                    if self.target.parent is not None:
                        self.target.parent.content.visited = True
        resultForms = []   
        
        transformlists = [None] * len(self.analysis.maps)
        lists = 0
        for map in self.analysis.maps.values():
            transformlists[lists] = map.potentialTransforms()
            if map.type == 'map-graphMap':
                transformlists[lists] = map.analyzeRHS(self.rule.rhs, self.form, transformlists[lists])
            if transformlists[lists] is not None:
                lists += 1

        paramsets = None
        if self.analysis.hasMap(parametricMap):
            paramsets = self.analysis.maps[parametricMap].potentialParamSets()
            #print('before:', len(paramsets))
            i = 0
            while i < len(paramsets):
                if not paramsets[i].resolveSimpleDelays():
                    paramsets.pop(i)
                else:
                    i += 1
            #print('after:', len(paramsets))

        indices = []
        totalcount = 1
        for i in range(lists):
            indices.append(0)
            totalcount *= len(transformlists[i])

        #print('\n*totalcount transform set:',totalcount)
        for i in range(totalcount):
            tfset = transformSet()
            for j in range(lists):
                tfset.add(transformlists[j][indices[j]])
            j = 0
            while indices[j] == len(transformlists[j])-1 and j < lists-1:
                indices[j] = 0
                j += 1
            if indices[j] == len(transformlists[j]):
                break
            indices[j] += 1
            
            """
            ####
            # For Testing: printing out the transformation matrix 
            #              and its respective transformed lhs
            ####
            print('*******************\n_Transformation matrix:')   
            for i in tfset.map:
                tfset.map[i].printTransf()
            print('_Transformed lhs:')
            tlhs = self.rule.lhs.transform(tfset) 
            #tlhs.printE('___')
            #print(tlhs)
            tlhs.printRhino()
            print('\ncheck:', self.rule.lhs.checkTransform(tfset, self.form) )
            print('\n______\n')
            """

            if paramsets is None:
                check = self.rule.lhs.checkTransform(tfset, self.form) 
                if check:
                    tlhs = self.rule.lhs.transform(tfset)
                    if not self.rule.evaluatePredicate(tfset, self.rule.lhs, tlhs, self.form):
                        continue
                    for tfS in self.rule.solveDirectives(tfset, self.rule.lhs, tlhs, self.rule.rhs, self.form):
                        trhs = self.rule.rhs.transform(tfS)
                        if self.rule.ruleType == 'ADDITIVE' and trhs.partOf(self.form):
                            continue
                        if [tlhs,trhs] in resultForms:
                            continue
                        resultForms.append([tlhs,trhs])
                        app = ruleApplication(self.rule, tfS, None, tlhs, trhs)
                        self.applications.append(app)                        
                        """
                        ####
                        # For Testing: printing out the transformation matrix 
                        #              and its respective transformed lhs
                        ####
                        print('*******************\n_Transformation matrix:')   
                        for i in tfS.map:
                            tfS.map[i].printTransf()
                        print('_Transformed lhs:')
                        #tlhs.printE('___')
                        #print(tlhs)
                        tlhs.printRhino()
                        print('\n______\n')
                        """
            else:
                description.clearDelays()
                check = self.rule.lhs.checkTransform(tfset, self.form) 
                if not check:
                    #print('geometry check failed')
                    continue
                tlhs = self.rule.lhs.transform(tfset) 
                if not self.rule.evaluatePredicate(tfset, self.rule.lhs, tlhs, self.form):
                    continue
                for tfS in self.rule.solveDirectives(tfset, self.rule.lhs, tlhs, self.rule.rhs, self.form):
                    trhs = self.rule.rhs.transform(tfset)
                    """
                    ####
                    # For Testing: printing out the transformation matrix 
                    #              and its respective transformed lhs
                    ####
                    print('*******************\n_Transformation matrix:')   
                    for i in tfset.map:
                        tfset.map[i].printTransf()
                    print('_Transformed lhs:')
                    tlhs.printE('~~~')
                    print('\n______\n')
                    """
                    #print('geometry check passed; num params:', len(paramsets))
                    for params in paramsets:
                        params.lhs = tlhs
                        params.rhs = trhs
                        if (params.resolveDelays()):
                            if not description.resolveDelays(params):
                                continue
                            try:
                                glhs = tlhs.edit(params)
                                grhs = trhs.edit(params)
                            except DescriptionEvaluatingError:
                                continue
                            
                            """
                            ####
                            # For Testing: printing out the transformation matrix 
                            #              and its respective transformed lhs
                            ####
                            print('*******************\n_Transformation matrix:')   
                            for i in tfset.map:
                                tfset.map[i].printTransf()
                            print('_Transformed lhs:')
                            glhs.printE('~~~')
                            print('\n______\n')
                            """
    
                            checkType = False
                            if self.rule.ruleType != 'ADDITIVE':
                                checkType = True if glhs.partOf(grhs) else False
    
                            if (glhs.partOf(self.form) 
                            and (checkType is False or not grhs.partOf(self.form))):
                                
                                app = ruleApplication(self.rule, tfS, params, glhs, grhs)
                                if app not in self.applications:
                                    """
                                    print('_Transformed lhs:')
                                    glhs.printE('~~~')
                                    print('\n______\n')
                                    """
                                    self.applications.append(app)

        self.applicationDict[self.rule.name].append([self.form.duplicate(), self.applications])
        return self.applications            

    def clear(self):
        self.form = None
        for map in self.analysis.maps.values():
            map.clear()
        self.target = None
        self.uptrail = []
        self.downTrail = []
        self.success = False
        self.applications = []


    def visitEnter(self, element, assignment):
        if 'ind' in element.type:
            if len(self.uptrail) == 0:
                try:
                    mapping = self.analysis.getMap(element.sort)
                except KeyError:
                    return False
                result = False
                fragment = self.target.content
                if len(self.downTrail) != 0:
                    fragment = self.downTrail[0].content
                for elm in fragment:
                    if assignment == 'refl' and mapping.type == 'map-parametric':
                        continue
                    try:
                        if mapping.map(elm, element):
                            result = True
                    except AttributeError:
                        pass
                if len(self.downTrail) == 0:
                    self.success = (result or self.success)
                return result
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                return True
            elif len(self.uptrail) != 0 and element.sort != self.uptrail[-1].content.sort:
                return False
            elif len(self.uptrail) != 0:
                self.uptrail.pop(0)
            else: 
                try:
                    fragment = self.downTrail[0]
                except IndexError: 
                    fragment = self.target
                for frag in fragment.children:
                    if element.sort == frag.content.sort:
                        self.downTrail.append(frag)
                        break
                if len(self.downTrail) >0 and self.downTrail[-1].content.visited:
                    return False

        return True
    
    def visitLeave(self, element):
        if 'ind' in element.type:
            if (self.success and len(self.uptrail) != 0):
                mapping = self.analysis.getMap(element.sort)
                for elm in self.uptrail[-1].parent.content:
                     mapping.map(elm, element)
                self.success = False
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                return
                
            elif len(self.downTrail) != 0:
                self.downTrail.pop(0)

            elif (len(self.uptrail) != 0 and self.uptrail[-1].parent is not None
                    and element.sort == self.uptrail[-1].parent.content.sort):
                self.uptrail.append(self.uptrail[-1].parent)
            elif element.sort == self.target.content.sort:
                self.uptrail.append(self.target)
        return