# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 12:36:49 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
"""
from future import standard_library
standard_library.install_aliases()
"""
import sys

from random import randint

from .ruling import ruling
from ..detection import detection


class rule(ruling):
    type = 'rule'
    
    register  = {}   
    
    """
    rule object constructor:
    input:  name <string>: name of rule, rule name must be unique
            description <string>: description of rule
            lhs <form>: left hand side of rule
            rhs <form>: right hand side of rule
    """
    def __init__(self, name, description, lhs, rhs, *args, **kwargs):
        super(rule, self).__init__()        
        self.name = name
        self.description = description
        lhs.maximalize()
        rhs.maximalize()
        self.lhs = lhs 
        self.rhs = rhs 
                
        self.ruleType = 'MIXED'
        if (rhs.partOf(lhs)):
            self.ruleType = 'SUBTRACTIVE'
        elif (lhs.partOf(rhs)):
            self.ruleType = 'ADDITIVE'

        self.detection = detection(self)
        if self.name in rule.register:
            print('a rule called '+self.name+' already exists in the rule register')
        else:
            rule.register[self.name] = self
        self.trace = []
            
    def __str__(self):
        return self.name
        
    def isValid(self):
        return self.detection.isValid()
            
        
    """
    change rule name
    input: newName <string>: new name of the rule, must be unique
    """
    def setName(self, newName):
        if newName in rule.register:
            raise ValueError('There already exists a rule with rule name '+str(newName))
        rule.register[newName] = rule.register[self.name]
        del rule.register[self.name]
        
        detection.applicationDict[newName] = detection.applicationDict[self.name]
        del detection.applicationDict[self.name]
        
        from ...io.sdlParser import sdlParser
        if self.name in sdlParser.rules:
            sdlParser.rules[newName] = sdlParser.rules[self.name]
            del sdlParser.rules[self.name]
        self.name = newName

        
    """
    change rule description
    input: newDesc <string>: new description for the rule
    """
    def setDescription(self, newDesc):
        self.description = newDesc
        
    """
    *static method*
    clear all existing rules
    """
    @staticmethod
    def cleanup():
        for i in list(rule.register.keys()):
            del rule.register[i]
        
    """
    *static method*
    get a specific rule from the rule register with rule name
    input:  ruleName <string>: name of the rule
    output: <rule>: the rule object with name %ruleName%
    """
    @staticmethod
    def getRule(ruleName):
        return rule.register[ruleName]
        
    """
    detect and return matches for a specific shape(form)
    input:  form <form>: the form to detect rule matches on
    output: <list> of <ruleApplication>: a list of rule applications, if there 
            is no matches for this rule, an empty list will be returned
    """
    def detect(self, form):
        form.maximalize()
        det = self.detection.detect(form)
        return det
        
    """
    detect matches and apply one of the results onto a specific shape(form),
    the rule application to be applied is selected at random so the result may
    differ when the code is rerun
    input:  form <form>: the form to apply the rule on
                         **this form will be modified to become the resulting form
                           at the end of this method
            random <boolean>: variable to indicate if the result will be selected
                              at random
                              random = True: select at random
                              random = False: select the first application
                                              and progressively increase in 
                                              order if backtrace
            backtrace <boolean>: variable to indicate if this is a backtrace from
                                 a flow
                                 *only used in flow*
            s <list>: a list of string indicating the process of a flow
                      *only used in flow*
    output: <form>: the resulting form after applying the rule
            ** the input form is also modified to become the resulting form
    """
    def apply(self, form = None, random = True, backtrace = False, s = []):
        if backtrace:
            if len(self.trace) == 0:
                print(' **'+self.name+'** ', end='')
                sys.stdout.flush()
                return False
            tracing = self.trace.pop(-1)
            app = tracing[0]
            form = tracing[2].duplicate()
        else:
            form.maximalize()
            app = self.detect(form) #detect the possible matches
            tracing = [app, list(range(len(app))), form.duplicate()]

            
        if len(tracing[1])>0:
            if random is True:
                randInt = randint(0, len(tracing[1])-1)
            else:
                randInt = 0
            appIndex = tracing[1].pop(randInt)
            app[appIndex].perform(form)
        else:
            print(' ##'+self.name+'## ', end='')
            sys.stdout.flush()
            return False
            
        self.trace.append(tracing)
        print(' '+self.name, end='')
        print('('+str(appIndex+1)+'/'+str(len(app))+')', end = '')
        sys.stdout.flush()

        s.append(' '+self.name+'('+str(appIndex+1)+'/'+str(len(app))+')')
            
        return form
        
    """
    detect matches and apply one of the results onto a specific shape(form),
    the rule application to be applied is specified by its index
    input:  form <form>: the form to apply the rule on
            choice <int>: the index of the application in the result list
    """
    def applySpecific(self, form = None, choice = 0):
        form.maximalize()
        app = self.detect(form)

        if choice >= len(app):
            print('no result')
            return form
        app[choice].perform(form)

        print(' '+self.name, end='')
        print('('+str(choice+1)+'/'+str(len(app))+')', end = '')
        sys.stdout.flush()

        return form
        
    def addPredicate(self, cond, keyList):
        for i in self.detection.analysis.maps:
            if i.type == 'map-graphMap':
                self.detection.analysis.maps[i].addPredicate(cond, keyList)
                return
        
    def addDirective(self, cond, keyList):
        for i in self.detection.analysis.maps:
            if i.type == 'map-graphMap':
                self.detection.analysis.maps[i].addDirective(cond, keyList)
                return

    def evaluatePredicate(self, transfSet, lhs, tlhs, form):   
        for i in self.detection.analysis.maps:
            if i.type == 'map-graphMap':
                if not self.detection.analysis.maps[i].evaluatePredicate(transfSet, lhs, tlhs, form):
                    return False
                break
        return True
                    
    def solveDirectives(self, transfSet, lhs, tlhs, rhs, form):   
        for i in self.detection.analysis.maps:
            if i.type == 'map-graphMap':
                a = self.detection.analysis.maps[i].solveDirectives(transfSet, lhs, tlhs, rhs, form)
                return a
        return [transfSet]
