# -*- coding: utf-8 -*-
"""
Created on Tue Mar 28 13:17:49 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .plane import plane2D, plane3D
from ..struct.coordinate import c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..io.parseReader import ParsingError

class circle2D(plane2D):
    type = 'ind-circle2D'
    def __init__(self, sort, center, start, end, *args, **kwargs):
        super(circle2D, self).__init__(sort)     
        self.start = start
        self.end = end
        self.center = center
        self.radius = self.start.subtract(self.center).length()
        
    def isNil(self):
        return False
        
    def printVal(self):
        return 'center:'+str(self.center)+', p1:'+str(self.start)+', p2:'+str(self.end)
        
    def __str__(self):
        return '<'+str(self.center)+','+str(self.start)+','+str(self.end)+'>'
        
    def equals(self, other):
        if self.center == other.center:
            return self.radius == other.radius
        return False
        
    def containsPoint(self, point):
        if point.subtract(self.center).length() == self.radius:
            return True
        return False
        
    def compareValue(self, other):
        c = self.center.compareValue(other.center)
        if c != 0:
            return c
        return self.radius.compareValue(other.radius)
            
    def concentric(self, other):
        if other.type != 'ind-circle2D':
            return False
        return self.center == other.center
        
    def diameter(self):
        return self.radius * 2
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        center = self.center.transform(affine)
        start = self.start.transform(affine)
        end = self.end.transform(affine)    
        return circle2D(self.sort, center, start, end)
        
    def replicateInd(self):
        return circle2D(self.sort, self.center, self.start, self.end)
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        x,y = vector2D.readVector(reader)
        center = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        return circle2D(sort, center, start, end)
        
class circle3D(plane3D):
    type = 'ind-circle3D'
    
    def __init__(self, sort, center, start, end, *args, **kwargs):
        super(circle3D, self).__init__(sort, [center, start, end])   
        self.radius = start.subtract(center).length()
        self.start = self.localize(start)
        self.end = self.localize(end)
        self.center = self.localize(center)
        
        
    def isNil(self):
        return False
        
    def printVal(self):
        return 'center:'+str(self.position(self.center))+', p1:'+str(self.position(self.start))+', p2:'+str(self.position(self.end))
    
    def __str__(self):
        return '<'+str(self.position(self.center))+','+str(self.position(self.start))+','+str(self.position(self.end))+'>'
        
    def equals(self, other):
        if self.normal == other.normal and self.center == other.center and self.radius == other.radius :
            return True
        return False
        
    def containsPoint(self, point):
        cent = self.position(self.center)
        if point == cent or not point.subtract(cent).dotproduct(self.normal) == 0:
            return False
        p = self.localize(point)
        if p == self.start or p == self.end:
            return True
        if p.subtract(self.center).length() == self.radius:
            return True
        return False
        
    def compareValue(self, other):
        c = super(circle3D, self).compareValue(other)
        if c != 0:
            return c
        c = self.center.compareValue(other.center)
        if c != 0:
            return c
        return self.radius.compareValue(other.radius)
            
    def concentric(self, other):
        if other.type != 'ind-circle3D':
            return False
        if not super(circle3D, self).__eq__(other):
            return False
        return self.center == other.center
        
    def diameter(self):
        return self.radius * 2
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        center = self.position(self.center).transform(affine)
        start = self.position(self.start).transform(affine)
        end = self.position(self.end).transform(affine)    
        return circle3D(self.sort, center, start, end)
        
    def replicateInd(self):
        return circle3D(self.sort, self.center, self.start, self.end)
        
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        x,y = vector3D.readVector(reader)
        center = vector3D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector3D.readVector(reader)
        start = vector3D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector3D.readVector(reader)
        end = vector3D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        return circle3D(sort, center, start, end)