# -*- coding: utf-8 -*-
"""
Created on Tue Jan 10 15:28:09 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .tree import tree
from .symbols import symbols
from ...io.parseReader import ParsingError, parseReader


class function(tree):
    FNAME = {}
    
    def __init__(self, name):
        self.type = 'description-function'
        self.name = name
        self.argument = None
        self.assist = None
    
    def __eq__(self, other):
        if other.type != 'description-function':
            return False
        if self.name != other.name or self.argument != other.argument:
            return False
        return self.assist is None or self.assist == other.assist
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def __hash__(self):
        hashNum = 5
        hashNum = 23 * hashNum + hash(self.name)
        hashNum = 23 * hashNum + hash(self.argument)
        hashNum = 23 * hashNum + hash(self.assist)
        return hashNum
    
    def compareValue(self, other):
        if other.type == 'description-function':
            if self.name > other.name:
                return 1
            elif self.name < other.name:
                return -1
            if self.argument > other.argument:
                return 1
            elif self.argument < other.argument:
                return -1
            if self.assist is None:
                if other.assist is None:
                    return 0
                else:
                    return -1
            if self.assist > other.assist:
                return 1
            elif self.assist < other.assist:
                return -1
            return 0
        elif (other.type == 'description-stringExpression' or
              other.type == 'description-numericExpression' or
              other.type == 'description-tupleExpression'):
                  return -other.compareValue(self)
        c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0
    
    def __str__(self):
        result = ''
        result += self.name + symbols.PAREN_OPEN + str(self.argument)
        if self.assist is not None:
            result += symbols.COMMA + ' ' + str(self.assist)
        result += symbols.PAREN_CLOSE
        return result
        
    def printVal(self):
        result = ''
        result += self.name + symbols.PAREN_OPEN + self.argument.printVal()
        if self.assist is not None:
            result += symbols.COMMA + ' ' + str(self.assist)
        result += symbols.PAREN_CLOSE
        return 'func:'+result
        
    def evaluate(self, assignment, sort):
        arg = self.argument.evaluate(assignment, sort)
        if self.assist is None:
            return arg.apply(self.name)
        return arg.apply(self.name, self.assist.evaluate(assignment, sort))
        
    @staticmethod
    def reserve(name):
        function.FNAME[name] = name
        
    @staticmethod
    def isFunctionName(name):
        if name in function.FNAME:
            return True
        return False
    
    @staticmethod
    def parse(desc, reader):
        name = reader.value()
        # token must be identifier with parenthesis after e.g. adjacencies(a4, a5 a6), length(‘‘room’’)
        if reader.token() != parseReader.IDENTIFIER or reader.newToken() != symbols.PAREN_OPEN:
            raise ParsingError(reader, 'expected an identifier followed by a "("')
        
        if not function.isFunctionName(name):
            raise ParsingError(reader, "function '" + name + "' not recognized")
        # create function with function name
        result = function(name)
        
        reader.newToken()        
        # parse the part inside the function
        result.argument = desc.parseEntity(desc, reader)

        # if token is not ')' or ',' --> could be part of a tuple expression --> parse again
        if reader.token() != symbols.PAREN_CLOSE and reader.token() != symbols.COMMA:
            nextEntity = desc.parseEntity(desc, reader)
            # if parse return None --> error
            if nextEntity is None:
                raise ParsingError(reader, "expected a ',' or ')'")
            # continue parsing as tuple expression, save it as argument
            result.argument = desc.tupleExpression.parseContinue(desc, result.argument, nextEntity, reader)
        # if token is ',' --> next parameter for the function --> parse again
        if reader.token() == symbols.COMMA:
            reader.newToken()
            # parse again as assist
            result.assist = desc.parseEntity(desc, reader)
            # if token is not ')' --> could be tuple expression --> parse again
            if reader.token() != symbols.PAREN_CLOSE:
                nextEntity = desc.parseEntity(desc, reader)
                # if next is None --> error
                if nextEntity is None:
                    raise ParsingError(reader, "expected a ',' or ')'")
                # set assist as the tuple expression
                result.assist = desc.tupleExpression.parseContinue(desc, result.assist, nextEntity, reader)
        reader.newToken()
        return result
    
