# -*- coding: utf-8 -*-
"""
Created on Tue Jan 10 13:41:26 2017

@author: akibdpt
"""
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import str
"""
from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)
"""
"""
from future import standard_library
standard_library.install_aliases()
"""
from mpmath import isint
from .tree import tree
from .symbols import symbols
from .stringLiteral import stringLiteral
from .literal import literal
from ...io.parseReader import parseReader, ParsingError
from ...sort.sort import sort

class stringExpression(tree):
    def __init__(self):
        self.type = 'description-stringExpression'
        self.concat = []
        
    def __eq__(self, other):
        if other.type != 'description-stringExpression':
            return False
        return self.concat == other.concat

    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def compareValue(self, other):
        if other.type == 'description-stringExpression':
            for i in range(len(self.concat)):
                if i >= len(other.concat):
                    break
                if self.concat[i] > other.concat[i]:
                    return 1
                elif self.concat[i] < other.concat[i]:
                    return -1
            c = len(self.concat) - len(other.concat)
            if c > 0:
                return 1
            elif c < 0:
                return -1
            return 0
        if other.type == 'description-tuple' or other.type == 'description-tupleExpression':
            c = hash(self.__class__) - hash(other.__class__)
            if c > 0:
                return 1
            elif c < 0:
                return -1
            return 0
        if self.concat[0] > other:
            return 1
        elif self.concat[0] < other:
            return -1
        return 0

    def __str__(self):
        result = ''
        for i in self.concat:
            try:
                if i.type == 'description-numericExpression' or i.type == 'description-reference':
                    result += symbols.PAREN_OPEN + str(i) + symbols.PAREN_CLOSE + symbols.CONCAT
                else:
                    result += str(i) + symbols.CONCAT
            except TypeError:
                result += str(i) + symbols.CONCAT
        return result[0:-1]

    def printVal(self):
        result = '('
        for i in self.concat:
            try:
                if i.type == 'description-numericExpression' or i.type == 'description-reference':
                    result += symbols.PAREN_OPEN + i.printVal() + symbols.PAREN_CLOSE + symbols.CONCAT
                else:
                    result += i.printVal() + symbols.CONCAT
            except TypeError:
                result += i.printVal() + symbols.CONCAT
        return 'strExpr:'+ result[0:-1]+')'

    def map(self, assignment, sort, onto):
        if onto.type != 'description-stringLiteral':
            return False
        target = onto.s
        index = 0
        for entIndex in range(len(self.concat)):
            entity = self.concat[entIndex]
            if entity.type == 'description-stringLiteral':
                source = entity.s
                if source == target[index:index+len(source)]:
                    index += len(source)
                else:
                    return False
            elif entity.type == 'description-parameter':
                if entIndex == len(self.concat)-1:
                    if len(target[index:]) == 0:
                        return False
                    return entity.map(assignment, sort, stringLiteral(target[index:]))
                    """
                elif entity.isEnumerated():
                    source = self.concat[entIndex + 1].s
                    try:
                        step = target.index(source, index)
                    except ValueError:
                        return False
                    print(index, step, target[index: step])
                    paramOnto = stringLiteral(target[index: step])
                    print(paramOnto)
                    if not entity.map(assignment, sort, paramOnto):
                        return False
                    index += len(paramOnto)
                    """

                elif self.concat[entIndex+1].type == 'description-stringLiteral':
                    source = self.concat[entIndex + 1].s
                    try:
                        step = target.index(source, index)
                    except ValueError:
                        return False
                    paramOnto = stringLiteral(target[index: step])
                    if not entity.map(assignment, sort, paramOnto):
                        return False
                    index += len(paramOnto)
                elif self.concat[entIndex+1].type == 'description-parameter' and entity.hasCondition():
                    length = len(entity.condition.operand.tuple[0])
                    for i in entity.condition.operand:
                        if len(i) != length:
                            raise TypeError('!!!!!!!')
                    paramOnto = stringLiteral(target[index: index+length])
                    if len(paramOnto) == 0:
                        return False
                    if not entity.map(assignment, sort, paramOnto):
                        return False
                    index += len(paramOnto)
                else:
                    raise TypeError('!!!!!!!')
        return True
    
    def evaluate(self, assignment, sort):
        s = '"'
        for entity in self.concat:
            ev = entity.evaluate(assignment, sort)
            if ev.type == 'description-literal':
                if ev != literal.getLiteral('e'):
                    s += str(ev)
            elif ev.type == 'description-numberLiteral':
                if isint(ev.value.value):
                    s += str(int(ev.value.value))
                else:
                    s += str(ev.value.value)
            elif ev.type == 'description-stringLiteral':
                s += ev.s
        s += '"'
        return stringLiteral(s)
        
    @staticmethod
    def parse(desc, reader):
        # if preview token is not '.' --> string literal e.g. "default"
        if reader.previewToken() != symbols.CONCAT:
            if reader.token() != parseReader.STRING:
                raise ParsingError(reader, "expected '\"'")
            literal = desc.stringLiteral(reader.value(), reader.value())    
            reader.newToken()
            return literal
        
        # if preview token is '.' --> string expression e.g. "with".(c+1)."columns", (n-1)."rows"
        # parse the first part of the expression, up to the first '.'
        # if token == string --> string literal
        if reader.token() == parseReader.STRING:
            entity = desc.stringLiteral(reader.stringVal(), reader.value())
            reader.newToken()
        # if token == identifier --> parameter
        elif reader.token() == parseReader.IDENTIFIER:
            entity = desc.parameter.parse(desc, reader, False)
        # neither string nor identifier --> error
        else:
            raise ParsingError(reader, "expected '\"' or identifier")
        # continue parsing the rest of the 
        return stringExpression.parseContinue(desc, entity, reader)
        
    @staticmethod
    def parseContinue(desc, entity, reader):
        # if token is not '.' --> error        
        if reader.token() != symbols.CONCAT:
            raise ParsingError(reader, "expected '.'")
        # add the first part (parsed above) into a new string expression
        result = stringExpression()
        result.concat.append(entity)
        # keep continuing while the string expression has not ended yet:
        while reader.token() == symbols.CONCAT:
            # if token is string --> string literal, then continue checking next token
            if reader.newToken() == parseReader.STRING:
                result.concat.append(desc.stringLiteral(reader.stringVal(), reader.value()))
                reader.newToken()
            # if token is number --> parse number literal (next token included)
            elif reader.token() == parseReader.NUMBER:
                result.concat.append(desc.numberLiteral.parse(desc, reader))
            # if token is identifier --> literal/function/reference/parameter then next token
            elif reader.token() == parseReader.IDENTIFIER:
                name = reader.value()
                # if there is a literal with this name --> literal
                if desc.literal.isLiteral(name):
                    result.concat.append(desc.literal.getLiteral(name))
                    reader.newToken()
                # if there is a function with this name --> parse function
                elif desc.function.isFunctionName(name) and not reader.previewToken() == '.':
                    result.concat.append(desc.function.parse(desc, reader))
                # if there is a sort with this name --> parse reference
                elif sort.checkSort(name):
                    result.concat.append(desc.reference.parse(desc, reader))
                # else --> parse parameter
                else:
                    result.concat.append(desc.parameter.parse(desc, reader, True))
            # if token = () --> numeric expression
            elif reader.token() == symbols.PAREN_OPEN:
                result.concat.append(desc.numericExpression.parse(desc, reader))
            # else --> error
            else: raise ParsingError(reader, 'expected anything but a tuple') 
        return result