# -*- coding: utf-8 -*-
"""
Created on Wed Jan 11 11:06:23 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import range
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""

from .tree import tree
from .symbols import symbols
from ...io.parseReader import parseReader, ParsingError
from ...struct.coordinate import coordinate, c0, c1

from .function import function
from .numberLiteral import numberLiteral
from .literal import literal


class tupleDesc(tree):
    LENGTH = "length"
    FIRST = "first"
    MINIMUM = "min"
    MAXIMUM = "max"
    LAST = "last"
    UNIQUE = "unique"
    SEGMENTS = "segments"
    PAIRWISE = "pairwise"
    LOOPS = "loops"
    ADJACENCIES = "adjacencies"
    
    EMPTY_TUPLE = []
    FUNCTIONS = [LENGTH, FIRST, LAST, MINIMUM, MAXIMUM, UNIQUE, SEGMENTS, PAIRWISE, LOOPS, ADJACENCIES]
    #ASSISTS = [None, None, None, None, None, None, None, None, None, tupleDesc]
    """
    to be put in __init__.py
    """
    ASSISTS = [None, None, None, None, None, None, None, None, None, None]
    """
    """
    for fname in FUNCTIONS:
        function.reserve(fname)
    
    def __init__(self, bracket, separator, entity = None):
        self.type = 'description-tuple'
        self.tuple = []
        self.bracket = bracket
        if bracket == symbols.PAREN_OPEN:
            self.endBracket = symbols.PAREN_CLOSE
        elif bracket == symbols.ANGLE_OPEN:
            self.endBracket = symbols.ANGLE_CLOSE
        elif bracket == symbols.SQUARE_OPEN:
            self.endBracket = symbols.SQUARE_CLOSE
        elif bracket == symbols.SET_OPEN:
            self.endBracket = symbols.SET_CLOSE
        elif bracket == symbols.NONE:
            self.endBracket = symbols.NONE
        else:
            raise TypeError('bracket not recognized')
        if (separator != symbols.COMMA and separator != symbols.SEMICOLON and 
            separator != symbols.NONE and separator != symbols.UNKNOWN):
            raise TypeError('comma not recognized')
        self.separator = separator
        if entity is not None:
            self.tuple.append(entity)
        
    @staticmethod
    def newTuple(first, second):
        tpl = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
        tpl.tuple.append(first)
        tpl.tuple.append(second)
        return tpl
        
    def isEmpty(self):
        return self.tuple is None or len(self.tuple) == 0
        
    def addEntity(self, entity, index = -1):
        if index > len(self.tuple):
            raise KeyError('index exceeds length of tuple')
        if len(self.tuple) > 0 and self.bracket is not None and (self.separator is None or 
                                                                 self.separator is symbols.UNKNOWN):
            self.separator = symbols.COMMA
        if index == -1:
            self.tuple.append(entity)
        else: 
            self.tuple.insert(index, entity)
            
    def __hash__(self):
        hashNum = 5
        hashNum = 97 * hashNum
        for i in self.tuple:
            hashNum += hash(i)
        return hashNum

        
    def getEntity(self, index):
        if self.tuple is None or index >= len(self.tuple):
            return None
        return self.tuple[index]
        
    def shareDelimiters(self, other):
        return self.separator == other.separator and self.bracket == other.bracket
        
    def __len__(self):
        return len(self.tuple)
        
    def __eq__(self, other):
        if other.type != 'description-tuple':
            return False
        if len(self.tuple) != len(other.tuple):
            return False
        for i in range(len(self.tuple)):
            if self.tuple[i] != other.tuple[i]:
                return False
        return True
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def compareValue(self, other):
        if other.type == 'description-tuple':
            for i in range(len(self.tuple)):
                if i >= len(other.tuple):
                    break
                if self.tuple[i] > other.tuple[i]:
                    return 1
                elif self.tuple[i] < other.tuple[i]:
                    return -1
            c = len(self.tuple) - len(other.tuple)
            if c > 0:
                return 1
            elif c < 0:
                return -1
            return 0
        elif other.type == 'description-literal' or other.type == 'description-tupleExpression':
            return -other.compareValue(self)
        else:
            c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0

    def getNumber(self, assignment, sort):
        num = 0
        for i in self.tuple:
            num += i.getNumber(assignment, sort)
        return num
            
    def __str__(self):
        s = ''
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.bracket
        if not self.isEmpty():
            for entity in self.tuple:
                s += str(entity)
                if self.separator == symbols.COMMA or self.separator == symbols.SEMICOLON:
                    s += self.separator
                s += symbols.SPACE
            if self.separator == symbols.COMMA or self.separator == symbols.SEMICOLON:
                s = s[:-2]
            else:
                s = s[:-1]
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.endBracket
        return s
        
    def printVal(self):
        s = ''
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.bracket
        if not self.isEmpty():
            for entity in self.tuple:
                s += entity.printVal()
                if self.separator == symbols.COMMA or self.separator == symbols.SEMICOLON:
                    s += self.separator
                s += symbols.SPACE
            if self.separator == symbols.COMMA or self.separator == symbols.SEMICOLON:
                s = s[:-2]
            else:
                s = s[:-1]
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.endBracket
        return 'tup'+s
        
    def duplicate(self):
        result = tupleDesc(self.bracket, self.separator)
        for i in self.tuple:
            result.addEntity(i)
        return result
    
    def __iter__(self):
        return iter(self.tuple)
        
    def map(self, assignment, sort, onto):
        if onto.type != 'description-tuple' or len(self) != len(onto):
            return False
        for i in range(len(self.tuple)):
            if not self.tuple[i].map(assignment, sort, onto.tuple[i]):
                return False
        return True
        
    def evaluate(self, assignment, sort):
        if self.isEmpty():
            return self
        result = tupleDesc(self.bracket, self.separator)
        for entity in self.tuple:
            result.addEntity(entity.evaluate(assignment, sort))
        return result
        
    def sum(self, other):
        if len(self) != len(other):
            raise KeyError('tuples must have the same length')
        for i in range(len(self.tuple)):
            if self.tuple[i].type == 'description-numberLiteral':
                self.tuple[i] = numberLiteral(self.tuple[i].value + other.tuple[i].value)
    
    def functionExists(self, name, assist= None):
        for i in range(len(tupleDesc.FUNCTION)):
            if tupleDesc.FUNCTION[i] == name and tupleDesc.ASSISTS[i] == assist:
                return True
                
        return False    
        
    def apply(self, function, value = None):
        if value is None:
            if function == self.MAXIMUM:
                return max(self.tuple).duplicate()
            elif function == self.MINIMUM:
                return min(self.tuple).duplicate()
            elif function == self.LENGTH:
                return numberLiteral(coordinate(len(self.tuple)))
            elif function == self.FIRST:
                return self.tuple[0].duplicate()
            elif function == self.LAST:
                return self.tuple[-1].duplicate()
            elif function == self.UNIQUE:
                result = tupleDesc(self.bracket, self.separator)
                for entity in self.tuple:
                    if len(result.tuple) == 0 or entity not in result.tuple:
                        result.addEntity(entity)
                return result
            elif function == self.SEGMENTS:
                result = tupleDesc(self.bracket, self.separator)
                for i in range(1,len(self.tuple)):
                    result.addEntity(tupleDesc.newTuple(self.tuple[i-1], self.tuple[i]))
                return result
            elif function == self.PAIRWISE:
                result = tupleDesc(self.bracket, self.separator)
                i = 1
                while i < len(self.tuple):
                    result.addEntity(tupleDesc.newTuple(self.tuple[i-1], self.tuple[i]))
                    i += 2
                return result
            elif function == self.LOOPS:
                result = tupleDesc(self.bracket, self.separator)
                loops = []
                for i in range(len(self.tuple)):
                    for j in range(i+1, len(self.tuple)):
                        if self.tuple[i] == self.tuple[j]:
                            loop = tupleDesc(self.bracket, self.separator)
                            for entity in range(i,j):
                                line = tupleDesc.newTuple(self.tuple[entity], self.tuple[entity+1])
                                loop.addEntity(line)
                            loops.append(loop)
                for i in loops:
                    j = 0
                    while j < len(i.tuple):
                        if i.tuple[j].tuple[0] == i.tuple[j-1].tuple[1] and i.tuple[j].tuple[1] == i.tuple[j-1].tuple[0]:
                            i.tuple.pop(j)
                            i.tuple.pop(j-1)
                            if j > 0:
                                j -= 1
                        else:
                            j += 1
                #loops.sort(key = lambda s: len(s.tuple))
                for rounds in range(2):
                    for i in range(len(loops)):
                        if loops[i] is None:
                            continue
                        if len(loops[i].tuple) <= 2:
                            loops[i] = None
                            continue
                        #print('....',loops)
                        del1 = False
                        for j in range(i+1, len(loops)):
                            if loops[j] is None:
                                continue
                            if len(loops[j].tuple) <= 2:
                                loops[j] = None
                                continue
                            common = []
                            lj = [False] * len(loops[j])
                            for ent1 in loops[i]:
                                for index, ent2 in enumerate(loops[j]):
                                    if lj[index]:
                                        continue
                                    if (ent1.tuple[0] == ent2.tuple[0] and ent1.tuple[1] == ent2.tuple[1]):
                                        common.append(ent1)
                                        lj[index] = True
                                        break
                                    elif (ent1.tuple[0] == ent2.tuple[1] and ent1.tuple[1] == ent2.tuple[0]):
                                        common.append(ent1)
                                        lj[index] = True
                                        break
    
                            if len(common) == 0:
                                continue
                            elif len(common) == len(loops[j]) == len(loops[i]):
                                loops[j] = None
                                continue
                            elif len(common) < min(len(loops[i]),len(loops[j]))/2:
                                continue
                            if len(loops[i]) > len(loops[j]):
                                loopRev = loops[i]
                                loopAdd = loops[j].duplicate()
                            else:
                                loopRev = loops[j]
                                loopAdd = loops[i].duplicate()
                            for c in common:
                                for lr in loopRev.tuple:
                                    if (c.tuple[0] == lr.tuple[0] and c.tuple[1] == lr.tuple[1]) or (c.tuple[0] == lr.tuple[1] and c.tuple[1] == lr.tuple[0]):
                                        loopRev.tuple.remove(lr)
                                        break
                                for la in loopAdd.tuple:
                                    if (c.tuple[0] == la.tuple[0] and c.tuple[1] == la.tuple[1]) or (c.tuple[0] == la.tuple[1] and c.tuple[1] == la.tuple[0]):
                                        loopAdd.tuple.remove(la)
                                        break
                            for add in loopAdd:
                                reverse = 0
                                normal = 0
                                for ent in range(len(loopRev)):
                                    if loopRev.tuple[ent].tuple[0] == add.tuple[0]:
                                        reverse += 1
                                        index = ent
                                    elif loopRev.tuple[ent].tuple[1] == add.tuple[1]:
                                        reverse += 1
                                        index = ent + 1
                                        if index == len(loopRev):
                                            index = 0
                                    elif loopRev.tuple[ent].tuple[1] == add.tuple[0]:
                                        normal += 1
                                        index = ent + 1
                                        if index == len(loopRev):
                                            index = 0
                                    elif loopRev.tuple[ent].tuple[0] == add.tuple[1]:
                                        normal += 1
                                        index = ent
                                if normal > reverse:
                                    loopRev.addEntity(add, index)
                                elif normal < reverse:
                                    loopRev.addEntity(tupleDesc.newTuple(add.tuple[1],add.tuple[0]), index)
                                else:
                                    raise ValueError('segment not recognized if reverse or not')
                            if len(loopRev.tuple) <= 2:
                                loopRev = None
                        if del1:
                            loops[i] = None
                            continue
                for i in range(len(loops)):
                    if loops[i] is None:
                        continue
                    for j in range(i+1, len(loops)):
                        if loops[j] is None:
                            continue
                        if len(loops[i].tuple) != len(loops[j].tuple):
                            continue
                        if set(loops[i].tuple) == set(loops[j].tuple):
                            loops[j] = None
                loops = filter(None, loops)
                for i in loops:
                    loop = tupleDesc(self.bracket, self.separator)
                    for j in i.tuple:
                        loop.addEntity(j.tuple[0])
                    result.addEntity(loop)
                if len(self.tuple) == 0:
                    return literal.getLiteral('e')
                return result
        elif function == self.ADJACENCIES:
            result = tupleDesc(self.bracket, self.separator)
            for i in range(len(self.tuple)+1):
                mat = tupleDesc(self.bracket, self.separator)
                for j in range(len(self.tuple)+1):
                    if i == j:
                        mat.addEntity(numberLiteral(c1))
                    else:
                        mat.addEntity(numberLiteral(c0))
                result.addEntity(mat)

            for door in value:
                try:
                    for i in range(len(self.tuple)):
                        for wall1 in range(len(self.tuple[i].tuple)):
                            nex = wall1+1
                            if nex == len(self.tuple[i].tuple):
                                nex = 0
                            if (self.tuple[i].tuple[wall1] == door.tuple[0] and  
                                (self.tuple[i].tuple[wall1-1] == door.tuple[1] or 
                                self.tuple[i].tuple[nex] == door.tuple[1])):
                                    outside = True
                                    for j in range(i+1,len(self.tuple)):
                                        for wall2 in range(len(self.tuple[j].tuple)):
                                            nex1 = wall2+1
                                            if nex1 == len(self.tuple[j].tuple):
                                                nex1 = 0
                                            if (self.tuple[j].tuple[wall2] == door.tuple[0] and  
                                                (self.tuple[j].tuple[wall2-1] == door.tuple[1] or 
                                                 self.tuple[j].tuple[nex1] == door.tuple[1])):
                                                     outside = False
                                                     result.tuple[j+1].tuple[i+1] = c1
                                                     result.tuple[i+1].tuple[j+1] = c1
                                                     break
                                    if outside:
                                        result.tuple[0].tuple[i+1] = c1
                                        result.tuple[i+1].tuple[0] = c1
                                    raise StopIteration
                except StopIteration:
                    pass
            return result
        raise NotImplementedError('function '+function+' not recognized or not implemented')
        
    def contains(self, other):
        if other.isEmpty():
            return True
        if self.isEmpty():
            return False
        for i in other.tuple:
            check = False
            for j in self.tuple:
                if i.type == j.type and j.contains(i):
                    check = True
                    break
            if not check:
                return False
            
        return True
        
    def similarFragment(self, other):
        if self.type != other.type:
            return False
        if self.bracket != other.bracket or self.separator != other.separator:
            return False
        if len(self.tuple) != len(other.tuple):
            return False
        for i in range(len(self.tuple)):
            if not self.tuple[i].similarFragment(other.tuple[i]):
                return False
        return True
        
    def mergeFragment(self, other):
        for i in range(len(self.tuple)):
            self.tuple[i].mergeFragment(other.tuple[i])

    @staticmethod
    def parse(desc, reader):
        try:
            result = tupleDesc(reader.token(), symbols.UNKNOWN)
        except TypeError:
            raise ParsingError(reader, "expected a '(', '<' or '['")
        if reader.newToken() == result.endbracket:
            reader.newToken()
            return result
        result.addEntity(tree.parseEntity(desc, reader))
        if reader.token() == result.endbracket:
            reader.newToken()
            return result
        if reader.token() != symbols.COMMA and reader.token() != symbols.SEMICOLON:
            raise ParsingError(reader, "expected a ',' or ';'")
        result.separator = reader.token()
        result.parseContinue(desc, reader)
        return result

    def parseContinue(self, desc, reader):
        while reader.token() == self.separator:
            reader.newToken()
            self.addEntity(desc.parseEntity(desc, reader))
        if reader.token() == self.endBracket:
            reader.newToken()
        else: 
            raise ParsingError(reader, "expected a '" + self.endBracket +  "'")

tupleDesc.ASSISTS[-1] = tupleDesc