# -*- coding: utf-8 -*-
"""
Created on Fri Apr 28 14:48:02 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .plane import plane2D, plane3D
from ..struct.coordinate import coordinate, c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..io.parseReader import ParsingError
from mpmath import sqrt


class ellipse2D(plane2D):
    type = 'ind-ellipse2D'
    def __init__(self, sort, focus1, focus2, start, end, *args, **kwargs):
        super(ellipse2D, self).__init__(sort)     
        self.start = start
        self.end = end
        
        if focus1 > focus2:
            self.foci = [focus2, focus1]
        else:
            self.foci = [focus1, focus2]
            
        self.center = focus1.midpoint(focus2)
        
        dist1 = self.start.subtract(self.foci[0])
        dist2 = self.start.subtract(self.foci[1])
        self.radii = [None,None]
        self.radii[0] = coordinate(sqrt(dist1.dotproduct(dist1).value) + sqrt(dist2.dotproduct(dist2).value)) / c2
        dist1 = self.foci[1].subtract(self.foci[0])
        self.radii[1] = coordinate(sqrt((self.radii[0] ** c2 - dist1.dotproduct(dist1) / 4).value))
                
    def isNil(self):
        return False
        
    def printVal(self):
        return 'foci:'+str(self.foci[0])+','+str(self.foci[1])+', p1:'+str(self.start)+', p2:'+str(self.end)
        
    def __str__(self):
        return '<'+str(self.foci[0])+','+str(self.foci[1])+','+str(self.start)+','+str(self.end)+'>'
        
    def equals(self, other):
        if self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
        
    def containsPoint(self, point):
        dist1 = point.subtract(self.foci[0])
        dist2 = point.subtract(self.foci[1])
        radius1 = coordinate(sqrt(dist1.dotproduct(dist1).value) + sqrt(dist2.dotproduct(dist2).value)) / c2
        distFoci = self.foci[1].subtract(self.foci[0])
        if (coordinate(sqrt((radius1 ** c2 - distFoci.dotproduct(distFoci) / 4).value)) == self.radii[1] and
            radius1 == self.radii[0]):
                return True
        return False
        
    def compareValue(self, other):
        c = self.foci[0].compareValue(other.foci[0])
        if c != 0:
            return c
        c = self.foci[1].compareValue(other.foci[1])
        if c != 0:
            return c
        c = self.radii[0].compareValue(other.radii[0])
        if c != 0:
            return c
        return self.radii[1].compareValue(other.radii[1])
            
    def concentric(self, other):
        if other.type != 'ind-ellipse2D':
            return False
        return self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        focus1 = self.foci[0].transform(affine)
        focus2 = self.foci[1].transform(affine)
        start = self.start.transform(affine)
        end = self.end.transform(affine)    
        return ellipse2D(self.sort, focus1, focus2, start, end)
        
    def replicateInd(self):
        return ellipse2D(self.sort, self.foci[0], self.foci[1], self.start, self.end)
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        x,y = vector2D.readVector(reader)
        focus1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector2D.readVector(reader)
        focus2 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        return ellipse2D(sort, focus1, focus2, start, end)
        
class ellipse3D(plane3D):
    type = 'ind-ellipse3D'
    
    def __init__(self, sort, focus1, focus2, start, end, takeSecondValue = False,*args, **kwargs):
        try:
            super(ellipse3D, self).__init__(sort, [focus1, focus2, start])
        except ZeroDivisionError:
            super(ellipse3D, self).__init__(sort, [focus1, focus2, end])
        
        if takeSecondValue:
            self.start = self.localize(end)
            self.end = self.localize(start)
        else:
            self.start = self.localize(start)
            self.end = self.localize(end)
        
        focus1 = self.localize(focus1)
        focus2 = self.localize(focus2)
        
        if focus1 > focus2:
            self.foci = [focus2, focus1]
        else:
            self.foci = [focus1, focus2]
            
        self.center = focus1.midpoint(focus2)
        
        dist1 = self.start.subtract(self.foci[0])
        dist2 = self.start.subtract(self.foci[1])
        self.radii = [None,None]
        self.radii[0] = coordinate(sqrt(dist1.dotproduct(dist1).value) + sqrt(dist2.dotproduct(dist2).value)) / c2
        dist1 = self.foci[1].subtract(self.foci[0])
        self.radii[1] = coordinate(sqrt((self.radii[0] ** c2 - dist1.dotproduct(dist1) / 4).value))

        
    def isNil(self):
        return False
        
    def printVal(self):
        return ('foci:'+str(self.position(self.foci[0]))+','+str(self.position(self.foci[1]))+
                ', p1:'+str(self.position(self.start))+', p2:'+str(self.position(self.end)))
    
    def __str__(self):
        return ('<'+str(self.position(self.foci[0]))+','+str(self.position(self.foci[1]))+
                ','+str(self.position(self.start))+','+str(self.position(self.end))+'>')
        
    def equals(self, other):
        if self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
    
    def containsPoint(self, point):
        fc = self.position(self.foci[0])
        if point == fc or not point.subtract(fc).dotproduct(self.normal) == 0:
            return False
            
        p = self.localize(point)
        if p == self.start or p == self.end:
            return True
            
        dist1 = p.subtract(self.foci[0])
        dist2 = p.subtract(self.foci[1])
        squareLength = dist1.dotproduct(dist1) + dist2.dotproduct(dist2)
        distFoci = self.foci[1].subtract(self.foci[0])
        if ((squareLength - distFoci.dotproduct(distFoci)) == self.radii[1] and
            squareLength == self.radii[0]):
                return True
        return False
    
    def compareValue(self, other):
        c = super(ellipse3D, self).compareValue(other)
        if c != 0:
            return c
        c = self.foci[0].compareValue(other.foci[0])
        if c != 0:
            return c
        c = self.foci[1].compareValue(other.foci[1])
        if c != 0:
            return c
        c = self.radii[0].compareValue(other.radii[0])
        if c != 0:
            return c
        return self.radii[1].compareValue(other.radii[1])
        
        
    def concentric(self, other):
        if other.type != 'ind-ellipse3D':
            return False
        if not super(ellipse3D, self).__eq__(other):
            return False
        return self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()        
        focus1 = self.position(self.foci[0]).transform(affine)
        focus2 = self.position(self.foci[1]).transform(affine)
        start = self.position(self.start).transform(affine)
        end = self.position(self.end).transform(affine)    
        return ellipse3D(self.sort, focus1, focus2, start, end)
        
    def replicateInd(self):
        return ellipse3D(self.sort, self.foci[0], self.foci[1], self.start, self.end)
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        x,y = vector3D.readVector(reader)
        focus1 = vector3D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector3D.readVector(reader)
        focus2 = vector3D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector3D.readVector(reader)
        start = vector3D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        x,y = vector3D.readVector(reader)
        end = vector3D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        return ellipse3D(sort, focus1, focus2, start, end)