# -*- coding: utf-8 -*-
"""
Created on Mon Nov  7 10:28:01 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .individual import individual
from ..struct.coordinate import cn1
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform3D

class plane2D(individual):
    
    type = 'ind-plane2D'
    
    def __init__(self, sort, *args, **kwargs):
        super(plane2D, self).__init__(sort)
    
    def __str__(self):
        return 'plane'
        
    def equals(self, other):
        return True
        
    def isNil(self):
        return False
    
    def compareValue(self, other):
        if other.type == 'ind-plane2D':
            return 0
        return False
        
    def findFloatValue(self, element):
        if element.type == 'ind-plane2D':
            return True
        return False
        
    @staticmethod
    def compareAngle(center, first, second):
        return (first.product(second) + center.product(first.subtract(second))).sign()
        
    @staticmethod
    def compareRelAngle(center, ref, first, second):
        a = plane2D.compareAngle(center, ref, first)
        b = plane2D.compareAngle(center, ref, second)
        if a == b:
            if a != 0:
                return plane2D.compareAngle(center, first, second)
            if (ref < center) == (second < center) and (ref < center) != (first < center):
                return 1
            if (ref < center) != (second < center) and (ref < center) == (first < center):
                return -1
            return 0
        if a == 0:
            if (ref < center) == (first < center):
                return -1
            else:
                return -1 * b
        if b == 0:
            if (ref < center) == (second < center):
                return 1
        return a

    @staticmethod
    def angleContainsPin(center, p1, pin, p2):
        a = plane2D.compareAngle(center, p1, pin)
        b = plane2D.compareAngle(center, p1, p2)
        if a == b:
            if a != 0:
                return plane2D.compareAngle(center, pin, p2)
            if (p1 < center) == (p2 < center) and (p1 < center) != (pin < center):
                return 1
            return 0
        if a == 0:
            if (p1 < center) == (pin < center):
                return 0
            else:
                return -1 * b
        if b == 0:
            if (p1 < center) == (p2 < center):
                return 1
        return a
        
    @staticmethod
    def findPin(center, start, end):
        angleComp = plane2D.compareAngle(center, start, end)
        if angleComp == -1:
            return start.midpoint(end)
        mid = start.midpoint(end)
        if angleComp == 1:
            pin = center.subtract(mid.subtract(center))
            return pin
        if mid == center:
            a = vector2D(1,0)
            pin = center.add(a)
            if plane2D.compareAngle(center, start, pin) == 0:
                a = vector2D(0,1)
                pin = center.add(a)
            if plane2D.compareAngle(center, start, pin) == 1:
                pin = center.subtract(a)
            return pin
        raise ValueError('Backend error: start == end when find pin')
        
    def parallel(self, other):
        if other.type == 'ind-plane2D':
            return True
        return False
        
class plane3D(individual):
    
    type = 'ind-plane3D'
    
    def __init__(self, sort, boundary, normal_n_origin = False, normal = None, origin = None, *args, **kwargs):
        if normal_n_origin is False:        
            super(plane3D, self).__init__(sort)
            self.normal = ((boundary[2].subtract(boundary[0])).product(
                            boundary[1].subtract(boundary[0]))).unit()
            if self.normal < vector3D(0,0,0):
                self.negNormal = True
                self.normal = self.normal.scale(cn1)
            else:
                self.negNormal = False
            self.scalar = self.normal.scalar(boundary[0])
            self.root = self.normal.scale(self.scalar)
            
        else:
            super(plane3D, self).__init__(sort)
            self.normal = normal
            if self.normal < vector3D(0,0,0):
                self.negNormal = True
                self.normal = self.normal.scale(cn1)
            else:
                self.negNormal = False
            self.scalar = self.normal.scalar(origin)
            self.root = self.normal.scale(self.scalar)
            
        if self.normal.parallel(vector3D(0,0,1)):
            self.u = vector3D(1,0,0)
            self.v = vector3D(0,1,0)
        else:
            self.u = self.normal.product(vector3D(0,0,1)).unit()
            self.v = self.normal.product(self.u).unit()
            
    def __str__(self):
        return 'normal: '+str(self.normal)+'root: '+str(self.root)
        
    def printRhino(self, string = '', displacement = 0):
        if not self.normal.x == 0:
            dir0 = self.normal.__class__(self.normal.y, -self.normal.x, 0)
            dir1 = self.normal.__class__(self.normal.z, 0, -self.normal.x)
        elif not self.normal.y == 0:
            dir0 = self.normal.__class__(self.normal.y, -self.normal.x, 0)
            dir1 = self.normal.__class__(0, self.normal.z, -self.normal.y)
        elif not self.normal.z == 0:
            dir0 = self.normal.__class__(self.normal.z, 0, -self.normal.x)
            dir1 = self.normal.__class__(0, self.normal.z, -self.normal.y)
        p = []
        p.append(dir0.scale(10).add(self.root))
        p.append(dir1.scale(10).add(self.root))
        p.append(dir0.scale(-10).add(self.root))
        p.append(dir1.scale(-10).add(self.root))
        s='_polyline '
        for j in p:
            s += str(j.x + displacement) + ',' + str(j.y) + ',' + str(j.z) + ' '
        s+= str(p[0].x + displacement) + ',' + str(p[0].y) + ',' + str(p[0].z) + '\n'+string+'   '
        s+='_SelPolyline _PlanarSrf _Delete'
        return s 
        
    def equals(self, other):
        return self.normal == other.normal and self.scalar == other.scalar
        
    def compareValue(self, other):
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        return self.scalar.compareValue(other.scalar)
                
    def isNil(self):
        return self.normal.isZero()
        
    def direction(self, coords):
        return self.u.scale(coords.x).add(self.v.scale(coords.y))
        
    def position(self, coords):
        return self.root.add(self.direction(coords))
        
    def localize(self, point):
        diff = point.subtract(self.root)
        return vector2D(self.u.scalar(diff), self.v.scalar(diff))
        
    def projection(self, point):
        return self.position(self.localize(point))
                
    def parallel(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plana may be nil')
        return self.normal == other.normal
    
    def perpendicular(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plana may be nil')
        return self.normal.perpendicular(other.normal)
            
    def parallelLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal.dotproduct(line.direction) == 0
    
    def perpendicularLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal == line.unitDir
            
    def findFloatValue(self, element):
        if element.type == 'ind-plane3D':
            return False
        if (self.normal.convertFloat() == element.normal and
            self.scale.convertFloat() == element.scale):
            return self
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        position0 = self.root.transform(affine);
        position1 = self.root.add(self.u).transform(affine)      
        position2 = self.root.add(self.v).transform(affine)      
        return plane3D(self.sort, [position0,position1,position2])
        
    def intersection(self, other):
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        if self.parallel(other):
            raise ValueError('The two planes are parallel to one another')
        prod = self.normal.product(other.normal)
        result = []
        result.append(prod.unit())
        result.append(self.root.add(prod.product(self.normal).scale(self.scalar)))
        return result #return a set of two vectors, 1st vector: direction vector,
                      #                             2nd vector: root vector

    def intersectionLine(self, line):
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        if self.parallelLine(line):
            raise ValueError('The two object are parallel to one another')
        d = (self.root.subtract(line.root).dotproduct(self.normal)) / (line.direction.dotproduct(self.normal))
        return line.direction.scale(d).add(line.root)
