# -*- coding: utf-8 -*-
"""
Created on Tue Sep 12 09:51:49 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .planeSegment import planeSegment2D
from .planeSegment3D import planeSegment3D
from ..struct.planeStruct import bSegment, boundary
from ..struct.vector import vector2D
from ..form.coForm import coForm
from ..io.parseReader import ParsingError

class planeSegmentP2D(planeSegment2D):
    type = 'ind-planeSegmentP2D'
    
    def __init__(self, sort, boundr, boundaryList = False, *args, **kwargs):
        super(planeSegmentP2D, self).__init__(sort, boundr, boundaryList)

    def codescriptor(self):
        return self
        
    @staticmethod
    def classifyBoundary(sort, boundary1, boundary2):
        form1 = coForm(sort)
        one = planeSegment2D(sort, [boundary1], True)
        form1.insertClassification(one)
        form2 = coForm(sort)
        two = planeSegment2D(sort, [boundary2], True)
        form2.insertClassification(two)
        result = planeSegment2D.classification(form1, form2)
        return result

    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a plane segment')
        if reader.newToken() != '<': #bracket
            raise ParsingError(reader,'expected a "<" ')
            
        bounds = []
        index = 0
        while reader.token() == '<':
            bounds.append(boundary())
            while True:
                reader.newToken() #bracket
                x,y = vector2D.readVector(reader) # comma
                start = vector2D(x,y) 
                    
                reader.newToken() #bracket
                
                x,y = vector2D.readVector(reader) # '>'
                end = vector2D(x,y) 
    
                bounds[index].add(bSegment(start, end))
    
                reader.newToken() # ','/'>'/'<'
                if reader.token() == ',':
                    reader.newToken()
                elif reader.token() == '>' or reader.token() == '<':
                    while reader.previewToken() == '<':
                        reader.newToken()
                    break
            
            index += 1
            
        for bdr in bounds:
            vertices = 2
            bdr.add(bdr.canonical[0].start)
            bdr.add(bdr.canonical[0].end)
            vert = bdr.canonical[0].end
            seg = bdr.canonical[0]
            while vertices < len(bdr.canonical):
                for segment in bdr.canonical:
                    if vert == segment.start and seg != segment:
                        bdr.add(segment.end)
                        vert = segment.end
                        seg = segment
                        vertices += 1
                        break
                    if vert == segment.end and seg != segment:
                        bdr.add(segment.start)
                        vert = segment.start
                        seg = segment
                        vertices += 1
                        break
        reader.newToken()

        ind = planeSegmentP2D(sort, bounds, True)
        classes = ind.classify(None)
        result = coForm(sort)
        ind.construct(classes.sumSet(), result)
        ind.boundaries = result.elements[0].boundaries
        
        
        return ind
        
        

class planeSegmentP3D(planeSegment3D):
    type = 'ind-planeSegmentP3D'
    
    def __init__(self, sort, boundr, boundaryList = False, *args, **kwargs):
        super(planeSegmentP2D, self).__init__(sort, boundr, boundaryList)

    def codescriptor(self):
        return self
        
    @staticmethod
    def classifyBoundary(sort, boundary1, boundary2):
        form1 = coForm(sort)
        one = planeSegment2D(sort, [boundary1], True)
        form1.insertClassification(one)
        form2 = coForm(sort)
        two = planeSegment2D(sort, [boundary2], True)
        form2.insertClassification(two)
        result = planeSegment2D.classification(form1, form2)
        return result

    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a plane segment')
        if reader.newToken() != '<': #bracket
            raise ParsingError(reader,'expected a "<" ')
            
        bounds = []
        index = 0
        while reader.token() == '<':
            bounds.append(boundary())
            while True:
                reader.newToken() #bracket
                x,y = vector2D.readVector(reader) # comma
                start = vector2D(x,y) 
                    
                reader.newToken() #bracket
                
                x,y = vector2D.readVector(reader) # '>'
                end = vector2D(x,y) 
    
                bounds[index].add(bSegment(start, end))
    
                reader.newToken() # ','/'>'/'<'
                if reader.token() == ',':
                    reader.newToken()
                elif reader.token() == '>' or reader.token() == '<':
                    while reader.previewToken() == '<':
                        reader.newToken()
                    break
            
            index += 1
            
        for bdr in bounds:
            vertices = 2
            bdr.add(bdr.canonical[0].start)
            bdr.add(bdr.canonical[0].end)
            vert = bdr.canonical[0].end
            seg = bdr.canonical[0]
            while vertices < len(bdr.canonical):
                for segment in bdr.canonical:
                    if vert == segment.start and seg != segment:
                        bdr.add(segment.end)
                        vert = segment.end
                        seg = segment
                        vertices += 1
                        break
                    if vert == segment.end and seg != segment:
                        bdr.add(segment.start)
                        vert = segment.start
                        seg = segment
                        vertices += 1
                        break
        reader.newToken()

        ind = planeSegmentP2D(sort, bounds, True)
        classes = ind.classify(None)
        result = coForm(sort)
        ind.construct(classes.sumSet(), result)
        ind.boundaries = result.elements[0].boundaries
        
        
        return ind
