# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 14:23:09 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .individual import individual
from copy import deepcopy
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate, c0, c1
from ..io.parseReader import parseReader, ParsingError
"""
 * A point is a 0-dimensional geometric data entity with a discrete
 * behavior.
 * 
 * The Point class extends on the Individual class.
 * It defines the characteristic individual for points.
 * A point is represented as a position vector.
 * This characteristic individual accepts no parameters.
 * It specifies a {sortal.map.similarityMap} as default.
 * Forms of points adhere to a discrete behavior.
"""

class point2D(individual):
    """
    initialize a point
    point contains: type: 'ind-point2D' and a coordinate vector e.g. (3,2)
    a point can either use vector2D or vector2Dfloat as it's coordinating vector
    vector2D: store numbers as symbolic expression e.g. sqrt(3), sin(pi/20), ...
    vector2Dfloat: store numbers as float values
    """
    
    type = 'ind-point2D'
    
    def __init__(self, sort, x, y = None, *args, **kwargs):
        super(point2D, self).__init__(sort)
        if isinstance(x, vector2D):
            self.value = x
        else:
            vector = vector2D(x,y)
            self.value = vector
                   
            
    """
    print out a string representing the value of this point,
    used in form.printE()
    """
    def printVal(self):
        return str(self.value)
        
    def printRhino(self, string = '', displacement = 0):
        return '_Point '+str(self.value.x + displacement)+','+str(self.value.y)
         
        
    """
    check equality
    """
    def equals(self, other):
        return (self.value == other.value)
        
    """
    comparing value of 2 points
    """
    def compareValue(self, other):
        return self.value.compareValue(other.value)   
        
    def replicateInd(self):
        return point2D(self.sort, self.value)    
        
    """
    transform a point with a transform set
    """
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            newPoint = self.duplicateInd()
        else:
            newPoint = self.__class__(self.sort, self.value.transform(affine))
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            newPoint = self.__class__(self.sort, transf.apply(self.value))
        return newPoint
                        

    @staticmethod
    def parse(reader, sort):
        num = c1
        inf = False
        if reader.token() == parseReader.MINUS:
            if not reader.previewToken() == parseReader.NUMBER:
                raise ParsingError(reader,'expected a number after "-"')
            num = '-' + reader.newValue() # number
            reader.newToken() # bracket
        elif reader.token() == parseReader.NUMBER: # e.g. 3(2,2,0)
            num = reader.value()
            reader.newToken() # bracket
        
        if reader.token() == '/' and reader.previewToken() == parseReader.NUMBER:
            denom = reader.newValue() # number
            if denom == 0:
                inf = True
                num = c0
            else:
                num = coordinate(num)/coordinate(denom)
            reader.newToken() # bracket
            
        if reader.token() == '(':
            x,y = vector2D.readVector(reader)
            if not isinstance(num, coordinate):
                num = coordinate(num)
            if not inf:
                x = x*num
                y = y*num
            ind = point2D(sort,vector2D(x,y,inf))
        return ind


class point3D(individual):
    """
    initialize a point
    point contains: type: 'ind-point3D' and a coordinate vector e.g. (3,2)
    a point can either use vector3D or vector3Dfloat as it's coordinating vector
    vector3D: store numbers as symbolic expression e.g. sqrt(3), sin(pi/20), ...
    vector3Dfloat: store numbers as float values
    """
    
    type = 'ind-point3D'
    
    def __init__(self, sort, x, y = None, z = None, *args, **kwargs):
        super(point3D, self).__init__(sort)
        if isinstance(x, vector3D):
            self.value = x
        else:
            vector = vector3D(x,y,z)
            self.value = vector
            
    """
    print out a string representing the value of this point,
    used in form.printE()
    """
    def printVal(self):
        return str(self.value)
        
    def printRhino(self, string = '', displacement = 0):
        return '_Point '+str(self.value.x + displacement)+','+str(self.value.y)+','+str(self.value.z)
         
    """
    check if the point's is (0,0)
    """
    def isNil(self):
        return self.value is None
        
    """
    check equality
    """
    def equals(self, other):
        return (self.value == other.value)
        
    """
    comparing value of 2 points
    """
    def compareValue(self, other):
        return self.value.compareValue(other.value)   
        
    """
    duplicating a point
    """
    def duplicateInd(self):
        duplicate = point3D(self.sort, self.value)
        if self.attribute is not None:
            for i in self.attribute:
                duplicate.addAttribute(i)
        return duplicate
        
    def replicateInd(self):
        return point3D(self.sort, self.value)    
        
    """
    transform a point with a transform set
    """
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newPoint = self.duplicateInd()
        else:
            newPoint = point3D(self.sort, self.value.transform(affine))
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            newPoint = point3D(self.sort, transf.apply(self.value))
        return newPoint
    

    @staticmethod
    def parse(reader, sort):
        num = c1
        inf = False
        if reader.token() == parseReader.MINUS:
            if not reader.previewToken() == parseReader.NUMBER:
                raise ParsingError(reader,'expected a number after "-"')
            num = '-' + reader.newValue() # number
            reader.newToken() # bracket
        elif reader.token() == parseReader.NUMBER: # e.g. 3(2,2,0)
            num = reader.value()
            reader.newToken() # bracket
        
        if reader.token() == '/' and reader.previewToken() == parseReader.NUMBER:
            denom = reader.newValue() # number
            if denom == 0:
                inf = True
                num = c0
            else:
                num = coordinate(num)/coordinate(denom)
            reader.newToken() # bracket
            
        if reader.token() == '(':
            x,y,z = vector3D.readVector(reader)
            if not isinstance(num, coordinate):
                num = coordinate(num)
            if not inf:
                x = x*num
                y = y*num
                z = z*num
            ind = point3D(sort,vector3D(x,y,z,inf))
        return ind
