# -*- coding: utf-8 -*-
"""
Created on Tue Feb 14 15:12:02 2017

@author: akibdpt
"""
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

from builtins import str

from os import path

from .parseReader import parseReader, ParsingError
from ..sort.sort import sort as sortClass
from ..sort.primitiveSort import primitiveSort
from ..sort.attributeSort import attributeSort
from ..sort.disjunctiveSort import disjunctiveSort

from ..grammar.rules.rule import rule
from ..grammar.rules.ruleSequence import ruleSequence
from ..grammar.rules.ruleDisjunction import ruleDisjunction
from ..grammar.rules.ruleDisjunctionP import ruleDisjunctionP
from ..grammar.rules.ruleNegation import ruleNegation
from ..grammar.rules.flow import flow

from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate
from ..struct.key import key

from ..ind.description import description as desc


class sdlParser():
    
    unnamedSortCount = 0
    
    sorts = {}
    forms = {}
    rules = {}
    flows = {}
    characteristics = primitiveSort.characteristicList
                 
    @staticmethod
    def parse(filePath):
            
        
        file = open(filePath, 'r')
        f = []
        cmdLine = ''
        numLines = 0
        for line in file:
            numLines += 1
            if line[0] == '#':
                continue
            
            cmdLine += line
            if len(line)>1 and ';' in line:
                cmdLine = cmdLine.split(';')[0]+';'
                f.append([cmdLine, numLines])
                numLines = 0
                cmdLine = ''
                
        totalLines = 0
        for i in f:
            totalLines += i[1]
            r = parseReader(i[0], totalLines)
            while r.token() != ';':
                if not r.token() == parseReader.IDENTIFIER:
                    r.newToken()
                    continue
                if r.value() == 'sort':
                    sname, sort = sdlParser.readSort(r)
                    sdlParser.sorts[sname] = sort
                elif r.value() == 'rule':
                    rname, rule = sdlParser.readRule(r)
                    sdlParser.rules[rname] = rule
                elif r.value() == 'flow':
                    rname, flow = sdlParser.readFlow(r)
                    sdlParser.flows[rname] = flow
                elif r.value() == 'form':
                    fname, form = sdlParser.readForm(r)
                    sdlParser.forms[fname] = form
                    
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#<============================================================================>
#<================================= Reading ==================================>
#<============================================================================>
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    #<========================================================================>
    #<================================= Sort =================================>
    #<========================================================================>
    @staticmethod
    def readSort(reader):
        # sort sortName: sortContent
        # read til 'sortName:' --> readsorts
        if reader.token() == '(':
            pass 
        reader.newToken()
        if reader.token() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as sort name')
        sortName = reader.value()
        if reader.newToken() != ':':
            raise ParsingError(reader, '":" expected after sort name')
        
        if sortClass.checkSort(sortName) == True:
            sort = sortClass.register[sortName]
            sdlParser.sorts[sortName] = sort
            reader.newToken()
            while True:
                if reader.newToken() == ';':
                    break
        else:
            sort = sdlParser.readSortStruct(reader, sortName)
            sdlParser.sorts[sortName] = sort
        return sortName, sort
    
    @staticmethod
    def readSortStruct(reader, sortName):
        compSorts = []
        # read the first sort in the structure
        # if sort component is a characteristic: [lineSegment2D]
        if reader.newToken() == '[':
            if reader.trailerToken(3) == ';':
                compSorts.append(sdlParser.readPrimitive(reader, sortName))
            else:
                sdlParser.unnamedSortCount += 1
                compSorts.append(sdlParser.readPrimitive(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount)))
        # if open bracket
        elif reader.token() == '(':
            if reader.trailerToken() == ':':
                sName, s = sdlParser.readSort(reader)
            else:
                sdlParser.unnamedSortCount += 1
                s = sdlParser.readSortStruct(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
            compSorts.append(s)
        # if identifier --> existing sort
        elif reader.token() == parseReader.IDENTIFIER:
            compSorts.append(sdlParser.sorts[reader.value()])
        # if the line is not ended yet /
        while reader.newToken() != ';' and reader.token() != ')':
            if reader.token() == '^':
                upperBase = None
                oriSort = compSorts.pop()
                base = oriSort

                while base.type == 'sort-attribute':
                    upperBase = base
                    base = base.weight

                """
                if reader.previewToken() == '[':
                    reader.newToken()
                    sdlParser.unnamedSortCount += 1
                    weight = sdlParser.readPrimitive(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
                el"""
                if reader.newToken() == '(':
                    if reader.trailerToken() == ':':
                        wName, weight = sdlParser.readSort(reader)
                    else:
                        sdlParser.unnamedSortCount += 1
                        weight = sdlParser.readSortStruct(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
                else:
                    weight = sdlParser.sorts[reader.value()]
                    
                if upperBase is None:
                    compSorts.append(attributeSort(sortName, base, weight))
                else:
                    upperBase.weight = attributeSort(sortName, base, weight)
                    compSorts.append(oriSort)
            elif reader.token() == '+':
                if reader.newToken() == '(':
                    if reader.trailerToken() == ':':
                        sName, s = sdlParser.readSort(reader)
                    else:
                        sdlParser.unnamedSortCount += 1
                        s = sdlParser.readSortStruct(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
                    compSorts.append(s)
                else:
                    try:
                        compSorts.append(sdlParser.sorts[reader.value()])
                    except KeyError:
                        raise ParsingError(reader, 'sort with name '+reader.value()+' does not exist')
        if len(compSorts) > 1:
            sort = disjunctiveSort(sortName, compSorts)
        elif len(compSorts) == 1:
            sort = compSorts[0]
        return sort
                 
    @staticmethod
    def readPrimitive(reader, sortName):
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as sort characteristic i.e. point, label, description, ...')
        chara = reader.value().lower()
        if reader.newToken() != ']':
            raise ParsingError(reader, '"]" expected after sort characteristic')
        arg = None
        if reader.previewToken() == '(':
            reader.newToken()
            reader.newToken()
            arg = reader.text[reader.index():-2]
            while reader.newToken() != ')':
                pass
        sort = primitiveSort(sortName, sdlParser.characteristics[chara], arg)
        return sort
        
    #<========================================================================>
    #<================================= Ind ==================================>
    #<========================================================================>
    @staticmethod
    def readInd(reader, sort):
        keyObj = None
        if reader.token() == '#':
            reader.newToken()
            keyValue = reader.value()
            try:
                base, offset = keyValue.split('-')
            except ValueError:
                raise ParsingError(reader, 'Key must be of structure: #identifier-number')
            keyObj = key(base, offset)
            reader.newToken()
        ind = sort.baseSort().characteristic.parse(reader, sort)
        if keyObj is not None:
            keyObj.ind = ind
            ind.key = keyObj
        return ind
        
    @staticmethod
    def readVector(reader):
        if reader.token() != '(':
            raise ParsingError(reader, '"(" expected')
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        x = coordinate(neg+reader.value())
        reader.newToken() # comma
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        y = coordinate(neg+reader.value())
        
        if reader.newToken() == ')':
            return vector2D(x,y)
            
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        z = coordinate(neg+reader.value())
        
        reader.newToken() # bracket

        reader.newToken() # next token
        return vector3D(x,y,z)
       
    #<========================================================================>
    #<================================= Form =================================>
    #<========================================================================>
    @staticmethod
    def readForm(reader):
        # current token being 'form'
        # read form name (for identification)
        if reader.newToken() != '$':
            raise ParsingError(reader, 'Expected a "$" before the form name')
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier after "$" as form name')
        formName = reader.value()
        
        # $formName = sortName: {form elements}
        if reader.newToken() != '=':
            raise ParsingError(reader, 'Expected "=" after form name')
        
        # read elements in form
        reader.newToken()
        form = sdlParser.readFormStruct(reader)
        
        # register form in form list
        sdlParser.forms[formName] = form
        return formName, form

    @staticmethod
    def readFormStruct(reader):
        # get the sort of the form
        try:
            sort = sdlParser.sorts[reader.value()]
        except:
            raise ParsingError(reader, 'Expected existing sort name')
        if reader.newToken() != ':':
            raise ParsingError(reader, '":" expected after sort name')
        # if next token is '{'
            # sortName: {form content} (creating new form)
        if reader.newToken() == '{':
            # initialize new form with the sort
            form = sort.newForm()
            # if sort is primitive --> form only contains one type of individual
            if sort.type == 'sort-primitive':
                # structure: { individual1 , individual2 , individual3 , ... }
                reader.newToken()
                while reader.token() != '}':
                    form.insert(sdlParser.readInd(reader, sort))
                    if reader.token() == ',':
                        reader.newToken()
                    elif reader.token() != '}':
                        raise ParsingError(reader, '"," or "}" expected after an individual')
            # if sort is attribute --> form containing one type of individual that has certain 
            #                          other individual as attribute
            elif sort.type == 'sort-attribute':
                reader.newToken()
                while reader.token() != '}':
                    base = sdlParser.readInd(reader, sort)
                    if reader.token() == '{':
                        reader.newToken()
                        while True:
                            if sort.weight.type == 'sort-disjunctive':
                                w = sdlParser.readFormStruct(reader)
                            else:
                                w = sdlParser.readInd(reader, sort.weight)
                            base.addAttribute(w)
                            if reader.token() == ',':
                                reader.newToken()
                            elif reader.token() == '}':
                                break
                        reader.newToken()
                    form.insert(base)
                    if reader.token() == ',':
                        reader.newToken()
                    elif reader.token() != '}':
                        raise ParsingError(reader, '"," or "}" expected after an individual')
            # if sort is disjunctive --> form containing other forms that 
            #                            contain different type of individuals
            elif sort.type == 'sort-disjunctive':
                reader.newToken()
                form = sort.newForm()
                check = True
                if reader.token() == '}':
                    check = False
                while check:
                    form.insert(sdlParser.readFormStruct(reader))
                    if reader.token() == ',':
                        reader.newToken()
                    elif reader.token() == '}':
                        break
                    else:
                        raise ParsingError(reader, '"," or "}" expected after an individual')
                
        # if next token is '$': referencing to existing form
            # sortName: $formName|ruleName (applying rule on form)
        elif reader.token() == '$':
            # get formName for reference
            if reader.newToken() != parseReader.IDENTIFIER:
                raise ParsingError(reader, 'Expected an identifier after "$" as form name')
            # get the form referenced
            try:
                form = sdlParser.forms[reader.value()].duplicateForm()
            except:
                raise ParsingError(reader, 'Expected existing form name after $')
            # if '|' then applying rule
            if reader.previewToken() == '|':
                while reader.previewToken() == '|':
                    # get the rule referenced
                    reader.newToken()
                    reader.newToken()
                    rule = sdlParser.rules[reader.value()]
                    print('\n======================================================\n___lhs:')
                    rule.lhs.printE('<<<')
                    #print(rule.lhs.elements[0].printVal())
                    #print(rule.rhs.elements[0].printVal())
                    print('apply on')
                    form.printE('...')
                    print('___rhs:')
                    rule.rhs.printE('>>>')
    
                    # apply rule on form
                    rule.apply(form)
                    print('result:')
                    form.printE('===')
            else:
                reader.newToken() ##################
        reader.newToken()
        return form
        
    #<========================================================================>
    #<================================= Rule =================================>
    #<========================================================================>
    @staticmethod
    def readRule(reader):
        # current token being 'rule'
        # read rule Name
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as rule name')
        ruleName = reader.value()
        
        # read rule description
        if reader.newToken() != parseReader.STRING:
            raise ParsingError(reader, 'string as description expected after sort name')
        ruleDesc = reader.value()[1:-1] # read description, without 1st and last characters (two doublequotes)
        
        reader.newToken()
            
        # read lhs
        if reader.token() != '<' or reader.newToken() != '<':
            raise ParsingError(reader, 'Expected "<<" to annotate left hand side of the rule')
        reader.newToken()

        try:
            lhs = sdlParser.readFormStruct(reader)
        except UnboundLocalError:
            raise ParsingError(reader,'Form structure is not complete')
        except ValueError as err:
            raise ParsingError(reader,'Error in reading, original error message: '+ err.args[0])
        
        # read predicates (only for graph rule)
        predicates = []
        if reader.token() == ':' and reader.newToken() == ':':
            comma = True
            while comma:
                reader.newToken()
                cond = reader.value()
                reader.newToken()
                if reader.newToken() != '{':
                    raise ParsingError(reader, '"{" expected')
                objList = []
                while reader.token() != '}':
                    reader.newToken()
                    prop = []
                    if reader.token() == '(':
                        obj = sdlParser.readVector(reader)
                    else:
                        if reader.token() == '>':
                            prop.append(0)
                            reader.newToken()
                        try:
                            base, offset = reader.value().split('-')
                        except ValueError:
                            raise ParsingError(reader, 'Key must be of structure: #identifier-number')
                        if reader.previewToken() == '<':
                            prop.append(1)
                            reader.newToken()
                        obj = key.retrieveKey(base, offset)
                        obj.addProperty(prop)
                    if obj is not None:
                        objList.append(obj)
                    if reader.newToken() != ',' and reader.token() !='}':
                        raise ParsingError(reader, '"," or "}" expected')
                predicates.append([cond, objList])
                if reader.newToken() !=',':
                    comma = False

        # read rhs
        if reader.token() != '>' or reader.newToken() != '>':
            raise ParsingError(reader, 'Expected ">>" to annotate right hand side of the rule')
        reader.newToken()
        
        try:
            rhs = sdlParser.readFormStruct(reader)
        except UnboundLocalError:
            raise ParsingError(reader,'Form structure is not complete')
        except ValueError as err:
            raise ParsingError(reader,'Error in reading, original error message: '+ err.args[0])
            
        # create rule
        newRule = rule(ruleName, ruleDesc, lhs, rhs)

        # read predicates (only for graph rule)
        directives = []
        if reader.token() == ':' and reader.newToken() == ':':
            comma = True
            while comma:
                reader.newToken()
                cond = reader.value()
                reader.newToken()
                
                if reader.newToken() != '{':
                    raise ParsingError(reader, '"{" expected')
                objList = []
                while reader.token() != '}':
                    reader.newToken()
                    if reader.token() != '(':
                        raise ParsingError(reader, '"(" expected')
                    obj = []
                    keyObj = []
                    while reader.previewToken() == 'a':
                        try:
                            base, offset = reader.newValue().split('-')
                            keyObj.append(key.retrieveKey(base, offset))
                            reader.newToken()
                        except ValueError:
                            raise ParsingError(reader, 'Key must be of structure: #identifier-number')
                    reader.newToken()
                    while reader.token() == parseReader.NUMBER or reader.token() == '`':
                        if reader.token() == parseReader.NUMBER:
                            obj.append(coordinate(reader.value()))
                            reader.newToken()
                        else:
                            NilDescSort = primitiveSort.getSort('__NIL_DESC__')
                            if NilDescSort is None:
                                NilDescSort = primitiveSort('__NIL_DESC__', desc)
                            obj.append(desc.parse(reader, NilDescSort))
                            """
                            d = desc.parse(reader, NilDescSort)
                            val = (d.evaluateInd(newRule)).getValue()[1:-1]
                            obj.append(coordinate(val))
                            """
                        if reader.token() != ',' and reader.token() !=')':
                            raise ParsingError(reader, '"," or ")" expected')
                        reader.newToken()
                    if reader.token() != ',' and reader.token() !='}':
                        raise ParsingError(reader, '"," or "}" expected')
                    for j in keyObj:
                        obj.append(j)
                    objList.append(obj)
                directives.append([cond, objList])
                if reader.newToken() !=',':
                    comma = False
            
            
        
        for i in predicates:
            newRule.addPredicate(i[0],i[1])
        for i in directives:
            newRule.addDirective(i[0],i[1])

        # register rule in rule list
        sdlParser.rules[ruleName] = newRule
        return ruleName, newRule
        
    @staticmethod
    def readFlow(reader):
        # current token being 'rule'
        # read rule Name
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as flow name')
        flowName = reader.value()
        
        # read rule description
        if reader.newToken() != parseReader.STRING:
            raise ParsingError(reader, 'string as description expected after flow name')
        flowDesc = reader.value()[1:-1] # read description, without 1st and last characters (two doublequotes)
        
        # read rule description
        if reader.newToken() != parseReader.STRING:
            raise ParsingError(reader, 'string as flow content expected after flow description')
        flowContent = reader.value()[1:-1] # read description, without 1st and last characters (two doublequotes)
        reader.newToken()        
        
        flowReader = parseReader(flowContent)

        r = sdlParser.readFlowRules(flowReader, starting = True)
        
        flowObj = flow(flowName, flowDesc, r)
        return flowName, flowObj
        
    @staticmethod
    def readFlowRules(reader, thisNegation = False, starting = False):
        rules = []
        paren = '('
        if not starting:
            paren = reader.token()
            if reader.newToken() == '*':
                paren += '*'
                reader.newToken()
        while reader.token() is not None and reader.token() not in (')', ']'):
            negation = False
            if reader.token() == '!':
                negation = True
                reader.newToken()
            if reader.token() in ('[','('):
                rules.append(sdlParser.readFlowRules(reader, negation))
            else:
                ruleName = reader.value()
                try:
                    singleRule = sdlParser.rules[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist before referencing')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = sdlParser.readQuantifier(reader)
                    singleRule = ruleSequence([singleRule], quantifier, matching)
                if negation:
                    singleRule = ruleNegation(singleRule)
                rules.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = sdlParser.readQuantifier(reader)
            
            
        if paren == '(' or len(rules) == 1:
            r = ruleSequence(rules, quantifier, matching)
        elif paren == '[*':
            r = ruleDisjunctionP(rules, quantifier, matching)
        elif paren == '[':
            r = ruleDisjunction(rules, quantifier, matching)
        if thisNegation:
            r= ruleNegation(r)
        if matching == 'PO':
            r.setPo()
        return r
        
    """
    @staticmethod
    def readRuleSequence(reader, starting = False):
        rules = []
        if not starting:
            reader.newToken()
        while reader.token() is not None and reader.token() != ')':
            if reader.token() == '[':
                rules.append(sdlParser.readRuleDisjunction(reader))
            elif reader.token() == '<':
                rules.append(sdlParser.readRuleDisjunctionP(reader))
            elif reader.token() == '(':
                rules.append(sdlParser.readRuleSequence(reader))
            else:
                ruleName = reader.value()
                try:
                    singleRule = sdlParser.rules[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist before referencing')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = sdlParser.readQuantifier(reader)
                    rules.append(ruleSequence([singleRule], quantifier, matching))
                else:
                    rules.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = sdlParser.readQuantifier(reader)
        r = ruleSequence(rules, quantifier, matching)
        return r
    
    @staticmethod
    def readRuleDisjunction(reader):
        rules = []
        reader.newToken()
        while reader.token() != ']':
            if reader.token() == '[':
                rules.append(sdlParser.readRuleDisjunction(reader))
            elif reader.token() == '<':
                rules.append(sdlParser.readRuleDisjunctionP(reader))
            elif reader.token() == '(':
                rules.append(sdlParser.readRuleSequence(reader))
            else:
                ruleName = reader.value()
                try:
                    singleRule = sdlParser.rules[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist before referencing')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = sdlParser.readQuantifier(reader)
                    rules.append(ruleSequence([singleRule], quantifier, matching))
                else:
                    rules.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = sdlParser.readQuantifier(reader)
        r = ruleDisjunction(rules, quantifier, matching)
    
        return r
        
    @staticmethod
    def readRuleDisjunctionP(reader):
        rules = []
        reader.newToken()
        while reader.token() != '>':
            if reader.token() == '[':
                rules.append(sdlParser.readRuleDisjunction(reader))
            elif reader.token() == '<':
                rules.append(sdlParser.readRuleDisjunctionP(reader))
            elif reader.token() == '(':
                rules.append(sdlParser.readRuleSequence(reader))
            else:                    
                ruleName = reader.value()
                try:
                    singleRule = sdlParser.rules[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist before referencing')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = sdlParser.readQuantifier(reader)
                    rules.append(ruleSequence([singleRule], quantifier, matching))
                else:
                    rules.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = sdlParser.readQuantifier(reader)
        r = ruleDisjunction(rules, quantifier, matching)
    
        return r
    """
    
    @staticmethod
    def readQuantifier(reader):
        quantifier = []
        matching = 'G'
        if reader.newToken() == '{':
            quantifier = []
            if reader.newToken() != parseReader.NUMBER:
                raise ParsingError(reader,'Number expected in the set bracket as quantifier for flow sequence')
            quantifier.append(int(reader.value()))
            if reader.newToken() == '}':
                quantifier.append(quantifier[0])
            else:
                if reader.token() != ',':
                    raise ParsingError(reader,'Number expected in the set bracket as quantifier for flow sequence')
                if reader.newToken() != '}':
                    if reader.token() != parseReader.NUMBER:
                        raise ParsingError(reader,'Number expected in the set bracket as quantifier for flow sequence')
                    quantifier.append(int(reader.value()))
                    if reader.newToken() != '}':
                        raise ParsingError(reader,'"}" expected')
                else:
                    quantifier.append(None)
        elif reader.token() == '?':
            quantifier = [0,1]
        elif reader.token() == '*':
            quantifier = [0,None]
        elif reader.token() == '+':
            quantifier = [1,None]

        if reader.previewToken() is not None and reader.previewToken() in ['?', '+', '*']:
            if reader.newToken() == '?':
                matching = 'L'
            elif reader.token() == '+':
                matching = 'PO'
            elif reader.token() == '*':
                matching = 'PR'
        
        return quantifier, matching
    
                    
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#<============================================================================>
#<================================= Writing ==================================>
#<============================================================================>
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
                    
                    
    #<========================================================================>
    #<================================= Sort =================================>
    #<========================================================================>
    @staticmethod
    def writeSort(sort, sortsWritten, file):
        sortsWritten.append(sort.name)
        if sort.type == 'sort-primitive':
            file.write('sort '+sort.name+' : '+sort.description+';\n')
        elif sort.type == 'sort-attribute':
            if sort.weight.name not in sortsWritten:
                sdlParser.writeSort(sort.weight, sortsWritten, file)
            if sort.base.name not in sortsWritten:
                sdlParser.writeSort(sort.base, sortsWritten, file)
            file.write('sort '+sort.name+' : '+sort.base.name + sort.separator + sort.weight.name +';\n')
        elif sort.type == 'sort-disjunctive':
            for s in sort.component:
                if s.name not in sortsWritten:
                    sdlParser.writeSort(s, sortsWritten, file)
            file.write('sort '+sort.name+' : ') #+sort.description+';\n'
            writeSeparator = False
            for i in sort.component:
                if writeSeparator:
                    file.write(sort.separator)
                else:
                    writeSeparator = True
                file.write(i.name)
            file.write(';\n')


    #<========================================================================>
    #<================================= Ind ==================================>
    #<========================================================================>
                    
                    
    @staticmethod
    def writeOneInd(ind, file):
        if ind.key is not None:
            file.write('#'+str(ind.key)+' ')
        file.write(str(ind))
        if ind.hasAttribute():
            sdlParser.writeAttributeForm(ind.attribute, file)
        
    #<========================================================================>
    #<================================= Form =================================>
    #<========================================================================>
    
    @staticmethod
    def writeOneForm(form, file):
        if form.type != 'form-coForm' or form.sort.baseSort().characteristic.type != 'ind-planeSegment3D':
            file.write(form.sort.name + ': {')            
        writeComma = False
        for element in form:
            if writeComma:
                file.write(', ')
            else:
                writeComma = True
            
            if 'form' in element.type:
                sdlParser.writeOneForm(element, file)
            else:
                sdlParser.writeOneInd(element, file)
        if form.type != 'form-coForm' or form.sort.baseSort().characteristic.type != 'ind-planeSegment3D':
            file.write('}')
            
        
    @staticmethod
    def writeAttributeForm(form, file):
        file.write(' {')
        
        writeComma = False
        for element in form:
            if writeComma:
                file.write(', ')
            else:
                writeComma = True
            
            index = 0
            if 'form' in element.type:
                if index > 0:
                    file.write(', ')
                file.write(element.sort.name + ':')            
                sdlParser.writeAttributeForm(element, file)
                #file.write('}')
            else:
                sdlParser.writeOneInd(element, file)
        
        
        """
        writeComma = False
        for element in form:
            if writeComma:
                file.write(', ')
            else:
                writeComma = True
            if 'form' in element.type:
                file.write(element.sort.name + ': ')
                sdlParser.writeAttributeForm(element, file)
            else:
                sdlParser.writeOneInd(element, file)
                """
        file.write('}')
        
    #<========================================================================>
    #<================================= Rule =================================>
    #<========================================================================>
    @staticmethod
    def writeOneRule(rule, file):
        s = ''
        file.write('\nrule ' + rule.name + ' "' + rule.description + '" '+ s +'\n')
        file.write('<< ')
        
        sdlParser.writeOneForm(rule.lhs, file)
        
        checkPredicates = False
        for i in rule.detection.analysis.maps:
            if i.type == 'map-graphMap':
                if len(rule.detection.analysis.maps[i].predicates) != 0:
                    gMap = rule.detection.analysis.maps[i]
                    checkPredicates = True

        if checkPredicates:
            file.write('\n:: ')
            writecomma = False
            for pred in gMap.predicates:
                if writecomma:
                    file.write(', ')
                else:
                    writecomma = True
                file.write(pred+': {'+str(gMap.predicates[pred][0]))
                if len(gMap.predicates[pred]) > 1:
                    for keyObj in range(1,len(gMap.predicates[pred])):
                        file.write(', '+str(gMap.predicates[pred][keyObj]))
                file.write('}')
        
        file.write('\n>> ')
        
        sdlParser.writeOneForm(rule.rhs, file)
        
        checkDirectives = False
        for i in rule.detection.analysis.maps:
            if i.type == 'map-graphMap':
                if len(rule.detection.analysis.maps[i].directives) != 0:
                    gMap = rule.detection.analysis.maps[i]
                    checkDirectives = True

        if checkDirectives:
            file.write('\n:: ')
            writecomma = False
            for pred in gMap.directives:
                if writecomma:
                    file.write(', ')
                else:
                    writecomma = True
                file.write(pred+': {'+str(gMap.directives[pred][0]))
                if len(gMap.directives[pred]) > 1:
                    for keyObj in range(1,len(gMap.directives[pred])):
                        file.write(', '+str(gMap.directives[pred][keyObj]))
                file.write('}')
        
        
        file.write(';\n\n')
                    
    @staticmethod
    def writeSeletedRules(rules, destination = path.abspath(path.curdir) +'\\rules.sdl', append = False):
        if append:
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        rules = sorted(rules)
        sortsWritten = []
        for r in rules:
            if r.lhs.sort.name not in sortsWritten:
                sdlParser.writeSort(r.lhs.sort, sortsWritten, file)
                file.write('\n')
            sdlParser.writeOneRule(r, file)
            
        file.close()

    #<========================================================================>
    #<================================ Grammar ===============================>
    #<========================================================================>
    @staticmethod
    def writeGrammar(rules, initialForms, destination = path.abspath(path.curdir) +'\\grammar.sdl', append = False):
        count = 0
        if append:
            file = open(destination, 'r')
            for line in file:
                for word in line.split('$initForm'):
                    count += 1
                count -= 1
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        rules = sorted(rules)
        sortsWritten = []
        for r in rules:
            if r.lhs.sort.name not in sortsWritten:
                sdlParser.writeSort(r.lhs.sort, sortsWritten, file)
                file.write('\n')
            sdlParser.writeOneRule(r, file)
        filename = path.splitext(destination)[0].split('\\')[-1].split('/')[-1] +'_initForm'
        for f in initialForms:
            file.write('\nform $'+filename+str(count)+' = ')
            sdlParser.writeOneForm(f, file)
            file.write(';')
            count+=1
        file.close()


    @staticmethod
    def writeDerivation(rules, initialForm, ruleApplications, destination = path.abspath(path.curdir) +'\\grammar.sdl', append = False):
        count = 0
        if append:
            file = open(destination, 'r')
            for line in file:
                for word in line.split('$initForm'):
                    count += 1
                count -= 1
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        rules = sorted(rules)
        sortsWritten = []
        for r in rules:
            if r.lhs.sort.name not in sortsWritten:
                sdlParser.writeSort(r.lhs.sort, sortsWritten, file)
                file.write('\n')
            sdlParser.writeOneRule(r, file)
            
        filename = path.splitext(destination)[0].split('\\')[-1].split('/')[-1] +'_initForm'
        
        file.write('\nform $'+filename+str(count)+' = ')
        sdlParser.writeOneForm(initialForm, file)
        file.write(';')
        
        count += 1
        for i in ruleApplications:
            file.write('\nform $'+filename+str(count)+' = ')
            file.write(initialForm.sort.name+': $initForm'+str(count-1)+'|'+i+';')
            count += 1
        file.close()
        
    @staticmethod
    def writeSorts(sorts, destination = path.abspath(path.curdir) +'\\sorts.sdl', append = False):
        sortsWritten = []
        if append:
            file = open(destination, 'r')
            for line in file:
                reader = parseReader(line)
                if reader.value() == 'sort':
                    sname = sdlParser.readSort(reader)[0]
                    sortsWritten.append(sname)
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        for sort in sorts:
            if sort.name not in sortsWritten:
                sdlParser.writeSort(sort, sortsWritten, file)
        file.close()
