# -*- coding: utf-8 -*-
"""
Created on Fri Jun 23 11:43:16 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

import itertools
from copy import copy
from random import randint

from .mapping import mapping
from .graphNode import graphNode
from .paramSet import paramSet

from ..transform.graphTransform import graphTransform

from ..ind.description import description
from ..ind.line import line2D
from ..ind.planeSegment import planeSegment2D

from ..struct.planeStruct import bSegment, boundary
from ..struct.coordinate import coordinate, c0, c1, pi, cos, sin
from ..struct.vector import vector2D


ANY = 0
PERPENDICULAR = 1
PARALLEL = 2

OUTSIDE = 0
BOUND = 1
INSIDE = 2
STRICT_BOUND = 3

NONE = '[None]'
"""
ANY = '~'
PARALLEL = '='
PERPENDICULAR = '+'

OUTSIDE = '0'
BOUND = 'b'
INSIDE = 'i'
STRICT_BOUND = 's'

NONE = '[--------]'
"""

class graphMap(mapping):
    type = 'map-graphMap'    
    
    def __init__(self):
        self.nodeList = graphNode()
        self.edgeMatrices = []
        self.intersections = []
        self.matchingNodeList = graphNode()
        self.matchingEdgeMatrices = []
        self.matchingIntersections = []
        self.postAnalysisProcess = True
        
        self.predicates = {}
        self.directives = {}
        self.postCheck = False

    def clear(self):
        self.matchingNodeList.clear()
        self.matchingEdgeMatrices = []
        self.matchingIntersections = []
        
    @staticmethod
    def getInstance():
        return graphMap()
        
    def addPredicate(self, cond, keyList):
        cond = cond.lower()
        if cond in self.predicates:
            self.predicates[cond] += keyList
            return
        self.predicates[cond] = keyList
        
    def addDirective(self, cond, keyList):
        cond = cond.lower()
        if cond in self.directives:
            self.directives[cond] += keyList
            return
        self.directives[cond] = keyList
    
    def generateNodes(self, ind, matching = False):
        if matching:
            nodeList = self.matchingNodeList
            edgeMatrices = self.matchingEdgeMatrices
            intersections = self.matchingIntersections
        else:
            nodeList = self.nodeList
            edgeMatrices = self.edgeMatrices
            intersections = self.intersections
            

        if ind.type == 'ind-pointP2D':
            nodeList.addPoint(ind)
            return nodeList
        if ind.type == 'ind-planeSegmentP2D':
            nodeList.addPlane(ind)
            return nodeList
            
        element = ind.parametricNode()
        index = nodeList.addInstance(element, ind)
        if index is True:
            edgeMatrices.append([None]*len(nodeList.lines))
            
            for j in range(len(nodeList.lines)-1):
                comparand = nodeList.lines[j].nodeElement
                edge = ANY
                if comparand.parallel(element):
                    edge = PARALLEL
                elif comparand.perpendicular(element):
                    edge = PERPENDICULAR
                    
                
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                if edge != PARALLEL:
                    intersection = comparand.intersection(element)
                    newIntersection = graphNode.intersection(intersection)
                    nodeList.lines[-1].addIntersection(newIntersection)
                    nodeList.lines[j].addIntersection(newIntersection)
                    intersections.append(newIntersection)
                    for k in nodeList.lines[-1].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            newIntersection.iType = 'i'
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                    
                edgeMatrices[j].append([edge, positionOne])
                edgeMatrices[-1][j] = [edge, positionTwo]
                
        else:
            for j in range(len(nodeList.lines)):
                if j == index:
                    continue
                if edgeMatrices[j][index][0] != PARALLEL:
                    
                    positionOne = OUTSIDE
                    positionTwo = OUTSIDE
                    comparand = nodeList.lines[j].nodeElement
                    intersection = comparand.intersection(element)
                    for k in nodeList.lines[index].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            break
                    
                    edgeMatrices[j][index][1] = positionOne
                    edgeMatrices[index][j][1] = positionTwo
        return nodeList
        
    
    def map(self, fromInd, toInd):
        self.generateNodes(toInd, matching = True)
        return True
        
        
    def distinguish(self, ind):
        node = self.generateNodes(ind)
        return node
        
    def potentialTransforms(self):
        """
        ### printing
        print('\nlhs Matrix:')
        for j in self.edgeMatrices:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        
        print('\nMatching Matrix:')
        for j in self.matchingEdgeMatrices:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print()
        ###
        """
            
        # Extra analysis processes:
        # _ addEndPoint: go through the lists and add all the end points of lines 
        #   that are not intersections into nodeList
        # _ classifyIntersections: classifying the relationship between all the
        #   intersections/end points on a line as connected/disconnected
        # _ organizeForm: order the line segments of a node and combine+separate
        #   them according to their attributes
        if self.postAnalysisProcess:
            self.postAnalysisProcess = False
            self.nodeList.classifyObjs(True)
            for node in self.nodeList:
                node.addEndPoints()
                #node.classifyIntersections()
                node.organizeForm()
        self.matchingNodeList.classifyObjs()
        for matchingNode in self.matchingNodeList:
            matchingNode.addEndPoints()
            #matchingNode.classifyIntersections()
            matchingNode.organizeForm()
        """
        ### printing
        print('\nGraph nodes:')
        for i in self.nodeList:
            print(i)
            for j in i.intersections:
                print('   ',j)
        ###
        """

        transforms = []
        matchLists = []
        # detect transform --> find all possible transformation matches for nodes
        # e.g. node 0 in lhs is matched with node 4 in the shape
        #      node 1 in lhs is matched with node 2 in the shape
        #      node 2 in lhs is matched with node 0 in the shape
        #      node 3 in lhs is matched with node 1 in the shape
        #      --> match = [4,2,1,0]
        #      --> matchLists = list of all possible matches
        self.detectTransform(matchLists)
        
        noInt = True
        # analyzing each match:
        ptransf = [graphTransform()]
        
        if not self.resolveFreePoints(ptransf):
            return []
        
        for match in matchLists:
            #print('\n-----------------------------------------------------------')
            #print('*match:', match)
            transf = [] # initialize a graph transform object for this particular match
            for tr in ptransf:
                transf.append(tr.duplicate())
                
            check = True              # check: by the end of this process, if check is still True --> valid transformation, if False --> invalid
            
            # for each node in lhs:
            singleIntNodes = []
            vectChecks = []
            
            
            for nodeIndex in range(len(self.nodeList)):
                
                nodeI = self.nodeList.lines[nodeIndex]
                matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
                                
                # find intersections transformation between the nodes and add it into the graph transform
                # e.g. node i in lhs is matched with node i0 in shape
                #      node j in lhs is matched with node j0 in shape
                #      node i  intersects j  at point vectFrom
                #      node i0 intersects j0 at point vectTo
                #      --> vectFrom is matched with vectTo
                for j in range(nodeIndex+1, len(self.nodeList)):
                    if self.matchingEdgeMatrices[match[nodeIndex]][match[j]][0] != PARALLEL:
                        noInt = False
                        vectFrom = self.getIntersection(nodeI,self.nodeList.lines[j]).value
                        vectTo = self.getIntersection(matchNodeI,self.matchingNodeList.lines[match[j]]).value
                        # the intersection match cannot be added to the graph transform
                        # (due to vectFrom already have a different match from VectTo, etc...),
                        # the graph transform is deemed invalid
                        ind = 0
                        while ind < len(transf):
                            tr = transf[ind]
                            if not tr.add(vectFrom, vectTo):
                                transf.pop(ind)
                                ind -=1
                            ind += 1
                                                   
                if noInt:
                    break
                if len(transf) == 0 or not self.orderingCheck(transf, nodeI):
                    check = False
                    break
                    
                
                # if the node doesnt have any end points --> all points on line have a match --> continue onto other ndoe
                if (not nodeI.endP) and len(nodeI.linePoints) == 0:
                    continue
                
                # check if the node and matchingNode are matched in reverse
                # i.e.  node:           --A---B------C----
                #       matchingNode:   ----X--Y--Z----W------
                #       A matches with W, B matches with Z, C matches with X
                #       --> the two nodes match in reverse order
                startI = 0
                while nodeI.intersections[startI].endP:
                    startI += 1
                start = transf[0].apply(nodeI.intersections[startI].value)
                
                endI = -1
                while nodeI.intersections[endI].endP:
                    endI -= 1
                end = transf[0].apply(nodeI.intersections[endI].value)
                
                if start == end:
                    if nodeI.intersections[startI] != nodeI.intersections[endI]:
                        check = False
                        break
                    singleIntNodes.append([nodeI, matchNodeI, startI])
                    self.addVectChecks(vectChecks, [vectFrom, vectTo])
                    if not self.vectChecking(vectChecks):
                        check = False
                        singleIntNodes = []
                        break
                    continue
                elif start > end:
                    reverse = True
                else:
                    reverse = False
                
                # resolve points that are on the node but not an intersection
                if len(nodeI.linePoints) > 0:
                    if len(matchNodeI.linePoints) < len(nodeI.linePoints):
                        check = False
                        break
                    self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
                        
                
                if not nodeI.endP:
                    continue
                    
                
                """
                ################################################
                """
                nodeInts = transf[0].getCheckPoints(nodeI.nodeElement)
                lineForm = []
                specialLineForm = []
                lineIndex = 0
                while lineIndex < len(nodeI.forms):
                    if transf[0].hasCheckPointOf(nodeI.forms[lineIndex]):
                        lineForm.append(nodeI.forms[lineIndex])
                        prevCheck = lineIndex
                        while prevCheck > 0 and nodeI.forms[prevCheck-1].end == nodeI.forms[prevCheck].start:
                            lineForm.append(nodeI.forms[prevCheck-1])
                            prevCheck -= 1
                        while lineIndex < len(nodeI.forms)-1 and nodeI.forms[lineIndex].end == nodeI.forms[lineIndex+1].start:
                            lineForm.append(nodeI.forms[lineIndex+1])
                            lineIndex += 1
                    lineIndex += 1
                for line in nodeI.forms:
                    if line not in lineForm:
                        specialLineForm.append(line)
                for line in specialLineForm:
                    posMatches = self.resolveSpecialLine(line, nodeInts, reverse, nodeI, matchNodeI)
                    if len(posMatches) == 0:
                        check = False
                    else:
                        transfIndex = 0
                        while transfIndex < len(transf):
                            tr = transf.pop(transfIndex)
                            for posM in posMatches:
                                additionalTransf = tr.duplicate()
                                if not reverse:
                                    additionalTransf.add(line.start, posM.start)
                                    additionalTransf.add(line.end, posM.end)                            
                                else:
                                    additionalTransf.add(line.start, posM.end)
                                    additionalTransf.add(line.end, posM.start)
                                transf.insert(transfIndex, additionalTransf)
                                transfIndex += 1
                if not check:
                    break
                
                ind = 0
                while ind < len(transf):
                    tr = transf[ind]
                    trCheck = True
                    ###################nodeInts = transf.getCheckPoints(nodeI.nodeElement)
                    for line in lineForm:
                        hasMatch = self.resolveLine(line, reverse, nodeI, matchNodeI, tr)
                        if not hasMatch:
                            transf.pop(ind)
                            trCheck = False
                            break
                    if not trCheck:
                        break
                    """
                    ################################################
                    """
                    for intersection in nodeI.intersections:
                        if not tr.contains(intersection.value):
                            transf.pop(ind)
                            ind -=1
                            break
                    ind += 1
            
            
            if not self.checkIntersectionPoints(transf):
                continue

            if noInt:
                transf = self.resolveNoIntersectionNodes(match, transf)
                if transf is False:
                    continue
                
            if len(singleIntNodes) == 2 and singleIntNodes[0][0].nodeElement.intersect(singleIntNodes[1][0].nodeElement):
                self.resolveSingleIntersectionNodes([singleIntNodes[0]], transf)
                self.resolveSingleIntersectionNodes([singleIntNodes[1]], transf)
            else:
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)
            
            self.resolvePlanes(transf, match)
            
            if check:
                for tr in transf:
                    #print('----------------')
                    #tr.printTransf()
                    transforms.append(tr)
        return transforms

    def detectTransform(self, matchLists):
        for subset in itertools.permutations([k for k in range(len(self.matchingNodeList))], len(self.nodeList)):
            check = True
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.edgeMatrices[indI][indJ]
                    matchEdge = self.matchingEdgeMatrices[i][j]
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            if check:
                matchLists.append(subset)
                
    def resolveSpecialLine(self, line, nodeInts, reverse, nodeI, matchNodeI):
        posMatch = []
        #print('~~~~~~')
        if line.start < nodeInts[0][0]:
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[0][0]) and nodeInts[0][0] != lineCheck.start:
                    checkEnd = True
            for mLine in matchNodeI.forms:
                #print(mLine)
                if line.hasAttribute() and not line.attribute.partOf(mLine.attribute):
                    continue
                if (not reverse) and mLine.start >= nodeInts[0][1]:
                    continue
                elif reverse and mLine.end <= nodeInts[0][1]:
                    continue
                if not checkEnd:
                    if not reverse:
                        matchStart = mLine.start
                        matchEnd = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.end
                    else:
                        matchStart = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.start
                        matchEnd = mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                else:
                    if checkEnd and mLine.containsPoint(nodeInts[0][1]):
                        continue
                    newLine = mLine
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                if not mergeable:
                    posMatch.append(newLine)
                    #print(posMatch)
            #raise ValueError('.... not implemented yet ....')
        elif line.start > nodeInts[-1][0]:
            checkStart = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[-1][0]) and nodeInts[-1][0] != lineCheck.end:
                    checkStart = True
            for mLine in matchNodeI.forms:
                #print(mLine)
                if line.hasAttribute() and not line.attribute.partOf(mLine.attribute):
                    continue
                if (not reverse) and mLine.end <= nodeInts[-1][1]:
                    continue
                elif reverse and mLine.start >= nodeInts[-1][1]:
                    continue
                if not checkStart:
                    if not reverse:
                        matchStart = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.start
                        matchEnd = mLine.end
                    else:
                        matchStart = mLine.start
                        matchEnd = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                    
                else:
                    if checkStart and mLine.containsPoint(nodeInts[-1][1]):
                        continue
                    newLine = mLine
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                if not mergeable:
                    posMatch.append(newLine)
            #raise ValueError('.... not implemented yet ....')
        else:
            index = 1
            while line.start > nodeInts[index][0]:
                index += 1
            if reverse:
                upperBound =  nodeInts[index-1][1]
                lowerBound =  nodeInts[index][1]
            else:
                upperBound =  nodeInts[index][1]
                lowerBound =  nodeInts[index-1][1]
            #print(' nodeI:', line)
            #print(' lowerBound-upperBound:', lowerBound, upperBound)
            checkStart = False
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[index-1][0]) and nodeInts[index-1][0] != lineCheck.end:
                    checkStart = True
                if lineCheck.containsPoint(nodeInts[index][0]) and nodeInts[index][0] != lineCheck.start:
                    checkEnd = True
            #print('   .....___',checkStart, checkEnd)
            for mLine in matchNodeI.forms:
                if line.hasAttribute() and not line.attribute.partOf(mLine.attribute):
                    continue
                if mLine.end <= lowerBound or mLine.start >= upperBound:
                    continue
                #print('   .....', mLine)
                if not (checkStart or checkEnd):
                    if not reverse:
                        matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.start
                        matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.end
                    else:
                        matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.end
                        matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.start
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                else:
                    if checkStart and mLine.containsPoint(nodeInts[index-1][1]):
                        continue
                    if checkEnd and mLine.containsPoint(nodeInts[index][1]):
                        continue
                    newLine = mLine
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                if not mergeable:
                    posMatch.append(newLine)
                    #print(posMatch)
        return posMatch
        
        
    def resolveLine(self, line, reverse, nodeI, matchNodeI, transf):
        checkStart = transf.contains(line.start)
        checkEnd = transf.contains(line.end)
        if checkStart and checkEnd:
            return True
        if checkStart:
            unmatchedInts = [line.end]
        elif checkEnd:
            unmatchedInts = [line.start]
        else:
            unmatchedInts = [line.start, line.end]
        
        
        hasMatch = False
        segmentInts = []


        nodeInts = []
        for nInt in nodeI.intersections:
            if nInt.endP:
                if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                    pass
                else:
                    continue
            matchingInt = transf.apply(nInt.value)
            if line.start == nInt.value or line.end == nInt.value:
                alignment = BOUND
                segmentInts.append([nInt.value, matchingInt, alignment])
            elif line.containsPoint(nInt.value):
                alignment = INSIDE
                segmentInts.append([nInt.value, matchingInt, alignment])
            nodeInts.append([nInt.value, matchingInt])
            
        """
        nodeInts = transf.getCheckPoints(nodeI.nodeElement)
        for nInt in nodeI.intersections:
            if line.containsPoint(nInt[0]):
                if line.start == nInt[0] or line.end == nInt[0]:
                    alignment = BOUND
                else:
                    alignment = INSIDE
                segmentInts.append([nInt[0], nInt[1], alignment])
                
        if len(segmentInts) == 0:
            return False
        """
        
        #print('\n____ line:',line, line.attribute)
        # checking against all line segments in the matching node
        for mLine in matchNodeI.forms:
            # if line's attribute is not a part of matching line's attribute --> continue
            if line.hasAttribute() and not line.attribute.partOf(mLine.attribute):
                continue
            
            #print('   ____ mLine:', mLine, mLine.attribute)
            
            # check if the matching line has the contains the matching intersections of intersections on line
            lineCheck = True
            for segInt in segmentInts:
                if (mLine.containsPoint(segInt[1]) 
                or (segInt[2] == BOUND and (mLine.start == segInt[1] or mLine.end == segInt[1]))):
                    continue
                lineCheck = False
                break
            if not lineCheck:
                continue
            hasMatch = True
            #print('     ~~~~ match ~~~~ ')
            # for each unmatched points
            for pointIndex, point in enumerate(unmatchedInts):
                posMatch = None # possible match
                orderingPriority = False
                if point < segmentInts[0][0]:
                    if point > nodeInts[0][0]:
                        if reverse:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching segment
                            #       smaller or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[0][1]  
                                and mInt.value <= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching segment
                            #       bigger or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[0][1]  
                                and mInt.value >= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                elif point > segmentInts[-1][0]:
                    if point < nodeInts[-1][0]:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching segment
                            #       bigger or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[-1][1]  
                                and mInt.value >= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching segment
                            #       smaller or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[-1][1]  
                                and mInt.value <= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                else:
                    # point is in between the first and last matched points on segment
                    # narrow down the range in which the point can be 
                    index = 1
                    while point > segmentInts[index][0] and index < len(segmentInts):
                        index += 1
                    # the point must be now between segmentInts[index-1] and segmentInts[index]
                    # i.e. point is bigger than point on node at (index-1) 
                    #          and smaller than point on node at (index) 
                    if reverse:
                        # the node is matched in reverse
                        # --> matching point must be:
                        #       smaller than matched point on matching node at (index-1)
                        #       bigger than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in reversed(matchNodeI.intersections):
                            if (mInt.value < segmentInts[index-1][1]  
                            and mInt.value > segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                    else:
                        # the node is NOT matched in reverse
                        # --> matching point must be:
                        #       bigger than matched point on matching node at (index-1)
                        #       smaller than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in matchNodeI.intersections:
                            if (mInt.value > segmentInts[index-1][1]  
                            and mInt.value < segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                if posMatch is None:
                    continue
                transf.add(point, posMatch, orderingPriority)
                """
                addCheck = False
                for j in range(len(nodeInts)):
                    if point < nodeInts[j][0]:
                        nodeInts.insert(j, [point, posMatch])
                        addCheck = True
                        break
                if addCheck is False:
                    nodeInts.append([point, posMatch])
                    """
        return hasMatch
        
    def resolveSingleIntersectionNodes(self, singleIntNodes, transf):
        if len(singleIntNodes) == 0:
            return transf
            
        transf1 = copy(transf)
        for sn in singleIntNodes:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
        transf += transf1
        return transf
    
    def resolveSingleIntersectionLine(self, nodeI, matchNodeI, tr, anchor, reverse):
        if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
            return []
                        
        anchorFrom = nodeI.intersections[anchor].value
        anchorTo = tr.apply(nodeI.intersections[anchor].value)
        anchorCheck = False
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                anchorCheck = True
                break

        lineMatches = []
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                lineCheck = 0
            elif line.start > anchorFrom:
                lineCheck = 1
            else:
                lineCheck = -1
            posMatch = []
            for mLine in matchNodeI.forms:
                if line.hasAttribute() and not line.attribute.partOf(mLine.attribute):
                    continue
                if not reverse:
                    if lineCheck == 0 and mLine.containsPoint(anchorTo):
                        if line.start == anchorFrom:
                            if anchorTo == mLine.end: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif line.end == anchorFrom:
                            if anchorTo == mLine.start: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        else:
                            newLine = mLine
                    elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                        newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                    elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                        newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                    elif lineCheck == 1 and mLine.start > anchorTo:
                        newLine = mLine
                    elif lineCheck == -1 and mLine.end < anchorTo:
                        newLine = mLine
                    else:
                        continue
                else:
                    if lineCheck == 0 and mLine.containsPoint(anchorTo):
                        if line.start == anchorFrom:
                            if anchorTo == mLine.start: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif line.end == anchorFrom:
                            if anchorTo == mLine.end: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        else:
                            newLine = mLine
                    elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                        newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                    elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                        newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                    elif lineCheck == 1 and mLine.end < anchorTo:
                        newLine = mLine
                    elif lineCheck == -1 and mLine.start > anchorTo:
                        newLine = mLine
                    else:
                        continue
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                    
                if not mergeable:
                    posMatch.append(newLine)
            
            if len(posMatch) == 0:
                return []
            lineMatches.append((line, posMatch))
            
        indexing = [0] * len(lineMatches)
        transf = []
        transfIndex = 0
        while True:
            additionalTransf = tr.duplicate()
            addCheck = True
            for l in range(len(indexing)):
                line = lineMatches[l][0]
                posM = lineMatches[l][1][indexing[l]]
                if not reverse:
                    if not (additionalTransf.add(line.start, posM.start)
                    and additionalTransf.add(line.end, posM.end)):
                        addCheck = False
                        break
                else:
                    if not (additionalTransf.add(line.start, posM.end)
                    and additionalTransf.add(line.end, posM.start)):
                        addCheck = False
                        break
                    
            if addCheck:
                transf.insert(transfIndex, additionalTransf)
            transfIndex += 1
            indexing[0] += 1
            for i in range(len(indexing)-1):
                if indexing[i] == len(lineMatches[i][1]):
                    indexing[i] = 0
                    indexing[i+1] += 1
                else:
                    break
            if indexing[-1] == len(lineMatches[-1][1]):
                break
        self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
        return transf
        
    def resolveNoIntersectionNodes(self, match, transf):
        if len(self.nodeList) == 0:
            return transf
            
        transf1 = copy(transf)
        
            
        line0 = self.nodeList.lines[0].nodeElement
        mLine0 = self.matchingNodeList.lines[match[0]].nodeElement
        intersector = matchIntersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(0,1))
        if not intersector.intersect(line0):
            intersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(1,0))
        if not matchIntersector.intersect(mLine0):
            matchIntersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(1,0))
            
        vectChecks = []
        for nodeIndex in range(len(self.nodeList)):
            if (len(self.nodeList.lines[nodeIndex].forms) > len(self.matchingNodeList.lines[match[nodeIndex]].forms)
            or len(self.nodeList.lines[nodeIndex].linePoints) > len(self.matchingNodeList.lines[match[nodeIndex]].linePoints)):
                return False
                
            lineI = self.nodeList.lines[nodeIndex].nodeElement
            matchLineI = self.matchingNodeList.lines[match[nodeIndex]].nodeElement
            
            v0 = lineI.intersection(intersector)
            v1 = matchLineI.intersection(matchIntersector)
            self.addVectChecks(vectChecks, [v0,v1])
        if not self.vectChecking(vectChecks):
            return False

        for nodeIndex in range(len(self.nodeList)):
            
            nodeI = self.nodeList.lines[nodeIndex]
            matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
            
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, True, transf)
                    
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, False, transf1)
        transf += transf1
        
        
        return transf
                    
    def resolveNoIntersectionLine(self, nodeI, matchNodeI, tr, reverse):
        transf = []
        for i in itertools.combinations(list(range(len(matchNodeI.forms))), len(nodeI.forms)):
            check = True
            for j in range(len(i)):
                if nodeI.forms[j].sort.type == 'sort-attribute' :
                    if (not reverse) and not nodeI.forms[j].attribute.partOf(matchNodeI.forms[i[j]].attribute):
                        check = False
                        break
                    elif reverse and not nodeI.forms[len(i) - j - 1].attribute.partOf(matchNodeI.forms[i[j]].attribute):
                        check = False
                        break
            if not check:
                continue
            newtr = tr.duplicate()
            if not reverse:
                for j in range(len(i)):
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[i[j]].start)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[i[j]].end)
            else:
                for j in range(len(i)):
                    k = len(i) - j - 1
                    newtr.add(nodeI.forms[k].start, matchNodeI.forms[i[j]].end)
                    newtr.add(nodeI.forms[k].end, matchNodeI.forms[i[j]].start)
            transf.append(newtr)
        return transf
        
    def resolveFreePoints(self, transf):
        fPoints = []
        for p in self.nodeList.points:
            if p.oType == 'f':
                fPoints.append(p.value)
        if len(fPoints) == 0:
            return True
                
        mFPoints = []
        for p in self.matchingNodeList.points:
            if p.oType == 'f':
                mFPoints.append(p.value)
        if len(fPoints)>len(mFPoints):
            return False
            
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(0)
            for i in itertools.permutations(mFPoints, len(fPoints)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(fPoints)):
                    if not newTr.add(fPoints[j], i[j]):
                        check = False
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True
        
        
    def resolvePointsOnLine(self, nodeI, matchNodeI, reverse, transf):
        for p in nodeI.linePoints:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                
                if not reverse:
                    if p.value < nodeInts[0][0]:
                        posRange = (None, nodeInts[0][1])
                    elif p.value > nodeInts[-1][0]:
                        posRange = (None, nodeInts[-1][1])
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex-1][1], nodeInts[rangeIndex][1])
                else:
                    if p.value < nodeInts[0][0]:
                        posRange = (nodeInts[0][1], None)
                    elif p.value > nodeInts[-1][0]:
                        posRange = (nodeInts[-1][1], None)
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex][1], nodeInts[rangeIndex-1][1])
                posMatch = []
                for mP in matchNodeI.linePoints:
                    if ((posRange[0] is not None and mP.value<posRange[0])
                    or (posRange[1] is not None and mP.value>posRange[1])):
                        continue
                    posMatch.append(mP.value)
                    
                for i in posMatch:
                    newTransf = tr.duplicate()
                    if not newTransf.add(p.value, i):
                        continue
                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1
                    
                
    def resolvePlanes(self, transf, match):
        for obj in self.nodeList.planes:
            plane = obj.value
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                for otherObj in self.matchingNodeList.planes:
                    otherPlane = otherObj.value
                    
                    if (otherPlane.sort != plane.sort or (plane.hasAttribute() 
                    and (not plane.attribute == otherPlane.attribute))
                    or len(plane.boundaries[0].polygonal) != len(otherPlane.boundaries[0].polygonal)):
                        continue

                    """
                    print('.............')
                    print(plane.printRhino())
                    print(otherPlane.printRhino())
                    """
                    
                    posMatch = []
                    check = True
                    for i in range(len(self.nodeList.lines)):
                        nodeI = self.nodeList.lines[i]
    
                        polygons = []
                        for p in plane.boundaries[0].polygonal:
                            if nodeI.nodeElement.containsPoint(p):
                                polygons.append(p)
                                
                        if len(polygons) == 0:
                            continue
                        polygons = sorted(polygons)
                        checkProceed = False
                        for p in polygons:
                            if not tr.contains(p):
                                checkProceed = True
                        if not checkProceed:
                            continue
                        matchNodeI = self.matchingNodeList.lines[match[i]]
                        
                        nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        if nodeInts[0][1] > nodeInts[-1][1]:
                            reverse = True
                        else:
                            reverse = False
                            
                        mPolygons = []
                        for p in otherPlane.boundaries[0].polygonal:
                            if matchNodeI.nodeElement.containsPoint(p):
                                mPolygons.append(p)
                        if len(mPolygons) != len(polygons):
                            check = False
                            break
                        mPolygons = sorted(mPolygons)
                        if reverse:
                            mPolygons = list(reversed(mPolygons))
                        for pI in range(len(polygons)):
                            mP = tr.transform(polygons[pI])
                            if mP is not False:
                                if mPolygons[pI] == mP:
                                    continue

                                check = False
                                break
                            if polygons[pI] < nodeInts[0][0]:
                                if ((not reverse and mPolygons[pI] > nodeInts[0][1])
                                or (reverse and mPolygons[pI] < nodeInts[0][1])):
                                    check = False
                                    break
                            elif polygons[pI] > nodeInts[-1][0]:
                                if ((not reverse and mPolygons[pI] < nodeInts[-1][1])
                                or (reverse and mPolygons[pI] > nodeInts[-1][1])):
                                    check = False
                                    break
                            else:
                                intI = 0
                                while polygons[pI] > nodeInts[intI][0]:
                                    intI += 1
                                if not ((not reverse and nodeInts[intI-1][1] <= mPolygons[pI] <= nodeInts[intI][1])
                                or (reverse and nodeInts[intI][1] <= mPolygons[pI] <= nodeInts[intI-1][1])):
                                    check = False
                                    break
                            posMatch.append([polygons[pI], mPolygons[pI]])
                    if not check:
                        continue
                    newTr = tr.duplicate()
                    for pM in posMatch:
                        newTr.add(pM[0], pM[1])
                    transf.insert(transfIndex, newTr)
                    transfIndex += 1      
                    
    def orderingCheck(self, transf, nodeI):
        ind = 0
        originalInts = []
        for ints in nodeI.intersections:
            if ints.iType == 'o':
                continue
            originalInts.append(ints.value)
        originalInts = sorted(originalInts)
        while ind < len(transf):
            tr = transf[ind]
            nodeInts = []
            for i in originalInts:
                nodeInts.append([i,tr.apply(i)])
            #nodeInts = tr.getCheckPoints(nodeI.nodeElement)
            if len(nodeInts) < 2:
                ind += 1
                continue
            reverse = False
            index = 1
            while index < len(nodeInts) and nodeInts[index][1] == nodeInts[index-1][1]:
                index += 1
            if index == len(nodeInts):
                ind += 1
                continue
            if nodeInts[index-1][1] < nodeInts[index][1]:
                reverse = False
            else:
                reverse = True
            for index in range(1,len(nodeInts)):
                if ((not reverse and nodeInts[index-1][1] > nodeInts[index][1])
                or (reverse and nodeInts[index-1][1] < nodeInts[index][1])):
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0 :
            return False
        return True
        
    """
    def orderingCheck(self, transf, nodeI):
        ind = 0
        while ind < len(transf):
            tr = transf[ind]
            nodeInts = tr.getCheckPoints(nodeI.nodeElement)
            if len(nodeInts) < 2:
                ind += 1
                continue
            reverse = False
            index = 1
            while index < len(nodeInts) and nodeInts[index][1] == nodeInts[index-1][1]:
                index += 1
            if index == len(nodeInts):
                ind += 1
                continue
            if nodeInts[index-1][1] < nodeInts[index][1]:
                reverse = False
            else:
                reverse = True
            for index in range(1,len(nodeInts)):
                if ((not reverse and nodeInts[index-1][1] > nodeInts[index][1])
                or (reverse and nodeInts[index-1][1] < nodeInts[index][1])):
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0 :
            return False
        return True
    """
    
    def addVectChecks(self, vectChecks, item):
        for i in range(len(vectChecks)):
            if vectChecks[i][0] > item[0]:
                vectChecks.insert(i, item)
                return
        vectChecks.append(item)
                
    def vectChecking(self, vectChecks):
        if len(vectChecks) > 2:
            a = vectChecks[0][0]<vectChecks[1][0]
            b = vectChecks[0][1]<vectChecks[1][1]
            if (a and b) or not (a or b):
                reverse = False
            else:
                reverse = True
            for index in range(2, len(vectChecks)):
                a = vectChecks[index-1][0]<vectChecks[index][0]
                b = vectChecks[index-1][1]<vectChecks[index][1]
                if (a and b) or not (a or b):
                    if reverse:
                        return False
                        break
                elif not reverse:
                    return False
                    break
        return True
        
    def checkIntersectionPoints(self, transf):
        ind = 0
        while ind < len(transf):
            tr = transf[ind]
            for p in self.nodeList.points:
                if p.oType != 'i':
                    continue
                transformedPoint = tr.apply(p.value)
                check = False
                for mP in self.matchingNodeList.points:
                    if mP.oType != 'i':
                        continue
                    if mP.value == transformedPoint:
                        check = True
                        break
                if not check:
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0:
            return False
        return True
                
    @staticmethod
    def getIntersection(line1, line2):
        for i in line1.intersections:
            for j in line2.intersections:
                if i == j:
                    return i
        return None

    def analyzeRHS(self, rhs, shape, transf):
        transfIndex = 0
        while transfIndex < len(transf):
            residuals = []
            if transf[transfIndex].transformable(rhs, residuals):
                transfIndex += 1
                continue
            tr = transf.pop(transfIndex)
            newTransf = [tr]
            for r in residuals:
                if r.type == 'ind-lineSegmentP2D':
                    containedPoints = []
                    for i in tr.vects:
                        if r.containsPoint(i[0]):
                            containedPoints.append(i)
                    if len(containedPoints) == 1:
                        parallelLines = self.getLineProperties('p', rhs, r)
                        i = 0
                        while i < len(parallelLines):
                            if (not tr.contains(r.start)) or (not tr.contains(r.end)):
                                parallelLines.pop(i)
                            else:
                                i+=1
                        if len(parallelLines) == 0:
                            for t in newTransf:
                                if not t.contains(r.start):
                                    t.addUnresolved(r.start)
                                if not t.contains(r.end):
                                    t.addUnresolved(r.end)
                            continue
                        intersectionLines = self.getLineProperties('iE', rhs, r)
                        posMatch = []
                        if len(intersectionLines) >= 2:
                            lr = line2D(r.sort, containedPoints[0][1], containedPoints[0][1].add((tr.apply(parallelLines[0].start).subtract(tr.apply(parallelLines[0].end)))))
                            l0 = line2D(r.sort, tr.apply(intersectionLines[0].start), tr.apply(intersectionLines[0].end))
                            l1 = line2D(r.sort, tr.apply(intersectionLines[-1].start), tr.apply(intersectionLines[-1].end))
                            if not (lr.intersect(l0) and lr.intersect(l1)):
                                newTransf = []
                                continue
                            start = lr.intersection(l0)
                            end = lr.intersection(l1)
                            if start == end:
                                intersectionLines = intersectionLines[0:1]
                            else:
                                posMatch.append([[r.start, start],[r.end, end]])
                        if len(intersectionLines) == 1:
                            lr = line2D(r.sort, containedPoints[0][1], containedPoints[0][1].add((tr.apply(parallelLines[0].start).subtract(tr.apply(parallelLines[0].end)))))
                            l0 = line2D(r.sort, tr.apply(intersectionLines[0].start), tr.apply(intersectionLines[0].end))
                            if not lr.intersect(l0):
                                newTransf = []
                                continue
                            intersectionP = lr.intersection(l0)
                            for pLine in parallelLines:
                                if pLine in residuals:
                                    continue
                                
                                intPoint = pLine.intersection(intersectionLines[0])
                                intProj = r.root.add(r.direction.scale(intPoint.subtract(r.root)))
                                if r.containsPoint(intProj):
                                    containsProj = True
                                else:
                                    containsProj = False
                                mPLine = pLine.__class__(pLine.sort, tr.apply(pLine.start), tr.apply(pLine.end))
                                mIntPoint = mPLine.intersection(l0)
                                mIntProj = lr.root.add(lr.direction.scale(mIntPoint.subtract(lr.root)))
                                
                                start = tr.apply(pLine.start)
                                end = tr.apply(pLine.end)

                                nStart = containedPoints[0][1].add(start.subtract(end).scale(start.subtract(containedPoints[0][1])))
                                nEnd = containedPoints[0][1].add(start.subtract(end).scale(end.subtract(containedPoints[0][1])))
                                                        
                                if nStart > nEnd:
                                    nStart, nEnd = nEnd, nStart

                                if (nStart < intersectionP) != (nEnd < intersectionP):
                                    if containsProj == (mIntProj == nStart):
                                        if intersectionP != nStart:
                                            posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                                    else:
                                        if intersectionP != nEnd:
                                            posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                elif nStart > intersectionP and containsProj == (mIntProj == nEnd):
                                    if intersectionP != nEnd:
                                        posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                elif containsProj == (mIntProj == nStart):
                                    if intersectionP != nStart:
                                        posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                        else:
                            for pLine in parallelLines:
                                start = tr.apply(pLine.start)
                                end = tr.apply(pLine.end)
                                nStart = containedPoints[0][1].add(start.subtract(end).scale(start.subtract(containedPoints[0][1])))
                                nEnd = containedPoints[0][1].add(start.subtract(end).scale(end.subtract(containedPoints[0][1])))
                                posMatch.append([[r.start, nStart],[r.end, nEnd]])
                        lenNewTransf = len(newTransf)
                        for count in range(lenNewTransf):
                            for m in posMatch:
                                newtr = newTransf[0].duplicate()
                                check = True
                                for item in m:
                                    if not newtr.add(item[0],item[1]):
                                        check = False
                                        break
                                if check:
                                    newTransf.append(newtr)
                                newtr = newTransf[0].duplicate()
                                check = True
                                for itemI in range(len(m)):
                                    if not newtr.add(m[itemI][0],m[itemI-1][1]):
                                        check = False
                                        break
                                if check:
                                    newTransf.append(newtr)
                            newTransf.pop(0)
                    elif len(containedPoints) > 1:
                        raise NotImplementedError
                    else:
                        for t in newTransf:
                            t.addUnresolved(r.start)
                            t.addUnresolved(r.end)
                if 'ind-point' in r.type:
                    for t in newTransf:
                        t.addUnresolved(r.value)
            for newTr in newTransf:
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return transf
        
    
            
    @staticmethod
    def getLineProperties(prop, element, lineSeg):
        if 'ind-lineSegment2D' == element.type or 'ind-lineSegmentP2D' == element.type:
            if (prop == 'p' and type(element) == type(lineSeg) and element != lineSeg 
            and element.parallel(lineSeg) and not element.start.colinear(lineSeg.start, lineSeg.end)):
                return element
            if prop == 'i' and type(element) == type(lineSeg) and element != lineSeg and element.intersectSegment(lineSeg):
                return element
            if prop == 'iE' and type(element) == type(lineSeg) and element != lineSeg:
                if (lineSeg.start.colinear(element.start, element.end)
                or lineSeg.end.colinear(element.start, element.end)):
                    return element
            return None
        elif 'ind-point2D' == element.type:
            if prop == 'cP' and lineSeg.containsPoint(element.value):
                return element
            return None
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                p = []
                for f in element.elements:
                    p += graphMap.getLineProperties(prop, f, lineSeg)
                return p
            else:
                p = []
                for l in element.elements:
                    check = graphMap.getLineProperties(prop, l, lineSeg)
                    if check is not None:
                        p.append(check)
                return p
            return []
            
            
    def evaluatePredicate(self, transfSet, lhs, tlhs, form):   
        graphTransf = transfSet.retrieveTransform(graphTransform)
        self.postCheck = True
        for cond in self.predicates:
            if cond == 'maxline':
                for keyI in self.predicates[cond]:
                    ind = keyI.ind
                    transfInd = ind.transform(transfSet)
                    form.accept(self, direction = [cond, transfInd])
            elif cond == 'bound':
                for keyI in self.predicates[cond]:
                    """
                    # bound for lhs after transformation
                    ind = keyI.ind.transform(transfSet)
                    form.accept(self, direction = [cond, ind, keyI.properties])
                    """
                    ind = keyI.ind.transform(transfSet)
                    transfStart = graphTransf.apply(keyI.ind.start)
                    transfEnd = graphTransf.apply(keyI.ind.end)
                    form.accept(self, direction = [cond, ind, keyI.properties, transfStart, transfEnd])
            elif cond == 'void':
                bound = boundary()
                for keyI in self.predicates[cond]:
                    newInd = graphTransf.transform(keyI)
                    if newInd is False:
                        raise ValueError('matching for vector '+str(keyI)+' is not found')
                    bound.add(newInd)
                for i in range(len(bound.polygonal)):
                    bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                planeSeg = planeSegment2D(lhs.sort, [bound], True)
                lines = []
                form.accept(self, direction = [cond, planeSeg, lines])
                check = planeSeg.classify_line(lines)
                if not check.notInside():
                    self.postCheck = False
            elif cond == 'shortest_line':
                for keyI in self.predicates[cond]:
                    ind = keyI.ind.transform(transfSet)
                    tlhs.accept(self, direction = [cond, ind.squareLength()])
            elif cond == 'nolabel':
                for keyI in self.predicates[cond]:
                    ind = keyI.ind.transform(transfSet)
                    tlhs.accept(self, direction = [cond, ind])
            if self.postCheck is False:
                return False
        return True
            
    def solveDirectives(self, transfSet, lhs, tlhs, rhs, form):
        transf = [transfSet]
        param = paramSet()
        param.lhs = tlhs
        param.rhs = rhs
        for cond in self.directives:
            if cond == 'point_on_line':
                for i0 in range(len(transf)):
                    tfSet = transf.pop(0)
                    graphTransf = tfSet.retrieveTransform(graphTransform)
                    for p in range(len(self.directives[cond])):
                        bounds = self.evaluateDesc(self.directives[cond][p], param)
                        if bounds[1].__class__ != coordinate:
                            bounds.insert(0, bounds[0])
                        if len(bounds) == 2:
                            continue
                        for i in range(2,len(bounds)):
                            obj = self.directives[cond][p][i].ind
                            self.solvePointOnLine(graphTransf, obj, bounds)
                    lhs.accept(self, direction = [cond, graphTransf])
                    if len(graphTransf.unresolved) > 0:
                        continue
                    transf.append(tfSet)
            elif cond == 'angle':
                for p in range(len(self.directives[cond])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        bounds = self.evaluateDesc(self.directives[cond][p], param)
                        posMatch = self.solveAngle(tfSet, bounds, lhs)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = ntfSet.retrieveTransform(graphTransform)
                            for i in j:
                                for uI in range(len(tr.unresolved)):
                                    u = tr.unresolved[uI]
                                    if u == i[0]:
                                        tr.add(i[0],i[1])
                                        tr.unresolved.pop(uI)
                                        break
                            transf.append(ntfSet)
            elif cond == 'distance':
                for p in range(len(self.directives[cond])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        bounds = self.evaluateDesc(self.directives[cond][p], param)
                        if 'lineSegment' in self.directives[cond][p][-2].ind.type and 'lineSegment' in self.directives[cond][p][-1].ind.type:
                            posMatch = self.solveLineLineDistance(tfSet, bounds)
                        elif 'lineSegment' in self.directives[cond][p][-2].ind.type and 'point' in self.directives[cond][p][-1].ind.type:
                            posMatch = self.solveLinePointDistance(tfSet, bounds)
                        elif 'point' in self.directives[cond][p][-2].ind.type and 'lineSegment' in self.directives[cond][p][-1].ind.type:
                            posMatch = self.solvePointLineDistance(tfSet, bounds)
                        elif 'point' in self.directives[cond][p][-2].ind.type and 'point' in self.directives[cond][p][-1].ind.type:
                            posMatch = self.solvePointPointDistance(tfSet, bounds)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = ntfSet.retrieveTransform(graphTransform)
                            for i in j:
                                for uI in range(len(tr.unresolved)):
                                    u = tr.unresolved[uI]
                                    if u == i[0]:
                                        tr.add(i[0],i[1], override = True)
                                        tr.unresolved.pop(uI)
                                        break
                            transf.append(ntfSet)
                
        if 'length' in self.directives:
            for p in range(len(self.directives['length'])):
                for i0 in range(len(transf)):
                    tfSet = transf.pop(0)
                    bounds = self.evaluateDesc(self.directives["length"][p], param)
                    posMatch = self.readjustLength(tfSet, bounds, lhs, rhs, )
                    for j in posMatch:
                        ntfSet = tfSet.duplicate()
                        tr = ntfSet.retrieveTransform(graphTransform)
                        for i in j:
                            tr.add(i[0],i[1], override = True)
                        transf.append(ntfSet)
        i0 = 0
        while i0 < len(transf):
            if len(transf[i0].retrieveTransform(graphTransform).unresolved) > 0:
                transf.pop(i0)
            else:
                i0 += 1
        return transf
        

    def visitEnter(self, obj, direction = 0):
        if direction[0] == 'maxline':
            if obj.__class__ == direction[1].__class__ and obj.contains(direction[1]) and obj != direction[1] :
                self.postCheck = False
                return False
        elif direction[0] == 'bound':
            if obj.__class__ == direction[1].__class__ and obj.contains(direction[1]):
                if 0 in direction[2] and obj.start != direction[3] and obj.end != direction[3]:
                    self.postCheck = False
                    return False
                if 1 in direction[2] and obj.start != direction[4] and obj.end != direction[4]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'shortest_line':
            if 'ind-lineSegment' in obj.type and obj.squareLength() < direction[1]:
                self.postCheck = False
                return False
        elif direction[0] == 'nolabel':
            if 'ind-lineSegment' in obj.type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
        elif direction[0] == 'nolabel*':
            if obj.type == 'ind-label':
                self.postCheck = False
                return False
        elif direction[0] == 'point_on_line':
            if 'ind-lineSegment' in obj.type:
                bounds = [c0, c1]
                self.solvePointOnLine(direction[1], obj, bounds)
        elif direction[0] == 'containsCheck':
            if 'ind-lineSegment' in obj.type:
                if len(direction) > 3 and (direction[3] is not None) and (obj == direction[3]):
                    return False
                elif obj.start == direction[1] or obj.end == direction[1]:
                    direction[2][0] = True
            if 'ind-point' in obj.type:
                if obj.value == direction[1]:
                    direction[2][0] = True
        return True
        
    def visitLeave(self, obj, direction = 0):
        return
        
        
    @staticmethod
    def ccw(A,B,C):
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
        
    @staticmethod
    def evaluateDesc(boundList, param):
        newList = []
        for i in boundList:
            if isinstance(i, description):
                val = i.evaluateInd(param).getValue()[1:-1]
                newList.append(coordinate(val))
            else:
                newList.append(i)
        return newList
        
    @staticmethod
    def solvePointOnLine(graphTransf, obj, bounds):
        i = 0
        while i < len(graphTransf.unresolved):
            u = graphTransf.unresolved[i]
            if obj.containsPoint(u) and (not obj.start == u) and (not obj.end == u):
                graphTransf.unresolved.pop(i)
                nStart = graphTransf.apply(obj.start)
                nEnd = graphTransf.apply(obj.end)
                if obj.reverse:
                    nStart, nEnd = nEnd, nStart
                rand = randint(0, 100000)
                ratio = (bounds[1]-bounds[0])*rand/100000 + bounds[0]
                p = nStart.add(nEnd.subtract(nStart).scale(ratio))
                graphTransf.add(u, p)
            else:
                i += 1
                
    def solveAngle(self, tfSet, bounds, lhs):
        angle = bounds[0]
        while angle > 180:
            angle -= 360
        angle = angle * pi / 180 
        base = bounds[1].ind
        new = bounds[2].ind
        graphTransf = tfSet.retrieveTransform(graphTransform)
    
        anchorVect = None
        containedPoints = []
        for i in graphTransf.vects:
            if i[0].subtract(new.root).parallel(new.direction):
                containedPoints.append(i[0])
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
        if len(containedPoints) == 0:
            raise ValueError('')
                
        intersections = []
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
        
        intsn = base.intersection(new)
        
        delayBase = False
        if base.start >= intsn:
            baseDir = True
            p0 = base.end
        elif base.end <= intsn:
            baseDir = False
            p0 = base.start
        else:
            delayBase = True
        
        if new.start >= intsn:
            p1 = new.end
        elif new.end <= intsn:
            p1 = new.start
        else:
            if delayBase:
                p0 = base.start
                delayBase = False
            endProj = base.projection(new.end)
            if (abs(angle)*2 < pi) == ((endProj > intsn) == (p0 > intsn)):
                p1 = new.end
            else:
                p1 = new.start
            
        if delayBase:
            p1Proj = base.projection(p1)
            if (abs(angle)*2 < pi) == ((p1Proj > intsn) == (base.start > intsn)):
                p0 = base.start
                baseDir = False
            else:
                p0 = base.end
                baseDir = True
        
        if angle > 0 and self.ccw(p0, intsn, p1) :
            angle = -angle
        elif angle < 0 and not self.ccw(p0, intsn, p1) :
            angle = -angle

        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
        
        if anchorVect is not None and self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            angle = -angle
        
            
        if nBaseS > nBaseE:
            baseDir = not baseDir
            angle = pi-angle
            angle = -angle

            
        if baseDir:
            d1 = nBaseE.subtract(nBaseS)
        else:
            d1 = nBaseS.subtract(nBaseE)
            
        cp = containedPoints[0]
        nCP = graphTransf.apply(cp)
        
                
        ca = cos(angle)
        sa = sin(angle)
        d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
        
        nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
        if graphTransf.contains(new.start):
            nStart = graphTransf.apply(new.start)
            nEnd = nStart.add(d2.scale(nLen))
        elif graphTransf.contains(new.end):
            nEnd = graphTransf.apply(new.end)
            nStart = nEnd.add(d2.scale(nLen))
        elif new.start == intsn:
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            nStart = nBase.intersection(pLine)
            nEnd = nStart.add(d2.scale(nLen))
        elif new.end == intsn:
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            nEnd = nBase.intersection(pLine)
            nStart = nEnd.add(d2.scale(nLen))
        else:
            nLen = nLen / 2
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            intersection = pLine.intersection(nBase)
            nStart = intersection.subtract(d2.scale(nLen))
            nEnd = intersection.add(d2.scale(nLen))
        
        nLine = new.__class__(new.sort, nStart, nEnd)
        check = True
        for intersectionSet in intersections:
            intersection = intersectionSet[0]
            if graphTransf.contains(intersection):
                intersection = graphTransf.apply(intersection)
            else:
                intersection = intersectionSet[1].edit(tfSet).intersection(nLine)
            if not intersection.subtract(nLine.root).parallel(nLine.direction):
                check = False
                break
            if intersection < nLine.start:
                nLine.start = intersection
            elif intersection > nLine.end:
                nLine.end = intersection
                
        if not check:
            return []
        
        if not nLine.reverse:
            nStart = nLine.start
            nEnd = nLine.end
        else:
            nStart = nLine.end
            nEnd = nLine.start
            
        posMatch = []
        posMatch.append(((new.start, nStart),(new.end, nEnd)))
        
        if anchorVect is None:
            sa = -sa
            d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
            
            nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
            if graphTransf.contains(new.start):
                nStart = graphTransf.apply(new.start)
                nEnd = nStart.add(d2.scale(nLen))
            elif graphTransf.contains(new.end):
                nEnd = graphTransf.apply(new.end)
                nStart = nEnd.subtract(d2.scale(nLen))
            elif new.start == intsn:
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                nStart = nBase.intersection(pLine)
                nEnd = nStart.add(d2.scale(nLen))
            elif new.end == intsn:
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                nEnd = nBase.intersection(pLine)
                nStart = nEnd.subtract(d2.scale(nLen))
            else:
                nLen = nLen / 2
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                intersection = pLine.intersection(nBase)
                nStart = intersection.subtract(d2.scale(nLen))
                nEnd = intersection.add(d2.scale(nLen))
            
            nLine = new.__class__(new.sort, nStart, nEnd)

            
            if not nLine.reverse:
                nStart = nLine.start
                nEnd = nLine.end
            else:
                nStart = nLine.end
                nEnd = nLine.start
                
            posMatch.append(((new.start, nStart),(new.end, nEnd)))
        return posMatch
        
    """
    
    def solveAngle(self, tfSet, bounds, lhs):
        angle = bounds[0]
        while angle > 180:
            angle -= 360
        while angle < -180:
            angle += 360
        angle = angle * pi / 180 
        base = bounds[1].ind
        new = bounds[2].ind
            
        graphTransf = tfSet.retrieveTransform(graphTransform)
        posMatch = []
    
        anchorVect = None
        containedPoints = []
        for i in graphTransf.vects:
            if i[0].subtract(new.root).parallel(new.direction):
                containedPoints.append(i[0])
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
        if len(containedPoints) == 0:
            raise ValueError('')
            
        #########################################################
        intsn = base.intersection(new)
        
        if base.start >= intsn:
            baseDir = base.direction
            p0 = base.end
        elif base.end <= intsn:
            baseDir = base.direction.inverse()
            p0 = base.start
        else:
            pass
        if new.start >= intsn:
            p1 = new.end
        elif new.end <= intsn:
            p1 = new.start
        else:
            pass
        
        if self.ccw(p0, intsn, p1):
            angle = -angle
            
        
        
        
        #########################################################
        
        return posMatch
    """
        
    
    def solveLineLineDistance(self, tfSet, bounds):
        distance = bounds[0]
        base = bounds[1].ind
        new = bounds[2].ind

        if (new.start.y > base.projection(new.start).y) != (distance > 0):
            distance = -distance

        graphTransf = tfSet.retrieveTransform(graphTransform)
        
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
                break
        
        intersections = []
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
        
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
            
        if anchorVect is not None and self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            distance = -distance
            
        if nBaseS > nBaseE:
            d1 = nBaseS.subtract(nBaseE)
            distance = -distance
        else:
            d1 = nBaseE.subtract(nBaseS)
        
        pDir = d1.__class__(-d1.y, d1.x)
        
        sc = ((distance*distance)/(pDir.dotproduct(pDir))).sqrt()
            
        if distance < 0:
            sc = -sc
        
        nStart = nBaseS.add(pDir.scale(sc))
        nEnd = nBaseE.add(pDir.scale(sc))
        
        nLine = new.__class__(new.sort, nStart, nEnd)
        check = True
        for intersectionSet in intersections:
            intersection = intersectionSet[0]
            if graphTransf.contains(intersection):
                intersection = graphTransf.apply(intersection)
            else:
                intersection = intersectionSet[1].edit(tfSet).intersection(nLine)
            if not intersection.subtract(nLine.root).parallel(nLine.direction):
                check = False
                break
            if intersection < nLine.start:
                nLine.start = intersection
            elif intersection > nLine.end:
                nLine.end = intersection
        if not check:
            return []
            
        if not nLine.reverse:
            nStart, nEnd = nLine.start, nLine.end
        else:
            nStart, nEnd = nLine.end, nLine.start
        posMatch = []
        posMatch.append([[new.start, nStart],[new.end, nEnd]])
        
        if anchorVect is None:
            sc = -sc
            
            nStart = nBaseS.add(pDir.scale(sc))
            nEnd = nBaseE.add(pDir.scale(sc))
            
            nLine = new.__class__(new.sort, nStart, nEnd)
            
            if not nLine.reverse:
                nStart, nEnd = nLine.start, nLine.end
            else:
                nStart, nEnd = nLine.end, nLine.start
            posMatch.append([[new.start, nStart],[new.end, nEnd]])
        return posMatch
    
    

    def solveLinePointDistance(self, tfSet, bounds):
        distance = bounds[0]
        base = bounds[1].ind
        new = bounds[2].ind
        graphTransf = tfSet.retrieveTransform(graphTransform)
        
        if (new.value.y > base.projection(new.value).y) != (distance > 0):
            distance = -distance

        nCline = None
        for i in self.nodeList:
            if i.nodeElement.containsPoint(new.value):
                nCline = i.forms[0].edit(tfSet)
                break
        if nCline is None:
            return []
            
        
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
                break
            
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
            
        if self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            distance = -distance
            
        if nBaseS > nBaseE:
            d1 = nBaseS.subtract(nBaseE)
            distance = -distance
        else:
            d1 = nBaseE.subtract(nBaseS)
            
        pDir = d1.__class__(-d1.y, d1.x)
        
        sc = ((distance*distance)/(pDir.dotproduct(pDir))).sqrt()
            
        if distance < 0:
            sc = -sc
        
        nStart = nBaseS.add(pDir.scale(sc))
        nLine = base.__class__(base.sort, nStart, nStart.add(d1))
        
        try:            
            nPoint = nLine.intersection(nCline)
        except ValueError:
            return []
            
        return [[[new.value, nPoint]]]
        
    def solvePointLineDistance(self, tfSet, bounds):
        distance = bounds[0]
        base = bounds[1].ind
        new = bounds[2].ind

        graphTransf = tfSet.retrieveTransform(graphTransform)
        
        cPoint = []
        lineN = new.parametricNode()
        for i in graphTransf.vects:
            if lineN.containsPoint(i[0]):
                cPoint.append(i[0])
                cPoint.append(i[1])
                if new.containsPoint(i[0]): cPoint.append(True)
                else: cPoint.append(False)
                break
        if len(cPoint) == 0:
            raise ValueError
            
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.value, cPoint[0]):
                anchorVect = i
                break
        
        intersections = []
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
                    
        nBase = graphTransf.apply(base.value)
        
        d1 = nBase.subtract(cPoint[1])
        d1len = d1.length()
        sa = distance / d1len
        
        try:
            ca = (d1len*d1len - distance*distance).sqrt() / d1len
        except ValueError:
            return []
        
        if anchorVect is not None:
            ccwC = self.ccw(base.value, cPoint[0], anchorVect[0])
            if ccwC:
                sa = -sa
            if ccwC != self.ccw(nBase, cPoint[1], anchorVect[1]):
                sa = -sa
            
        d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
            
        nLine = line2D(new.sort, cPoint[1], cPoint[1].add(d2))
        
        newInts = []
        for i in intersections:
            l = i[1].edit(tfSet)
            newInts.append(l.intersection(nLine))
            
        posMatch = []
        if cPoint[-1] is True:
            if cPoint[1] not in newInts:
                newInts.append(cPoint[1])
            if len(newInts) < 2:
                newInts.append(nLine.projection(nBase))
            newInts = sorted(newInts)
            if new.start == cPoint[0]:
                if cPoint[1] == newInts[-1]:
                    posMatch.append([[new.end, newInts[0]]])
                else:
                    posMatch.append([[new.end, newInts[-1]]])
            elif new.end == cPoint[0]:
                if cPoint[1] == newInts[0]:
                    posMatch.append([[new.start, newInts[-1]]])
                else:
                    posMatch.append([[new.start, newInts[0]]])
            else:
                posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
        else:
            newInts = sorted(newInts)
            if len(newInts) < 2:
                newInts.append(nLine.projection(nBase))
                if len(newInts) < 2:
                    newInts.append(cPoint[1])
            posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
            
        if anchorVect is None:
            sa = -sa
            d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
            nLine = line2D(new.sort, cPoint[1], cPoint[1].add(d2))
            if new.start == cPoint[0]:
                posMatch.append([[new.end, nLine.projection(nBase)]])
            elif new.end == cPoint[0]:
                posMatch.append([[new.start, nLine.projection(nBase)]])
            else:
                posMatch.append([[new.start, cPoint[1]],[new.end, nLine.projection(nBase)]])
        return posMatch
        
    def solvePointPointDistance(self, tfSet, bounds):
        distance = bounds[0]
        base = bounds[1].ind
        new = bounds[2].ind
        graphTransf = tfSet.retrieveTransform(graphTransform)
        
        nCline = None
        for i in self.nodeList:
            if i.nodeElement.containsPoint(new.value):
                cline = i.forms[0]
                nCline = cline.edit(tfSet)
                break
        if nCline is None:
            return []
        
        proj = cline.projection(base.value)
        if proj > new.value:
            distance = -distance
        if nCline.reverse:
            distance = -distance
        
        nBase = graphTransf.apply(base.value)
        nProj = nCline.projection(nBase)
        nProsBaseDist = nProj.subtract(nBase)
        try:
            dProjP = (distance**2-(nProsBaseDist.dotproduct(nProsBaseDist))).sqrt()
        except ValueError:
            return []
        
        if distance < 0:
            dProjP = -dProjP
            
        nPoint = nProj.add(nCline.unitDir.scale(dProjP))
        
        return [[[new.value, nPoint]]]
        
    def readjustLength(self, tfSet, bounds, lhs, rhs):
        length = bounds[0]
        line = bounds[1].ind
        graphTransf = tfSet.retrieveTransform(graphTransform)
        
        adjust = None
        rhsCheckE = [False]
        rhs.accept(self, direction = ['containsCheck', line.end, rhsCheckE, line])
        if rhsCheckE[0] is False:
            adjust = [line.end, line.start]
        else:
            rhsCheckS = [False]
            rhs.accept(self, direction = ['containsCheck', line.start, rhsCheckS, line])
        if adjust is not None:
            pass
        elif rhsCheckS[0] is False:
            adjust = [line.start, line.end]
        else:
            lhsCheckE = [False]
            lhs.accept(self, direction = ['containsCheck', line.end, lhsCheckE, line])
            lhsCheckS = [False]
            lhs.accept(self, direction = ['containsCheck', line.start, lhsCheckS, line])
            if lhsCheckE[0] is False and lhsCheckS[0] is False:
                for i in self.directives:
                    lhsCheckE = False
                    lhsCheckS = False
                    if i == 'length' or i == 'point_on_line':
                        continue
                    for j in self.directives[i]:
                        ind = j[1]
                        if 'lineSegment' in ind.type:
                            if line.end == ind.start or line.end == ind.end:
                                lhsCheckE = True
                            if line.start == ind.start or line.start == ind.end:
                                lhsCheckS = True
                        elif 'point' in ind.type:
                            if line.end == ind.value:
                                lhsCheckE = True
                            if line.start == ind.value:
                                lhsCheckS = True
                if lhsCheckE and lhsCheckS:
                    raise ValueError('cannot apply directive "Length" onto line '+str(line))
                elif lhsCheckE is False:
                    adjust = [line.end, line.start]
                else:
                    adjust = [line.start, line.end]
            elif lhsCheckE[0] is False:
                adjust = [line.end, line.start]
            elif lhsCheckS[0] is False:
                adjust = [line.start, line.end]
            else:
                raise ValueError('cannot apply directive "Length" onto line '+str(line))
        
        new0 =  graphTransf.apply(adjust[0])
        new1 =  graphTransf.apply(adjust[1])
        d0 = new0.subtract(new1)
        newPoint = new1.add(d0.scale(length/d0.length()))
        #print(adjust[0], new0, new1)
        return [[[adjust[0], newPoint]]]
        
