# -*- coding: utf-8 -*-
"""
Created on Thu Oct 12 09:56:47 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str


import itertools

from .mapping import mapping
from .graphNode import graphNode3D

"""
from copy import copy
from random import randint

from .paramSet import paramSet

from ..transform.graphTransform import graphTransform

from ..ind.description import description
from ..ind.line import line2D
from ..ind.planeSegment import planeSegment2D
from ..ind.planeSegment3D import planeSegment3D

from ..struct.planeStruct import bSegment, boundary
from ..struct.coordinate import coordinate, c0, c1, pi, cos, sin
from ..struct.vector import vector2D
"""



ANY = 0
PERPENDICULAR = 1
PARALLEL = 2
SKEW = 3
EMBEDDED = 4

OUTSIDE = 0
BOUND = 1
INSIDE = 2
STRICT_BOUND = 3

NONE = '[None]'
"""
ANY = '~'
PARALLEL = '='
PERPENDICULAR = '+'

OUTSIDE = '0'
BOUND = 'b'
INSIDE = 'i'
STRICT_BOUND = 's'

NONE = '[--------]'
"""



class graphMap3D(mapping):
    type = 'map-graphMap'    
    
    def __init__(self):
        self.nodeList = graphNode3D()
        self.matrices = [[],[],[],[]]
        self.matchingNodeList = graphNode3D()
        self.matchingMatrices = [[],[],[],[]]
        self.postAnalysisProcess = True
        
        self.predicates = {}
        self.directives = {}
        self.postCheck = False

    def clear(self):
        self.matchingNodeList.clear()
        self.matchingMatrices = [[],[],[],[]]
        
    @staticmethod
    def getInstance():
        return graphMap3D()
        
    def addPredicate(self, cond, keyList):
        cond = cond.lower()
        if cond in self.predicates:
            self.predicates[cond] += keyList
            return
        self.predicates[cond] = keyList
        
    def addDirective(self, cond, keyList):
        cond = cond.lower()
        if cond in self.directives:
            self.directives[cond] += keyList
            return
        self.directives[cond] = keyList
    
    def generateNodes(self, ind, matching = False):
        if matching:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices
        else:
            nodeList = self.nodeList
            matrices = self.matrices

        if ind.type == 'ind-pointP3D':
            nodeList.addPoint(ind)
            return nodeList
        elif ind.type == 'ind-planeSegmentP3D':
            element = ind.parametricNode()
            index = nodeList.addPlane(element, ind)
            self.updatePlaneMatrix(element, nodeList, matrices, index)
            self.updatePlaneLineMatrix(element, None, nodeList, matrices, index, None)
            return nodeList
        elif ind.type == 'ind-lineSegmentP3D':
            element = ind.parametricNode()
            
            index, newplanes = nodeList.addLine(element, ind)
            self.updatePlaneLineMatrix(None, element, nodeList, matrices, None, index)
            if len(newplanes) > 0:
                for i in newplanes:
                    pIndex = nodeList.addPlane(i)
                    self.updatePlaneMatrix(i, nodeList, matrices, pIndex)
                    self.updatePlaneLineMatrix(i, None, nodeList, matrices, pIndex, None)
            self.updateLineMatrix(element, nodeList, matrices, index)
        return nodeList
        
    def updatePlaneMatrix(self, element, nodeList, matrices, index):
        planeMatrix = matrices[0]
        if index is True:
            planeMatrix.append([None]*len(nodeList.planes))
            for j in range(len(nodeList.planes)-1):
                comparand = nodeList.planes[j].nodeElement
                edge = ANY
                if comparand.parallel(element):
                    edge = PARALLEL
                elif comparand.perpendicular(element):
                    edge = PERPENDICULAR
                planeMatrix[j].append(edge)
                planeMatrix[-1][j] = edge
                
        """
                
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                if edge != PARALLEL:
                    intersection = comparand.intersection(element)
                    newIntersection = graphNode3D.intersection(intersection)
                    nodeList.planes[-1].addIntersection(newIntersection)
                    nodeList.planes[j].addIntersection(newIntersection)
                    for k in nodeList.planes[-1].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.planes[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            break
                planeMatrix[j].append([edge, positionOne])
                planeMatrix[-1][j] = [edge, positionTwo]
                """
        """
        else:
            for j in range(len(nodeList.planes)):
                if j == index:
                    continue
                if planeMatrix[j][index][0] != PARALLEL:
                    
                    positionOne = OUTSIDE
                    positionTwo = OUTSIDE
                    comparand = nodeList.planes[j].nodeElement
                    intersection = comparand.intersection(element)
                    for k in nodeList.planes[index].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.planes[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            break
                    
                    planeMatrix[j][index][1] = positionOne
                    planeMatrix[index][j][1] = positionTwo
                """
        
    def updatePlaneLineMatrix(self, plane, line, nodeList, matrices, planeIndex, lineIndex):
        planeLineMatrix = matrices[1]
        linePlaneMatrix = matrices[2]
        
        if line is not None:
            if lineIndex is True:
                linePlaneMatrix.append([None]*len(nodeList.planes))
                for j in range(len(nodeList.planes)):
                    comparand = nodeList.planes[j].nodeElement
                    edge = ANY
                    if comparand.parallelLine(line):
                        edge = PARALLEL
                    elif comparand.perpendicularLine(line):
                        edge = PERPENDICULAR
                        
                    positionOne = OUTSIDE
                    positionTwo = OUTSIDE
    
                    if edge < PARALLEL:
                        intersection = comparand.intersectionLine(line)
                        newIntersection = graphNode3D.lineIntersection(intersection)
                        nodeList.lines[-1].addIntersection(newIntersection)
                        for k in nodeList.lines[-1].forms:
                            if k.start == intersection or k.end == intersection:
                                if positionOne == BOUND:
                                    positionOne = INSIDE
                                    break
                                else:
                                    positionOne = BOUND
                            elif k.containsPoint(intersection):
                                positionOne = INSIDE
                                break
                        for k in nodeList.planes[j].forms:
                            if k.containsPoint(intersection):
                                positionTwo = INSIDE
                                break

                    planeLineMatrix[j].append([edge, positionTwo])
                    linePlaneMatrix[-1][j] = [edge, positionOne]
                    
            else:
                for j in range(len(nodeList.planes)):
                    if planeLineMatrix[j][lineIndex][0] < PARALLEL:
                        
                        positionOne = OUTSIDE
                        positionTwo = OUTSIDE
        
                        if edge < PARALLEL:
                            intersection = comparand.intersectionLine(line)
                            newIntersection = graphNode3D.lineIntersection(intersection)
                            nodeList.lines[-1].addIntersection(newIntersection)
                            for k in nodeList.lines[-1].forms:
                                if k.start == intersection or k.end == intersection:
                                    if positionOne == BOUND:
                                        positionOne = INSIDE
                                        break
                                    else:
                                        positionOne = BOUND
                                elif k.containsPoint(intersection):
                                    positionOne = INSIDE
                                    break
                            for k in nodeList.planes[j].forms:
                                if k.containsPoint(intersection):
                                    positionTwo = INSIDE
                                    break
                            
                        planeLineMatrix[j][lineIndex][1] = positionTwo
                        linePlaneMatrix[lineIndex][j][1] = positionOne
                        
        if plane is not None:
            if planeIndex is True:
                planeLineMatrix.append([None]*len(nodeList.lines))
                for j in range(len(nodeList.lines)):
                    comparand = nodeList.lines[j].nodeElement
                    edge = ANY
                    if plane.parallelLine(comparand):
                        edge = PARALLEL
                    elif plane.perpendicularLine(comparand):
                        edge = PERPENDICULAR
                        
                    
                    positionOne = OUTSIDE
                    positionTwo = OUTSIDE
    
                    if edge != PARALLEL:
                        intersection = plane.intersectionLine(comparand)
                        newIntersection = graphNode3D.lineIntersection(intersection)
                        nodeList.planes[-1].addIntersection(newIntersection)
                        for k in nodeList.planes[-1].forms:
                            if k.containsPoint(intersection):
                                positionOne = INSIDE
                                break
                        for k in nodeList.lines[j].forms:
                            if k.start == intersection or k.end == intersection:
                                if positionTwo == BOUND:
                                    positionTwo = INSIDE
                                    break
                                else:
                                    positionTwo = BOUND
                            elif k.containsPoint(intersection):
                                positionTwo = INSIDE
                                break
                    linePlaneMatrix[j].append([edge, positionTwo])
                    planeLineMatrix[-1][j] = [edge, positionOne]
                    
            else:
                for j in range(len(nodeList.lines)):
                    if linePlaneMatrix[j][planeIndex][0] < PARALLEL:
                        positionOne = OUTSIDE
                        positionTwo = OUTSIDE
                        comparand = nodeList.lines[j].nodeElement
                        intersection = plane.intersectionLine(comparand)
                        for k in nodeList.planes[planeIndex].forms:
                            if k.containsPoint(intersection):
                                positionOne = INSIDE
                                break
                        for k in nodeList.lines[j].forms:
                            if k.start == intersection or k.end == intersection:
                                if positionTwo == BOUND:
                                    positionTwo = INSIDE
                                    break
                                else:
                                    positionTwo = BOUND
                            elif k.containsPoint(intersection):
                                positionTwo = INSIDE
                                break
                        
                            
                        linePlaneMatrix[j][planeIndex][1] = positionTwo
                        planeLineMatrix[planeIndex][j][1] = positionOne

    
    def updateLineMatrix(self, element, nodeList, matrices, index):
        lineMatrix = matrices[3]
        if index is True:
            lineMatrix.append([None]*len(nodeList.lines))
            
            for j in range(len(nodeList.lines)-1):
                comparand = nodeList.lines[j].nodeElement
                edge = ANY
                if comparand.skew(element):
                    edge = SKEW
                elif comparand.parallel(element):
                    edge = PARALLEL
                elif comparand.perpendicular(element):
                    edge = PERPENDICULAR
                    
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                if edge < PARALLEL:
                    intersection = comparand.intersection(element)
                    newIntersection = graphNode3D.lineIntersection(intersection)
                    nodeList.lines[-1].addIntersection(newIntersection)
                    nodeList.lines[j].addIntersection(newIntersection)
                    for k in nodeList.lines[-1].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            break
                    
                lineMatrix[j].append([edge, positionOne])
                lineMatrix[-1][j] = [edge, positionTwo]
                
        else:
            for j in range(len(nodeList.lines)):
                if j == index:
                    continue
                if lineMatrix[j][index][0] < PARALLEL:
                    
                    positionOne = OUTSIDE
                    positionTwo = OUTSIDE
                    comparand = nodeList.lines[j].nodeElement
                    intersection = comparand.intersection(element)
                    for k in nodeList.lines[index].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            break
                    
                    lineMatrix[j][index][1] = positionOne
                    lineMatrix[index][j][1] = positionTwo
            
            
        
    def map(self, fromInd, toInd):
        self.generateNodes(toInd, matching = True)
        return True
        
    def distinguish(self, ind):
        node = self.generateNodes(ind)
        return node
        
    def potentialTransforms(self):
        
        print('\n1. Plane matrix')
        for j in self.matrices[0]:
            for k in j:
                if k is None:
                    print('X', end='  ')
                else:
                    print (k,end='  ')
            print()
        print('\n2. Plane Line matrix')
        for j in self.matrices[1]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n3. Line Plane matrix')
        for j in self.matrices[2]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n4. Line matrix')
        for j in self.matrices[3]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print()
        
        return []
        
    def detectTransform(self, matchLists):
        for subset in itertools.permutations([k for k in range(len(self.matchingNodeList))], len(self.nodeList)):
            check = True
            ###########################################################
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.edgeMatrices[indI][indJ]
                    matchEdge = self.matchingEdgeMatrices[i][j]
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            ###########################################################
            if check:
                matchLists.append(subset)

        
    def analyzeRHS(self, rhs, shape, transf):
        return []        
        
        
        
        
        
