# -*- coding: utf-8 -*-
"""
Created on Thu Jun 29 10:22:54 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str
from builtins import object

import itertools
from ..form.coForm import coForm
from ..ind.plane import plane3D
from ..sort.sort import sort

CONNECTED = True
DISCONNECTED = False

class graphNode(object):
    
    class auxiliaryObj(object):
        def __init__(self, coord):
            self.value = coord
            self.oType = 'f'
            self.assocLine = None
        
        def __str__(self):
            return self.oType+str(self.value)
            
            
    class intersection(object):
        type = 'graphNode-intersection'
        
        def __init__(self, coord):
            self.value = coord
            self.elms = []
            self.endP = False
            self.iType = 'o'
            
        def __str__(self):
            return str(self.value)
            
        def __repr__(self):
            return str(self.value)
            
        def __eq__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value == other.value
            
        def __ne__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value != other.value
            
        def __lt__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value < other.value
            
        def __gt__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value > other.value
            
        def __le__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value <= other.value
            
        def __ge__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value >= other.value
            
        def addElm(self, elm):
            if elm not in self.elms:
                self.elms.append(elm)
            
    class elm(object):
        def __init__(self, nodeElement, ind):
            self.nodeElement = nodeElement
            self.forms = [ind]
            self.intersections = []
            self.intersectionMatrix = None
            self.endP = False
            self.linePoints = []
            
        def __repr__(self):
            return str(self)
            
        def __str__(self):
            return str(self.nodeElement)+str(self.forms)
            
        def add(self, line):
            for i in range(len(self.forms)):
                if self.forms[i].compareInterval(line) == 1:
                    self.forms.insert(i, line)
                    return
            self.forms.append(line)
            return
            
        def addIntersection(self, intersection):
            intersection.addElm(self)
            for i in range(len(self.intersections)):
                if self.intersections[i] >= intersection:
                    if self.intersections[i] == intersection:
                        return
                    self.intersections.insert(i,intersection)
                    return
            self.intersections.append(intersection)
            
        def addEndPoints(self):
            for indiv in self.forms:
                checkStart = False
                checkEnd = False
                for intsn in self.intersections:
                    if indiv.start == intsn.value:
                        checkStart = True
                    if indiv.end == intsn.value:
                        checkEnd = True
                if not checkStart:
                    start = graphNode.intersection(indiv.start)
                    start.endP = True
                    self.endP = True
                    self.addIntersection(start)
                if not checkEnd:
                    end = graphNode.intersection(indiv.end)
                    end.endP = True
                    self.endP = True
                    self.addIntersection(end)
                    
        
        def classifyIntersections(self):
            self.intersectionMatrix = [[None for i in range(len(self.intersections))] 
                                             for j in range(len(self.intersections))]
            for i in range(len(self.intersections)):
                for j in range(i+1,len(self.intersections)):
                    check = False
                    for line in self.forms:
                        if (line.containsPoint(self.intersections[i].value)
                        and line.containsPoint(self.intersections[j].value)):
                            check = True
                            self.intersectionMatrix[i][j] = CONNECTED
                            self.intersectionMatrix[j][i] = CONNECTED
                            break
                    if not check:
                        self.intersectionMatrix[i][j] = DISCONNECTED
                        self.intersectionMatrix[j][i] = DISCONNECTED
        
        def organizeForm(self):
            if self.forms[0].sort.type != 'sort-attribute':
                return
            i = 0
            while i < len(self.forms):
                ind = self.forms[i]
                j = i+1
                while j < len(self.forms):
                    if ind.end == self.forms[j].start:
                        newline = ind.__class__(ind.sort, ind.start, self.forms[j].end)
                        if ind.attribute == self.forms[j].attribute:
                            newline.addAttribute(ind.attribute)
                            self.forms[i] = newline
                            ind = newline
                            self.forms.pop(j)
                            j-=1
                        else:
                            attr = ind.attribute.duplicate()
                            attr.product(self.forms[j].attribute)

                            if attr is not None:
                                newline.addAttribute(attr)
                            if newline.attribute == ind.attribute:
                                self.forms[i] = newline
                                ind = newline
                            elif newline.attribute == self.forms[j].attribute:
                                self.forms[j] = newline
                            else:
                                check = False
                                for k in self.forms:
                                    if k.contains(newline):
                                        check = True
                                if not check:
                                    self.forms.insert(j+1,newline)
                                    j+=1
                    elif ind.end < self.forms[j].start:
                        break
                    elif ind.attribute == self.forms[j].attribute and not ind.disjoint(self.forms[j]):
                        self.forms[j].combine(ind)
                        self.forms.pop(i)
                        i -= 1
                        break
                    j+=1
                i+=1
            """
            print('\n.....')
            for l in self.forms:
                print('@@@====',l, l.attribute)
            """
            
        def __eq__(self, other):
            if not isinstance(other, graphNode.elm):
                return False
            return self.nodeElement == other.nodeElement
    
    def __init__(self):
        self.lines = []
        self.points = []
        self.planes = []
        
    def __iter__(self):
        return iter(self.lines)
        
    def __len__(self):
        return len(self.lines)
        
    def __repr__(self):
        return str(self)
            
    def __str__(self):
        return str(self.lines)
            
    def addInstance(self, codesc, ind):
        for j in range(len(self.lines)):
            if self.lines[j].nodeElement == codesc:
                self.lines[j].add(ind)
                return j
        newNode = graphNode.elm(codesc, ind)
        self.lines.append(newNode)
        return True
        
    def addPoint(self, p):
        newP = self.auxiliaryObj(p.value)
        self.points.append(newP)
        
    def addPlane(self, p):
        newP = self.auxiliaryObj(p)
        self.planes.append(newP)
        
    def classifyObjs(self, ruleAnalysis = False):
        for p in self.points:
            check = 0
            for node in self.lines:
                if node.nodeElement.containsPoint(p.value):
                    check+=1
                    if not ruleAnalysis:
                        node.linePoints.append(p)
                    elif check == 1:
                        p.assocLine = node
            if check == 1:
                p.oType = 'l'
                if ruleAnalysis:
                    p.assocLine.linePoints.append(p)
            elif check > 1:
                p.oType = 'i'
        if ruleAnalysis:
            index = 0
            while index < len(self.planes):
                plane = self.planes[index]
                plane.oType = 'i'
                for bound in plane.value.boundaries:
                    for p in bound.polygonal:
                        check = 0
                        for node in self.lines:
                            if node.nodeElement.containsPoint(p):
                                check+=1
                                if check == 1:
                                    p.assocLine = node
                        if check == 0:
                            plane.oType = 'f'
                            break
                        if check == 1:
                            plane.oType = 'l'
                    if plane.oType == 'f':
                        break
                if plane.oType == 'i':
                    self.planes.pop(index)
                    index -= 1
                index += 1
            return
        splitPlanes = []
        for plane in self.planes:
            if not plane.value.hasAttribute():
                splitPlanes.append(plane.value)
                continue
            splitAttributes = plane.value.attribute.splitElements()
            splitPlanes.append(plane.value.replicateInd())
            for i in range(1,len(splitAttributes)+1):
                for comp in itertools.combinations(splitAttributes,i):
                    newPlane = plane.value.replicateInd()
                    for att in comp:
                        newPlane.addAttribute(att)
                    splitPlanes.append(newPlane)
                    
        combinedPlanes = []
        while len(splitPlanes)!=0:
            plane1 = splitPlanes.pop(0)
            
            index = 0
            while index < len(combinedPlanes):
                plane = combinedPlanes[index]
                if (not plane1.sort == plane.sort) or (not plane1.attribute == plane.attribute):
                    index += 1
                    continue
                classes = plane1.classify(plane)
                if not classes.disjoint():
                    resultPlane = coForm(plane.sort)
                    plane.construct(classes.sumSet(), resultPlane)
                    plane1 = resultPlane.elements[0]
                    if plane.hasAttribute():
                        plane1.addAttribute(plane.attribute)
                    combinedPlanes.remove(plane)
                    index -= 1
                index += 1
            
            index = 0
            while index < len(splitPlanes):
                plane = splitPlanes[index]
                
                if (not plane1.sort == plane.sort) or (not plane1.attribute == plane.attribute):
                    index += 1
                    continue
                classes = plane1.classify(plane)
                
                if not classes.disjoint():
                    resultPlane = coForm(plane.sort)
                    plane.construct(classes.sumSet(), resultPlane)
                    plane1 = resultPlane.elements[0]
                    if plane.hasAttribute():
                        plane1.addAttribute(plane.attribute)
                    splitPlanes.remove(plane)
                    index -= 1
                index += 1
            combinedPlanes.append(plane1)
        self.planes = []
        for i in combinedPlanes:
            self.addPlane(i)

    def clear(self):
        self.lines = []
        self.points = []
        self.planes = []
        


class graphNode3D(object):
    
    class auxiliaryObj(object):
        def __init__(self, coord):
            self.value = coord
            self.oType = 'f'
            self.assocLine = None
        
        def __str__(self):
            return self.oType+str(self.value)
            
            
    class lineIntersection(object):
        type = ''
        def __init__(self, coord):
            self.value = coord
            self.assocLines = []
            self.endP = False
            
        def __str__(self):
            return str(self.value)
            
        def __repr__(self):
            return str(self.value)
            
        def __eq__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value == other.value
            
        def __ne__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value != other.value
            
        def __lt__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value < other.value
            
        def __gt__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value > other.value
            
        def __le__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value <= other.value
            
        def __ge__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value >= other.value
            
        def addAssocLine(self, line):
            if line not in self.assocLines:
                self.assocLines.append(line)
            
    class maxline(object):
        def __init__(self, nodeElement, ind):
            self.nodeElement = nodeElement
            self.forms = [ind]
            self.intersections = []
            self.intersectionMatrix = None
            self.endP = False
            self.linePoints = []
            
        def __repr__(self):
            return str(self)
            
        def __str__(self):
            return str(self.nodeElement)+str(self.forms)
            
        def add(self, line):
            for i in range(len(self.forms)):
                if self.forms[i].compareInterval(line) == 1:
                    self.forms.insert(i, line)
                    return
            self.forms.append(line)
            return
            
        def addIntersection(self, intersection):
            intersection.addAssocLine(self)
            for i in range(len(self.intersections)):
                if self.intersections[i] >= intersection:
                    if self.intersections[i] == intersection:
                        return
                    self.intersections.insert(i,intersection)
                    return
            self.intersections.append(intersection)
            
        def addEndPoints(self):
            for indiv in self.forms:
                checkStart = False
                checkEnd = False
                for intsn in self.intersections:
                    if indiv.start == intsn.value:
                        checkStart = True
                    if indiv.end == intsn.value:
                        checkEnd = True
                if not checkStart:
                    start = graphNode3D.lineIntersection(indiv.start)
                    start.endP = True
                    start.iType = 'i'
                    self.endP = True
                    self.addIntersection(start)
                if not checkEnd:
                    end = graphNode3D.lineIntersection(indiv.end)
                    end.endP = True
                    end.iType = 'i'
                    self.endP = True
                    self.addIntersection(end)
                    
        
        def classifyIntersections(self):
            self.intersectionMatrix = [[None for i in range(len(self.intersections))] 
                                             for j in range(len(self.intersections))]
            for i in range(len(self.intersections)):
                for j in range(i+1,len(self.intersections)):
                    check = False
                    for line in self.forms:
                        if (line.containsPoint(self.intersections[i].value)
                        and line.containsPoint(self.intersections[j].value)):
                            check = True
                            self.intersectionMatrix[i][j] = CONNECTED
                            self.intersectionMatrix[j][i] = CONNECTED
                            break
                    if not check:
                        self.intersectionMatrix[i][j] = DISCONNECTED
                        self.intersectionMatrix[j][i] = DISCONNECTED
        
        def organizeForm(self):
            if self.forms[0].sort.type != 'sort-attribute':
                return
            i = 0
            while i < len(self.forms):
                ind = self.forms[i]
                j = i+1
                while j < len(self.forms):
                    if ind.end == self.forms[j].start:
                        newline = ind.__class__(ind.sort, ind.start, self.forms[j].end)
                        if ind.attribute == self.forms[j].attribute:
                            newline.addAttribute(ind.attribute)
                            self.forms[i] = newline
                            ind = newline
                            self.forms.pop(j)
                            j-=1
                        else:
                            attr = ind.attribute.duplicate()
                            attr.product(self.forms[j].attribute)

                            if attr is not None:
                                newline.addAttribute(attr)
                            if newline.attribute == ind.attribute:
                                self.forms[i] = newline
                                ind = newline
                            elif newline.attribute == self.forms[j].attribute:
                                self.forms[j] = newline
                            else:
                                check = False
                                for k in self.forms:
                                    if k.contains(newline):
                                        check = True
                                if not check:
                                    self.forms.insert(j+1,newline)
                                    j+=1
                    elif ind.end < self.forms[j].start:
                        break
                    elif ind.attribute == self.forms[j].attribute and not ind.disjoint(self.forms[j]):
                        self.forms[j].combine(ind)
                        self.forms.pop(i)
                        i -= 1
                        break
                    j+=1
                i+=1
            
        def __eq__(self, other):
            if not isinstance(other, graphNode3D.maxline):
                return False
            return self.nodeElement == other.nodeElement
            
            
    class maxplane(object):
        def __init__(self, nodeElement, ind = None):
            self.nodeElement = nodeElement
            self.forms = []
            if ind is not None:
                self.forms.append(ind)
            self.intersections = []
            self.intersectionMatrix = None
            self.endP = False
            self.linePoints = []
            
        def __repr__(self):
            return str(self)
            
        def __str__(self):
            return str(self.nodeElement)+str(self.forms)
            
        def add(self, plane):
            pass
        
        def addLine(self, line):
            pass
            
        def addIntersection(self, intersection):
            pass
                                
        def classifyIntersections(self):
            pass
        
        def organizeForm(self):
            pass
            
        def __eq__(self, other):
            if not isinstance(other, graphNode3D.maxline):
                return False
            return self.nodeElement == other.nodeElement
    
    
    def __init__(self):
        self.lines = []
        self.points = []
        self.planes = []
        
    def __iter__(self):
        return iter(self.planes)
        
    def __len__(self):
        return len(self.planes)
        
    def __repr__(self):
        return str(self)
            
    def __str__(self):
        return str(self.planes)+str(self.lines)
        
    def addPlane(self, codesc, ind = None):
        if ind is None:
            for j in range(len(self.planes)):
                if self.planes[j].nodeElement == codesc:
                    return j
            newPlaneNode = graphNode3D.maxplane(codesc)
            self.planes.append(newPlaneNode)
            return True
        for j in range(len(self.lines)):
            if self.lines[j].nodeElement == codesc:
                self.lines[j].add(ind)
                return j
        newNode = graphNode3D.maxplane(codesc, ind)
        self.lines.append(newNode)
        return True
            
    def addLine(self, codesc, ind):
        for j in range(len(self.lines)):
            if self.lines[j].nodeElement == codesc:
                self.lines[j].add(ind)
                return j
        newLineNode = graphNode3D.maxline(codesc, ind)
        p1 ,p2 = ind.start, ind.end
        
        newplanes = []
        for j in range(len(self.lines)):
            if self.lines[j].nodeElement.skew(codesc):
                continue
            p3 = self.lines[j].forms[0].start
            if p3 == p1 or p3 == p2:
                p3 = self.lines[j].forms[0].end
                
            dummyPlaneSort3D = sort.getSort('__dummyPlane3D__')
            plane = plane3D(dummyPlaneSort3D, [p1,p2,p3])
            newplanes.append(plane)
        for planeNode in self.planes:
            planeNode.addLine(newLineNode)
        self.lines.append(newLineNode)
        return True, newplanes
        
    def addPoint(self, p):
        newP = self.auxiliaryObj(p.value)
        self.points.append(newP)
        
    def classifyObjs(self, ruleAnalysis = False):
        for p in self.points:
            check = 0
            for node in self.lines:
                if node.nodeElement.containsPoint(p.value):
                    check+=1
                    if not ruleAnalysis:
                        node.linePoints.append(p)
                    elif check == 1:
                        p.assocLine = node
            if check == 1:
                p.oType = 'l'
                if ruleAnalysis:
                    p.assocLine.linePoints.append(p)
            elif check > 1:
                p.oType = 'i'

    def clear(self):
        self.lines = []
        self.points = []
        self.planes = []
        

