# -*- coding: utf-8 -*-
"""
@author: akidribm
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import Rhino
import System

# Python built-ins
from builtins import range

# Sortal Library Imports
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  
                    vector3D, vector2D, planeSegment3D, sort)
                    
# Import for transformations
from mpmath import pi, sin, cos
from copy import deepcopy

# Conversion Imports
from convert.extract import extract
from convert.convert_rhino_to_agnostic import convertR
from sortalgi.sortal_lib_api.convert.convert_agnostic_to_sortal import convertA

# Sortal API Imports
import sortalgi as sgi

# API Initialization
convertR = convertR()
convertA = convertA()
extract = extract()

"""
===========
SCRIPT INFO
===========

This script is run after 'set_up_all.py', and creates a rule by prompting 
for rule name, rule description, and selecting frames (LHS and RHS). 
When running this code, click the green 'Play' button. Do not reset the engine as it will empty the sort register.

Created rules are stored in the sortal library's rule register ('rule.register'), 
a dictionary where rule instances are stored under the keys corresponding
to their rule names.
"""

# Prompts for rule name
ruleName = raw_input('Enter name of rule')

# If rule name already exists in rule.register, then prompts for overwrting pre-existing rule instance/new name
if ruleName in rule.register:
    print ('A rule with the same name already exists. Overwrite pre-existing rule?')
    ans = raw_input('Enter y for yes, n for no')
    if ans == 'y':
        del rule.register[ruleName]
    else:
        while True:
            ruleName = raw_input('Enter new name of rule')
            if ruleName not in rule.register:
                break
            else:
                print ('The rule name %s has already been used for a previous rule. Please enter a different name for the new rule.'%ruleName)

# Prompts for frames (block instances) that contain the LHS, RHS
# Select each frame by clicking on them with the mouse cursor
# Input: Prompting Text; this can be empty
# Returns: List of Rhino GUIDs, reference point of frame block instance, name of layer where objects are located
lhs, refPointL, layerNameL = extract.getIndividuals('Select the frame containing the LHS')
rhs, refPointR, layerNameR = extract.getIndividuals('Select the frame containing the RHS')

# Sorts through and converts list of Rhino GUIDs to agnostic tuple form, adjusts coordinates using frame insertion point as reference point
# Input: list of Rhino GUIDs, reference point (default point is (0,0,0) if no reference point is given)
# Output: agnostic tuple form of list of Rhino GUIDs, list of Rhino GUIDs that correspond to the agnostic tuple
lhsAgnostic, lhsObjects = convertR.toAgnostic(lhs, refPointL)
rhsAgnostic, rhsObjects = convertR.toAgnostic(rhs, refPointR)

# Prompts for description of rule
ruleDesc = raw_input('Enter rule description')   

# Creates new rule
# The code for this function can be found in api -> agnostic_api -> create_rule
# Input: rule name, rule description (string), agnostic tuple of LHS, agnostic tuple of rhs
# Output: new rule instance, rule name of new rule instance
# Note: The new rule instance created is stored in the sortal library's rule register under the key corresponding to its rule name
ruleNew, ruleName = sgi.create_rule(ruleName, ruleDesc, lhsAgnostic, rhsAgnostic)
    
# To check if your rule exists in the sortal library register
"""
print ('%s has been created'%ruleName)
print (lhsAgnostic)
print ('\n')
print (rhsAgnostic)
"""
