from visualAgnostic import visualA as visualA
from rhino_specific_attributes import rhino_methods as rm

"""
=============
FUNCTION INFO
=============
Deletes previous main shape with function 'clearShape'  
Redraws new main shape with the class visualA's method, 'plotAgnostic'
The specific function for drawing a form, visuaA(); it can be found in the file 'visualAgnostic.py'

INPUT 
agnostic form of main shape after rule application
layerNameMain - layer where main shape is drawn
refPointName - insertion point of frame instance of main shape

OUTPUT
Rhino GUIDs of new shape drawing
"""

def draw_result(shapeIds, choice, ruleAppns, layerNameMain, refPointMain):
    if len(ruleAppns) != 0:
        # CLEARING SHAPE
        # Clears previous shape IDs inside frame; input is a list of GUIDs
        rm.clearShape(shapeIds)                                            
        
        # PLOTTING SHAPE
        # Assigns which agnostic tuple (in this case, the main shape after rule application) to plot
        shapeNew = ruleAppns[choice][3]     

        # Specific function for plotting is 'visualA', housed inside 'visualAgnostic'
        # To plot a form, the function visualA() takes the following inputs:
        #      (<agnostic form to be plotted>, <layer name where agnostic form is to be plotted>, <reference point)
        # The method visualA.plotAgnostic() is called to plot the form
        vis1 = visualA(shapeNew, layerNameMain, refPointMain)           # Input: element of ruleAppns; creates an instance of visualA
        shapeNewAgnosticIds = vis1.plotAgnostic()                       # Redraws shape after rule application based on agnostic tuple input 'shapeNew'  

        return shapeNewAgnosticIds
    else:
        print ('No rule applications can be found. Ending run')
        return
