# -*- coding: utf-8 -*-
"""
@author: akidribm
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs

# Python built-ins
from builtins import range

# Sortal Library Imports
import builtins
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  
                    vector3D, vector2D, planeSegment3D, sort, rule)

# Import for transformations
from mpmath import pi, sin, cos
from copy import deepcopy

# API imports
import sortalgi as sgi
from setup.grammar import Grammar as g

"""
===========
SCRIPT INFO
===========

'set_up_all.py' has the following functions:
    1. Setting up empty grammar frames for initial shape and rule; this can be commented out if one wishes 
       to use a pre-existing .3dm grammar file
    2. Initializing the sort types in the sortal library; this is necessary for rule creation and application;
       this MUST BE RUN FIRST before doing anything else with the sortal library API

===================    
RUNNING THE SCRIPT:
===================
    To run this script, click on the small downward pointing arrow next to the green 'Play' button and select 'Reset engine and Debug'.
    This empties the sort register and rule register, and initializes new sort types.
"""

# GRAMMAR SET UP
# Sets up empty frames: initial shape, LHS and RHS (indicated by an arrow in between them)
# Add hash to turn off this line of code
#g.set_up_grammar() 

# SORT TYPE SET UP
# The function used to set up the sortal grammar interpreter is a method called 'sortal_setup' inside the package sortalgi
# It is necessary to run sgi.sortal_setup() before doing anything else with the sortal library and its API
# It takes as input a single tuple with two sub-tuples as elements
# It creates disjunctive sort type  based on user input
# This example enables:
#       Index 0: Points with label, description, weight, weight255, enumerative, color
#       Index 1: Line segments with label, description, weight, weight255, enumerative, color
# However, note that only weight 255 (as grayscale) is visualized for points, and only weight (line width as viewed in Print Preview), are visualized in this demo 
# The function returns the Boolean value True which can be used to check if the setup was successful

if sgi.sortal_setup(((True, True, True, True, True, True, True), (True, True, True, True, True, True, True))):
    print ('Setup complete')

# A simpler way of initializing sort types:
# sgi.sortal_setup((True, True), (True, False, True, False, True))
# This example enables:
#       Points with label
#       Line segments with description, weight255
# If the inputs for a geometric object's attributes are consecutive False values, they do not need to be inputted
# However, if there is a False value followed by a True value, it needs to be written explicitly, as in the line segment section

# For example, to enable a line segment with color as its only attribute, the tuple of True/False values needs to be: (True, False, False, False, False, False, True)
# If that is the only object to be enabled, then the function call would be as follows:
# sgi.sortal_setup((False,), (True, False, True, False, True))
# This enables only line segments with color, but no points; do take note that the False value for point should be enclosed by parentheses
# and followed by a comma, so that it is correctly read as a tuple

                                                     