#-*- coding: utf-8 -*-
"""
@author: akidribm
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs

# Python built-ins
from builtins import range

# Sortal Library Imports
from sortal import (transformSet, rule, primitiveSort, point3D, point2D, coordinate, lineSegment3D, planeSegment2D, 
                    lineSegment2D, line3D, label, rWeight, disjunctiveSort, attributeSort, affineTransform,  
                    vector3D, vector2D, planeSegment3D, sort)
                    
# Import for transformations
from mpmath import pi, sin, cos
from copy import deepcopy

# Conversion Imports
from convert.convert_rhino_to_agnostic import convertR
from sortalgi.sortal_lib_api.convert.convert_sortal_to_agnostic import convertS

# API Imports
import sortalgi as sgi
from rhino_specific.visualAgnostic import visualA as visualA
from rhino_specific.rhino_specific_attributes import rhino_methods as rm
from convert.extract import extract

convertR = convertR()
convertS = convertS()
extract = extract()

"""
=====================
Function: maximalize()
=====================

This function maximalizes a shape.
Input: list of Rhino GUIDs
Output: maximalized agnostic tuple form of shape
"""

shapeRhino = rs.GetObjects('Select shape to maximalize')
shapeLayer = rs.ObjectLayer(shapeRhino[0])
shapeAgnostic, shapeObjects = convertR.toAgnostic(shapeRhino)
objectsMaximalized =  sgi.maximalize(shapeAgnostic)

rm.clearShape(shapeObjects)
vis1 = visualA(objectsMaximalized, shapeLayer)
vis1.plotAgnostic()


"""
==================
Function: partOf()
==================

This function checks if a subshape is part of a larger shape.
Input: agnostic tuple forms of subshape and shape
Output: Boolean value True/False
"""
"""
obj1 = rs.GetObjects('Select subshape')
obj2 = rs.GetObjects('Select main shape')
obj1Agnostic, obj1Objects = convertR.toAgnostic(obj1, (0,0,0))
obj2Agnostic, obj2Objects = convertR.toAgnostic(obj2, (0,0,0))
print ('Is subshape part of shape?', sgi.partOf(obj1Agnostic, obj2Agnostic))
"""

"""
=====================
Function: getRuleName(), getDesc(), getLHS(), getRHS()
=====================

These functions return a specific aspect of a rule object, as per the name.
The rule sides are returned in the form of agnostic tuples.
Input: rule object
Output: string or agnostic tuple
"""
"""
ruleName = 'rule_1'
ruleObj = sgi.getRule(ruleName)
print ('Name:', sgi.getRuleName(ruleObj))
print ('Description:', sgi.getRuleDesc(ruleObj))
print ('LHS:', sgi.getLHS(ruleObj))
print ('RHS:', sgi.getRHS(ruleObj))
"""

"""
=====================
Function: setRuleName(), setDesc()
=====================

These functions change the rule name/rule description
The altered rule object is returned.
Input: rule object, name/description as string
Output: altered rule object
"""
"""
# Gets name of first rule present in register
for i in rule.register:
    ruleName = i
    break
    
ruleObj = sgi.getRule(ruleName)

# Shows original rule name and new rule name
print ('Original Rule Name: ', sgi.getRuleName(ruleObj))
newRuleName = 'rule_new'
sgi.setRuleName(ruleObj, newRuleName)
print ('New Rule Name: ', sgi.getRuleName(ruleObj))

# Shows original rule description and new rule description
print ('Original Rule Description:', sgi.getRuleDesc(ruleObj))
newRuleDesc = 'new rule desc'
sgi.setRuleDesc(ruleObj, newRuleDesc)
print ('New Rule Description: ', sgi.getRuleDesc(ruleObj)) 
"""

"""
=====================
Function: overwriteRule()
=====================

This function changes the shapes of a rule object, while maintaining the rule name and the rule description.
The altered rule object is returned.
Input: rule object, LHS and RHS (agnostic objects)
Output: altered rule object
"""
"""
ruleName = 'rule_1'
ruleObj = sgi.getRule(ruleName)

# Obtains new LHS agnostic object
lhs, refPointL, layerNameL = extract.getIndividuals('Select the frame containing the new LHS')
lhsAgnostic, lhsObjects = convertR.toAgnostic(lhs, refPointL)

# Shows original rule sides
print ('Name:', sgi.getRuleName(ruleObj))
print ('Description:', sgi.getRuleDesc(ruleObj))
print ('Original LHS:', sgi.getLHS(ruleObj))
print ('Original RHS:', sgi.getRHS(ruleObj))

rhsAgnostic = sgi.getRHS(ruleObj)

# Changes LHS of ruleObj to new LHS agnostic object
newRule = sgi.overwriteRule(ruleObj, lhsAgnostic, rhsAgnostic)

# Shows altered rule sides
print ('\nAltered rule')
print ('Name:', sgi.getRuleName(newRule))
print ('Description:', sgi.getRuleDesc(newRule))
print ('New LHS:', sgi.getLHS(newRule))
print ('RHS:', sgi.getRHS(newRule))
"""
