from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import


from .ruleAnalysis import ruleAnalysis
from .ruleApplication import ruleApplication
from ..struct.planeStruct import PlaneTransformationError
from ..map.similarityMap import similarityMap2D
from ..map.similarityMap3D import similarityMap3D
from ..map.parametricMap import parametricMap
from ..map.exactMap import exactMap
from ..transform.paramSet import paramSet
from ..transform.transformSet import transformSet
from ..ind.description import description
from ..ind.descriptions.tree import DescriptionEvaluatingError
from ..visit.elementVisitor import elementVisitor

class detection(elementVisitor):
    applicationDict = {}
    
    def __init__(self, rule):
        self.rule = rule
        self.applicationDict[rule.name] = []
        ###############################################
        self.rule.lhs.maximalize()
        self.analysis= ruleAnalysis(self.rule.lhs)
        """
        if not self.isValid():
            warnings.warn("\nrule "+self.rule.name+" does not meet the minimum requirement for similarity transformation\n")
        """
        ###############################################
        
        self.form = None
        self.target = None
        self.uptrail = []
        self.downTrail = []
        self.success = False
        self.applications = []
        
    def isValid(self):
        for i in self.analysis.maps:
            if not self.analysis.maps[i].isValid():
                return False
        return True
        
    def count(self):
        return len(self.applications)
        
        
    def getApplication(self, index = 0):
        if index >= len(self.applications):
            raise IndexError
        return self.applications[index]
        
    def matches(self):
        matches = []
        for i in self.applications:
            match = i.matchForm
            if len(matches) == 0 or match not in matches:
                matches.append(match)
        return matches
    
    def results(self, match):
        r = []
        for i in self.applications:
            if match is i.matchForm or match == i.matchForm:
                r.append(i.resultForm)
        return r
        
    #..................
    def detect(self, form, printing = None):
        self.clear()
        ###########################
        self.form = form
        ###########################
        for i in self.applicationDict[self.rule.name]:
            if i[0] == form:
                    
                self.applications = i[1]
                if printing:
                    print('---------------------------------------------\nExisting matches found:', len(i[1]),'results')
                return i[1]

        self.analysis.unsetVisitedAll()
        fragments = self.analysis.fragments
        
        for i in fragments:
            self.target = i
            self.uptrail= []
            if not self.target.content.visited:
                visit = True
                self.uptrail.append(self.target)
                while(self.uptrail[-1].parent is not None):
                    self.uptrail.append(self.uptrail[-1].parent)
                    if self.uptrail[-1].content.visited:
                        visit = False
                if visit:
                    self.form.accept(self)
                    self.target.content.visited = True
                    if self.target.parent is not None:
                        self.target.parent.content.visited = True
        resultForms = []   
        
        if printing:
            print('==================================================================')
            print('                    GENERATING MATCHINGS')
            print('==================================================================')
            print('_ Geometric matchings:')
            printGeoMatch, printBase, printPredicate, printDirective, printDescription = printing
            frame = self.form.geometricFrame()
            dispX = frame[1][0] - frame[0][0] + 1
            dispY = frame[1][1] - frame[0][1] + 1
            
            
        transformlists = [None] * len(self.analysis.maps)
        lists = 0
        for map in self.analysis.maps.values():
            transformlists[lists] = map.potentialTransforms()
            if printing is not None and not map.type == 'map-parametric':
                print('   +',map.type+': '+str(len(transformlists[lists]))+' results')
            if map.type == 'map-graphMap':
                transformlists[lists] = map.analyzeRHS(self.rule.rhs, self.form, transformlists[lists])
            transfId = map.id.split('=|=')
            if len(transfId) > 1:
                for i in transformlists[lists]:
                    i.id = transfId[0]
            if transformlists[lists] is not None:
                lists += 1

        paramsets = None
        if self.analysis.hasMap(parametricMap):
            paramsets = self.analysis.maps[parametricMap.type].potentialParamSets()
            #paramsets = self.analysis.maps[parametricMap].potentialParamSets()
            if printing:
                print('_ Description matchings:')
                if len(paramsets) == 0:
                    paramsets = [paramSet()]
                    print('   + No description matches --> creating an empty description transform set')
                else:
                    print('   + Description matches (before simple resolving):', len(paramsets),'results')
            elif len(paramsets) == 0:
                paramsets = [paramSet()]
            i = 0
            while i < len(paramsets):
                if not paramsets[i].resolveSimpleDelays():
                    paramsets.pop(i)
                else:
                    i += 1
            if printing:
                print('   + Description matches (after simple resolving):', len(paramsets),'results')
            
        indices = []
        totalcount = 1
        for i in range(lists):
            indices.append(0)
            totalcount *= len(transformlists[i])
            
        sGeo = ''
        sBase = ''
        sPredicate = ''
        sDirective = ''
        
        count = 0
        countGeo = 0
        countBase = 0
        mergedCount = 0
        predicateCount = 0
        directiveCount = 0
        firstDescResolvedCount = 0
        secondDescResolvedCount = 0
        for i in range(totalcount):
            tfset = transformSet()
            for j in range(lists):
                tfset.add(transformlists[j][indices[j]])
            j = 0
            while indices[j] == len(transformlists[j])-1 and j < lists-1:
                indices[j] = 0
                j += 1
            if indices[j] == len(transformlists[j]):
                break
            indices[j] += 1
            
            if printing and printGeoMatch:
                #tfset.printTransf()
                try:
                    print()
                    tlhs = self.rule.lhs.transform(tfset) 
                    tlhs.printRhino(countGeo*dispX, -dispY)
                    #sGeo += tlhs.rhinoString(countBase*dispX, -dispY) + '\n'
                except Exception as ex:
                    print(ex)
            countGeo += 1
            check = self.rule.lhs.checkTransform(tfset, self.form) 
                        
            if not check:
                continue
            
            if printing and printBase:
                #tfset.printTransf()
                tlhs = self.rule.lhs.transform(tfset) 
                sBase += tlhs.rhinoString(countBase*dispX, -dispY) + '\n'
            countBase += 1
                
            if paramsets is None:
                mergedCount += 1
                try:
                    tlhs = self.rule.lhs.transform(tfset) 
                except PlaneTransformationError:
                    continue
                if not self.rule.evaluatePredicate(tfset, self.rule.lhs, tlhs, self.form, self.rule.keys):
                    continue
                predicateCount += 1
                for tfS in self.rule.solveDirectives(tfset, self.rule.lhs, tlhs, self.rule.rhs, self.form, self.rule.keys):
                    try:
                        directiveCount += 1
                        try:
                            trhs = self.rule.rhs.transform(tfS)
                        except ValueError:
                            continue
                        if self.rule.ruleType == 'ADDITIVE' and trhs.partOf(self.form):
                            continue
                        if [tlhs,trhs] in resultForms:
                            continue
                        if not tlhs.partOf(self.form):
                            continue
                        resultForms.append([tlhs,trhs])
                        app = ruleApplication(self.rule, tfS, None, tlhs, trhs)
                        if app not in self.applications:
                            self.applications.append(app)
                    except PlaneTransformationError:
                        continue
            else:
                description.clearDelays()      
                try:
                    tlhs = self.rule.lhs.transform(tfset) 
                except PlaneTransformationError:
                    continue
                
                for paramset in paramsets:
                    try:
                        mergedCount += 1
                        params = paramset.duplicate()
                        params.lhs = tlhs
                        params.transform = tfset
                        params.keys = self.rule.keys
                        if not self.rule.evaluatePredicate(tfset, self.rule.lhs, tlhs, self.form, params):
                            continue
                        
                        ''' ------------------------- ------------------------- '''
                        if printing and printPredicate:
                            sPredicate += tlhs.rhinoString(predicateCount*dispX, -2*dispY) + '\n'
                        predicateCount += 1
                        ''' ------------------------- ------------------------- '''
                        
                        if not (params.resolveDelays()):
                            continue
                        firstDescResolvedCount += 1
                        for tfS in self.rule.solveDirectives(tfset, self.rule.lhs, tlhs, self.rule.rhs, self.form, params):
                            try:
                                trhs = self.rule.rhs.transform(tfS)
                            except ValueError:
                                continue
                            if printing and printDirective:
                                sDirective += tlhs.rhinoString(directiveCount*dispX, -3*dispY)
                                sDirective += trhs.rhinoString(directiveCount*dispX, -4*dispY) + '\n'
                            directiveCount += 1
    
                            params.rhs = trhs
                            if not description.resolveDelays(params):
                                continue
                            secondDescResolvedCount += 1
                            try:
                                glhs = tlhs.edit(params)
                                grhs = trhs.edit(params)
                            except DescriptionEvaluatingError as ERROR:
                                print('Description Error:', str(ERROR))
                                continue
                            """
                            try:
                                tlhs.maximalize()
                                grhs.maximalize()
                            except Exception as ERROR:
                                print('Transformation Error:', str(ERROR))
                                continue
                            """
                            
                            checkType = True if (self.rule.ruleType == 'ADDITIVE' or glhs.partOf(grhs)) else False
                            '''
                            print('\n\n\n-------------------------------------')
                            print(glhs.partOf(self.form))
                            print('=====================================\n\n\n')
                            '''
                            if (glhs.partOf(self.form)
                            and (checkType is False or not grhs.partOf(self.form))):
                                app = ruleApplication(self.rule, tfS, params, glhs, grhs)
                                if app in self.applications:
                                    continue
                                '''
                                print('_Transformed lhs:')
                                glhs.printE('~~~')
                                print('\n______\n')
                                '''
                                self.applications.append(app)
                    except PlaneTransformationError:
                        continue
        
        if printing:
            if paramsets is None:
                print('\n\n==================================================================')
                print('                    EVALUATE MATCHINGS')
                print('==================================================================')
                print('  _ Total geometric transformations:', countBase)
                if printBase:
                    print('*** base geometric results: ***\n\n')
                    print(sBase, end = '')
                    print('*******************************\n')
                print('  _ Total transformations after geometric check:',mergedCount)
                print('\n<===================== Evaluate Predicates ======================>\n')
                if printPredicate:
                    print('*** predicate results: ***\n')
                    print(sPredicate, end = '')
                    print('**************************\n')
                print('  _ Total geometric transformations after predicate check:',predicateCount)
                print('\n<===================== Evaluate Directives ======================>\n')
                if printDirective:
                    print('*** directive results: ***\n')
                    print(sDirective, end = '')
                    print('**************************\n')
                print('  _ Number of transformations after directive check:',directiveCount)
                print('\n<=================== Final Containment Check ====================>\n')
                print('  _ Number of final results:', len(self.applications),'\n\n')
            else:
                print('\n\n==================================================================')
                print('                    EVALUATE MATCHINGS')
                print('==================================================================')
                if printBase:
                    print('*** base geometric results: ***\n')
                    print(sBase, end = '')
                    print('*******************************\n')
                print('  _ Total geometric transformations:', countBase)
                print('  _ Total description transformations:', len(paramsets))
                print('\n<=============== Merging Description and geometric ==============>\n')
                print('  _ Total transformations after merging:',mergedCount)
                print('\n<===================== Evaluate Predicates ======================>\n')
                if printPredicate:
                    print('*** predicate results: ***\n')
                    print(sPredicate, end = '')
                    print('**************************\n')
                print('  _ Total geometric transformations after predicate check:',predicateCount)
                print('\n<==================== Evaluate Descriptions =====================>\n')
                print('  _ Number of transformations after first description check:',firstDescResolvedCount)
                print('\n<===================== Evaluate Directives ======================>\n')
                if printDirective:
                    print('*** directive results: ***\n')
                    print(sDirective, end = '')
                    print('**************************\n')
                print('  _ Number of transformations after directive check:',directiveCount)
                print('\n<==================== Evaluate Descriptions =====================>\n')
                print('  _ Number of transformations after second description check:',secondDescResolvedCount)
                print('\n<=================== Final Containment Check ====================>\n')
                print('  _ Number of final results:', len(self.applications),'\n\n')
        self.applicationDict[self.rule.name].append([self.form.duplicate(), self.applications])
        return self.applications        


    def clear(self):
        self.form = None
        for map in self.analysis.maps.values():
            map.clear()
        self.target = None
        self.uptrail = []
        self.downTrail = []
        self.success = False
        self.applications = []


    def visitEnter(self, element, assignment):
        if 'ind' in element.type:
            if len(self.uptrail) == 0:
                try:
                    mapping = self.analysis.getMap(element.sort, assignment)
                except KeyError:
                    return False
                result = False
                fragment = self.target.content
                if len(self.downTrail) != 0:
                    fragment = self.downTrail[0].content
                for elm in fragment:
                    if assignment == 'refl' and mapping.type == 'map-parametric':
                        continue
                    try:
                        if mapping.map(elm, element):
                            result = True
                    except AttributeError:
                        pass
                if len(self.downTrail) == 0:
                    self.success = (result or self.success)
                return result
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                return True
            elif len(self.uptrail) != 0 and element.sort != self.uptrail[-1].content.sort:
                return False
            elif len(self.uptrail) != 0:
                self.uptrail.pop(0)
            else: 
                try:
                    fragment = self.downTrail[0]
                except IndexError: 
                    fragment = self.target
                for frag in fragment.children:
                    if element.sort == frag.content.sort:
                        self.downTrail.append(frag)
                        break
                if len(self.downTrail) >0 and self.downTrail[-1].content.visited:
                    return False

        return True
    
    def visitLeave(self, element):
        if 'ind' in element.type:
            if (self.success and len(self.uptrail) != 0):
                mapping = self.analysis.getMap(element.sort)
                for elm in self.uptrail[-1].parent.content:
                     mapping.map(elm, element)
                self.success = False
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                return
                
            elif len(self.downTrail) != 0:
                self.downTrail.pop(0)

            elif (len(self.uptrail) != 0 and self.uptrail[-1].parent is not None
                    and element.sort == self.uptrail[-1].parent.content.sort):
                self.uptrail.append(self.uptrail[-1].parent)
            elif element.sort == self.target.content.sort:
                self.uptrail.append(self.target)
        return