# -*- coding: utf-8 -*-
"""
Created on Tue May 30 10:21:21 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .rule import rule
from .ruleDisjunction import ruleDisjunction
from .ruleDisjunctionP import ruleDisjunctionP
from .ruleNegation import ruleNegation
from .ruleSequence import ruleSequence

from ...io.parseReader import parseReader, ParsingError

"""
flow: a set of rules that are to be performed in a certain order
"""
class flow(rule):
    
    type = 'rule-flow'
    register = {}
    
    def __init__(self, name, description, rules, *args, **kwargs):
        self.name = name
        self.description = description
        self.rules = rules
        self.process = None
        flow.register[name] = self
        
    @staticmethod
    def existingFlows():
        return flow.register.keys()
        
    @staticmethod
    def cleanup():
        for i in list(flow.register.keys()):
            del flow.register[i]
            
    def setName(self, newName):
        if newName in flow.register:
            raise ValueError('There already exists a flow with flow name '+str(newName))
        flow.register[newName] = flow.register[self.name]
        del flow.register[self.name]
        
        from ...io.sdlParser import sdlParser
        if self.name in sdlParser.flows:
            sdlParser.flows[newName] = sdlParser.flows[self.name]
            del sdlParser.flows[self.name]
        self.name = newName
    
    @staticmethod
    def getFlow(flowName):
        return flow.register[flowName]

    def __str__(self):
        return str(self.rules)
    
    def detect(self, *args, **kwargs):
        raise NotImplementedError

    def apply(self, form, printProcess = False, random = True, *args, **kwargs):
        self.process = []
        form = self.rules.apply(form, s = self.process, random = random)
        print('\n\ncomplete flow:\n<Begin>', end='')
        for i in self.process:
            print(i,end='')
        print(' <end>')
        return form
        #raise NotImplementedError

    @staticmethod
    def readQuantifier(reader):
        quantifier = []
        matching = 'G'
        if reader.newToken() == '{':
            quantifier = []
            if reader.newToken() != parseReader.NUMBER:
                raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
            quantifier.append(int(reader.value()))
            if reader.newToken() == '}':
                quantifier.append(quantifier[0])
            else:
                if reader.token() != ',':
                    raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
                if reader.newToken() != '}':
                    if reader.token() != parseReader.NUMBER:
                        raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
                    quantifier.append(int(reader.value()))
                    if reader.newToken() != '}':
                        raise ParsingError(reader,'"}" expected')
                else:
                    quantifier.append(None)
        elif reader.token() == '?':
            quantifier = [0,1]
        elif reader.token() == '*':
            quantifier = [0,None]
        elif reader.token() == '+':
            quantifier = [1,None]

        if reader.previewToken() is not None and reader.previewToken() in ['?', '+', '*']:
            if reader.newToken() == '?':
                matching = 'L'
            elif reader.token() == '+':
                matching = 'PO'
            elif reader.token() == '*':
                matching = 'PR'
        
        return quantifier, matching
        
        
    @staticmethod
    def readFlowRules(reader, thisNegation = False, starting = False):
        ruleList = []
        paren = '('
        if not starting:
            paren = reader.token()
            if reader.newToken() == '*':
                paren += '*'
                reader.newToken()
        while reader.token() is not None and reader.token() not in (')', ']'):
            negation = False
            if reader.token() == '!':
                negation = True
                reader.newToken()
            if reader.token() in ('[','('):
                ruleList.append(flow.readFlowRules(reader, negation))
            else:
                ruleName = reader.value()
                try:
                    singleRule = rule.register[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist in rule register')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = flow.readQuantifier(reader)
                    singleRule = ruleSequence([singleRule], quantifier, matching)
                if negation:
                    singleRule = ruleNegation(singleRule)
                ruleList.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = flow.readQuantifier(reader)
            
            
        if paren == '(' or len(ruleList) == 1:
            r = ruleSequence(ruleList, quantifier, matching)
        elif paren == '[*':
            r = ruleDisjunctionP(ruleList, quantifier, matching)
        elif paren == '[':
            r = ruleDisjunction(ruleList, quantifier, matching)
        if thisNegation:
            r= ruleNegation(r)
        if matching == 'PO':
            r.setPo()
        return r



    @staticmethod
    def createFlow(name, description, string):
        reader = parseReader(string)
        r = flow.readFlowRules(reader, starting = True)
        flowObj = flow(name, description, r)
        return flowObj
