# -*- coding: utf-8 -*-
"""
Created on Mon Jan  9 12:10:15 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .tree import tree, DescriptionEvaluatingError
from .symbols import symbols
from .numberLiteral import numberLiteral
from ...io.parseReader import parseReader, ParsingError
from ...struct.coordinate import coordinate

class condition():
    
    type = 'description-condition'
    def __init__(self, operator, operand):
        self.operator = operator
        self.operand = operand
        
    def printVal(self):
        s = ''
        if self.operator == symbols.SET_OPEN:
            return str(self.operand) 
        elif self.operator in [symbols.EQUAL_, symbols.LESS_, symbols.GREATER_]:
            s += self.operator
        elif self.operator == parseReader.LESSQ:
            s += symbols.LESS_ + symbols.EQUAL_
        elif self.operator == parseReader.GREATERQ:
            s += symbols.GREATER_ + symbols.EQUAL_
        elif self.operator == parseReader.NOTEQUAL:
            s += symbols.LESS_ + symbols.GREATER_
        if self.operand.type == 'description-numericExpression':
            s += symbols.PAREN_OPEN + self.operand.printVal() + symbols.PAREN_CLOSE
        else:
            s += self.operand.printVal()
        return 'condition:'+s
        
    def isEnumerative(self):
        return self.operator == symbols.SET_OPEN
        
    def isEquational(self):
        return self.operator != symbols.SET_OPEN
        
    def __eq__(self, other):
        if other.type != 'description-condition':
            return False
        return self.operator == other.operator and self.operand == other.operand
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def compareValue(self, other):
        if self.operator > other.operator:
            return 1
        if self.operator < other.operator:
            return -1
        return self.operand.compareValue(other.operand)
        
    def __repr__(self):
        return str(self)

    def __str__(self):
        s = ''
        if self.operator == symbols.SET_OPEN:
            return str(self.operand)
        elif self.operator in [symbols.EQUAL_, symbols.LESS_, symbols.GREATER_]:
            s += self.operator
        elif self.operator == parseReader.LESSQ:
            s += symbols.LESS_ + symbols.EQUAL_
        elif self.operator == parseReader.GREATERQ:
            s += symbols.GREATER_ + symbols.EQUAL_
        elif self.operator == parseReader.NOTEQUAL:
            s += symbols.LESS_ + symbols.GREATER_
        if self.operand.type == 'description-numericExpression':
            s += symbols.PAREN_OPEN + str(self.operand) + symbols.PAREN_CLOSE
        else:
            s += str(self.operand)
        return s
    
    def matches(self, params, sort, value, preDelay = True):
        if self.operator == symbols.SET_OPEN:
            if value.type == 'description-stringLiteral' and len(self.operand.tuple) != 0 and self.operand.tuple[0].type == 'description-numberLiteral':
                value = numberLiteral(coordinate(value.s))
            for i in self.operand:
                try:
                    op = i.evaluate(params, sort)
                except (ValueError, KeyError, DescriptionEvaluatingError):
                    op = None
                if op is None:
                    if preDelay:
                        params.delay(self, sort, value)
                        return True
                    else:
                        return False
                if op == value:
                    return True
            return False
        try:
            comparand = self.operand.evaluate(params, sort)
        except (ValueError, KeyError, DescriptionEvaluatingError):
            comparand = None
        if comparand is None:
            if preDelay:
                params.delay(self, sort, value)
                return True
            else:
                return False
        #params.resolveDelay(self, value)
        if value.type == 'description-stringLiteral' and comparand is not None and comparand.type == 'description-numberLiteral':
            try:
                value = numberLiteral(coordinate(value.s))
            except ValueError:
                return False
        if self.operator == symbols.EQUAL_:
            return value.compareValue(comparand) == 0
        if self.operator == parseReader.NOTEQUAL:
            return value.compareValue(comparand) != 0
        if self.operator == parseReader.LESSQ:
            return value.compareValue(comparand) != 1
        if self.operator == symbols.LESS_:
            return value.compareValue(comparand) == -1
        if self.operator == parseReader.GREATERQ:
            return value.compareValue(comparand) != -1
        if self.operator == symbols.GREATER_:
            return value.compareValue(comparand) == 1
        return False

        #comparand = self.operand.evaluate(params, sort)
        #if comparand is None:
            
    def simpleDelayedMatch(self, params, reference):
        sort = reference[1]
        value = reference[2]
        if self.operator == symbols.SET_OPEN:
            if value.type == 'description-stringLiteral' and len(self.operand.tuple) != 0 and self.operand.tuple[0].type == 'description-numberLiteral':
                value = numberLiteral(coordinate(value.s))
            for i in self.operand:
                try:
                    op = i.evaluate(params, sort)
                except (ValueError, KeyError, DescriptionEvaluatingError):
                    op = None
                if op is None:
                    return 1
                if op == value:
                    return True
            return False
        try:
            comparand = self.operand.evaluate(params, sort)
        except (ValueError, KeyError, DescriptionEvaluatingError):
            comparand = None
            
        if comparand is None:
            return 1
            
        #params.resolveDelay(self, value)
        if value.type == 'description-stringLiteral' and comparand is not None and comparand.type == 'description-numberLiteral':
            try:
                value = numberLiteral(coordinate(value.s))
            except ValueError:
                return False
        if self.operator == symbols.EQUAL_:
            return value.compareValue(comparand) == 0
        if self.operator == parseReader.NOTEQUAL:
            return value.compareValue(comparand) != 0
        if self.operator == parseReader.LESSQ:
            return value.compareValue(comparand) != 1
        if self.operator == symbols.LESS_:
            return value.compareValue(comparand) == -1
        if self.operator == parseReader.GREATERQ:
            return value.compareValue(comparand) != -1
        if self.operator == symbols.GREATER_:
            return value.compareValue(comparand) == 1
        return False

        #comparand = self.operand.evaluate(params, sort)
        #if comparand is None:
        
            
    def delayedMatch(self, map, reference):
        return self.matches(map, reference[1], reference[2], preDelay = False)
        
    @staticmethod
    def parse(desc, reader):
        if reader.token() == symbols.SET_OPEN:
            type = reader.previewToken()
            if type == parseReader.MINUS: 
                type = parseReader.NUMBER
            if type == parseReader.IDENTIFIER: 
                type = parseReader.STRING
            if type != parseReader.STRING and type != parseReader.NUMBER:
                raise ParsingError(reader, "expected a number, identifier or '\"'")
            tupleSet = desc.tupleDesc(symbols.SET_OPEN, symbols.COMMA)
            while True:
                reader.newToken()
                if type == '#':
                    operand = desc.reference.parse(desc, reader)
                elif type == parseReader.NUMBER:
                    tupleSet.addEntity(desc.numberLiteral.parse(desc, reader))
                elif reader.token() == parseReader.STRING:
                    tupleSet.addEntity(desc.stringLiteral(reader.stringVal(), reader.value()))
                    reader.newToken()
                elif reader.token() == parseReader.IDENTIFIER:
                    name = reader.value()
                    if desc.literal.isLiteral(name):
                        operand = desc.literal.getLiteral(name)
                        reader.newToken()
                    elif desc.function.isFunctionName(name):
                        operand = desc.function.parse(desc, reader)
                    elif desc.sort.checkSort(name) or name == 'rhs':
                        operand = desc.reference.parse(desc, reader)
                    else:
                        operand = desc.parameter.parse(desc, reader, True)
                    tupleSet.addEntity(operand)
                else:
                    raise ParsingError(reader, "expected a literal")
                if reader.token() != symbols.COMMA:
                    break
            if reader.token() != symbols.SET_CLOSE:
                raise ParsingError(reader, "expected a '}'")
            reader.newToken()
            return condition(symbols.SET_OPEN, tupleSet)
        op = reader.token()
        if op == symbols.LESS_:
            if reader.previewToken() == symbols.GREATER_:
                op = parseReader.NOTEQUAL
                reader.newToken()
        if op in [symbols.EQUAL_, symbols.GREATER_, symbols.LESS_, parseReader.NOTEQUAL, parseReader.GREATERQ, parseReader.LESSQ]:
            reader.newToken()
            if reader.token() == '#':
                operand = desc.reference.parse(desc, reader)
            elif reader.token() == parseReader.NUMBER or reader.token() == parseReader.MINUS:
                operand = desc.numberLiteral.parse(desc, reader)
            elif reader.token() == parseReader.IDENTIFIER:
                if reader.value() == 'pi':
                    operand = desc.literal.createLiteral('pi')
                    reader.newToken()
                elif desc.sort.checkSort(reader.value()) or reader.value() == 'rhs':
                    operand = desc.reference.parse(desc, reader)
                elif desc.function.isFunctionName(reader.value()):
                    operand = desc.function.parse(desc, reader)
                else: 
                    operand = desc.parameter.parse(desc, reader, False)
            else:
                operand = desc.parseEntity(desc, reader)
                #raise ParsingError(reader, "expected a (positive or negative) number or '('")
            return condition(op, operand)
        raise ParsingError(reader, "expected a conditional operator or a set")
             
