# -*- coding: utf-8 -*-
"""
Created on Tue Jan 30 16:13:48 2018

@author: akibdpt
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .ellipse import ellipse2D, ellipse3D
from ..struct.coordinate import c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError


class ellipseP2D(ellipse2D):
    
    type = 'ind-ellipseP2D'
    
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        super(ellipseP2D, self).__init__(sort, focus1, focus2, pin)
        
    def equals(self, other):
        if self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
    
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        ellipse = ellipseP2D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        ellipse.assoc = self
        return ellipse
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        focus1 = transf.apply(self.foci[0])
        focus2 = transf.apply(self.foci[1])
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) == 0:
            return self
        return ellipseP2D(self.sort, focus1, focus2, newPts[0])
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a ellipse')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y = vector2D.readVector(reader)
        focus1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        focus2 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        pin = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return ellipseP2D(sort, focus1, focus2, pin)



class ellipseP3D(ellipse3D):
    
    type = 'ind-ellipseP3D'
    
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        super(ellipseP3D, self).__init__(sort, focus1, focus2, pin)
        
    def codescriptor(self):
        return self
        
    def parametricNode(self):
        return self
        

    def transformInd(self, transformSet):
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        focus1 = transf.apply(self.foci[0])
        focus2 = transf.apply(self.foci[1])
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) == 0:
            return self
        return ellipseP3D(self.sort, focus1, focus2, newPts[0])
        
        
    def checkTransform(self, transform, compareForm):
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a ellipse')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        focus1 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        focus2 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        pin = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        
        return ellipseP3D(sort, focus1, focus2, pin)