# -*- coding: utf-8 -*-
"""
Created on Thu Dec 29 12:40:51 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import range
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .plane import plane3D
from .lineSegment import lineSegment3D
from .line import line3D
from ..struct.planeStruct import bSegment, boundary
from ..struct.classes import classes as cls, lineClasses
from ..struct.splayTree import splayTree
from ..struct.splayable import sSeg, sRef, vertex
from ..struct.coordinate import c1, cn1, c0p5
from ..struct.vector import vector3D
from ..form.coForm import coForm
from ..transform.affineTransform import affineTransform3D
from ..transform.graphTransform import graphTransform3D
from ..io.parseReader import ParsingError

class planeSegment3D(plane3D):
    
    type = 'ind-planeSegment3D'
    
    def __init__(self, sort, boundr, boundaryList = False, *args, **kwargs):
        if boundaryList:
            plane3D.__init__(self, sort, boundr[0].polygonal)
            self.boundaries = boundr
        else:
            i = 0
            while i < len(boundr):
                if boundr[i] == boundr[i-1]:
                    boundr.pop(i)
                else:
                    i += 1
            plane3D.__init__(self, sort, boundr)
            self.boundaries = [boundary()]
            #add line segments defining the boundary: 1st vector + 2nd vector = 1st line
            #                                         2nd vector + 3rd vector = 2nd line, etc
            for i in range(1,len(boundr)):
                self.boundaries[0].add(bSegment(boundr[i-1],boundr[i]))
            #add the last line segment defining the boundary: last vector + 1st vector = last line
            self.boundaries[0].add(bSegment(boundr[-1],boundr[0]))

            classes = self.classify(None)
            result = coForm(self.sort)
            self.construct(classes.sumSet(), result)
            self.boundaries = result.first().boundaries

    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        for i in self.boundaries:
            for j in range(len(i.polygonal)):
                if j > 0:
                    s += '>, <'
                else:
                    s += '<<'
                s += str(i.polygonal[j-1])+', '+str(i.polygonal[j])
            s += '>>'
        return s
        
    def codescriptor(self):
        c = plane3D(self.sort, self.boundaries[0].polygonal)
        c.assoc = self
        return c
            
    def equals(self, other):
        if len(self.boundaries) != len(other.boundaries):
            return False
        for i in self.boundaries:
            check = False
            for j in other.boundaries:
                if i == j:
                    check = True
            if check == False:
                return False
        return True
            
    def __ne__(self, other):
        return not self == other
        
    def getValue(self):
        return str(self)
        
    def geometricFrame(self):
        minCoord = [self.boundaries[0].polygonal[0].x, self.boundaries[0].polygonal[0].y, self.boundaries[0].polygonal[0].z]
        maxCoord = [self.boundaries[0].polygonal[0].x, self.boundaries[0].polygonal[0].y, self.boundaries[0].polygonal[0].z]
        for i in self.boundaries[0].polygonal:
            if minCoord[0] > i.x:
                minCoord[0] = i.x
            if minCoord[1] > i.y:
                minCoord[1] = i.y
            if minCoord[2] > i.z:
                minCoord[2] = i.z
            if maxCoord[0] < i.x:
                maxCoord[0] = i.x
            if maxCoord[1] < i.y:
                maxCoord[1] = i.y
            if maxCoord[2] < i.z:
                maxCoord[2] = i.z
        return [minCoord, maxCoord]

    def printRhino(self, dX = 0, dY = 0):
        s = ''
        for i in self.boundaries:
            s+='_polyline '
            for j in i.polygonal:
                s += str(j.x + dX) + ',' + str(j.y + dY) + ',' + str(j.z) + ' '
            s+= str(i.polygonal[0].x + dX) + ',' + str(i.polygonal[0].y + dY) + ',' + str(i.polygonal[0].z) + '\n '
        s+='_SelPolyline _PlanarSrf _Delete'
        
        return s 

        
    def compareAreal(self, other):
        if 'ind-planeSegment' not in other.type:
            raise Exception('other is not a plane segment')
        bounds = other.boundaries
        c = len(self.boundaries) - len(bounds)
        if c != 0:
            return c/abs(c)
        for i in range(min(len(self.boundaries), len(bounds))):
            c = len(self.boundaries[i]) - len(bounds[i])
            if c != 0:
                return c/abs(c)
            for j in range(min(len(self.boundaries[i]), len(bounds[i]))):
                c = self.boundaries[i].canonical[j].compareValue(bounds[i].canonical[j])
                if (c != 0):
                    return c
        return 0
        
    def compareValue(self, other):
        if 'ind-planeSegment' not in other.type:
            return False
        else:
            return self.codescriptor().compareValue(other.codescriptor())
            
    def coEquals(self, other):
        if other.type == 'ind-planeSegment3D':
            return True
        else:
            return self.codescriptor().compareValue(other.codescriptor()) == 0
            
    def boundaryArea(self, boundary):
        y = boundary.polygonal[0].y
        for i in boundary.polygonal:
            if y < i.y:
                y = i.y
        area = 0
        prev = self.localize(boundary.polygonal[-1])
        for i in range(len(boundary.polygonal)):
            current = self.localize(boundary.polygonal[i])
            area += ((prev.y - y) + (current.y - y)) * (current.x - prev.x) * c0p5
            prev = current
        return abs(area)
            
            
    def area(self):
        area = 0
        for i in range(len(self.boundaries)):
            if i == 0:
                area += self.boundaryArea(self.boundaries[i])
            else:
                area -= self.boundaryArea(self.boundaries[i])
        return abs(area)
        
    def clockwiseBoundary(self):
        vect = vector3D(0,0,0)
        for i in self.boundaries[0].polygonal:
            vect = vect.add(i)
        vect = vect.scale(c1 / len(self.boundaries[0].polygonal))
        if self.clockwiseAngle(vect, self.boundaries[0].polygonal[0],self.boundaries[0].polygonal[1]) == True:
            polygonal = self.boundaries[0].polygonal
            polygonal.append(polygonal[0])
            return polygonal
        polygonal = list(reversed(self.boundaries[0].polygonal))
        polygonal.append(polygonal[0])
        return polygonal
        
    def ccwBoundary(self):
        vect = vector3D(0,0,0)
        for i in self.boundaries[0].polygonal:
            vect = vect.add(i)
        vect = vect.scale(c1 / len(self.boundaries[0].polygonal))
        if self.clockwiseAngle(vect, self.boundaries[0].polygonal[0],self.boundaries[0].polygonal[1]) == False:
            polygonal = self.boundaries[0].polygonal
            polygonal.append(polygonal[0])
            return polygonal
        polygonal = list(reversed(self.boundaries[0].polygonal))
        polygonal.append(polygonal[0])
        return polygonal
            
            
            
    """
    def transformInd(self, transforms):
        affine = transforms.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        bounds = []
        for i in self.boundaries:
            bounds.append(i.transform(affine))
        return planeSegment3D(self.sort, bounds, True)
        
    """
    def transformInd(self, transforms):
        affine = transforms.retrieveTransform(affineTransform3D)
        if affine is None:
            newPlane = self.duplicateInd()
        else:
            bounds = []
            for i in self.boundaries:
                bounds.append(i.transform(affine))
            newPlane = self.__class__(self.sort, bounds, True)
            
        transf = transforms.retrieveTransform(graphTransform3D)
        if transf is not None:
            bounds = []
            for i in self.boundaries:
                check = i.graphTransform(transf)
                if check is False:
                    return newPlane
                bounds.append(check)
            newPlane = self.__class__(self.sort, bounds, True)
        return newPlane
    
        
        
    def replicateInd(self):
        bdr = []
        for i in self.boundaries:
            bound = boundary()
            for j in i.canonical:
                bound.canonical.append(j)
            for j in i.polygonal:
                bound.polygonal.append(j)
            bdr.append(bound)
        result = planeSegment3D(self.sort, bdr, True)
        return result
        
                
    def convertFloat(self):
        bdrs = []
        for i in self.boundaries:
            bdr =  boundary()
            for j in i.canonical:
                bSeg = bSegment(j.start.convertFloat(),j.end.convertFloat())
                bdr.canonical.append(bSeg)
            for j in i.polygonal:
                bdr.polygonal.append(j.convertFloat())
            bdrs.append(bdr)
        result = planeSegment3D(self.sort, bdrs, True)
        if self.hasAttribute():
            result.addAttribute(self.attribute)
        return result
        
    def findFloatValue(self, element):
        if element.type == "ind-lineSegment3D":
            for i in self.boundaries:
                for j in i.canonical:
                    if j.start.floatValue == element.start and j.end.floatValue == element.end:
                        return lineSegment3D(self.sort, j.start, j.end)
        elif element.type == "ind-line3D":
            for i in self.boundaries:
                for j in i.canonical:
                    line = line3D(self.sort, j.start, j.end)
                    if line.findFloatValue(element) is not False:
                        return line
        elif element.type == "ind-planeSegment3D":
            for i in self.boundaries:
                check = False
                for j in element.boundaries:
                    if i.findFloatValue(j):
                        check = True
                if check == False:
                    return False
        if self.hasAttribute():
            return self.attribute.findFloatValue(element)
        return False
        
    def containsPoint(self, point):
        for i in self.boundaries[0].canonical:
            if i.containsPoint(point):
                return True
        if self.boundaries[0].containsPointInside(point):
            for i in range(1,len(self.boundaries)):
                for j in self.boundaries[i].canonical:
                    if j.containsPoint(point):
                        return True
                if self.boundaries[i].containsPointInside(point):
                    return False
                return True
        return False

    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        """
        f1 = coForm(self.sort)
        f1.insertClassification(self)
        f2 = coForm(ind.sort)
        f2.insertClassification(ind)
        c = planeSegment3D.classification(f1,f2)
        if not c.contains():
            return False
        """
        if self.sort.type == 'sort-attribute':
            if not self.hasAttribute():
                if ind.hasAttribute():
                    return False
                else:
                    return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
    """
    def accept(self, visitor, direction = -1):
        if self.isNil():
            return
        for i in self.boundaries:
            for j in i:
                lineSeg = lineSegment3D(self.sort, j.start, j.end)
                if self.hasAttribute():
                    lineSeg.addAttribute(self.attribute)
                lineSeg.accept(visitor, direction)
    """
    
    def merge(self, other):
        thisBound = self.boundaries[0]
        otherBound = other.boundaries[0]
        commonLines = []
        
        descAtt = None
        if self.hasAttribute():
            if self.attribute.type == 'form-metaForm':
                for i in self.attribute:
                    if i.sort.baseSort().characteristic.type == 'ind-description':
                        for j in other.attribute:
                            if j.sort.baseSort().characteristic.type == 'ind-description':
                                if i == j:
                                    descAtt = self.attribute
                                else:
                                    return None
                        if descAtt is None:
                            return None
                        break
            else:
                if self.attribute.sort.baseSort().characteristic.type == 'ind-description':
                    if self.attribute == other.attribute:
                        descAtt = self.attribute
                    else:
                        return None

        for thisLine in thisBound.canonical:
            for otherLine in otherBound.canonical:
                if not thisLine.direction.parallel(otherLine.direction):
                    continue
                if not thisLine.start.colinear(otherLine.start, otherLine.end):
                    continue
                if thisLine.end <= otherLine.start or thisLine.start >= otherLine.end:
                    continue
                end = thisLine.end
                if thisLine.end > otherLine.end:
                    end = otherLine.end
                start = thisLine.start
                if thisLine.start < otherLine.start:
                    start = otherLine.start
                commonLines.append(bSegment(start, end))
        if len(commonLines) == 0:
            return None
        mergedCanon = []
        for i in thisBound.canonical:
            mergedCanon.append(i)
        for i in otherBound.canonical:
            mergedCanon.append(i)
        for i in commonLines:
            index = 0
            while index < len(mergedCanon):
                line = mergedCanon[index]
                if i.partOf(line):
                    mergedCanon.pop(index)
                    if i == line:
                        continue
                    compl = line.complement(i)
                    for l in compl:
                        mergedCanon.insert(index, l)
                        index += 1
                    continue
                index += 1
        result = coForm(self.sort)
        self.construct(mergedCanon, result)
        result = result.elements[0]
        for i in range(1,len(self.boundaries)):
            result.boundaries.append(self.boundaries[i])
        for i in range(1,len(other.boundaries)):
            result.boundaries.append(other.boundaries[i])
        if descAtt is not None:
            result.addAttribute(descAtt)
        return result



                        
    def isNil(self):
        if len(self.boundaries) == 0:
            return True
        for i in self.boundaries:
            if len(i.canonical) == 0 and len(i.polygonal) == 0:
                return True
        return False
        
    def printPlaneSegment(self, string = ''):
        print(string,'_ plane Segment:')
        for j in range(len(self.boundaries)):
            print(string,'__ boundary',j+1)
            """
            for i in self.boundaries[j]:
                print(string,'___ +segment:',i)
            """
            for i in self.boundaries[j].polygonal:
                print(string,'    _position:',i)
    
    def polygon(self):
        if len(self.boundaries) == 1:
            return self.boundaries[0].polygonal
        polygon = []
        polygon += self.boundaries[0].polygonal
        polygon.append(polygon[0])
        return iter(polygon)
        
    def outer(self):
        p = self.boundaries[0].polygonal + [self.boundaries[0].polygonal[0]]
        return p
        
    """
     * Classifies the boundary segments of two coequal plane-segments,
     * each contained within a co-form, wrt each other into the categories of
     * inner and outer boundary segments for each plane-segment and the
     * categories of same-shared and oppositely-shared boundary segments.
    """        
    @staticmethod
    def classification(thisForm, otherForm, printing = False):
        """=========="""
        if printing:
            print('\n\n_____________________________________________________________')
        """=========="""
        classes = cls()
        # create the status of the sweepline as a splay tree
        # create a splayTree with two boundary segments: 1 from negative y infinity (0, -inf) to positive x infinity(inf, 0)
        #                                                1 from positive y infinity (0,  inf) to positive x infinity(inf, 0)
        """
                           \ |      /
                            \|     /
                      (0,inf)\    /
                             |\  /
                 ____________|_\/____________
                             | /\ (inf, 0)
                             |/  \
                     (0,-inf)/    \
                            /|     \
                           / |      \
                           
           tree:
                           |
                   (0,-inf),(inf,0)   <--- root
                    /            \
                  None     (0,inf),(inf,0)     
        """
        sweepline = splayTree()
        neginf = sSeg(bSegment(bSegment.NEG_Y_INF_3D, bSegment.POS_X_INF_3D))
        posinf = sSeg(bSegment(bSegment.POS_Y_INF_3D, bSegment.POS_X_INF_3D))
        neginf.alignedWithOne = False
        posinf.alignedWithOne = True
        sweepline.initSentinelTree(neginf, posinf)
        
        # create the sweep schedule of transition points as a splay tree
        transitions = splayTree()
        memberOfOther = False
        coform = thisForm
        while coform is not None:
            #for each of the bsegment in each of the boundary in the coform:
            for segment in coform.elements:
                for i in range(len(segment.boundaries)):
                    for val in segment.boundaries[i]:
                        #create a new splay segment of the bSegment with membership: False for thisForm, True for otherForm
                        interval = sSeg(val)
                        interval.membership = memberOfOther
                        #add start and end position of the splay segment into transition splay tree
                        transitions.insert(sRef.startPosition(interval))
                        transitions.insert(sRef.endPosition(interval))
            #switch to check otherform if currently checking thisForm, else break while loop
            if coform is thisForm:
                coform = otherForm
            else:
                coform = None
            memberOfOther = True
            
        # sweep across each of the transition points
        while not transitions.empty():
            #################################
            """=========="""
            if printing:
                print('\n****************************************************\n===transitions===')
                transitions.printPredSucc()
                print('---sweepline---')
                sweepline.printPredSucc()
            """=========="""
            ##################################
            shared = False
            point = transitions.extractMin()
            # locate point within plane slice defined by sweepline status
            if printing:
                print('\n---determine boundary---')
                print('__ point:', point)
            node = sweepline.find(point.position)
            #splay compare the point with the bSegment of the tree node:
            #                       side == 0 --> point is colinear with the segment
            if printing:
                print('__ node:',node.content)
            side = node.content.splayCompare(point.position) * cn1
            ##################################
            """=========="""
            if printing:
                print('__ side:', side)
            """=========="""
            ##################################
            if side > 0:
                low = node
                high = node.succ
            elif side < 0:
                low = node.pred
                high = node
            else:
                # determine set of edges containing current transition point
                # and delete edges ending at transition point
                low = node.pred
                prev = low
                node = low.succ
                """=========="""
                if printing:
                    print('    __low:', low.content)
                    print('    __node:', node.content)
                """=========="""
                # while point is still colinear with node bSegment (either being an endpoint of the line segment or on the segment):
                while node.content.splayCompare(point.position) == 0:
                    interval = node.content
                    #print('    __interval:', interval)
                    #extract (remove) the node from sweepline if point is the end position of segment
                    if interval.segment.end == point.position:
                        if printing:
                            print("    __point is interval's end -> extract node from sweepline")
                        sweepline.extract(node)
                    else: # if the point is on the line segment but not an end position --> split the segment into two
                        node.content = sSeg(bSegment(point.position, interval.segment.end),interval)
                        interval = sSeg(bSegment(interval.segment.start, point.position),interval)
                        if printing:
                            print('    __point not end point --> split segment')                        
                            print('      __node:    ', node.content)
                            print('      __interval:', interval)
                        prev = node
                        
                    # classify edge into one of the categories:
                    # inner, outer, same-shared or opposite-shared
                    """=========="""
                    if printing:
                        print('       +++ interval \t\t| prev.succ')
                        print('       +++', interval.segment,'\t|', prev.succ.content.segment)
                        print('       +++ shared\t| same start | paralel\t| interval align other')
                        print('       +++',shared,'\t|', interval.segment.start == prev.succ.content.segment.start,'\t    |', interval.segment.direction.parallel(prev.succ.content.segment.direction),'\t|', interval.membership)
                    """=========="""
                    if shared: #both plane segments contain this bSegment
                        """=========="""
                        if printing:
                            print('       +++ interval shared')
                        """=========="""
                        if interval.alignedWithOne == interval.alignedWithOther:
                            #segment is align with both plane segments --> same shared
                            """=========="""
                            if printing:
                                print('       +++ alignedWithOne == alignedWithOther --> same shared')
                            """=========="""
                            classes.addSameShared(interval.segment)
                        else:
                            #segment have different alignment with the two plane segments --> opposite shared
                            """=========="""
                            if printing:
                                print('       +++ alignedWithOne != alignedWithOther --> opp shared')
                            """=========="""
                            classes.addOppShared(interval.segment)
                        shared = False #reset shared for next segment 
                    #
                    elif (interval.segment.start == prev.succ.content.segment.start 
                            and interval.segment.direction.parallel(prev.succ.content.segment.direction)):
                        """=========="""
                        if printing:
                            print('       @@ next occurence of this interval -> shared')
                        """=========="""
                        shared = True
                    elif interval.membership: #segment is a member of other plane segment only
                        """=========="""
                        if printing:
                            print('       +++ interval member of other')
                        """=========="""
                        if interval.alignedWithOne: # align with this plane segment
                            """=========="""
                            if printing:
                                print('       +++ interval alignedWithOne --> inner two')
                            """=========="""
                            classes.addInnerTwo(interval.segment) # --> I2
                        else: # does not a lign with this plane segment
                            """=========="""
                            if printing:
                                print('       +++ interval not alignedWithOne --> outer two')
                            """=========="""
                            classes.addOuterTwo(interval.segment) # --> O2
                    else: # segment is a member of this plane segment only
                        """=========="""
                        if printing:
                            print('       +++ interval member of self')
                        """=========="""
                        if interval.alignedWithOther: # align with other plane segment
                            """=========="""
                            if printing:
                                print('       +++ interval alignedWithOther --> inner one')
                            """=========="""
                            classes.addInnerOne(interval.segment) # --> I1
                        else:  # does not align with other plane segment
                            """=========="""
                            if printing:
                                print('       +++ interval not alignedWithOther --> outer one')
                            """=========="""
                            classes.addOuterOne(interval.segment) # --> O1
                    node = prev.succ 
                    """=========="""
                    if printing:
                        print('    __node:    ', node.content)
                    """=========="""
                high = node
                # reverse the order of all other edges in the set
                if prev != low: 
                    sweepline.reverse(low.succ, prev)
                
            #insert all edges with transition point as start point
            ref = point.position
            if point.isStartPosition():
                sweepline.insert(point.segment)
            while (not transitions.empty()) and ref == transitions.findMin().content.position:
                point = transitions.extractMin()
                if point.isStartPosition():
                    #insert edge if transition point of type start
                    sweepline.insert(point.segment)
                    
            if printing:
                print('\n---new sweepline---')
                sweepline.printPredSucc()
                print()

            # update alignment information
            alignedWithOne = low.content.alignedWithOne
            alignedWithOther = low.content.alignedWithOther  
            """=========="""
            if printing:
                print('---alignment checking---')
                print('== low:',low.content)
                print('== high:', high.content)
                print('++ low  align:',low.content.alignedWithOne, low.content.alignedWithOther)
                print('++ high  align:', high.content.alignedWithOne, high.content.alignedWithOther)
            """=========="""
            node = low.succ
            while node != high:
                """=========="""
                if printing:
                    print('  _node:', node)
                """=========="""
                if node.content.membership:
                    """=========="""
                    if printing:
                        print('   __node align: other')
                    """=========="""
                    alignedWithOther = not alignedWithOther
                else:
                    """=========="""
                    if printing:
                        print('   __node align: one')
                    """=========="""
                    alignedWithOne = not alignedWithOne
                """=========="""
                if printing:
                    print('   __alignment:', alignedWithOne, alignedWithOther)
                """=========="""
                node.content.alignedWithOne = alignedWithOne
                node.content.alignedWithOther = alignedWithOther
                node = node.succ
            """=========="""
            if printing:
                print('  -high:', node)
            """=========="""
            if high.content.membership:
                """=========="""
                if printing:
                    print('   --high align: other')
                    """=========="""
                alignedWithOther = not alignedWithOther
            else:
                """=========="""
                if printing:
                    print('   --high align: one')
                """=========="""
                alignedWithOne = not alignedWithOne
            """=========="""
            if printing:
                print('   --alignment:',high.content.alignedWithOne, high.content.alignedWithOther)
                print('   --compare:  ',alignedWithOne, alignedWithOther)
            """=========="""
            if (high.content.alignedWithOne != alignedWithOne or
                high.content.alignedWithOther != alignedWithOther):
                    raise Exception('alignment error')
                    
            # check for intersection point between neighboring segment 'low'
            # and immediate successor
            """=========="""
            if printing:
                print('\n---finding intersection---')
                print('** low:', low.content, low.succ.content)
            """=========="""
            intersection = low.content.segment.intersection(low.succ.content.segment, bSegment.OPEN);
            if intersection is not None:
                transitions.insert(sRef.intersection(intersection))
                if printing:
                    print('**** intersection found:', intersection)
            # idem for neighboring segment 'high' and immediate predecessor
            if high.pred != low:
                """=========="""
                if printing:
                    print('** high:', high.content, high.pred.content)
                """=========="""
                intersection = high.content.segment.intersection(high.pred.content.segment, bSegment.OPEN);
                if intersection is not None:
                    transitions.insert(sRef.intersection(intersection))
                    if printing:
                        print('**** intersection found:', intersection)
        if len(sweepline) > 2:
            print("Sweep status not empty")
        sweepline.purge()
        transitions.purge()
    
        return classes
            
    def classify(self, other):
        thisForm = coForm(self.sort)
        thisForm.insertClassification(self)
        if other is not None:
            otherForm = coForm(other.sort)
            otherForm.insertClassification(other)
        else:
            otherForm = coForm(self.sort)
        result = self.classification(thisForm, otherForm)
        thisForm.remove(self)
        if other is not None:
            otherForm.remove(other)
        return result
        
    def classify_line(self, lines):
        thisForm = coForm(self.sort)
        thisForm.insertClassification(self)
        bSeg = []
        for i in lines:
            if type(i) == bSegment:
                bSeg.append(i)
            else:
                bSeg.append(bSegment(i.start, i.end))
        result = self.classification_line(thisForm, bSeg)
        return result

    @staticmethod
    def classifyBoundary(sort, boundary1, boundary2):
        form1 = coForm(sort)
        one = planeSegment3D(sort, [boundary1], True)
        form1.insertClassification(one)
        form2 = coForm(sort)
        two = planeSegment3D(sort, [boundary2], True)
        form2.insertClassification(two)
        result = planeSegment3D.classification(form1, form2)
        return result


    @staticmethod
    def classification_line(thisForm, bS, printing = False):
        """=========="""
        if printing:
            print('\n\n_____________________________________________________________')
        """=========="""
        classes = lineClasses()
                    
        # create the status of the sweepline as a splay tree
        # create a splayTree with two boundary segments: 1 from negative y infinity (0, -inf) to positive x infinity(inf, 0)
        #                                                1 from positive y infinity (0,  inf) to positive x infinity(inf, 0)
        """
                           \ |      /
                            \|     /
                      (0,inf)\    /
                             |\  /
                 ____________|_\/____________
                             | /\ (inf, 0)
                             |/  \
                     (0,-inf)/    \
                            /|     \
                           / |      \
                           
           tree:
                           |
                   (0,-inf),(inf,0)   <--- root
                    /            \
                  None     (0,inf),(inf,0)     
        """
        sweepline = splayTree()
        neginf = sSeg(bSegment(bSegment.NEG_Y_INF_3D, bSegment.POS_X_INF_3D))
        posinf = sSeg(bSegment(bSegment.POS_Y_INF_3D, bSegment.POS_X_INF_3D))
        neginf.alignedWithOne = False
        posinf.alignedWithOne = True
        sweepline.initSentinelTree(neginf, posinf)
        
        # create the sweep schedule of transition points as a splay tree
        transitions = splayTree()
        
        #for each of the bsegment in each of the boundary in the coform:
        for segment in thisForm.elements:
            for i in range(len(segment.boundaries)):
                for val in segment.boundaries[i]:
                    #create a new splay segment of the bSegment with membership: False for thisForm, True for otherForm
                    interval = sSeg(bSegment(val.start, val.end))
                    interval.membership = False
                    #add start and end position of the splay segment into transition splay tree
                    transitions.insert(sRef.startPosition(interval))
                    transitions.insert(sRef.endPosition(interval))
            #switch to check otherform if currently checking thisForm, else break while loop
        
        for s in bS:
            interval = sSeg(s)
            interval.membership = True
            transitions.insert(sRef.startPosition(interval))
            transitions.insert(sRef.endPosition(interval))

        # sweep across each of the transition points
        while not transitions.empty():
            #################################
            """=========="""
            if printing:
                print('\n===transitions===')
                transitions.printPredSucc()
                print('---sweepline---')
                sweepline.printPredSucc()
                print('---------------')
            """=========="""
            ##################################
            shared = False
            point = transitions.extractMin()
            # locate point within plane slice defined by sweepline status
            node = sweepline.find(point.position)
            #splay compare the point with the bSegment of the tree node:
            #                       side == 0 --> point is colinear with the segment
            side = node.content.splayCompare(point.position) * -1
            ##################################
            """=========="""
            if printing:
                print('__ point:', point)
                print('__ node :', node.content)
                print('__ side :', side)
            """=========="""            
            
            ##################################
            if side > 0:
                low = node
                high = node.succ
            elif side < 0:
                low = node.pred
                high = node
            else:
                # determine set of edges containing current transition point
                # and delete edges ending at transition point
                low = node.pred
                prev = low
                node = low.succ
                """=========="""
                if printing:
                    print('    __low:', low.content)
                    print('    __node:', node.content)
                """=========="""
                # while point is still colinear with node bSegment (either being an endpoint of the line segment or on the segment):
                while node.content.splayCompare(point.position) == 0:
                    interval = node.content
                    #print('    __interval:', interval)
                    #extract (remove) the node from sweepline if point is the end position of segment
                    if interval.segment.end == point.position:
                        if printing:
                            print("    __point is interval's end -> extract node")
                        sweepline.extract(node)
                    else: # if the point is on the line segment but not an end position --> split the segment into two
                        node.content = sSeg(bSegment(point.position, interval.segment.end, interval.segment.assoc),interval)
                        interval = sSeg(bSegment(interval.segment.start, point.position, interval.segment.assoc),interval)
                        if printing:
                            print('    __point not end point --> split segment')                        
                            print('      __node:    ', node.content)
                            print('      __interval:', interval)
                        prev = node
                        
                    # classify edge into one of the categories:
                    # inner, outer, same-shared or opposite-shared
                    """=========="""
                    if printing:
                        print('       +++', interval.segment, prev.succ.content.segment)
                        print('       +++',shared, interval.segment.start == prev.succ.content.segment.start, interval.segment.direction.parallel(prev.succ.content.segment.direction), interval.membership)
                    """=========="""
                    if shared: #both plane segments contain this bSegment
                        """=========="""
                        if printing:
                            print('       +++interval shared')
                        """=========="""
                        classes.addShared(interval.segment)
                        shared = False #reset shared for next segment 
                        
                    #
                    elif (interval.segment.start == prev.succ.content.segment.start 
                            and interval.segment.direction.parallel(prev.succ.content.segment.direction)):
                        """=========="""
                        if printing:
                            print('       @@ next -> shared')
                        """=========="""
                        shared = True
                    elif interval.membership: #segment is a member of other plane segment only
                        """=========="""
                        if printing:
                            print('       +++interval member of other')
                        """=========="""
                        if interval.alignedWithOne: # align with this plane segment
                            classes.addInner(interval.segment) # --> I2
                        else: # does not a lign with this plane segment
                            classes.addOuter(interval.segment) # --> O2
                    node = prev.succ 
                    """=========="""
                    if printing:
                        print('    __node:    ', node.content)
                    """=========="""
                high = node
                # reverse the order of all other edges in the set
                if prev != low: 
                    sweepline.reverse(low.succ, prev)
                
            #insert all edges with transition point as start point
            ref = point.position
            if point.isStartPosition():
                sweepline.insert(point.segment)
            while (not transitions.empty()) and ref == transitions.findMin().content.position:
                
                point = transitions.extractMin()
                if point.isStartPosition():
                    #insert edge if transition point of type start
                    sweepline.insert(point.segment)

            # update alignment information
            alignedWithOne = low.content.alignedWithOne
            """=========="""
            if printing:
                print('== low:',low.content)
                print('== high:', high.content)
                print('++ low  align:',alignedWithOne)
                print('++ high align:',alignedWithOne)
            """=========="""
            node = low.succ
            while node != high:
                if not node.content.membership:
                    """=========="""
                    if printing:
                        print('    node align: one')
                    """=========="""
                    alignedWithOne = not alignedWithOne
                node.content.alignedWithOne = alignedWithOne
                node = node.succ
            if not high.content.membership:
                """=========="""
                if printing:
                    print('    high align: one__high')
                """=========="""
                alignedWithOne = not alignedWithOne
            """=========="""
            if printing:
                print('^^ ',high)
                print('^^ ',high.content.alignedWithOne)
                print('^^ ',alignedWithOne)
            """=========="""
            if (high.content.alignedWithOne != alignedWithOne):
                    raise Exception('alignment error')
                    
            # check for intersection point between neighboring segment 'low'
            # and immediate successor
            """=========="""
            if printing:
                print('** find intersection low:', low.content, low.succ.content)
            """=========="""
            intersection = low.content.segment.intersection(low.succ.content.segment, bSegment.OPEN);
            if intersection is not None:
                transitions.insert(sRef.intersection(intersection))
            # idem for neighboring segment 'high' and immediate predecessor
            if high.pred != low:
                """=========="""
                if printing:
                    print('** find intersection high:', high.content, high.pred.content)
                """=========="""
                intersection = high.content.segment.intersection(high.pred.content.segment, bSegment.OPEN);
                if intersection is not None:
                    transitions.insert(sRef.intersection(intersection))
        if len(sweepline) > 2:
            print("Sweep status not empty")
        sweepline.purge()
        transitions.purge()
    
        return classes
            




    
    @staticmethod
    def clockwiseAngle(r, p, q):
        cross = p.subtract(r).product(q.subtract(r))
        if cross.z != 0:
            return cross.z < 0
        if cross.y != 0:
            return cross.y > 0
        return cross.z < 0

    
    @staticmethod
    def counterclockwiseAngle(r, p, q):
        cross = p.subtract(r).product(q.subtract(r))
        if cross.z != 0:
            return cross.z > 0
        if cross.y != 0:
            return cross.y < 0
        return cross.z > 0
    
    """/
     * Finds among all adjacency boundary segments of a vertex, the <b>starting
     * segment</b> with the most clockwise polar angle (is ccw from 'below'
     * reference line). This segment is removed from the vertex's adjacency
     * boundary segments list.
    """
    @staticmethod
    def startingSegment(vertex):
        # assume first edge is possible continuation edge
        interval = vertex.edges[0]

        # find edge with most clockwise polar angle
        for i in vertex.edges:
            # if new edge clockwise wrt to current, then update current
            if planeSegment3D.clockwiseAngle(vertex.position, interval.end, i.end):
                interval = i
        vertex.edges.remove(interval)
        return interval
    
    """
     * Finds among all adjacency boundary segments of a vertex, the
     * <b>continuation segment</b> with the smallest clockwise polar angle from
     * the last visited boundary segment. Both segments are removed from the
     * vertex's adjacency boundary segments list.
    """
    @staticmethod
    def continuationSegment(vertex):
        # if there are only two edges left then one has to be the
        # last visited edge and the other the continuation edge
        if len(vertex.edges) == 2:
            edge = vertex.edges[0]
            if edge == vertex.lastVisited():
                edge = vertex.edges[1]
            del vertex.edges[:]
            return edge
            
        # find and remove last visited edge from vertex
        i = 0
        for i in vertex.edges:
            if i == vertex.lastVisited():
                vertex.edges.remove(i)
                break
        # determine reference point for calculation of edge angle
        refpos = vertex.lastVisited().start
        if refpos == vertex.position:
            refpos = vertex.lastVisited().end;

        # assume first edge is possible result
        oldpos = vertex.edges[0].start
        if oldpos == vertex.position:
            oldpos = vertex.edges[0].end
        cw = not planeSegment3D.counterclockwiseAngle(vertex.position, oldpos, refpos)

        # find edge with the smallest clockwise polar angle from reference edge
        index = 0
        for i in range(1, len(vertex.edges)):
            newpos = vertex.edges[i].start
            if newpos == vertex.position:
                newpos = vertex.edges[i].end
            t = not planeSegment3D.counterclockwiseAngle(vertex.position, newpos, refpos)
            r = planeSegment3D.clockwiseAngle(vertex.position, newpos, oldpos)
            # if new edge counterclockwise wrt to current
            # and clockwise wrt reference, then update current
            if ((cw or r) and not t) or (cw and r):
                index = i
                oldpos = newpos
                cw = t
        result = vertex.edges[index]
        vertex.edges.remove(vertex.edges[index])
        return result
    
    @staticmethod
    def removeRedundantVertices(graph):
        for i in list(vertex.vertices):
            if len(vertex.vertices[i].edges) > 2:
                continue
            for j in range(len(vertex.vertices[i].edges)):
                check = False
                for k in range(j+1,len(vertex.vertices[i].edges)):
                    if vertex.vertices[i].edges[j].direction == vertex.vertices[i].edges[k].direction:
                        if vertex.vertices[i].edges[j].start == vertex.vertices[i].position:
                            newSegment = bSegment(vertex.vertices[i].edges[k].start, vertex.vertices[i].edges[j].end)
                            one = str(vertex.vertices[i].edges[j].end.x)+'|'+str(vertex.vertices[i].edges[j].end.y)+'|'+str(vertex.vertices[i].edges[j].end.z)
                            two = str(vertex.vertices[i].edges[k].start.x)+'|'+str(vertex.vertices[i].edges[k].start.y)+'|'+str(vertex.vertices[i].edges[k].start.z)
                            oneEdge = vertex.vertices[i].edges[j]
                            twoEdge = vertex.vertices[i].edges[k]
                        else:
                            newSegment = bSegment(vertex.vertices[i].edges[j].start, vertex.vertices[i].edges[k].end)
                            one = str(vertex.vertices[i].edges[j].start.x)+'|'+str(vertex.vertices[i].edges[j].start.y)+'|'+str(vertex.vertices[i].edges[j].start.z)
                            two = str(vertex.vertices[i].edges[k].end.x)+'|'+str(vertex.vertices[i].edges[k].end.y)+'|'+str(vertex.vertices[i].edges[k].end.z)
                            oneEdge = vertex.vertices[i].edges[j]
                            twoEdge = vertex.vertices[i].edges[k]
                        m = 0
                        while m < len(vertex.vertices[one].edges):
                            if vertex.vertices[one].edges[m] == oneEdge:
                                vertex.vertices[one].edges.pop(m)
                                vertex.vertices[one].edges.append(newSegment)
                                m -= 1
                            m += 1
                        m = 0
                        while m < len(vertex.vertices[two].edges):
                            if vertex.vertices[two].edges[m] == twoEdge:
                                vertex.vertices[two].edges.pop(m)
                                vertex.vertices[two].edges.append(newSegment)
                                m -= 1
                            m += 1
                        m = 0
                        while m < len(graph):
                            if graph[m].position == vertex.vertices[i].position:
                                graph.pop(m)
                                m -= 1
                            m += 1
                        del vertex.vertices[i]
                        check = True
                if check:
                    break
        return graph    
    
    """
     * Returns whether the start of a boundary segment is inside a
     * given boundary.
    """
    @staticmethod
    def insideplaneboundary(segment, boundary, normal):
        left = 0
        inside = True
        
        # take a scanline coincident with the segment
        # and check for intersectionpoints with the boundary
        for n in range(len(boundary)):
            if not inside:
                break
            bound = boundary[n]
            for current in bound:
                if segment.partOf(current):
                    return segment.alignedPositive == current.alignedPositive
                ipt = segment.intersection(current, bSegment.INFINITE)
                # only keep track of the # of intersections left of (or below) start point
                if ipt is not None and ipt <= segment.start:
                    # if ipt is start or end point of the current line, only count the line with the positive dot product with the normal
                    if ipt == current.start:
                        val = normal.dotproduct(current.end.subtract(current.start)) #xvdot(normal, copoint(current->Head), val);
                        if val.sign() < 0:
                            continue
                    elif ipt == current.end:
                        val = normal.dotproduct(current.start.subtract(current.end)) #xvdot(normal, copoint(current->Tail), val);
                        if val.sign() < 0:
                            continue
                    left+=1
            #if ((ptr->prev == NULL) == !(left % 2)) inside = FALSE;
            if n == 0 and left % 2 == 0: 
                inside = False 
        return inside
    
    
    """
     * <b>Determines the boundary type</b> of a plane-segment boundary wrt a
     * list of plane-segment boundaries as inner or outer. If the plane-segment
     * boundary is determined as inner wrt one of the plane-segment boundaries
     * in the list, then its outer boundary will be added as an inner boundary
     * (hole) to this plane-segment boundary, while its inner boundaries (holes)
     * will be added as (outer) boundaries to the list of plane-segment
     * boundaries; otherwise, it will be added to the list of plane-segment
     * boundaries.
     * @param boundaries the boundary (list) of a single plane segment;
     * this list is emptied
     * @param segments the list of plane segment boundaries (boundary lists);
     * this list is updated
    """
    @staticmethod
    def determineboundarytype (boundaries, segments):
        if len(segments) == 0:
            segment = []
            segment += boundaries
            segments.append(segment)
            del boundaries[:]
            return
        
        boundary = boundaries[0]
                
        refval = boundary.canonical[0] #refl = bounds->FirstEl->FirstEl;
        
        # a non-colinear point is sufficient to define the normal vector
        point = boundary.canonical[-1].start #pt = bounds->FirstEl->LastEl->Tail;
        
        
        if point == refval.start or point == refval.end:
            point = boundary.canonical[-1].end #pt = bounds->FirstEl->LastEl->Head;
        
        # compute the normal vector to the plane
        normalv = point.subtract(refval.start); #xvsub(copoint(pt), copoint(refl->Tail), normv);
        planev = refval.direction.product(normalv) #vprod(coline(refl), normv, planev);
        
        # determine comparator plane equation (normal vector)
        normalv = refval.direction.product(planev)
        #Rational factor = normalv.dotProduct(refval.tail()); #vdot(normv, copoint(refl->Tail), normv->h);
        #factor = factor.negate(); #cneg(normv->h, normv->h);
        #normalv = normalv.scale(refval.start.w) #vscal(normv, refl->Tail->Wcoord);
        
        n = 0
        while n < len(segments):
            if planeSegment3D.insideplaneboundary(refval, segments[n], normalv):
                break
            n+=1
        if n < len(segments):
            # new boundary is an inner boundary
            segments[n].append(boundaries.pop(0)) #shappend_and_extract(ptr->el, bounds);
            # inner boundaries become outer boundaries
            while len(boundaries) != 0:
                bounds = []
                bounds.append(boundaries.pop(0))
                segments.append(bounds)
        else:
            segment = []
            segment += boundaries
            segments.append(segment)
            del boundaries[:]
        
    """
     * Extracts the polygonal boundaries that define the plane
     * segments corresponding to a list of (non-intersecting) boundary
     * segments. The list of boundary segments is emptied.
    """
    @staticmethod
    def extractSegments (intervals, maximal):
        segments = []
        boundaries = []
        normal = abs(intervals[0].direction.product(intervals[-1].direction))
        
        # create a sorted set of endposition vectors
        positions = []
        for interval in intervals:
            positions.append(interval.start)
            positions.append(interval.end)
        positions.sort()
        # create a graph of vertices from the set of endposition vectors
        graph = []
        vertex.vertices.clear()
        ref = vertex(positions.pop(0))
        graph.append(ref)
        
        while len(positions)!=0: 
            vector = positions.pop(0)
            if ref.position!=vector:
                ref = vertex(vector)
                graph.append(ref)
            
            string = str(vector.x)+'|'+str(vector.y)+'|'+str(vector.z)
            vertex.vertices[string] = ref
        
        # creates two edge-halves for each line segment,
        # with an entry into the adjacency list of each boundary segment
        while len(intervals)!=0:
            interval = intervals.pop(0)
            
            stringStart = str(interval.start.x)+'|'+str(interval.start.y)+'|'+str(interval.start.z)
            stringEnd = str(interval.end.x)+'|'+str(interval.end.y)+'|'+str(interval.end.z)
            
            vertex.vertices[stringStart].edges.append(interval)
            vertex.vertices[stringEnd].edges.append(interval)
                
                
        planeSegment3D.removeRedundantVertices(graph)
                
        cycle = []
        #ArrayList<BSegment> boundary = new ArrayList<BSegment>();
        bdr = boundary()
        while len(graph)!=0: 
            
            # first vertex is leftmost bottom vertex (always start position)
            vert = graph[0]
            vert.setPartOfCycle(True)
            bdr.add(vert.position)
                        
            # determine starting edge ccw from vertical down direction
            interval = planeSegment3D.startingSegment(vert)

            cycle.append(interval) # add interval into cycle

            refv = normal.product(interval.direction).add(vert.position)
            alignedPositive = True
            
            if planeSegment3D.clockwiseAngle(vert.position, interval.end, refv):
                alignedPositive = False
                
            
            interval.alignedPositive = alignedPositive
            stringEnd = str(interval.end.x)+'|'+str(interval.end.y)+'|'+str(interval.end.z)
            vert = vertex.vertices[stringEnd]	# next vertex is always end position (since first vertex is always start position)

            vert.setLastVisited(interval)
            bdr.add(vert.position)
            while len(cycle) != 0 :
                                
                vert.setPartOfCycle(True)
                # determine next edge cw from current edge
                interval = planeSegment3D.continuationSegment(vert)
                cycle.append(interval)
                # next vert
                if interval.start == vert.position:
                    stringEnd = str(interval.end.x)+'|'+str(interval.end.y)+'|'+str(interval.end.z)
                    vert = vertex.vertices[stringEnd]
                    interval.alignedPositive = alignedPositive
                else:
                    stringStart = str(interval.start.x)+'|'+str(interval.start.y)+'|'+str(interval.start.z)
                    vert = vertex.vertices[stringStart]
                    interval.alignedPositive = not alignedPositive
                vert.setLastVisited(interval)
                if vert.isPartOfCycle():
                    # if vert is already in cycle then a polygon is found
                    # create polygon (current) as a set of lines
                    segment = cycle.pop(-1)
                    stringStart = str(segment.start.x)+'|'+str(segment.start.y)+'|'+str(segment.start.z)
                    stringEnd = str(segment.end.x)+'|'+str(segment.end.y)+'|'+str(segment.end.z)
            
                    vertex.vertices[stringStart].setPartOfCycle(False)
                    vertex.vertices[stringEnd].setPartOfCycle(False)
                    bdr.add(segment)
                    while True:
                        segment = cycle.pop(-1)
                        stringStart = str(segment.start.x)+'|'+str(segment.start.y)+'|'+str(segment.start.z)
                        stringEnd = str(segment.end.x)+'|'+str(segment.end.y)+'|'+str(segment.end.z)
                        tail = vertex.vertices[stringStart]
                        tail.setPartOfCycle(False)
                        head = vertex.vertices[stringEnd]
                        head.setPartOfCycle(False)
                        bdr.add(segment)
                        if len(cycle) == 0 or tail is vert or head is vert:
                            break
                    bdr.sort()
                    #if (!maximal) boundary.maximalize();
                    
                    if len(cycle) == 0:
                        boundaries.insert(0, bdr)
                        bdr = boundary()
                        
                        # add boundaries to list of polygons
                        planeSegment3D.determineboundarytype(boundaries, segments)
                        """
                        if len(segments) == 2:
                            raise ValueError
                        """
                        # remove edge from new vert
                        vert.edges.remove(interval)
                    else:
                        tempbdr = boundary()
                        v = bdr.polygonal[0]
                        while v != vert.position:
                            tempbdr.add(bdr.polygonal.pop(0))
                            v = bdr.polygonal[0]
                        if len(bdr.polygonal) > 0:
                            tempbdr.add(vert.position)
                        boundaries.append(bdr)
                        bdr = tempbdr
                    
                else: 
                    bdr.add(vert.position)
            # find first vertex with non-empty adjacency list
            while len(graph) != 0 and len(graph[0].edges) == 0:
                graph.pop(0)
        
        if len(cycle) != 0:
            raise KeyError("Some vertices are left on the cycle\n")
        
        del cycle[:]
        vertex.vertices.clear()
        
        return segments


    """
     * <b>Constructs</b> a plane-segment from a collection of boundary segments.
     * This collection of boundary segments must correspond to a proper
     * plane-segment boundary (list).
     * @param intervals a <tt>TreeSet</tt> collection of boundary segments
     * defining a plane-segment
     * @param result a {@link CoForm} object to contain the resulting
     * plane-segment
    """
    def construct(self, intervals, result):
        result.clear()
        if len(intervals) == 0:
            return
        segments = planeSegment3D.extractSegments(intervals, True)
        while len(segments) != 0:
            result.insert(self.__class__(self.sort, segments.pop(0), boundaryList = True))
        result.maximalize()

    """        
    def contains(self, other):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).contains()
        
    """
    def contains(self, other):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        selfBounds = self.convertBoundaries()
        otherBounds = other.convertBoundaries()
        if not selfBounds[0].containsBoundary(otherBounds[0]):
            return False
        if len(selfBounds) > 1:
            for i in range(1,len(selfBounds)):
                classes = self.classifyBoundary(self.sort, self.boundaries[i], other.boundaries[0])
                if not (classes.disjoint() or classes.touches()):
                    if (len(otherBounds) > 1 and classes.isContained()):
                        check = False
                        for j in range(1,len(otherBounds)):
                            if otherBounds[j].containsBoundary(selfBounds[i]):
                                check = True
                                break
                        if not check:
                            return False
                    else:
                        return False
        return True
        
    def convertBoundaries(self):
        selfBounds = []
        for i in self.boundaries:
            b = boundary()
            for v in i.polygonal:
                b.polygonal.append(self.localize(v))
            for l in i.canonical:
                b.canonical.append(bSegment(self.localize(l.start),self.localize(l.end)))
            selfBounds.append(b)
        return selfBounds
    
    def touches(self, other):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).touches()

    def disjoint(self, other):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).disjoint()

    def aligns(self, other):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        return self.classify(other).aligns()

    def combine(self, other, result):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        classes = self.classify(other)
        if classes.disjoint():
            return False
        self.construct(classes.sumSet(), result)
        
    def common(self, other, result):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        classes = self.classify(other);
        if classes.disjoint() or classes.touches():
            return False
        self.construct(classes.productSet(), result)
        return True
    
    def complement(self, other, result):
        if 'ind-planeSegment' not in other.type:
            raise ValueError('other individual must be a plane')
        if self.isNil() or other.isNil():
            raise ValueError('Neither plane-segment may be nil')
        classes = self.classify(other)
        if classes.disjoint() or classes.touches():
            result.insert(self)
            return True
        if classes.isContained():
            return False
        self.construct(classes.diffSet(), result)
        return True
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a plane segment')
        if reader.newToken() != '<': #bracket
            raise ParsingError(reader,'expected a "<" ')
            
        bounds = []
        index = 0
        while reader.token() == '<':
            bounds.append(boundary())
            while True:
                reader.newToken() #bracket
                x,y,z = vector3D.readVector(reader) # comma
                start = vector3D(x,y,z) 
                    
                reader.newToken() #bracket
                
                x,y,z = vector3D.readVector(reader) # '>'
                end = vector3D(x,y,z) 
    
                bounds[index].add(bSegment(start, end))
    
                reader.newToken() # ','/'>'/'<'
                if reader.token() == ',':
                    reader.newToken()
                elif reader.token() == '>' or reader.token() == '<':
                    while reader.previewToken() == '<':
                        reader.newToken()
                    break
            
            index += 1
            

        for bdr in bounds:
            vertices = 2
            bdr.add(bdr.canonical[0].start)
            bdr.add(bdr.canonical[0].end)
            vert = bdr.canonical[0].end
            seg = bdr.canonical[0]
            while vertices < len(bdr.canonical):
                for segment in bdr.canonical:
                    if vert == segment.start and seg != segment:
                        bdr.add(segment.end)
                        vert = segment.end
                        seg = segment
                        vertices += 1
                        break
                    if vert == segment.end and seg != segment:
                        bdr.add(segment.start)
                        vert = segment.start
                        seg = segment
                        vertices += 1
                        break


        ind = planeSegment3D(sort, bounds, True)
        classes = ind.classify(None)
        result = coForm(sort)
        ind.construct(classes.sumSet(), result)
        ind.boundaries = result.elements[0].boundaries
        
        reader.newToken()
        return ind
        