# -*- coding: utf-8 -*-
"""
Created on Tue Feb 14 15:12:02 2017

@author: akibdpt
"""
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

from builtins import str

from os import path

from .parseReader import parseReader, ParsingError
from ..sort.sort import sort as __sortClass__
from ..sort.primitiveSort import primitiveSort
from ..sort.attributeSort import attributeSort
from ..sort.disjunctiveSort import disjunctiveSort
from ..sort.compoundSort import compoundSort

from ..grammar.rules.rule import rule
from ..grammar.rules.ruleSequence import ruleSequence
from ..grammar.rules.ruleDisjunction import ruleDisjunction
from ..grammar.rules.ruleDisjunctionP import ruleDisjunctionP
from ..grammar.rules.ruleNegation import ruleNegation
from ..grammar.rules.flow import flow

from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate
from ..struct.key import key

from ..form.form import form as __formClass__

from ..ind.description import description as desc
from ..ind.descriptions.function import function as descFunc


class sdlParser():
    
    unnamedSortCount = 0
    
    sorts = {}
    forms = {}
    rules = {}
    flows = {}
    
    tempSorts = {}
    tempForms = {}
    tempRules = {}
    tempFlows = {}
    characteristics = primitiveSort.characteristicList


    @staticmethod
    def clear():
        sdlParser.sorts = {}
        sdlParser.forms = {}
        sdlParser.rules = {}
        sdlParser.flows = {}
        
    @staticmethod
    def cleanup():
        sdlParser.clear()

    @staticmethod
    def parse(filePath):
            
        sdlParser.tempSorts = {}
        sdlParser.tempRules = {}
        sdlParser.tempFlows = {}
        sdlParser.tempForms = {}
        
        
        file = open(filePath, 'r')
        f = []
        cmdLine = ''
        numLines = 0
        for line in file:
            numLines += 1
            if line[0] == '#':
                continue
            
            cmdLine += line
            if len(line)>1 and ';' in line:
                cmdLine = cmdLine.split(';')[0]+';'
                f.append([cmdLine, numLines])
                numLines = 0
                cmdLine = ''
        file.close()
        totalLines = 0
        for i in f:
            totalLines += i[1]
            r = parseReader(i[0], totalLines)
            while r.token() != ';':
                if not r.token() == parseReader.IDENTIFIER:
                    r.newToken()
                    continue
                if r.value().lower() == 'sort':
                    sname, sort = sdlParser.readSort(r)
                    sdlParser.tempSorts[sname] = sort
                elif r.value().lower() == 'rule':
                    rname, rule = sdlParser.readRule(r)
                    sdlParser.tempRules[rname] = rule
                elif r.value().lower() == 'flow':
                    rname, flow = sdlParser.readFlow(r)
                    sdlParser.tempFlows[rname] = flow
                elif r.value().lower() == 'form':
                    fname, form = sdlParser.readForm(r)
                    sdlParser.tempForms[fname] = form
                else:
                    raise ParsingError(r, 'Unexpected end of command, check for extra or lack of ";" before the error line')
        for i in sdlParser.tempSorts:
            sdlParser.sorts[i] = sdlParser.tempSorts[i]
        for i in sdlParser.tempRules:
            sdlParser.rules[i] = sdlParser.tempRules[i]
        for i in sdlParser.tempFlows:
            sdlParser.flows[i] = sdlParser.tempFlows[i]
        for i in sdlParser.tempForms:
            sdlParser.forms[i] = sdlParser.tempForms[i]
            __formClass__.formRegister[i] = sdlParser.tempForms[i]
        
                    
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#<============================================================================>
#<================================= Reading ==================================>
#<============================================================================>
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    #<========================================================================>
    #<================================= Sort =================================>
    #<========================================================================>
    @staticmethod
    def readSort(reader):
        # sort sortName: sortContent
        # read til 'sortName:' --> readsorts
        if reader.token() == '(':
            pass 
        reader.newToken()
        if reader.token() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as sort name')
        sortName = reader.value()
        if reader.newToken() != ':':
            raise ParsingError(reader, '":" expected after sort name')
        sort = sdlParser.readSortStruct(reader, sortName)
        sdlParser.tempSorts[sortName] = sort
        return sortName, sort
    
    @staticmethod
    def readSortStruct(reader, sortName):
        compSorts = []
        # read the first sort in the structure
        # if sort component is a characteristic: [lineSegment2D]
        if reader.newToken() == '[':
            compSorts.append(sdlParser.readPrimitive(reader, sortName))
            """
            if reader.trailerToken(3) == ';':
                compSorts.append(sdlParser.readPrimitive(reader, sortName))
            else:
                sdlParser.unnamedSortCount += 1
                compSorts.append(sdlParser.readPrimitive(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount)))
            """
        # if open bracket
        elif reader.token() == '(':
            if reader.trailerToken() == ':':
                sName, s = sdlParser.readSort(reader)
            else:
                sdlParser.unnamedSortCount += 1
                s = sdlParser.readSortStruct(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
            compSorts.append(s)
        # if identifier --> existing sort
        elif reader.token() == parseReader.IDENTIFIER:
            eSort = reader.value()
            if eSort in sdlParser.sorts:
                eSort = sdlParser.sorts[eSort]
            else:
                eSort = sdlParser.tempSorts[eSort]
            compSorts.append(eSort)
        # if the line is not ended yet /
        while reader.newToken() != ';' and reader.token() != ')':
            if reader.token() == '^':
                upperBase = None
                oriSort = compSorts.pop()
                base = oriSort

                while base.type == 'sort-attribute':
                    upperBase = base
                    base = base.weight

                """
                if reader.previewToken() == '[':
                    reader.newToken()
                    sdlParser.unnamedSortCount += 1
                    weight = sdlParser.readPrimitive(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
                """
                if reader.newToken() == '(':
                    if reader.trailerToken() == ':':
                        wName, weight = sdlParser.readSort(reader)
                    else:
                        sdlParser.unnamedSortCount += 1
                        weight = sdlParser.readSortStruct(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
                else:
                    weight = reader.value()
                    if weight in sdlParser.sorts:
                        weight = sdlParser.sorts[weight]
                    else:
                        weight = sdlParser.tempSorts[weight]
                    
                if upperBase is None:
                    compSorts.append(attributeSort.create(sortName, base, weight))
                else:
                    upperBase.weight = attributeSort.create(sortName, base, weight)
                    compSorts.append(oriSort)
            elif reader.token() == '+':
                if reader.newToken() == '(':
                    if reader.trailerToken() == ':':
                        sName, s = sdlParser.readSort(reader)
                    else:
                        sdlParser.unnamedSortCount += 1
                        s = sdlParser.readSortStruct(reader, 'unnamedSort'+str(sdlParser.unnamedSortCount))
                    compSorts.append(s)
                else:
                    try:
                        eSort = reader.value()
                        if eSort in sdlParser.sorts:
                            eSort = sdlParser.sorts[eSort]
                        else:
                            eSort = sdlParser.tempSorts[eSort]
                        compSorts.append(eSort)
                    except KeyError:
                        raise ParsingError(reader, 'sort with name '+reader.value()+' does not exist')
        if len(compSorts) > 1:
            """
            sort = disjunctiveSort.create(sortName, compSorts)
            """
            compoundCheck = True
            for i in compSorts:
                if i.type != 'sort-disjunctive':
                    compoundCheck = False
            if compoundCheck:
                sort = compoundSort.create(sortName, compSorts)
            else:
                sort = disjunctiveSort.create(sortName, compSorts)
        elif len(compSorts) == 1:
            sort = compSorts[0]
        return sort
                 
    @staticmethod
    def readPrimitive(reader, sortName):
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as sort characteristic i.e. point, label, description, ...')
        chara = reader.value().lower()
        if reader.newToken() != ']':
            raise ParsingError(reader, '"]" expected after sort characteristic')
        arg = None
        if reader.previewToken() == '(':
            reader.newToken()
            reader.newToken()
            arg = reader.text[reader.index():-2]
            while reader.newToken() != ')':
                pass
        sort = primitiveSort.create(sortName, sdlParser.characteristics[chara], arg)
        return sort
        
    #<========================================================================>
    #<================================= Ind ==================================>
    #<========================================================================>
    @staticmethod
    def readInd(reader, sort, keyList):
        keyCheck = False
        if reader.token() == '#':
            reader.newToken()
            keyValue = reader.value()
            keyCheck = True
            reader.newToken()
        ind = sort.baseSort().characteristic.parse(reader, sort)
        if keyCheck and keyList is not None:
            if keyValue in keyList and not keyList[keyValue].ind == ind:
                raise ParsingError(reader, 'Individual with the same tag name: ' + keyValue +' already exist in this rule')
            keyObj = key(keyValue)
            keyObj.ind = ind
            ind.key = keyObj
            keyList[keyValue] = keyObj
        return ind
        
    @staticmethod
    def readVector(reader):
        if reader.token() != '(':
            raise ParsingError(reader, '"(" expected')
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        x = coordinate(neg+reader.value())
        reader.newToken() # comma
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        y = coordinate(neg+reader.value())
        
        if reader.newToken() == ')':
            return vector2D(x,y)
            
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        z = coordinate(neg+reader.value())
        
        reader.newToken() # bracket

        return vector3D(x,y,z)
       
    #<========================================================================>
    #<================================= Form =================================>
    #<========================================================================>
    @staticmethod
    def readForm(reader):
        # current token being 'form'
        # read form name (for identification)
        if reader.newToken() != '$':
            raise ParsingError(reader, 'Expected a "$" before the form name')
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier after "$" as form name')
        formName = reader.value()
        
        # $formName = sortName: {form elements}
        if reader.newToken() != '=':
            raise ParsingError(reader, 'Expected "=" after form name')
        
        # read elements in form
        reader.newToken()
        form = sdlParser.readFormStruct(reader)
        form.name = formName
        # register form in form list
        sdlParser.tempForms[formName] = form
        return formName, form

    @staticmethod
    def readFormStruct(reader, keyList = None):
        # get the sort of the form
        try:
            sort = reader.value()
            if sort in sdlParser.sorts:
                sort = sdlParser.sorts[sort]
            else:
                sort = sdlParser.tempSorts[sort]
        except:
            raise ParsingError(reader, 'Expected existing sort name')
        if reader.newToken() != ':':
            raise ParsingError(reader, '":" expected after sort name')
        # if next token is '{'
            # sortName: {form content} (creating new form)
        if reader.newToken() == '{':
            # initialize new form with the sort
            form = sort.newForm()
            # if sort is primitive --> form only contains one type of individual
            if sort.type == 'sort-primitive':
                # structure: { individual1 , individual2 , individual3 , ... }
                reader.newToken()
                while reader.token() != '}':
                    form.insert(sdlParser.readInd(reader, sort, keyList))
                    if reader.token() == ',':
                        reader.newToken()
                    elif reader.token() != '}':
                        raise ParsingError(reader, '"," or "}" expected after an individual')
            # if sort is attribute --> form containing one type of individual that has certain 
            #                          other individual as attribute
            elif sort.type == 'sort-attribute':
                reader.newToken()
                while reader.token() != '}':
                    base = sdlParser.readInd(reader, sort, keyList)
                    if reader.token() == '{':
                        reader.newToken()
                        while True:
                            if sort.weight.type == 'sort-disjunctive':
                                w = sdlParser.readFormStruct(reader, keyList)
                            else:
                                w = sdlParser.readInd(reader, sort.weight, keyList)
                            base.addAttribute(w)
                            if reader.token() == ',':
                                reader.newToken()
                            elif reader.token() == '}':
                                break
                        reader.newToken()
                    form.insert(base)
                    if reader.token() == ',':
                        reader.newToken()
                    elif reader.token() != '}':
                        print(base)
                        raise ParsingError(reader, '"," or "}" expected after an individual')
            # if sort is disjunctive --> form containing other forms that 
            #                            contain different type of individuals
            elif sort.type == 'sort-disjunctive' or sort.type == 'sort-compound':
                reader.newToken()
                form = sort.newForm()
                check = True
                if reader.token() == '}':
                    check = False
                while check:
                    form.insert(sdlParser.readFormStruct(reader, keyList))
                    if reader.token() == ',':
                        reader.newToken()
                    elif reader.token() == '}':
                        break
                    else:
                        raise ParsingError(reader, '"," or "}" expected after an individual')
        # if next token is '$': referencing to existing form
            # sortName: $formName|ruleName (applying rule on form)
        elif reader.token() == '$':
            # get formName for reference
            if reader.newToken() != parseReader.IDENTIFIER:
                raise ParsingError(reader, 'Expected an identifier after "$" as form name')
            # get the form referenced
            try:
                form = reader.value()
                if form in sdlParser.forms:
                    form = sdlParser.forms[form].duplicateForm()
                else:
                    form = sdlParser.tempForms[form].duplicateForm()
            except:
                raise ParsingError(reader, 'Expected existing form name after $')
            # if '|' then applying rule
            if reader.previewToken() == '|':
                while reader.previewToken() == '|':
                    # get the rule referenced
                    reader.newToken()
                    reader.newToken()
                    
                    rule = reader.value()
                    if rule in sdlParser.rules:
                        rule = sdlParser.rules[rule]
                    else:
                        rule = sdlParser.tempRules[rule]
                    
                    print('\n======================================================\n___lhs:')
                    rule.lhs.printE('<<<')
                    #print(rule.lhs.elements[0].printVal())
                    #print(rule.rhs.elements[0].printVal())
                    print('___rhs:')
                    rule.rhs.printE('>>>')
                    print('\napply on')
                    form.printE('...')
                    print()
                    # apply rule on form
                    rule.apply(form)
                    print('\nresult:')
                    form.printE('===')
            else:
                reader.newToken() ##################
        reader.newToken()
        return form
        
    #<========================================================================>
    #<================================= Rule =================================>
    #<========================================================================>
    @staticmethod
    def readRule(reader):
        # current token being 'rule'
        # read rule Name
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as rule name')
        ruleName = reader.value()

        # read rule description
        if reader.newToken() != parseReader.STRING:
            raise ParsingError(reader, 'string as description expected after sort name')
        ruleDesc = reader.value()[1:-1] # read description, without 1st and last characters (two doublequotes)
        
        reader.newToken()
            
        # read lhs
        if reader.token() != '<' or reader.newToken() != '<':
            raise ParsingError(reader, 'Expected "<<" to annotate left hand side of the rule')
        reader.newToken()

        keyList = {}
        try:
            lhs = sdlParser.readFormStruct(reader, keyList)
        except UnboundLocalError:
            raise ParsingError(reader,'Form structure is not complete')
        except Exception as err:
            raise ParsingError(reader,'Error in reading, original error message: '+ err.args[0])
        # read predicates (only for graph rule)
        predicates = []
        if reader.token() == ':' and reader.newToken() == ':':
            comma = True
            while comma:
                reader.newToken()
                if reader.value() == '{':
                    cond = 'description'
                    objList = sdlParser.readDescPred(reader, keyList)
                else:
                    cond = reader.value()
                    reader.newToken()
                    if reader.newToken() != '{':
                        raise ParsingError(reader, '"{" expected')
                    objList = []
                    while reader.token() != '}':
                        reader.newToken()
                        prop = []
                        if reader.token() == '(':
                            obj = []
                            while reader.token() != ')':
                                if reader.newToken() == '(' or reader.token() == '{':
                                    if (reader.previewToken() == parseReader.MINUS and reader.trailerToken() == parseReader.NUMBER) or reader.previewToken() == parseReader.NUMBER:
                                        obj.append(sdlParser.readVector(reader))
                                    else:
                                        subObj = []
                                        while reader.token() != ')' and reader.token() != '}':
                                            subObj.append(reader.newValue())
                                            if reader.newToken() != ',' and reader.token() != ')' and reader.token() != '}':
                                                raise ParsingError(reader, '"," or ")" or "}" expected')
                                        obj.append(subObj)
                                elif reader.token() == '#':
                                    keyName = reader.newValue()
                                    if keyName in __sortClass__.register:
                                        obj.append(keyName)
                                    elif keyName in keyList:
                                        obj.append(keyList[keyName])
                                    else:
                                        raise ParsingError(reader, 'Invalid sort or tag name: '+keyName)
                                else:
                                    raise ParsingError(reader, '# required before tag name or invalid predicate structure')
                                    keyName = reader.value()
                                    if keyName in __sortClass__.register:
                                        obj.append(keyName)
                                    elif keyName in keyList:
                                        obj.append(keyList[keyName])
                                    else:
                                        raise ParsingError(reader, 'Invalid sort or tag name: '+keyName)
                                if reader.newToken() != ',' and reader.token() != ')':
                                    raise ParsingError(reader, '"," or ")" expected')
                            """
                            if (reader.previewToken() == parseReader.MINUS and reader.trailerToken() == parseReader.NUMBER) or reader.previewToken() == parseReader.NUMBER:
                                obj = sdlParser.readVector(reader)
                            else:
                                obj = []
                                while reader.token() != ')':
                                    obj.append(reader.newValue())
                                    if reader.newToken() != ',' and reader.token() != ')':
                                        raise ParsingError(reader, '"," or ")" expected')
                            """
                        else:
                            if reader.token() == '>':
                                prop.append(0)
                                reader.newToken()
                            if reader.token() == '#':
                                reader.newToken()
                            else:
                                raise ParsingError(reader, '# required before tag name or invalid predicate structure')
                            keyName = reader.value()
                            if reader.previewToken() == '<':
                                prop.append(1)
                                reader.newToken()
                            if keyName not in keyList:
                                raise ParsingError(reader, 'Tag name '+keyName+' does not exist in the lhs of the rule')
                            obj = keyList[keyName]
                            obj.addProperty(prop)
                        if obj is not None:
                            objList.append(obj)
                        if reader.newToken() != ',' and reader.token() !='}':
                            raise ParsingError(reader, '"," or "}" expected')
                predicates.append([cond, objList])
                if reader.newToken() !=',':
                    comma = False

        # read rhs
        if reader.token() != '>' or reader.newToken() != '>':
            raise ParsingError(reader, 'Expected ">>" to annotate right hand side of the rule')
        reader.newToken()
        
        try:
            rhs = sdlParser.readFormStruct(reader, keyList)
        except UnboundLocalError:
            raise ParsingError(reader,'Form structure is not complete')
        except ValueError as err:
            raise ParsingError(reader,'Error in reading, original error message: '+ err.args[0])
            
        # create rule
        newRule = rule(ruleName, ruleDesc, lhs, rhs, keyList)
        
        # read directives (only for graph rule)
        directives = []
        if reader.token() == ':' and reader.newToken() == ':':
            comma = True
            while comma:
                reader.newToken()
                if reader.value() == '{':
                    cond = 'description'
                    objList = sdlParser.readDescDir(reader, keyList)
                    """
                    for obj in objList:
                        if len(directives) > 0 and directives[-1][0] == obj[0]:
                            directives[-1][1].append(obj[1][0])
                            continue
                        directives.append(obj)
                    """
                else:
                    cond = reader.value()
                    reader.newToken()
                    
                    if reader.newToken() != '{':
                        raise ParsingError(reader, '"{" expected')
                    objList = []
                    while reader.token() != '}':
                        reader.newToken()
                        if reader.token() != '(':
                            raise ParsingError(reader, '"(" expected')
                        obj = []
                        while reader.previewToken() == '#':
                            if reader.trailerToken(3) != ',' and reader.trailerToken(3) != '}' :
                                break
                            try:
                                reader.newToken()
                                keyName = reader.newValue()
                                if keyName not in keyList:
                                    raise ParsingError(reader, 'Tag name '+keyName+' does not exist in the lhs or the rhs of the rule')
                                obj.append(keyList[keyName])
                                reader.newToken()
                            except ValueError:
                                raise ParsingError(reader, 'Key does not exist')
                        reader.newToken()
                        while ((reader.token() == parseReader.MINUS and reader.previewToken() == parseReader.NUMBER) 
                               or reader.token() == parseReader.NUMBER 
                               or reader.token() == parseReader.IDENTIFIER 
                               or reader.token() == parseReader.STRING 
                               or reader.token() == '`'):
                            if (reader.token() == parseReader.MINUS and reader.previewToken() == parseReader.NUMBER):
                                reader.newToken()
                                obj.append(coordinate(reader.value()).__neg__())
                                reader.newToken()
                            elif reader.token() == parseReader.NUMBER:
                                obj.append(coordinate(reader.value()))
                                reader.newToken()
                            elif reader.token() == parseReader.STRING:
                                obj.append(reader.value()[1:-1])
                                reader.newToken()
                            elif reader.token() == parseReader.IDENTIFIER:
                                if reader.value() != 'distance' and reader.value() != 'angle':
                                    raise ParsingError(reader, 'Key does not exist')
                                objs = [reader.value()]
                                bracketCheck = False
                                if reader.newToken() == '(':
                                    bracketCheck = True
                                    reader.newToken()
                                
                                objs.append(reader.value())
                                while reader.newToken() != ')':
                                    if reader.token() == '#':
                                        reader.newToken()
                                        objs.append(reader.value())
                                    elif reader.token() != ',':
                                        objs.append(reader.value())
                                if bracketCheck:
                                    reader.newToken()
                                obj.append(objs)
                                #reader.newToken()
                            else:
                                NilDescSort = __sortClass__.getSort('__dummyDesc__')
                                obj.append(desc.parse(reader, NilDescSort))
                                
                                """
                                d = desc.parse(reader, NilDescSort)
                                val = (d.evaluateInd(newRule)).getValue()[1:-1]
                                obj.append(coordinate(val))
                                """
                            if reader.token() != ',' and reader.token() !=')':
                                raise ParsingError(reader, '"," or ")" expected')
                            reader.newToken()
                        if reader.token() != ',' and reader.token() !='}':
                            raise ParsingError(reader, '"," or "}" expected')
                        objList.append(obj)
                directives.append([cond, objList])
                if reader.newToken() !=',':
                    comma = False
        for i in predicates:
            newRule.addPredicate(i[0],i[1])
        for i in directives:
            newRule.addDirective(i[0],i[1])

        # register rule in rule list
        sdlParser.tempRules[ruleName] = newRule
        return ruleName, newRule
        
    @staticmethod
    def assembleDesc(reader, keyList):
        val = ''
        bracketCount = 0
        while bracketCount != 0 or reader.token() not in (',','}',']','?',':'):
            if reader.token() in ('(','{','[','<'):
                bracketCount += 1
                val += reader.value()
            elif reader.token() in ('}',')',']','>'):
                bracketCount -= 1
                val += reader.value()
            elif reader.token() == '-':
                val+= ' - '
            else:
                val+= reader.value()
            reader.newToken()
        return val
        
    @staticmethod
    def readDescPred(reader, keyList):
        reader.newToken()
        objList = []
        while reader.token() != '}':
            obj = ['`'+ sdlParser.assembleDesc(reader, keyList) + '`']
            if reader.token() != '?':
                raise ParsingError(reader, '"?" expected')
            reader.newToken()
            if reader.token() in ('[','{'):
                comparandList = []
                obj.append(reader.token())
                while True:
                    reader.newToken()
                    comparandList.append('`'+sdlParser.assembleDesc(reader, keyList)+'`')
                    if reader.token() == ',':
                        continue
                    if reader.token() in (']','}'):
                        break
                reader.newToken()
                obj.append(comparandList)
            else:
                if reader.token() == '<' and reader.previewToken() == '>':
                    obj.append('<>')
                    reader.newToken()
                else:
                    obj.append(reader.value())
                reader.newToken()
                obj.append('`'+ sdlParser.assembleDesc(reader, keyList) + '`')
            objList.append(obj)
            if reader.token() == ',':
                reader.newToken()
                continue
            elif reader.token() =='}':
                break
            else:
                raise ParsingError(reader, '"," or "}" expected')
        return objList
        
    @staticmethod
    def readDescDir(reader, keyList):
        reader.newToken()
        objList = []
        while reader.token() != '}':
            cond = reader.value()
            if cond == 'angle':
                if reader.newToken() != '(':
                    raise ParsingError(reader, '"(" expected')
                if reader.newToken() != '#':
                    raise ParsingError(reader, '"#" expected')
                l1 = keyList[reader.newValue()]
                reader.newToken()
                reader.newToken()
                if reader.newToken() != ',':
                    raise ParsingError(reader, '"," expected')
                if reader.newToken() != '#':
                    raise ParsingError(reader, '"#" expected')
                l2 = keyList[reader.newValue()]
                reader.newToken()
                reader.newToken()
                if reader.newToken() != ')':
                    raise ParsingError(reader, '")" expected')
                reader.newToken()
                obj = [l1, l2]
            elif cond == '#':
                l1 = keyList[reader.newValue()]
                reader.newToken()
                cond = reader.newValue()
                reader.newToken()
                obj = [l1]

            if reader.token() != ':' or reader.newToken() != '=':
                raise ParsingError(reader, '":=" expected')
            reader.newToken()
            obj.append(desc(__sortClass__.getSort('__dummyDesc__'),'`'+ sdlParser.assembleDesc(reader, keyList) + '`'))
            objList.append([cond, obj])
            if reader.token() == ',':
                reader.newToken()
                continue
            elif reader.token() =='}':
                break
            else:
                raise ParsingError(reader, '"," or "}" expected')
        return objList
        
    @staticmethod
    def readFlow(reader):
        # current token being 'rule'
        # read rule Name
        if reader.newToken() != parseReader.IDENTIFIER:
            raise ParsingError(reader, 'Expected an identifier as flow name')
        flowName = reader.value()
        
        # read rule description
        if reader.newToken() != parseReader.STRING:
            raise ParsingError(reader, 'string as description expected after flow name')
        flowDesc = reader.value()[1:-1] # read description, without 1st and last characters (two doublequotes)
        
        # read rule description
        if reader.newToken() != parseReader.STRING:
            raise ParsingError(reader, 'string as flow content expected after flow description')
        flowContent = reader.value()[1:-1] # read description, without 1st and last characters (two doublequotes)
        reader.newToken()        
        
        flowReader = parseReader(flowContent)

        r = sdlParser.readFlowRules(flowReader, starting = True)
        
        flowObj = flow(flowName, flowDesc, r)
        return flowName, flowObj
        
    @staticmethod
    def readFlowRules(reader, thisNegation = False, starting = False):
        ruleList = []
        paren = '('
        if not starting:
            paren = reader.token()
            if reader.newToken() == '*':
                paren += '*'
                reader.newToken()
        while reader.token() is not None and reader.token() not in (')', ']'):
            negation = False
            if reader.token() == '!':
                negation = True
                reader.newToken()
            if reader.token() in ('[','('):
                ruleList.append(sdlParser.readFlowRules(reader, negation))
            else:
                ruleName = reader.value()
                try:
                    if ruleName in sdlParser.rules:
                        singleRule = sdlParser.rules[ruleName]
                    else:
                        singleRule = sdlParser.tempRules[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist before referencing')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = sdlParser.readQuantifier(reader)
                    singleRule = ruleSequence([singleRule], quantifier, matching)
                if negation:
                    singleRule = ruleNegation(singleRule)
                ruleList.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = sdlParser.readQuantifier(reader)
            
            
        if paren == '(' or len(ruleList) == 1:
            r = ruleSequence(ruleList, quantifier, matching)
        elif paren == '[*':
            r = ruleDisjunctionP(ruleList, quantifier, matching)
        elif paren == '[':
            r = ruleDisjunction(ruleList, quantifier, matching)
        if thisNegation:
            r= ruleNegation(r)
        if matching == 'PO':
            r.setPo()
        return r
        
    
    @staticmethod
    def readQuantifier(reader):
        quantifier = []
        matching = 'G'
        if reader.newToken() == '{':
            quantifier = []
            if reader.newToken() != parseReader.NUMBER:
                raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
            quantifier.append(int(reader.value()))
            if reader.newToken() == '}':
                quantifier.append(quantifier[0])
            else:
                if reader.token() != ',':
                    raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
                if reader.newToken() != '}':
                    if reader.token() != parseReader.NUMBER:
                        raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
                    quantifier.append(int(reader.value()))
                    if reader.newToken() != '}':
                        raise ParsingError(reader,'"}" expected')
                else:
                    quantifier.append(None)
        elif reader.token() == '?':
            quantifier = [0,1]
        elif reader.token() == '*':
            quantifier = [0,None]
        elif reader.token() == '+':
            quantifier = [1,None]

        if reader.previewToken() is not None and reader.previewToken() in ['?', '+', '*']:
            if reader.newToken() == '?':
                matching = 'L'
            elif reader.token() == '+':
                matching = 'PO'
            elif reader.token() == '*':
                matching = 'PR'
        
        return quantifier, matching
    
                    
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
#<============================================================================>
#<================================= Writing ==================================>
#<============================================================================>
#<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<------>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
                    
                    
    #<========================================================================>
    #<================================= Sort =================================>
    #<========================================================================>
    @staticmethod
    def writeSort(sort, sortsWritten, file):
        sortsWritten.append(sort.name)
        if sort.type == 'sort-primitive':
            file.write('sort '+sort.name+' : '+sort.description+';\n')
        elif sort.type == 'sort-attribute':
            if sort.weight.name not in sortsWritten:
                sdlParser.writeSort(sort.weight, sortsWritten, file)
            if sort.base.name not in sortsWritten:
                sdlParser.writeSort(sort.base, sortsWritten, file)
            file.write('sort '+sort.name+' : '+sort.base.name + sort.separator + sort.weight.name +';\n')
        elif sort.type == 'sort-disjunctive' or sort.type == 'sort-compound':
            for s in sort.component:
                if s.name not in sortsWritten:
                    sdlParser.writeSort(s, sortsWritten, file)
            file.write('sort '+sort.name+' : ') #+sort.description+';\n'
            writeSeparator = False
            for i in sort.component:
                if writeSeparator:
                    file.write(sort.separator)
                else:
                    writeSeparator = True
                file.write(i.name)
            file.write(';\n')


    #<========================================================================>
    #<================================= Ind ==================================>
    #<========================================================================>
                    
                    
    @staticmethod
    def writeOneInd(ind, file):
        file.write(str(ind))
        if ind.hasAttribute():
            sdlParser.writeAttributeForm(ind.attribute, file)
        
    #<========================================================================>
    #<================================= Form =================================>
    #<========================================================================>
    
    @staticmethod
    def writeOneForm(form, file, writekey = True):
        file.write(form.sort.name + ': ')
        s = form.toString('', writekey)
        file.write(s)
            
        
    @staticmethod
    def writeAttributeForm(form, file):
        file.write(' {')
        
        writeComma = False
        for element in form:
            if writeComma:
                file.write(', ')
            else:
                writeComma = True
            
            index = 0
            if 'form' in element.type:
                if index > 0:
                    file.write(', ')
                file.write(element.sort.name + ':')            
                sdlParser.writeAttributeForm(element, file)
                #file.write('}')
            else:
                sdlParser.writeOneInd(element, file)
        
        
        """
        writeComma = False
        for element in form:
            if writeComma:
                file.write(', ')
            else:
                writeComma = True
            if 'form' in element.type:
                file.write(element.sort.name + ': ')
                sdlParser.writeAttributeForm(element, file)
            else:
                sdlParser.writeOneInd(element, file)
                """
        file.write('}')
        
    #<========================================================================>
    #<================================= Rule =================================>
    #<========================================================================>
    @staticmethod
    def writeOneRule(rule, file):
        s = ''
        file.write('\nrule ' + rule.name + ' "' + rule.description + '" '+ s +'\n')
        file.write('<< ')
        
        sdlParser.writeOneForm(rule.lhs, file)
        
        checkPredicates = False
        for i in rule.detection.analysis.maps:
            if 'map-graphMap' in i:
                if len(rule.detection.analysis.maps[i].predicates) != 0:
                    gMap = rule.detection.analysis.maps[i]
                    checkPredicates = True

        if checkPredicates:
            file.write('\n:: ')
            writecomma = False
            for pred in gMap.predicates:
                if writecomma:
                    file.write(', ')
                else:
                    writecomma = True
                if pred[0] == 'description':
                    file.write('{')
                    commaCheck = False
                    for descS in pred[1]:
                        if commaCheck:
                            file.write(', ')
                        else:
                            commaCheck = True
                        sdlParser.writeDescPredDir(file, descS[0])
                        file.write("?")
                        if descS[1] == '[':
                            file.write('[')
                            c = False
                            for subObj in descS[-1]:
                                if c:
                                    file.write(', ')
                                else:
                                    c = True
                                sdlParser.writeDescPredDir(file, subObj[1:-1])
                            file.write(']')
                        elif descS[1] == '{':
                            file.write('{')
                            c = False
                            for subObj in descS[-1]:
                                if c:
                                    file.write(', ')
                                else:
                                    c = True
                                sdlParser.writeDescPredDir(file, subObj[1:-1])
                            file.write('}')
                        else:
                            sdlParser.writeDescPredDir(file, descS[1])
                            sdlParser.writeDescPredDir(file, descS[-1][1:-1])
                    file.write('}')
                            
                else:
                    file.write(pred[0]+': {')
                    checkComma = False
                    for obj in pred[1]:
                        if not checkComma:
                            checkComma = True
                        else:
                            file.write(', ')
                        if isinstance(obj, list):
                            writeComma = False
                            file.write('(')
                            for i in obj:
                                if not writeComma:
                                    writeComma = True
                                else:
                                    file.write(', ')
                                if isinstance(i,list):
                                    if pred[0] == 'void':
                                        file.write('(' + str(i[0]))
                                        for j in i[1:]:
                                            file.write(', '+str(j))
                                        file.write(')')
                                    else:
                                        file.write(str(i[0])+'('+str(i[1]))
                                        for j in i[2:]:
                                            file.write(', '+str(j))
                                        file.write(')')
                                else:
                                    file.write(str(i))
                            file.write(')')
                        else:
                            file.write(str(obj))
                    file.write('}')
        
        file.write('\n>> ')
        
        sdlParser.writeOneForm(rule.rhs, file)
        
        checkDirectives = False
        for i in rule.detection.analysis.maps:
            if 'map-graphMap' in i:
                if len(rule.detection.analysis.maps[i].directives) != 0:
                    gMap = rule.detection.analysis.maps[i]
                    checkDirectives = True

        if checkDirectives:
            file.write('\n:: ')
            writecomma = False
            for pred in gMap.directives:
                if writecomma:
                    file.write(', ')
                else:
                    writecomma = True
                if pred[0] == 'description':
                    file.write('{')
                    c = False
                    for cond in pred[1]:
                        if not c:
                            c = True
                        else:
                            file.write(', ')
                        func = cond[0]
                        if func == 'angle':
                            file.write('angle('+cond[1][0].name+'.direction, '+cond[1][1].name+'.direction):=')
                            file.write(str(cond[1][2]))
                        elif func == 'length':
                            file.write(cond[1][0].name+'.length:=')
                            file.write(str(cond[1][1]))
                    file.write('}')
                else:
                    file.write(pred[0]+': {')
                    checkComma = False
                    index = 0
                    while index < len(pred[1]):
                        obj = pred[1][index]
                        if isinstance(obj, list) and any(x is None for x in obj):
                            for ind0, i in enumerate(obj):
                                if i is None:
                                    ind = ind0
                                    break
                            pred[1].pop(index)
                            pred[1].insert(index, obj[ind+1:])
                            pred[1].insert(index, obj[:ind])
                        index += 1
                    for obj in pred[1]:
                        if not checkComma:
                            checkComma = True
                        else:
                            file.write(', ')
                        if isinstance(obj, list):
                            writeComma = False
                            file.write('(')
                            for i in obj:
                                if not writeComma:
                                    writeComma = True
                                else:
                                    file.write(', ')
                                if isinstance(i,list):
                                    file.write(str(i[0])+'('+str(i[1]))
                                    for j in i[2:]:
                                        file.write(', '+str(j))
                                    file.write(')')
                                else:
                                    file.write(str(i))
                            file.write(')')
                        else:
                            file.write(str(obj))
                    file.write('}')
        file.write(';\n\n')
                    
    @staticmethod
    def writeSeletedRules(destination = path.abspath(path.curdir) +'\\rules.sdl', ruleList = [], append = False):
        if append:
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        ruleList = sorted(ruleList)
        sortsWritten = []
        for r in ruleList:
            if r.lhs.sort.name not in sortsWritten:
                sdlParser.writeSort(r.lhs.sort, sortsWritten, file)
                file.write('\n')
            sdlParser.writeOneRule(r, file)
            
        file.close()

    @staticmethod
    def writeDescPredDir(file, string):
        reader = parseReader(string)
        while reader.token() is not None:
            if reader.token() == '`':
                reader.newToken()
                continue
            if reader.token() == '-':
                file.write(' - ')
                reader.newToken()
                continue
            if __sortClass__.checkSort(reader.value()):
                if reader.trailerToken(3) == ':' and reader.trailerValue(4) == 'tag':
                                            # sortname (current)
                    reader.newToken()       # .
                    prop = reader.newValue()# property
                    reader.newToken()       # :
                    reader.newToken()       # key
                    reader.newToken()       # .
                    reader.newToken()       # value
                    reader.newToken()       # =
                    keyName = reader.newValue()[1:-1]
                    file.write(keyName+'.'+prop)
                    reader.newToken()
                    continue
            file.write(reader.value())
            reader.newToken()

    #<========================================================================>
    #<================================ Grammar ===============================>
    #<========================================================================>
    @staticmethod
    def writeGrammar(destination = path.abspath(path.curdir) +'\\grammar.sdl', ruleList = [], initialForms = [], flows = [], sorts = [], append = False, writeSort = True):
        count = 0
        sortsWritten = []
        if append:
            try:
                file = open(destination, 'r')
                for line in file:
                    l = line.strip()
                    if len(l) == 0 or l[0] == '#':
                        continue
                    reader = parseReader(line)
                    if reader.value().lower() == 'sort':
                        sname = sdlParser.readSort(reader)[0]
                        sortsWritten.append(sname)
                        
                    for word in line.split('$initForm'):
                        count += 1
                    count -= 1
                file = open(destination, 'a')
            except:
                file = open(destination, 'w')
        else:
            file = open(destination, 'w')
        ruleList = sorted(ruleList)
        
        if writeSort:
            for r in ruleList:
                if r.lhs.sort.name not in sortsWritten:
                    sdlParser.writeSort(r.lhs.sort, sortsWritten, file)
                    file.write('\n')
            for f in initialForms:
                if f.sort.name not in sortsWritten:
                    sdlParser.writeSort(f.sort, sortsWritten, file)
                    file.write('\n')
        for s in sorts:
            if s.name not in sortsWritten:
                sdlParser.writeSort(s, sortsWritten, file)
                file.write('\n')
                    
        for r in ruleList:
            sdlParser.writeOneRule(r, file)
            
        for f in flows:
            flowString = str(f.rules)
            if flowString[0] == '(' and flowString[-1] == ')':
                flowString = flowString[1:-1]
            file.write('flow ' + str(f.name) + '"'+str(f.description)+'" "' + flowString + '"' + ';\n')
            
        filename = path.splitext(destination)[0].split('\\')[-1].split('/')[-1] +'_initForm'
        for f in initialForms:
            if f.name is not None:
                file.write('\nform $'+str(f.name)+' = ')
            else:
                file.write('\nform $'+filename+str(count)+' = ')
                count+=1
            sdlParser.writeOneForm(f, file, False)
            file.write(';')
        file.close()

    """
    @staticmethod
    def writeDerivation(ruleList, initialForm, ruleApplications, destination = path.abspath(path.curdir) +'\\grammar.sdl', append = False, writeSort = True):
        count = 0
        if append:
            file = open(destination, 'r')
            for line in file:
                for word in line.split('$initForm'):
                    count += 1
                count -= 1
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        ruleList = sorted(ruleList)
        sortsWritten = []
        for r in ruleList:
            if r.lhs.sort.name not in sortsWritten:
                sdlParser.writeSort(r.lhs.sort, sortsWritten, file)
                file.write('\n')
            sdlParser.writeOneRule(r, file)
            
        filename = path.splitext(destination)[0].split('\\')[-1].split('/')[-1] +'_initForm'
        
        file.write('\nform $'+filename+str(count)+' = ')
        if initialForm.name is not None:
            file.write('\nform $'+str(initialForm.name)+' = ')
        else:
            file.write('\nform $'+filename+str(count)+' = ')
            count+=1
        sdlParser.writeOneForm(initialForm, file, False)
        file.write(';')
        
        for i in ruleApplications:
            if initialForm.name is not None:
                file.write('\nform $'+str(initialForm.name)+' = ')
            else:
                file.write('\nform $'+filename+str(count)+' = ')
                count+=1
            file.write(initialForm.sort.name+': $initForm'+str(count-1)+'|'+i+';')
        file.close()
    """
        
    @staticmethod
    def writeSorts(destination, sorts, append = False):
        sortsWritten = []
        if append:
            file = open(destination, 'r')
            for line in file:
                reader = parseReader(line)
                if reader.value().lower() == 'sort':
                    sname = sdlParser.readSort(reader)[0]
                    sortsWritten.append(sname)
            file = open(destination, 'a')
        else:
            file = open(destination, 'w')
        for sort in sorts:
            if sort.name not in sortsWritten:
                sdlParser.writeSort(sort, sortsWritten, file)
        file.close()
