# -*- coding: utf-8 -*-
"""
Created on Fri Jun 23 11:43:16 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

import itertools
from copy import copy
from random import randint

from .mapping import mapping
from .graphNode import graphNode

from ..io.parseReader import parseReader
from ..transform.graphTransform import graphTransform
from ..transform.paramSet import paramSet

from ..ind.description import description
from ..ind.line import line2D
from ..ind.planeSegment import planeSegment2D

from ..struct.planeStruct import bSegment, boundary
from ..struct.coordinate import coordinate, c0, c1, pi, cos, sin, acos
from ..struct.vector import vector2D

vz = vector2D(0,0)

from ..struct.key import key

from ..sort.sort import sort as sortClass

ANY = 0
PERPENDICULAR = 1
PARALLEL = 2

OUTSIDE = 0
BOUND = 1
INSIDE = 2
STRICT_BOUND = 3

NONE = '[None]'
"""
ANY = '~'
PARALLEL = '='
PERPENDICULAR = '+'

OUTSIDE = '0'
BOUND = 'b'
INSIDE = 'i'
STRICT_BOUND = 's'

NONE = '[--------]'
"""

class graphMap(mapping):
    type = 'map-graphMap'    
    
    def __init__(self, *args, **kwargs):
        super(graphMap, self).__init__()     
        self.nodeList = graphNode()
        self.edgeMatrices = []
        self.intersections = []
        self.matchingNodeList = graphNode()
        self.matchingEdgeMatrices = []
        self.matchingIntersections = []
        self.postAnalysisProcess = True
        
        self.rhsLines = []
        
        self.predicates = []
        self.directives = []
        self.postCheck = False

    def clear(self):
        self.matchingNodeList.clear()
        self.matchingEdgeMatrices = []
        self.matchingIntersections = []
        
    @staticmethod
    def getInstance():
        return graphMap()
        
        
    def addPredicate(self, cond, keyList):
        cond = cond.lower()
        self.predicates.append([cond,keyList])
        
    def addDirective(self, cond, keyList):
        cond = cond.lower()
        self.directives.append([cond,keyList])
    
    def generateNodes(self, ind, matching = False):
        if matching:
            nodeList = self.matchingNodeList
            edgeMatrices = self.matchingEdgeMatrices
            intersections = self.matchingIntersections
        else:
            nodeList = self.nodeList
            edgeMatrices = self.edgeMatrices
            intersections = self.intersections
            

        if ind.type == 'ind-pointP2D':
            nodeList.addPoint(ind)
            return nodeList
        if ind.type == 'ind-planeSegmentP2D':
            nodeList.addPlane(ind)
            return nodeList
        if ind.type == 'ind-circleP2D' or ind.type == 'ind-ellipseP2D' or ind.type == 'ind-bezierCurveP2D':
            nodeList.addCurve(ind)
            return nodeList
            
        element = ind.parametricNode()
        index = nodeList.addInstance(element, ind)
        if index is True:
            edgeMatrices.append([None]*len(nodeList.lines))
            
            for j in range(len(nodeList.lines)-1):
                comparand = nodeList.lines[j].nodeElement
                edge = ANY
                if comparand.parallel(element):
                    edge = PARALLEL
                elif comparand.perpendicular(element):
                    edge = PERPENDICULAR
                    
                
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                if edge != PARALLEL:
                    intersection = comparand.intersection(element)
                    newIntersection = graphNode.intersection(intersection)
                    nodeList.lines[-1].addIntersection(newIntersection)
                    nodeList.lines[j].addIntersection(newIntersection)
                    intersections.append(newIntersection)
                    for k in nodeList.lines[-1].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            newIntersection.iType = 'i'
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                    
                edgeMatrices[j].append([edge, positionOne])
                edgeMatrices[-1][j] = [edge, positionTwo]
                
        else:
            for j in range(len(nodeList.lines)):
                if j == index:
                    continue
                if edgeMatrices[j][index][0] != PARALLEL:
                    
                    positionOne = OUTSIDE
                    positionTwo = OUTSIDE
                    comparand = nodeList.lines[j].nodeElement
                    intersection = comparand.intersection(element)
                    for k in nodeList.lines[index].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                break
                            else:
                                positionOne = BOUND
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                break
                            else:
                                positionTwo = BOUND
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            break
                    
                    edgeMatrices[j][index][1] = positionOne
                    edgeMatrices[index][j][1] = positionTwo
        return nodeList
        
    
    def map(self, fromInd, toInd):
        self.generateNodes(toInd, matching = True)
        return True
        
        
    def distinguish(self, ind):
        node = self.generateNodes(ind)
        return node
        
    def potentialTransforms(self):
        """
        ### printing
        print('\nlhs Matrix:')
        for j in self.edgeMatrices:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\nMatching Matrix:')
        for j in self.matchingEdgeMatrices:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print()
        """
        
        ###
            
        # Extra analysis processes:
        # _ addEndPoint: go through the lists and add all the end points of lines 
        #   that are not intersections into nodeList
        # _ classifyIntersections: classifying the relationship between all the
        #   intersections/end points on a line as connected/disconnected
        # _ organizeForm: order the line segments of a node and combine+separate
        #   them according to their attributes
        if self.postAnalysisProcess:
            self.postAnalysisProcess = False
            self.nodeList.classifyObjs(True)
            for node in self.nodeList:
                node.addEndPoints()
                #node.classifyIntersections()
                node.organizeForm()
        self.matchingNodeList.classifyObjs()
        for matchingNode in self.matchingNodeList:
            matchingNode.addEndPoints()
            #matchingNode.classifyIntersections()
            matchingNode.organizeForm()
            
        """
        print('Nodes:')
        for i in self.nodeList:
            print(i)
        print('Matching nodes:')
        for i in self.matchingNodeList:
            print(i)
        """
        
        

        transforms = []
        matchLists = []
        # detect transform --> find all possible transformation matches for nodes
        # e.g. node 0 in lhs is matched with node 4 in the shape
        #      node 1 in lhs is matched with node 2 in the shape
        #      node 2 in lhs is matched with node 0 in the shape
        #      node 3 in lhs is matched with node 1 in the shape
        #      --> match = [4,2,1,0]
        #      --> matchLists = list of all possible matches
        self.detectTransform(matchLists)
        noInt = True
        # analyzing each match:
        ptransf = [graphTransform()]

        if not self.resolveFreePoints(ptransf):
            return []
        for match in matchLists:
            # printRhino the matching lines with matching color
            """
            print('\n-----------------------------------------------------------')
            print('*match:', match)
            s = [0,0,0,2, int(765/len(match))]
            for index, num in enumerate(match):
                print(self.nodeList.lines[index].nodeElement.printRhino(), end='')
                print(' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone', end=' ')
                print(self.matchingNodeList.lines[num].nodeElement.printRhino(), end='')
                print(' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone')
                s[s[3]] += s[4]
                if s[s[3]] > 255:
                    s[3] -= 1
                    s[s[3]] = s[s[3]+1] - 255
                    s[s[3]+1] = 255
            print()
            """
                    
            transf = [] # initialize a graph transform object for this particular match
            for tr in ptransf:
                transf.append(tr.duplicate())
            check = True              # check: by the end of this process, if check is still True --> valid transformation, if False --> invalid
            
            # for each node in lhs:
            singleIntNodes = []
            vectChecks = []
            for nodeIndex in range(len(self.nodeList)):
                    
                nodeI = self.nodeList.lines[nodeIndex]
                matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
                # find intersections transformation between the nodes and add it into the graph transform
                # e.g. node i in lhs is matched with node i0 in shape
                #      node j in lhs is matched with node j0 in shape
                #      node i  intersects j  at point vectFrom
                #      node i0 intersects j0 at point vectTo
                #      --> vectFrom is matched with vectTo
                for j in range(nodeIndex+1, len(self.nodeList)):
                    if self.matchingEdgeMatrices[match[nodeIndex]][match[j]][0] != PARALLEL:
                        noInt = False
                        vectFrom = self.getIntersection(nodeI,self.nodeList.lines[j]).value
                        vectTo = self.getIntersection(matchNodeI,self.matchingNodeList.lines[match[j]]).value
                        # the intersection match cannot be added to the graph transform
                        # (due to vectFrom already have a different match from VectTo, etc...),
                        # the graph transform is deemed invalid
                        ind = 0
                        while ind < len(transf):
                            tr = transf[ind]
                            if not tr.add(vectFrom, vectTo, index = 0):
                                transf.pop(ind)
                            else:
                                ind += 1
                
                if noInt:
                    break
                if len(transf) == 0 or not self.orderingCheck(transf, nodeI):
                    check = False
                    break
                # if the node doesnt have any end points --> all points on line have a match --> continue onto other ndoe
                if (not nodeI.endP) and len(nodeI.linePoints) == 0:
                    continue
                
                # check if the node and matchingNode are matched in reverse
                # i.e.  node:           --A---B------C----
                #       matchingNode:   ----X--Y--Z----W------
                #       A matches with W, B matches with Z, C matches with X
                #       --> the two nodes match in reverse order
                startI = 0
                while (not transf[0].contains(nodeI.intersections[startI].value)):
                #while (not transf[0].contains(nodeI.intersections[startI].value)) or nodeI.intersections[startI].value < nodeI.forms[0].start:
                    startI += 1
                    if startI >= len(nodeI.intersections):
                        noInt = True
                        break
                if noInt:
                    break
                start = transf[0].apply(nodeI.intersections[startI].value)
                
                endI = -1
                while not transf[0].contains(nodeI.intersections[endI].value):
                #while not transf[0].contains(nodeI.intersections[endI].value) or nodeI.intersections[endI].value > nodeI.forms[-1].end:
                    endI -= 1
                    if startI < -len(nodeI.intersections):
                        noInt = True
                        break
                if noInt:
                     break
                end = transf[0].apply(nodeI.intersections[endI].value)
                

                if start == end:
                    if nodeI.intersections[startI] != nodeI.intersections[endI]:
                        check = False
                        break
                    singleIntNodes.append([nodeI, matchNodeI, startI])
                    for vec in tr.vects:
                        if nodeI.nodeElement.containsPoint(vec[0]):
                            break
                    self.addVectChecks(vectChecks, vec)
                    if not self.vectChecking(vectChecks):
                        check = False
                        singleIntNodes = []
                        break
                    continue
                elif start > end:
                    reverse = True
                else:
                    reverse = False
                    
                # resolve points that are on the node but not an intersection
                if len(nodeI.linePoints) > 0:
                    if len(matchNodeI.linePoints) < len(nodeI.linePoints):
                        check = False
                        break
                    self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
                if not nodeI.endP:
                    continue
                
                if len(transf) == 0:
                    check = False
                    break
                """
                ################################################
                """
                lineForm = []
                specialLineForm = []
                lineIndex = 0
                while lineIndex < len(nodeI.forms):
                    if transf[0].hasCheckPointOf(nodeI.forms[lineIndex]):
                        lineForm.append(nodeI.forms[lineIndex])
                        prevCheck = lineIndex
                        while prevCheck > 0 and nodeI.forms[prevCheck-1].end == nodeI.forms[prevCheck].start:
                            lineForm.append(nodeI.forms[prevCheck-1])
                            prevCheck -= 1
                        while lineIndex < len(nodeI.forms)-1 and nodeI.forms[lineIndex].end == nodeI.forms[lineIndex+1].start:
                            lineForm.append(nodeI.forms[lineIndex+1])
                            lineIndex += 1
                    lineIndex += 1
                for line in nodeI.forms:
                    if line not in lineForm:
                        specialLineForm.append(line)
                        
                transfIndex = 0
                for line in specialLineForm:
                    while transfIndex < len(transf):
                        tr = transf.pop(transfIndex)
                        nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        posMatches = self.resolveSpecialLine(line, nodeInts, reverse, nodeI, matchNodeI)
                        if len(posMatches) == 0:
                            continue
                        for posM in posMatches:
                            additionalTransf = tr.duplicate()
                            if not reverse:
                                additionalTransf.add(line.start, posM.start)
                                additionalTransf.add(line.end, posM.end)                            
                            else:
                                additionalTransf.add(line.start, posM.end)
                                additionalTransf.add(line.end, posM.start)
                            transf.insert(transfIndex, additionalTransf)
                            transfIndex += 1
                """
                nodeInts = transf[0].getCheckPoints(nodeI.nodeElement)
                for line in specialLineForm:
                    print('\n',match)
                    print(nodeInts)
                    posMatches = self.resolveSpecialLine(line, nodeInts, reverse, nodeI, matchNodeI)
                    if len(posMatches) == 0:
                        check = False
                        break
                    else:
                        transfIndex = 0
                        while transfIndex < len(transf):
                            tr = transf.pop(transfIndex)
                            for posM in posMatches:
                                additionalTransf = tr.duplicate()
                                if not reverse:
                                    additionalTransf.add(line.start, posM.start)
                                    additionalTransf.add(line.end, posM.end)                            
                                else:
                                    additionalTransf.add(line.start, posM.end)
                                    additionalTransf.add(line.end, posM.start)
                                transf.insert(transfIndex, additionalTransf)
                                transfIndex += 1
                """
                                
                                
                
                if not check:
                    break
                ind = 0
                while ind < len(transf):
                    tr = transf[ind]
                    trCheck = True
                    ###################nodeInts = transf.getCheckPoints(nodeI.nodeElement)
                    for line in lineForm:
                        hasMatch = self.resolveLine(line, reverse, nodeI, matchNodeI, tr)
                        if not hasMatch:
                            transf.pop(ind)
                            trCheck = False
                            break
                    if not trCheck:
                        break
                    for intersection in nodeI.intersections:
                        if intersection.iType != 'o' and not tr.contains(intersection.value):
                            transf.pop(ind)
                            ind -=1
                            break
                    ind += 1
            if noInt:
                transf = self.resolveNoIntersectionNodes(match, transf)
                if transf is False:
                    continue
            if not self.checkIntersectionPoints(transf):
                continue

            # if there are 2 lines with single intersection and they intersect each other
            if len(singleIntNodes) == 2 and singleIntNodes[0][1].nodeElement.intersect(singleIntNodes[1][1].nodeElement):
                self.resolveSingleIntersectionNodes([singleIntNodes[0]], transf)
                self.resolveSingleIntersectionNodes([singleIntNodes[1]], transf)
            # if there is only 1 line with 1 intersection --> there are 2 other lines that intersect this line at a single point
            elif len(singleIntNodes) == 1:
                if len(transf) == 0:
                    continue
                nodeI = singleIntNodes[0][0]
                matchNodeI = singleIntNodes[0][1]
                anchor = nodeI.intersections[singleIntNodes[0][2]].value
                vec = []
                for v in transf[0].vects:
                    if v[0] == anchor:
                        nAnchor = v[1]
                        continue
                    vec.append(v)

                while vec[0][0].colinear(vec[1][0], anchor):
                    vec.pop(1)

                p = nodeI.forms[0].end if nodeI.forms[0].start == anchor else nodeI.forms[0].start
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)
                for index in range(len(transf)):
                    tr = transf.pop(0)
                    nP = tr.apply(p)
                    ccw1 = graphMap.ccw(vec[0][0], vec[1][0], anchor)
                    ccw2 = graphMap.ccw(vec[0][0], anchor, p)
                    ccw3 = graphMap.ccw(vec[0][1], vec[1][1], nAnchor)
                    ccw4 = graphMap.ccw(vec[0][1], nAnchor, nP)
                    if (ccw1 == ccw2) == (ccw3 == ccw4):
                        transf.append(tr)
                """
            # if there are multiple lines that intersect one another at a single point
            elif len(vectChecks) == 1:
                pass
                """
            # all other cases
            else:
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)
            self.resolvePlanes(transf, match)
            self.resolveCurves(transf)
            if check:
                for tr in transf:
                    #print('----------------')
                    #tr.printTransf()
            
                    transforms.append(tr)
        i = 0
        while i < len(transforms):
            j = i+1
            while j < len(transforms):
                if transforms[i] == transforms[j]:
                    transforms.pop(j)
                else:
                    j += 1
            i += 1
        return transforms

    def detectTransform(self, matchLists):
        for subset in itertools.permutations([k for k in range(len(self.matchingNodeList))], len(self.nodeList)):
            check = True
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.edgeMatrices[indI][indJ]
                    matchEdge = self.matchingEdgeMatrices[i][j]
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            if check:
                matchLists.append(subset)
                
    def resolveSpecialLine(self, line, nodeInts, reverse, nodeI, matchNodeI):
        posMatch = []
        if line.start < nodeInts[0][0]:
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[0][0]) and nodeInts[0][0] != lineCheck.start:
                    checkEnd = True
            for mLine in matchNodeI.forms:
                
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if (not reverse) and mLine.start >= nodeInts[0][1]:
                    continue
                elif reverse and mLine.end <= nodeInts[0][1]:
                    continue
                if not checkEnd:
                    if not reverse:
                        matchStart = mLine.start
                        matchEnd = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.end
                    else:
                        matchStart = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.start
                        matchEnd = mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                else:
                    if checkEnd and mLine.containsPoint(nodeInts[0][1]):
                        continue
                    newLine = mLine
                mergeable = False
                """
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                """
                if not mergeable:
                    posMatch.append(newLine)
                    #print(posMatch)
            #raise ValueError('.... not implemented yet ....')
        elif line.start > nodeInts[-1][0]:
            checkStart = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[-1][0]) and nodeInts[-1][0] != lineCheck.end:
                    checkStart = True
            for mLine in matchNodeI.forms:
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if (not reverse) and mLine.end <= nodeInts[-1][1]:
                    continue
                elif reverse and mLine.start >= nodeInts[-1][1]:
                    continue
                if not checkStart:
                    if not reverse:
                        matchStart = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.start
                        matchEnd = mLine.end
                    else:
                        matchStart = mLine.start
                        matchEnd = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                    
                else:
                    if checkStart and mLine.containsPoint(nodeInts[-1][1]):
                        continue
                    newLine = mLine
                mergeable = False
                """
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                """
                if not mergeable:
                    posMatch.append(newLine)
            #raise ValueError('.... not implemented yet ....')
        else:
            index = 1
            while line.start > nodeInts[index][0]:
                index += 1
            if reverse:
                upperBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                lowerBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
                """
                upperBound =  nodeInts[index-1][1]
                lowerBound =  nodeInts[index][1]
                """
            else:
                lowerBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
                upperBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                """
                upperBound =  nodeInts[index][1]
                lowerBound =  nodeInts[index-1][1]
                """
            #print(' nodeI:', line)
            #print(' lowerBound-upperBound:', lowerBound, upperBound)
            checkStart = False
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[index-1][0]) and nodeInts[index-1][0] != lineCheck.end:
                    checkStart = True
                if lineCheck.containsPoint(nodeInts[index][0]) and nodeInts[index][0] != lineCheck.start:
                    checkEnd = True
            #print('   .....___',checkStart, checkEnd)
            for mLine in matchNodeI.forms:
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if mLine.end <= lowerBound or mLine.start >= upperBound:
                    continue
                if not (checkStart or checkEnd):
                    if not reverse:
                        matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.start
                        matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.end
                    else:
                        matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.end
                        matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.start
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                else:
                    if checkStart and mLine.containsPoint(nodeInts[index-1][1]):
                        continue
                    if checkEnd and mLine.containsPoint(nodeInts[index][1]):
                        continue
                    newLine = mLine
                mergeable = False
                """
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                    """
                if not mergeable:
                    posMatch.append(newLine)
                    #print(posMatch)
        return posMatch
        
    def resolveLine(self, line, reverse, nodeI, matchNodeI, transf):
        checkStart = transf.contains(line.start)
        checkEnd = transf.contains(line.end)
        if checkStart and checkEnd:
            return True
        if checkStart:
            unmatchedInts = [line.end]
        elif checkEnd:
            unmatchedInts = [line.start]
        else:
            unmatchedInts = [line.start, line.end]
        
        
        hasMatch = False
        segmentInts = []

        nodeInts = []
        for nInt in nodeI.intersections:
            if nInt.endP:
                if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                    pass
                else:
                    continue
            matchingInt = transf.apply(nInt.value)
            
            if line.start == nInt.value or line.end == nInt.value:
                alignment = BOUND
                segmentInts.append([nInt.value, matchingInt, alignment])
            elif line.containsPoint(nInt.value):
                alignment = INSIDE
                segmentInts.append([nInt.value, matchingInt, alignment])
            nodeInts.append([nInt.value, matchingInt])
        
        # checking against all line segments in the matching node
        for mLine in matchNodeI.forms:
            # if line's attribute is not a part of matching line's attribute --> continue
            if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                continue
            #print('   ____ mLine:', mLine, mLine.attribute)
            # check if the matching line contains the matching intersections of intersections on line
            lineCheck = True
            for segInt in segmentInts:
                if (mLine.containsPoint(segInt[1]) 
                or (segInt[2] == BOUND and (mLine.start == segInt[1] or mLine.end == segInt[1]))):
                    continue
                lineCheck = False
                break
            if not lineCheck:
                continue
            hasMatch = True
            #print('     ~~~~ match ~~~~ ')
            # for each unmatched points
            for pointIndex, point in enumerate(unmatchedInts):
                posMatch = None # possible match
                orderingPriority = False
                if point < segmentInts[0][0]:
                    if point > nodeInts[0][0]:
                        if reverse:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching segment
                            #       smaller or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[0][1]  
                                and mInt.value <= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching segment
                            #       bigger or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[0][1]  
                                and mInt.value >= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                elif point > segmentInts[-1][0]:
                    if point < nodeInts[-1][0]:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching segment
                            #       bigger or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[-1][1]  
                                and mInt.value >= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching segment
                            #       smaller or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[-1][1]  
                                and mInt.value <= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                else:
                    # point is in between the first and last matched points on segment
                    # narrow down the range in which the point can be 
                    index = 1
                    while point > segmentInts[index][0] and index < len(segmentInts):
                        index += 1
                    # the point must be now between segmentInts[index-1] and segmentInts[index]
                    # i.e. point is bigger than point on node at (index-1) 
                    #          and smaller than point on node at (index) 
                    if reverse:
                        # the node is matched in reverse
                        # --> matching point must be:
                        #       smaller than matched point on matching node at (index-1)
                        #       bigger than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in reversed(matchNodeI.intersections):
                            if (mInt.value < segmentInts[index-1][1]  
                            and mInt.value > segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                    else:
                        # the node is NOT matched in reverse
                        # --> matching point must be:
                        #       bigger than matched point on matching node at (index-1)
                        #       smaller than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in matchNodeI.intersections:
                            if (mInt.value > segmentInts[index-1][1]  
                            and mInt.value < segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                if posMatch is None:
                    continue
                transf.add(point, posMatch, orderingPriority)
        for i in unmatchedInts:
            if not transf.contains(i):
                hasMatch= False
        return hasMatch
        
        
    def resolveSingleIntersectionNodes(self, singleIntNodes, transf):
        if len(singleIntNodes) == 0:
            return transf
        transf1 = copy(transf)
        for sn in singleIntNodes:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
        transf += transf1
        return transf
    
    def resolveSingleIntersectionLine(self, nodeI, matchNodeI, tr, anchor, reverse):
        if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
            return []
        # the single intersection on the line
        anchorFrom = nodeI.intersections[anchor].value
        anchorTo = tr.apply(nodeI.intersections[anchor].value)
        
        # check if the anchor is on the line
        anchorCheck = False
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                anchorCheck = True
                break
        lineMatches = []

        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                lineCheck = 0
            elif line.start > anchorFrom:
                lineCheck = 1
            else:
                lineCheck = -1
            posMatch = []
            for mLine in matchNodeI.forms:
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if not reverse:
                    if lineCheck == 0 and mLine.containsPoint(anchorTo):
                        if line.start == anchorFrom:
                            if anchorTo == mLine.end: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif line.end == anchorFrom:
                            if anchorTo == mLine.start: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        else:
                            newLine = mLine
                    elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                        if mLine.end == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                    elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                        if mLine.start == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                    elif lineCheck == 1 and mLine.start > anchorTo:
                        newLine = mLine
                    elif lineCheck == -1 and mLine.end < anchorTo:
                        newLine = mLine
                    else:
                        continue
                else:
                    if lineCheck == 0 and mLine.containsPoint(anchorTo):
                        if line.start == anchorFrom:
                            if anchorTo == mLine.start: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif line.end == anchorFrom:
                            if anchorTo == mLine.end: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        else:
                            newLine = mLine
                    elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                        if mLine.start == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                    elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                        if mLine.end == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                    elif lineCheck == 1 and mLine.end < anchorTo:
                        newLine = mLine
                    elif lineCheck == -1 and mLine.start > anchorTo:
                        newLine = mLine
                    else:
                        continue
                mergeable = False
                """
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                """
                if not mergeable:
                    posMatch.append(newLine)
            if len(posMatch) == 0:
                return []
            lineMatches.append((line, posMatch))
        indexing = [0] * len(lineMatches)
        transf = []
        transfIndex = 0
        while True:
            additionalTransf = tr.duplicate()
            addCheck = True
            for l in range(len(indexing)):
                line = lineMatches[l][0]
                posM = lineMatches[l][1][indexing[l]]
                if not reverse:
                    additionalTransf.add(line.start, posM.start)
                    additionalTransf.add(line.end, posM.end)
                    """
                    if not (additionalTransf.add(line.start, posM.start)
                    and additionalTransf.add(line.end, posM.end)):
                        addCheck = False
                        break
                    """
                else:
                    additionalTransf.add(line.start, posM.end)
                    additionalTransf.add(line.end, posM.start)
                    """
                    if not (additionalTransf.add(line.start, posM.end)
                    and additionalTransf.add(line.end, posM.start)):
                        addCheck = False
                        break
                    """
            if not self.orderingCheck([additionalTransf], nodeI, True):
                addCheck = False
            
            if addCheck:
                transf.insert(transfIndex, additionalTransf)
            transfIndex += 1
            indexing[0] += 1
            for i in range(len(indexing)-1):
                if indexing[i] == len(lineMatches[i][1]):
                    indexing[i] = 0
                    indexing[i+1] += 1
                else:
                    break
            if indexing[-1] == len(lineMatches[-1][1]):
                break
        self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
        return transf
        
    def resolveNoIntersectionNodes(self, match, transf):
        if len(self.nodeList) == 0:
            return transf
        transf1 = copy(transf)
            
        line0 = self.nodeList.lines[0].nodeElement
        mLine0 = self.matchingNodeList.lines[match[0]].nodeElement
        intersector = matchIntersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(0,1))
        if not intersector.intersect(line0):
            intersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(1,0))
        if not matchIntersector.intersect(mLine0):
            matchIntersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(1,0))
        vectChecks = []
        for nodeIndex in range(len(self.nodeList)):
            if (len(self.nodeList.lines[nodeIndex].forms) > len(self.matchingNodeList.lines[match[nodeIndex]].forms)
            or len(self.nodeList.lines[nodeIndex].linePoints) > len(self.matchingNodeList.lines[match[nodeIndex]].linePoints)):
                return False
                
            lineI = self.nodeList.lines[nodeIndex].nodeElement
            matchLineI = self.matchingNodeList.lines[match[nodeIndex]].nodeElement
            
            v0 = lineI.intersection(intersector)
            v1 = matchLineI.intersection(matchIntersector)
            self.addVectChecks(vectChecks, [v0,v1])
        if not self.vectChecking(vectChecks):
            return False
        for nodeIndex in range(len(self.nodeList)):
            
            nodeI = self.nodeList.lines[nodeIndex]
            matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, True, transf)
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, False, transf1)
        transf += transf1
        
        i = 0
        while i < len(transf):
            j = i+1
            while j <len(transf):
                if transf[i] == transf[j]:
                    transf.pop(j)
                else:
                    j += 1
            i += 1
        """
        for i in transf:
            print()
            i.printTransf()
        """
        return transf
                    
    def resolveNoIntersectionLine(self, nodeI, matchNodeI, tr, reverse):
        transf = []
        for i in itertools.permutations(list(range(len(matchNodeI.forms))), len(nodeI.forms)):
            check = True
            for j in range(len(i)):
                if nodeI.forms[j].sort.type == 'sort-attribute' :
                    if (not reverse) and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[i[j]].attribute, flagDesc = False):
                        check = False
                        break
                    elif reverse and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[i[len(i) - j - 1]].attribute, flagDesc = False):
                        check = False
                        break
            if not check:
                continue
            newtr = tr.duplicate()
            if not reverse:
                for j in range(len(i)):
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[i[j]].start)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[i[j]].end)
            else:
                for j in range(len(i)):
                    k = len(i) - j - 1
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[i[k]].end)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[i[k]].start)
            if self.orderingCheck([newtr], nodeI, True):
                transf.append(newtr)
        return transf
        
    def resolveFreePoints(self, transf):
        fPoints = []
        for p in self.nodeList.points:
            if p.oType == 'f':
                fPoints.append(p.value)
        if len(fPoints) == 0:
            return True
        mFPoints = []
        for p in self.matchingNodeList.points:
            if p.oType == 'f':
                mFPoints.append(p.value)
        if len(fPoints)>len(mFPoints):
            return False
            
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(mFPoints, len(fPoints)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(fPoints)):
                    if not newTr.add(fPoints[j], i[j]):
                        check = False
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True
        
        
    def resolvePointsOnLine(self, nodeI, matchNodeI, reverse, transf):
        for p in nodeI.linePoints:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                if not reverse:
                    if p.value < nodeInts[0][0]:
                        posRange = (None, nodeInts[0][1])
                    elif p.value > nodeInts[-1][0]:
                        posRange = (nodeInts[-1][1], None)
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex-1][1], nodeInts[rangeIndex][1])
                else:
                    if p.value < nodeInts[0][0]:
                        posRange = (nodeInts[0][1], None)
                    elif p.value > nodeInts[-1][0]:
                        posRange = (None, nodeInts[-1][1])
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex][1], nodeInts[rangeIndex-1][1])
                if (posRange[0] is not None and posRange[1] is not None 
                and posRange[0] > posRange[1]):
                    continue
                posMatch = []
                for mP in matchNodeI.linePoints:
                    if ((posRange[0] is not None and mP.value<posRange[0])
                    or (posRange[1] is not None and mP.value>posRange[1])):
                        continue
                    posMatch.append(mP.value)
                for i in posMatch:
                    newTransf = tr.duplicate()
                    if not newTransf.add(p.value, i):
                        continue
                    """
                    if (newTransf.containsMatch(i) and not newTransf.contains(p.value)) or not newTransf.add(p.value, i):
                        continue
                    """
                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1
                    
    def resolvePlanes(self, transf, match):
        for obj in self.nodeList.planes:
            plane = obj.value
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                
                pCheck = True
                
                for poly in plane.boundaries[0].polygonal:
                    if not tr.contains(poly):
                        pCheck = False
                        break
                if pCheck:
                    transf.insert(transfIndex, tr)
                    transfIndex += 1      
                    continue
                
                index = 0
                while index < len(plane.boundaries[0].polygonal):
                    if tr.contains(plane.boundaries[0].polygonal[index]):
                        pCheck = True
                        break
                    index += 1
                if pCheck:
                    mPoly = tr.apply(plane.boundaries[0].polygonal[index])
                    
                for otherObj in self.matchingNodeList.planes:
                    otherPlane = otherObj.value
                    if (otherPlane.sort.baseSort() != plane.sort.baseSort() or (plane.hasAttribute() 
                    and (not plane.attribute == otherPlane.attribute))
                    or len(plane.boundaries[0].polygonal) != len(otherPlane.boundaries[0].polygonal)):
                        continue
                    
                        
                    if not pCheck:
                        continue
                    posMatch = []
                    check = False
                    
                    
                    mIndex = 0
                    while mIndex < len(otherPlane.boundaries[0].polygonal):
                        if otherPlane.boundaries[0].polygonal[mIndex] == mPoly:
                            check = True
                            break
                        mIndex += 1
                    if not check:
                        continue
                    
                    i = index + 1
                    j = mIndex + 1
                    for k in range(1, len(plane.boundaries[0].polygonal)):
                        if i >= len(plane.boundaries[0].polygonal):
                            i -= len(plane.boundaries[0].polygonal)
                        if j >= len(plane.boundaries[0].polygonal):
                            j -= len(plane.boundaries[0].polygonal)
                            
                        if tr.contains(plane.boundaries[0].polygonal[i]):
                            newPoly = tr.apply(plane.boundaries[0].polygonal[i])
                            if otherPlane.boundaries[0].polygonal[j] != newPoly:
                                check = False
                                break
                        else:
                            posMatch.append([plane.boundaries[0].polygonal[i], otherPlane.boundaries[0].polygonal[j]])
                        i += 1
                        j += 1
                    if not check:
                        check = True
                        i = index + 1
                        j = mIndex - 1
                        for k in range(1, len(plane.boundaries[0].polygonal)):
                            if i >= len(plane.boundaries[0].polygonal):
                                i -= len(plane.boundaries[0].polygonal)
                            if j < 0:
                                j += len(plane.boundaries[0].polygonal)
                            if tr.contains(plane.boundaries[0].polygonal[i]):
                                newPoly = tr.apply(plane.boundaries[0].polygonal[i])
                                if otherPlane.boundaries[0].polygonal[j] != newPoly:
                                    check = False
                                    break
                            else:
                                posMatch.append([plane.boundaries[0].polygonal[i], otherPlane.boundaries[0].polygonal[j]])
                            i += 1
                            j -= 1

                    
                    if not check:
                        continue
                    newTr = tr.duplicate()
                    for pM in posMatch:
                        if not newTr.add(pM[0], pM[1]):
                            check = False
                            break
                    if not check:
                        continue
                    
                    for i in range(len(self.nodeList.lines)):
                        nodeI = self.nodeList.lines[i]                        
                        nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        if len(nodeInts) < 2:
                            continue
                        
                        if nodeInts[0][1] > nodeInts[-1][1]:
                            reverse = True
                        else:
                            reverse = False
                            
                        polygons = []
                        for p in plane.boundaries[0].polygonal:
                            if nodeI.nodeElement.containsPoint(p):
                                polygons.append(p)
                        if len(polygons) == 0:
                            continue
                        
                        for i in polygons:
                            j = newTr.transform(i)
                            if j is False:
                                print('! ERROR: missing vector !',i)
                                newTr.printTransf()
                                raise ValueError('Missing vector in Transformation')
                            if i < nodeInts[0][0]:
                                if (reverse and j < nodeInts[0][1]) or (not reverse and j > nodeInts[0][1]):
                                    check = False
                                    break
                            elif i > nodeInts[-1][0]:
                                if (reverse and j > nodeInts[-1][1]) or (not reverse and j < nodeInts[-1][1]):
                                    check = False
                                    break
                            else:
                                index = 0
                                while index < len(nodeInts) and nodeInts[index][0] < i:
                                    index += 1
                                if not (nodeInts[index-1][1] <= j <= nodeInts[index][1]
                                or nodeInts[index][1] <= j <= nodeInts[index-1][1]):
                                    check = False
                                    break
                        if not check:
                            break
                    if not check:
                        continue
                    
                    transf.insert(transfIndex, newTr)
                    transfIndex += 1      


                    
    def resolveCurves(self, transf):
        if len(self.nodeList.curves) == 0:
            return True
        if len(self.nodeList.curves) > len(self.matchingNodeList.curves):
            for i in range(len(transf)):
                transf.pop(0)
        tr = transf[0]
        circleList = []
        ellipseList = []
        bezierList = []
        for i in self.nodeList.curves:
            containedVects = []
            if i.value.type == 'ind-bezierCurveP2D':
                bezierList.append(i)
                continue
            for j in tr.vects:
                if i.value.containsPoint(j[0]):
                    containedVects.append(j[0])
            if len(containedVects) >= 1:
                i.oType = 't'
                if i.value.type == 'ind-ellipseP2D':
                    ellipseList.insert(0,i)
                if i.value.type == 'ind-circleP2D' and not tr.contains(i.value.center):
                    circleList.insert(0,i)
            else:
                if i.value.type == 'ind-ellipseP2D':
                    ellipseList.append(i)
                if i.value.type == 'ind-circleP2D':
                    circleList.append(i)
            i.assocObj = containedVects
            
        otherCircleList = []
        otherEllipseList = []
        otherBezierList = []
        for i in self.matchingNodeList.curves:
            if i.value.type == 'ind-bezierCurveP2D':
                otherBezierList.append(i)
            elif i.value.type == 'ind-circleP2D':
                otherCircleList.append(i)
            elif i.value.type == 'ind-ellipseP2D':
                otherEllipseList.append(i)
                
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherBezierList, len(bezierList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(bezierList)):
                    mp1 = i[j].value.start
                    mp2 = i[j].value.controlPoint
                    mp3 = i[j].value.end
                    if newTr.contains(bezierList[j].value.controlPoint):
                        if newTr.apply(bezierList[j].value.controlPoint) != mp2:
                            check = False
                            break
                        mp2 = None
                    if newTr.contains(bezierList[j].value.start):
                        p1 = newTr.apply(bezierList[j].value.start)
                        if p1 == mp1:
                            mp1 = None
                        elif p1 == mp3:
                            mp3 = mp1
                            mp1 = None
                        else:
                            check = False
                            break
                    if newTr.contains(bezierList[j].value.end):
                        p3 = newTr.apply(bezierList[j].value.end)
                        if p3 == mp3:
                            mp3 = None
                        elif mp1 is not None and p3 == mp1:
                            mp1 = mp3
                            mp3 = None
                        else:
                            check = False
                            break
                    if mp1 is not None:
                        newTr.add(bezierList[j].value.start, mp1)
                    if mp2 is not None:
                        newTr.add(bezierList[j].value.controlPoint, mp2)
                    if mp3 is not None:
                        newTr.add(bezierList[j].value.end, mp3)
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1

        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherCircleList, len(circleList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(circleList)):
                    if circleList[j].oType == 't':
                        for v0 in circleList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):
                                check = False
                                break
                        if not check:
                            break
                        newTr.add(circleList[j].value.center, i[j].value.center, override = True)
                    else:
                        if (not newTr.add(circleList[j].value.center, i[j].value.center)
                        or not newTr.add(circleList[j].value.start, i[j].value.start) 
                        or not newTr.add(circleList[j].value.end, i[j].value.end)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherEllipseList, len(ellipseList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(ellipseList)):
                    if ellipseList[j].oType == 't':
                        for v0 in ellipseList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):
                                check = False
                                break
                        if not check:
                            break
                        newTr.add(ellipseList[j].value.foci[0], i[j].value.foci[0], override = True)
                        newTr.add(ellipseList[j].value.foci[1], i[j].value.foci[1], override = True)
                    else:
                        if (not newTr.add(ellipseList[j].value.foci[0], i[j].value.foci[0])
                        or not newTr.add(ellipseList[j].value.foci[1], i[j].value.foci[1])
                        or not newTr.add(ellipseList[j].value.start, i[j].value.start) 
                        or not newTr.add(ellipseList[j].value.end, i[j].value.end)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True
        
    def orderingCheck(self, transf, nodeI, outsideCheck = False):
        ind = 0
        originalInts = []
        for ints in nodeI.intersections:
            if (not outsideCheck) and ints.iType == 'o':
                continue
            originalInts.append(ints.value)
        originalInts = sorted(originalInts)
        while ind < len(transf):
            tr = transf[ind]
            nodeInts = []
            for i in originalInts:
                if not tr.contains(i):
                    continue
                nodeInts.append([i,tr.apply(i)])
            #nodeInts = tr.getCheckPoints(nodeI.nodeElement)

            if len(nodeInts) < 2:
                ind += 1
                continue
            reverse = False
            index = 1
            while index < len(nodeInts) and nodeInts[index][1] == nodeInts[index-1][1]:
                index += 1
            if index == len(nodeInts):
                ind += 1
                continue
            if nodeInts[index-1][1] < nodeInts[index][1]:
                reverse = False
            else:
                reverse = True
            for index in range(1,len(nodeInts)):
                if ((not reverse and nodeInts[index-1][1] > nodeInts[index][1])
                or (reverse and nodeInts[index-1][1] < nodeInts[index][1])):
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0 :
            return False
        return True
            
    def addVectChecks(self, vectChecks, item):
        for i in range(len(vectChecks)):
            if vectChecks[i][0] > item[0]:
                vectChecks.insert(i, item)
                return
            elif vectChecks[i] == item:
                return
        vectChecks.append(item)
                
    """
    checking if intersection points on a line is in correct ordering
    e.g. lhs:     A---B---C---D     (line1 has 4 intersection points in this order: A, B, C, D)
         matching: a matches with A; b - B; c - C; d - D
         shape1:  a----b-c-----d    (match1 for line1 has 4 matching intersection points in order: a,b,c,d --> correct)
         shape2:  d--c--b--a        (match2 for line1 has 4 matching intersection points in reverse order: d,c,b,a --> correct)
         shape3:  d--a--c--b        (match3 for line1 has 4 matching intersection points in order: d,a,c,b --> wrong)
    """
    def vectChecking(self, vectChecks):
        if len(vectChecks) > 2:
            a = vectChecks[0][0] < vectChecks[1][0]
            b = vectChecks[0][1] < vectChecks[1][1]
            if (a and b) or not (a or b):
                reverse = False
            else:
                reverse = True
            for index in range(2, len(vectChecks)):
                a = vectChecks[index-1][0] < vectChecks[index][0]
                b = vectChecks[index-1][1] < vectChecks[index][1]
                if (a and b) or not (a or b):
                    if reverse:
                        return False
                        break
                elif not reverse:
                    return False
                    break
        return True
        
    def checkIntersectionPoints(self, transf):
        ind = 0
        while ind < len(transf):
            tr = transf[ind]
            for p in self.nodeList.points:
                if p.oType != 'i':
                    continue
                transformedPoint = tr.apply(p.value)
                check = False
                for mP in self.matchingNodeList.points:
                    if mP.oType != 'i':
                        continue
                    if mP.value == transformedPoint:
                        check = True
                        break
                if not check:
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0:
            return False
        return True
                
    @staticmethod
    def getIntersection(line1, line2):
        for i in line1.intersections:
            for j in line2.intersections:
                if i == j:
                    return i
        return None

    def analyzeRHS(self, rhs, shape, transf):
        transfIndex = 0
        while transfIndex < len(transf):
            residuals = []
            if transf[transfIndex].transformable(rhs, residuals):
                transfIndex += 1
                continue
            tr = transf.pop(transfIndex)
            newTransf = [tr]
            for r in residuals:
                if r.type == 'ind-lineSegmentP2D':
                    containedPoints = []
                    ######
                    check = True
                    for i in self.nodeList:
                        if i.nodeElement.compareValue(r) == 0:
                            check = False
                    if check and r not in self.rhsLines:
                        self.rhsLines.append(r)
                    ######
                        
                    for i in tr.vects:
                        if r.containsPoint(i[0]):
                            containedPoints.append(i)
                    if len(containedPoints) == 1:
                        parallelLines = self.getLineProperties('p', rhs, r)
                        i = 0
                        while i < len(parallelLines):
                            if (not tr.contains(parallelLines[i].start)) or (not tr.contains(parallelLines[i].end)):
                                parallelLines.pop(i)
                            else:
                                i+=1
                        if len(parallelLines) == 0:
                            for t in newTransf:
                                if not t.contains(r.start):
                                    t.addUnresolved(r.start)
                                if not t.contains(r.end):
                                    t.addUnresolved(r.end)
                            continue
                        intersectionLines = self.getLineProperties('iE', rhs, r)
                        posMatch = []
                        while len(intersectionLines) >= 2:
                            lr = line2D(r.sort, containedPoints[0][1], containedPoints[0][1].add((tr.apply(parallelLines[0].start).subtract(tr.apply(parallelLines[0].end)))))
                            l0 = line2D(r.sort, tr.apply(intersectionLines[0].start), tr.apply(intersectionLines[0].end))
                            l1 = line2D(r.sort, tr.apply(intersectionLines[-1].start), tr.apply(intersectionLines[-1].end))
                            if not (lr.intersect(l0) and lr.intersect(l1)):
                                newTransf = []
                                continue
                            start = lr.intersection(l0)
                            end = lr.intersection(l1)
                            if start == end:
                                intersectionLines = intersectionLines[1:]
                            else:
                                posMatch.append([[r.start, start],[r.end, end]])
                        if len(intersectionLines) == 1:
                            lr = line2D(r.sort, containedPoints[0][1], containedPoints[0][1].add((tr.apply(parallelLines[0].start).subtract(tr.apply(parallelLines[0].end)))))
                            l0 = line2D(r.sort, tr.apply(intersectionLines[0].start), tr.apply(intersectionLines[0].end))
                            if not lr.intersect(l0):
                                newTransf = []
                                continue
                            # intersection between r and 1st intersection line
                            intR = r.intersection(intersectionLines[0])
                            # intersection between new line r and the transformed 1st intersection line 
                            nIntR = lr.intersection(l0)
                            if r.start < intR < r.end:
                                for p in parallelLines:
                                    if p in residuals:
                                        continue
                                    proj1 = lr.projection(tr.apply(p.start))
                                    proj2 = lr.projection(tr.apply(p.end))
                                    if proj1 < proj2:
                                        proj1, proj2 = proj2, proj1
                                    if not proj1 < nIntR < proj2:
                                        continue
                                    posMatch.append([[r.start, proj1],[r.end, proj2]])
                                    break
                            else:
                                pointR = r.end if r.start == intR else r.start
                                # for each of the parallel lines (p)
                                for p in parallelLines:
                                    # if p is a line that has not been matched yet --> continue onto others
                                    if p in residuals:
                                        continue
                                    # intPoint = intersection point between p and the 1st intersection line
                                    intP = p.intersection(intersectionLines[0])
                                    pointP = p.end if p.start == intP else p.start
                                    checkCCW = graphMap.ccw(pointR, intR, intP) == graphMap.ccw(pointP, intP, intR)
                                    
                                    nPointP = tr.apply(pointP)
                                    lp = line2D(p.sort, tr.apply(p.start), tr.apply(p.end))
                                    if not lp.intersect(l0):
                                        continue
                                    nIntP = lp.intersection(l0)
                                    if checkCCW:
                                        nDir = nIntP.subtract(nPointP)
                                    else:
                                        nDir = nPointP.subtract(nIntP)
                                    nPointR = nIntR.add(nDir)
                                    if r.start == intR or r.end == intR:
                                        posMatch.append([[intR, nIntR],[pointR, nPointR]])
                                    else:
                                        posMatch.append([[intR, nIntR],[r.start, nIntR],[r.end, nPointR]])
                                    break
                                    """
                                    # intPoint = intersection point between p and the 1st line that intersect the new line
                                    intPoint = p.intersection(intersectionLines[0])
                                    # intProj = projection of intPoint onto the new line r
                                    intProj = r.root.add(r.direction.scale(intPoint.subtract(r.root)))
                                    # check if the new line r contains the projection
                                    if r.containsPoint(intProj):
                                        containsProj = True
                                    else:
                                        containsProj = False
                                        
                                    # mP = transformed p
                                    mP = p.__class__(p.sort, tr.apply(p.start), tr.apply(p.end))
                                    #mIntPoint = transformed intPoint
                                    mIntPoint = mP.intersection(l0)
                                    #mIntProj = transformed intProj
                                    mIntProj = lr.root.add(lr.direction.scale(mIntPoint.subtract(lr.root)))
                                    
                                    # start/end = transformed parallel line start/end
                                    start = tr.apply(p.start)
                                    end = tr.apply(p.end)
                                    
                                    # nstart/nEnd = projection of start/end
                                    nStart = containedPoints[0][1].add(start.subtract(end).scale(start.subtract(containedPoints[0][1])))
                                    nEnd = containedPoints[0][1].add(start.subtract(end).scale(end.subtract(containedPoints[0][1])))
                                                     
                                                
                                    # swap nStart, nEnd if nStart > nEnd
                                    if nStart > nEnd:
                                        nStart, nEnd = nEnd, nStart
                                    
                                    # if nStart and nEnd are not on the same side to intersectionP
                                    if (nStart < intersectionP) != (nEnd < intersectionP):
                                        # if 
                                        if containsProj == (mIntProj == nEnd):
                                            if intersectionP != nEnd:
                                                posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                        elif containsProj == (mIntProj == nStart):
                                            if intersectionP != nStart:
                                                posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                                        else:
                                            if intersectionP != nEnd:
                                                posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                    elif nStart > intersectionP and containsProj == (mIntProj == nEnd):
                                        if intersectionP != nEnd:
                                            posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                    elif containsProj == (mIntProj == nStart):
                                        if intersectionP != nStart:
                                            posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                                    """
                        elif len(intersectionLines) == 0:
                            for p in parallelLines:
                                start = tr.apply(p.start)
                                end = tr.apply(p.end)
                                nStart = containedPoints[0][1].add(start.subtract(end).scale(start.subtract(containedPoints[0][1])))
                                nEnd = containedPoints[0][1].add(start.subtract(end).scale(end.subtract(containedPoints[0][1])))
                                posMatch.append([[r.start, nStart],[r.end, nEnd]])
                        lenNewTransf = len(newTransf)
                        for count in range(lenNewTransf):
                            for m in posMatch:
                                newtr = newTransf[0].duplicate()
                                check = True
                                for item in m:
                                    if not newtr.add(item[0],item[1]):
                                        check = False
                                        break
                                if check:
                                    newTransf.append(newtr)
                                newtr = newTransf[0].duplicate()
                                check = True
                                for itemI in range(len(m)):
                                    if not newtr.add(m[itemI][0],m[itemI-1][1]):
                                        check = False
                                        break
                                if check:
                                    newTransf.append(newtr)
                            newTransf.pop(0)
                    elif len(containedPoints) > 1:
                        cStart = True
                        cEnd = True
                        for i in containedPoints:
                            if i[0] == r.start:
                                cStart = False
                            if i[0] == r.end:
                                cEnd = False
                        for t in newTransf:
                            if cStart:
                                t.addUnresolved(r.start)
                            if cEnd:
                                t.addUnresolved(r.end)
                    else:
                        for t in newTransf:
                            t.addUnresolved(r.start)
                            t.addUnresolved(r.end)
                if r.type == 'ind-pointP2D':
                    for t in newTransf:
                        t.addUnresolved(r.value)
            for newTr in newTransf:
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return transf
        
    
            
    @staticmethod
    def getLineProperties(prop, element, lineSeg):
        if 'ind-lineSegment2D' == element.type or 'ind-lineSegmentP2D' == element.type:
            if (prop == 'p' and type(element) == type(lineSeg) and element != lineSeg 
            and element.parallel(lineSeg) and not element.start.colinear(lineSeg.start, lineSeg.end)):
                return element
            if prop == 'i' and type(element) == type(lineSeg) and element != lineSeg and element.intersectSegment(lineSeg):
                return element
            if prop == 'iE' and type(element) == type(lineSeg) and element != lineSeg:
                if (lineSeg.start.colinear(element.start, element.end)
                or lineSeg.end.colinear(element.start, element.end)):
                    return element
            return None
        elif 'ind-point2D' == element.type:
            if prop == 'cP' and lineSeg.containsPoint(element.value):
                return element
            return None
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                p = []
                for f in element.elements:
                    p += graphMap.getLineProperties(prop, f, lineSeg)
                return p
            else:
                p = []
                for l in element.elements:
                    check = graphMap.getLineProperties(prop, l, lineSeg)
                    if check is not None:
                        p.append(check)
                return p
            return []
            
    def getTransform(self, tfs):
        string = self.id.split("=|=")
        if len(string) > 1:
            tfs.current =  string[0]
        return tfs.retrieveTransform(graphTransform)
    
    def evaluatePredicate(self, transfSet, lhs, tlhs, form, param):   
        graphTransf = self.getTransform(transfSet)
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        self.postCheck = True
        for cond in self.predicates:
            if cond[0] == 'maxline':
                for keyI in cond[1]:
                    ind = keyI.ind
                    transfInd = ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], transfInd])
            elif cond[0] == 'bound':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    transfStart = graphTransf.apply(keyI.ind.start)
                    transfEnd = graphTransf.apply(keyI.ind.end)
                    form.accept(self, direction = [cond[0], ind, keyI.properties, transfStart, transfEnd])
                    
            elif cond[0] == 'void':
                for bd in cond[1]:
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            newInd = graphTransf.transform(keyI)
                            if newInd is False:
                                graphTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        lines = []
                        form.accept(self, direction = ['voidof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    else:
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            newInd = graphTransf.transform(keyI)
                            if newInd is False:
                                graphTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        lines = []
                        form.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            elif cond[0] == 'embeds':
                for keyI in cond[1]:
                    bigObj = keyI[0].ind
                    smallObj = keyI[1].ind
                    if 'ind-planeSegment' in bigObj.type:
                        if 'ind-lineSegment' in smallObj.type:
                            check = bigObj.classify_line([smallObj])
                            if check.notInside():
                                self.postCheck = False
                        elif 'ind-point' in smallObj.type:
                            check = bigObj.containsPoint(smallObj.value)
                            if not check:
                                self.postCheck = False
                    elif 'ind-lineSegment' in bigObj.type and 'ind-point' in smallObj.type:
                        if not bigObj.containsPoint(smallObj.value):
                            self.postCheck = False
                    else:
                        self.postCheck = False
            elif cond[0] == 'shortest_line' or cond[0] == 'longest_line':
                for keyI in cond[1]:
                    if isinstance(keyI, list):
                        sorts = keyI[0]
                        if isinstance(sorts, str):
                            sorts = [sorts]
                        for i in keyI[1:]:
                            nInd = i.ind.transform(transfSet)
                            tlhs.accept(self, direction = [cond[0], nInd.squareLength(), sorts])
                    else:
                        ind = keyI.ind.transform(transfSet)
                        tlhs.accept(self, direction = [cond[0], ind.squareLength()])
            elif cond[0] == 'nolabel':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], ind])
            elif cond[0] == 'description':
                for keyI in cond[1]:
                    val = description(sortClass.getSort('__dummyDesc__'), keyI[0])
                    val = val.evaluateInd(param)
                    if keyI[1] == '{':
                        check = False
                        for comp in keyI[-1]:
                            
                            comparand = description(sortClass.getSort('__dummyDesc__'), comp)
                            comparand = comparand.evaluateInd(param)
                            if comparand == val:
                                check = True
                                break
                    elif keyI[1] == '[':
                        if len(keyI[-1]) != 2:
                            raise ValueError('[ ] condition only accepts 2 arguments as min and max. e.g. [1,3] --> value between 1 and 3')
                        comp1 = description(sortClass.getSort('__dummyDesc__'), keyI[-1][0])
                        comp1 = comp1.evaluateInd(param)
                        comp2 = description(sortClass.getSort('__dummyDesc__'), keyI[-1][-1])
                        comp2 = comp2.evaluateInd(param)
                        if comp1 < val < comp2:
                            check = True
                        else:
                            check = False
                    else:
                        comparand = description(sortClass.getSort('__dummyDesc__'), keyI[-1])
                        comparand = comparand.evaluateInd(param)
                        if keyI[1] == '=':
                            check = val == comparand
                        if keyI[1] == '<':
                            check = val < comparand
                        if keyI[1] == '>':
                            check = val > comparand
                        if keyI[1] == '<=':
                            check = val <= comparand
                        if keyI[1] == '>=':
                            check = val < comparand
                        if keyI[1] == '<>':
                            check = not (val == comparand)
                    if not check:
                        self.postCheck = False
            if self.postCheck is False:
                return False
        return True
            
    def solveDirectives(self, transfSet, lhs, tlhs, rhs, form, param):
        transf = [transfSet]
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        param.rhs = rhs
        for cond in self.directives:
            if cond[0] != 'distance':
                continue
            i = 0
            while i < len(cond[1]):
                if len(cond[1][i]) > 4:
                    i+=1
                    continue
                for j in range(i+1, len(cond[1])):
                    if cond[1][i][1].ind == cond[1][j][1].ind:
                        dir2 = cond[1].pop(j)
                        cond[1][i].append(None)
                        for k in dir2:
                            cond[1][i].append(k)
                        break
                i+=1
        for cond in self.directives:
            if cond[0] == 'point_on_line':
                for i0 in range(len(transf)):
                    tfSet = transf.pop(0)
                    param.transform = tfSet
                    graphTransf = self.getTransform(tfSet)
                    for p in cond[1]:
                        bounds, evalCheck = self.evaluateDesc(p, param)
                        if not evalCheck:
                            break
                        if bounds[-2].__class__ != coordinate:
                            bounds.append(bounds[-1])
                        if len(bounds) == 2:
                            continue
                        for i in range(0,len(bounds)-2):
                            obj = p[i].ind
                            self.solvePointOnLine(graphTransf, obj, bounds)
                    if not evalCheck:
                        continue
                    transf.append(tfSet)
            elif cond[0] == 'normal':
                for p in cond[1]:
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        param.transform = tfSet
                        newSet = tfSet.duplicate()
                        tr = self.getTransform(newSet)
                        
                        # evaluate descriptions in the directive
                        bounds, evalCheck = self.evaluateDesc(p, param)
                        if not evalCheck:
                            continue
                        
                        # line = new notmal line to be added
                        line = bounds[0].ind
                        # oPt = rhs point the new line contains; pt = oPt match in shape
                        if bounds[1].type == 'struct-key':
                            oPt = bounds[1].ind
                            pt = oPt.transform(newSet)
                            strdLen = 6
                        else:
                            check = False
                            for i in tr.vects:
                                if line.containsPoint(i[0]):
                                    oPt = i[0]
                                    pt = i[1]
                                    check = True
                                    strdLen = 5
                                    break
                            if not check:
                                continue
                            
                        
                        # nDir = normal direction i.e. direction of the new line
                        nDir = vector2D(bounds[3], c0 - bounds[2])
                        # nLine = 2D line with pt and nDir
                        nLine = line2D(line.sort, pt.value, pt.value.add(nDir))
                        
                        # determine length of the normal line
                        if len(bounds) == strdLen - 2:
                            length = line.length()
                        elif len(bounds) == strdLen - 1:
                            length = bounds[-1]
                        elif len(bounds) == strdLen:
                            rand = randint(0, 100000)
                            length = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
                        
                        cStart = False
                        cEnd = False
                        
                        
                        for i in self.nodeList.lines:
                            if i.nodeElement.containsPoint(line.start):
                                cStart = True
                                #cContains = i.containsPoint(line.start)
                                break
                            if i.nodeElement.containsPoint(line.end):
                                cEnd = True
                                #cContains = i.containsPoint(line.end)
                                break
                        
                        if oPt.value == line.start:
                            cStart = None
                            oPt1 = line.end
                        elif oPt.value == line.end:
                            cEnd = None
                            oPt1 = line.start
                            
                        if not cStart and not cEnd:
                            # check if there is any other line that does not contains oPt
                            cc = False
                            for nodeLine in self.nodeList.lines:
                                if not nodeLine.nodeElement.containsPoint(oPt.value):
                                    cc = True
                                    break
                            # if there is no other line --> use nLine's unitDir
                            if not cc:
                                nPt = pt.value.add(nLine.unitDir.scale(length))
                                if cStart is None:
                                    tr.add(line.end, nPt)
                                elif cEnd is None:
                                    tr.add(line.start, nPt)
                            if cStart is None or cEnd is None:
                                l0 = nodeLine.forms[0]
                                sP1 = tr.apply(l0.start)
                                eP1 = tr.apply(l0.end)
                                if graphMap.ccw(l0.end, l0.start, oPt.value) == graphMap.ccw(eP1, sP1, pt.value):
                                    inverted = False
                                else:
                                    inverted = True
                                nPt1 = pt.value.add(nLine.unitDir.scale(length))
                                nPt2 = pt.value.subtract(nLine.unitDir.scale(length))
                                count1 = 0
                                count2 = 0
                                for i in tr.vects:
                                    if i[0].colinear(oPt.value, oPt1) or i[1].colinear(pt.value, nPt1):
                                        continue
                                    oPtccw = graphMap.ccw(oPt1, oPt.value, i[0])
                                    if (oPtccw == graphMap.ccw(nPt1, pt.value, i[1])) != inverted:
                                        count1 += 1
                                    if (oPtccw == graphMap.ccw(nPt2, pt.value, i[1])) != inverted:
                                        count2 += 1
                                nPt = nPt1 if count1 >= count2 else nPt2
                                if cStart is None:
                                    tr.add(line.end, nPt)
                                elif cEnd is None:
                                    tr.add(line.start, nPt)
                            else:
                                splitLength = length/2
                                splitVec = nLine.unitDir.scale(splitLength)
                                nPt1 = pt.value.add(splitVec)
                                nPt2 = pt.value.subtract(splitVec)
                                tr.add(line.start, nPt1)
                                tr.add(line.end, nPt2)
                                
                            """
                            intPt = None
                            print('')
                            for nodeLine in self.nodeList.lines:
                                if nodeLine.nodeElement.intersect(line):
                                    l = nodeLine.forms[0].transform(newSet)
                                    if not line2D.intersect(l, nLine):
                                        continue
                                    intersection = nodeLine.nodeElement.intersection(line)
                                    if intersection == oPt.value:
                                        continue
                                    nIntersection = line2D.intersection(l, nLine)
                                    tr.add(intersection, nIntersection)
                                    if intersection in tr.unresolved:
                                        tr.unresolved.remove(intersection)
                                    if intPt is None:
                                        intPt = (intersection, nIntersection)
                                    elif line.containsPoint(intersection) and ((not nodeLine.containsPoint(intersection)) or nodeLine.transfContainsPoint(newSet, nIntersection)):
                                        if intersection.subtract(bounds[1].ind.value).squareLength() < intPt[0].subtract(bounds[1].ind.value).squareLength():
                                            intPt = (intersection, nIntersection)
                            if intPt is None:
                                continue
                                #intPt = (intersection, nIntersection)
                            if cStart is None:
                                nPt = pt.value.add(intPt[1].subtract(pt.value).unit().scale(length))
                                tr.add(line.end, nPt)
                            elif cEnd is None:
                                nPt = pt.value.subtract(intPt[1].subtract(pt.value).unit().scale(length))
                                tr.add(line.start, nPt)
                            else:
                                splitLength = (length - intPt[1].subtract(pt.value).length())/2
                                splitVec = intPt[1].subtract(pt.value).unit().scale(splitLength)
                                nPt1 = intPt[1].add(splitVec)
                                nPt2 = pt.value.subtract(splitVec)
                                tr.add(line.start, nPt1)
                                tr.add(line.end, nPt2)
                            """
                        else:
                            nIntersector = i.forms[0].transform(newSet)
                            if not nLine.intersect(nIntersector):
                                continue
                            nInt = nLine.intersection(nIntersector)
                            nPt = nInt.add(pt.value.subtract(nInt).unit().scale(length))
                            """
                            if cContains and not nodeLine.transfContainsPoint(newSet, nIntersection):
                                continue
                            """
                            if cStart:
                                tr.add(line.start, nInt)
                                tr.add(line.end, nPt)
                            else:
                                tr.add(line.start, nPt)
                                tr.add(line.end, nInt)
                            
                            for nodeLine in self.nodeList.lines:
                                if nodeLine.nodeElement.intersect(line):
                                    l = nodeLine.forms[0].transform(newSet)
                                    if not line2D.intersect(l, nLine):
                                        continue
                                    intersection = nodeLine.nodeElement.intersection(line)
                                    nIntersection = line2D.intersection(l, nLine)
                                    tr.add(intersection, nIntersection)
                                    if intersection in tr.unresolved:
                                        tr.unresolved.remove(intersection)
                        if line.start in tr.unresolved:
                            tr.unresolved.remove(line.start)
                        if line.end in tr.unresolved:
                            tr.unresolved.remove(line.end)
                        for lhsl in self.nodeList:
                            if lhsl.nodeElement.intersect(line):
                                lhsl1 = lhsl.forms[0].edit(newSet)
                                if not nLine.intersect(lhsl1):
                                    continue
                                tr.add(line.intersection(lhsl.nodeElement), lhsl1.intersection(nLine))
                        for rhsl in self.rhsLines:
                            if tr.contains(rhsl.start) and tr.contains(rhsl.end) and rhsl.intersect(line):
                                rhsl1 = rhsl.edit(newSet)
                                if not nLine.intersect(rhsl1):
                                    continue
                                tr.add(line.intersection(rhsl), rhsl1.intersection(nLine))
                        transf.append(newSet)
            elif cond[0] == 'angle':
                for p in cond[1]:
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        param.transform = tfSet
                        bounds, evalCheck = self.evaluateDesc(p, param)
                        if not evalCheck:
                            continue
                        
                        posMatch = self.solveAngle(tfSet, bounds, lhs)
                        
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = self.getTransform(ntfSet)
                            adding = True
                            count = 0
                            for i in j:
                                if count >=2:
                                    tr.add(i[0],i[1])
                                    if i[0] in tr.unresolved:
                                        tr.unresolved.remove(i[0])
                                    continue
                                present = False
                                if i[0] in tr.unresolved:
                                    tr.unresolved.remove(i[0])
                                    present = True
                                count += 1
                                if not present:
                                    vCheck = False
                                    for vec in tr.vects:
                                        if vec[0] == i[0] and vec[1] == i[1]:
                                            vCheck = True
                                            break
                                    if not vCheck:
                                        adding = self.unifyDirectives(i, j, tr)
                                else:
                                    tr.add(i[0],i[1])
                            if adding:
                                tr.addDirectiveResult(j)
                                transf.append(ntfSet)
            elif cond[0] == 'distance':
                for p in cond[1]:
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        param.transform = tfSet
                        bounds, evalCheck = self.evaluateDesc(p, param)
                        if not evalCheck:
                            continue
                        if len(bounds) > 4:
                            cond1 = []
                            cond2 = []
                            index = 0
                            while bounds[index] is not None:
                                cond1.append(bounds[index])
                                index += 1
                            index += 1
                            while index < len(bounds):
                                cond2.append(bounds[index])
                                index += 1
                            d1 = self.evaluateBounds(cond1, tfSet)
                            d2 = self.evaluateBounds(cond2, tfSet)
                            if 'lineSegment' in cond1[1].ind.type:
                                posMatch = self.solveDoublePointLineDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind)
                            elif cond1[0].ind.type != cond2[0].ind.type:
                                if 'lineSegment' in cond1[0].ind.type:
                                    cond1, cond2 = cond2, cond1
                                    d1, d2 = d2, d1
                                posMatch = self.solvePointPointLinePointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind)
                            elif 'point' in cond1[0].ind.type:
                                posMatch = self.solveDoublePointPointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind)
                            elif 'lineSegment' in cond1[0].ind.type:
                                posMatch = self.solveDoubleLinePointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind)
                        elif 'lineSegment' in p[0].ind.type and 'lineSegment' in p[1].ind.type:
                            posMatch = self.solveLineLineDistance(tfSet, bounds)
                        elif 'lineSegment' in p[0].ind.type and 'point' in p[1].ind.type:
                            posMatch = self.solveLinePointDistance(tfSet, bounds)
                        elif 'point' in p[0].ind.type and 'lineSegment' in p[1].ind.type:
                            posMatch = self.solvePointLineDistance(tfSet, bounds)
                        elif 'point' in p[0].ind.type and 'point' in p[1].ind.type:
                            posMatch = self.solvePointPointDistance(tfSet, bounds)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = self.getTransform(ntfSet)
                            for i in j:
                                for uI in range(len(tr.unresolved)):
                                    u = tr.unresolved[uI]
                                    if u == i[0]:
                                        tr.add(i[0],i[1], override = True)
                                        tr.unresolved.pop(uI)
                                        break
                            transf.append(ntfSet)
            elif cond[0] == 'description':
                for subCond in cond[1]:
                    p = subCond[1]
                    if subCond[0] == 'angle':
                        for i0 in range(len(transf)):
                            tfSet = transf.pop(0)
                            param.transform = tfSet
                            bounds, evalCheck = self.evaluateDesc(p, param)
                            if not evalCheck:
                                continue

                            posMatch = self.solveAngle(tfSet, bounds, lhs)
                            
                            for j in posMatch:
                                ntfSet = tfSet.duplicate()
                                tr = self.getTransform(ntfSet)
                                adding = True
                                count = 0
                                for i in j:
                                    if count >=2:
                                        tr.add(i[0],i[1])
                                        if i[0] in tr.unresolved:
                                            tr.unresolved.remove(i[0])
                                        continue
                                    present = False
                                    if i[0] in tr.unresolved:
                                        tr.unresolved.remove(i[0])
                                        present = True
                                    count += 1
                                    if not present:
                                        vCheck = False
                                        for vec in tr.vects:
                                            if vec[0] == i[0] and vec[1] == i[1]:
                                                vCheck = True
                                                break
                                        if not vCheck:
                                            adding = self.unifyDirectives(i, j, tr)
                                    else:
                                        tr.add(i[0],i[1])
                                if adding:
                                    tr.addDirectiveResult(j)
                                    transf.append(ntfSet)
                    elif subCond[0] == 'length':
                        for i0 in range(len(transf)):
                            tfSet = transf.pop(0)
                            param.transform = tfSet
                            bounds, evalCheck = self.evaluateDesc(p, param)
                            if not evalCheck:
                                continue
                            posMatch = self.readjustLength(tfSet, bounds, lhs, rhs)
                            for j in posMatch:
                                ntfSet = tfSet.duplicate()
                                tr = self.getTransform(ntfSet)
                                for i in j:
                                    tr.add(i[0],i[1], override = True)
                                    if i[0] in tr.unresolved:
                                        tr.unresolved.remove(i[0])
                                transf.append(ntfSet)
                pass
            
        for directive in self.directives:
            if 'length' == directive[0]:
                for p in range(len(directive[1])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        param.transform = tfSet
                        bounds, evalCheck = self.evaluateDesc(directive[1][p], param)
                        if not evalCheck:
                            continue
                        posMatch = self.readjustLength(tfSet, bounds, lhs, rhs)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = self.getTransform(ntfSet)
                            for i in j:
                                tr.add(i[0],i[1], override = True)
                                if i[0] in tr.unresolved:
                                    tr.unresolved.remove(i[0])
                            transf.append(ntfSet)
        i0 = 0
        while i0 < len(transf):
            if len(self.getTransform(transf[i0]).unresolved) > 0:
                tfSet = transf.pop(i0)
                graphTransf = self.getTransform(tfSet)
                """
                x = 0
                while x < len(graphTransf.unresolved):
                    if graphTransf.contains(graphTransf.unresolved[x]):
                        graphTransf.unresolved.pop(x)
                    else:
                        x += 1
                """
                #graphTransf.printTransf()

                lhs.accept(self, direction = ['point_on_line', graphTransf, True])
                if len(graphTransf.unresolved) > 0:
                    continue
                transf.append(tfSet)
            i0 += 1
        return transf
        
    def unifyDirectives(self, coordMatch, posMatch, transf):
        if len(coordMatch) < 2:
            return False
        otherResults = transf.getDirectiveResults(coordMatch)
        intersection = None
        for i in otherResults:
            l1 = line2D(sortClass.getSort('__dummyLine2D__'), i[0][1], i[1][1])
            l2 = line2D(sortClass.getSort('__dummyLine2D__'), posMatch[0][1], posMatch[1][1])
            if not l1.intersect(l2):
                continue
            ints = l1.intersection(l2)
            if intersection is None:
                intersection = ints
            elif intersection != ints:
                return False
        if intersection is None:
            return False
        transf.add(coordMatch[0], intersection, override = True)
        return True
        

    def visitEnter(self, obj, direction = 0):
        if direction[0] == 'maxline':
            if (obj.__class__ == direction[1].__class__ 
            and obj.sort == direction[1].sort 
            and obj != direction[1] 
            and obj.contains(direction[1])):
                self.postCheck = False
                return False
        elif direction[0] == 'bound':
            if (obj.__class__ == direction[1].__class__ 
            and obj.sort == direction[1].sort 
            and obj.contains(direction[1])):
                if 0 in direction[2] and obj.start != direction[3] and obj.end != direction[3]:
                    self.postCheck = False
                    return False
                if 1 in direction[2] and obj.start != direction[4] and obj.end != direction[4]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'voidof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'shortest_line':
            if len(direction) == 3:
                if 'ind-lineSegment' not in obj.type:
                    return True
                if (obj.sort.name in direction[-1] or
                obj.sort.baseSort().name in direction[-1]) and obj.squareLength() < direction[1]:
                    self.postCheck = False
                    return False
            else:
                if 'ind-lineSegment' in obj.type and obj.squareLength() < direction[1]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'longest_line':
            if len(direction) == 3:
                if 'ind-lineSegment' not in obj.type:
                    return True
                if (obj.sort.name in direction[-1] or
                obj.sort.baseSort().name in direction[-1]) and obj.squareLength() > direction[1]:
                    self.postCheck = False
                    return False
            else:
                if 'ind-lineSegment' in obj.type and obj.squareLength() > direction[1]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'nolabel':
            if 'ind-lineSegment' in obj.type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
        elif direction[0] == 'nolabel*':
            if obj.type == 'ind-label':
                self.postCheck = False
                return False
        elif direction[0] == 'point_on_line':
            if 'ind-lineSegment' in obj.type:
                bounds = [c0, c1]
                self.solvePointOnLine(direction[1], obj, bounds, direction[2])
        elif direction[0] == 'containsCheck':
            if 'ind-lineSegment' in obj.type:
                if len(direction) > 3 and (direction[3] is not None) and (obj == direction[3]):
                    return False
                elif obj.containsPoint(direction[1]):
                    direction[2][0] = True
            if 'ind-point' in obj.type:
                if obj.value == direction[1]:
                    direction[2][0] = True
            """
        elif direction[0] == 'containsCheckLine':
            if 'ind-lineSegment' in obj.type:
                if len(direction) > 3 and (direction[3] is not None) and (obj == direction[3]):
                    return False
                elif obj.start == direction[1] or obj.end == direction[1]:
                    direction[2][0] = True
            """
        return True
        
    def visitLeave(self, obj, direction = 0):
        return
        
        
    @staticmethod
    def ccw(A,B,C):
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
        
    @staticmethod
    def evaluateDesc(boundList, param):
        newList = []
        check = True
        try:
            for i in boundList:
                if isinstance(i, description):
                    val = i.evaluateInd(param).getValue()[1:-1]
                    try:
                        val = i.evaluateInd(param).getValue()[1:-1]
                    except ValueError as E:
                        check = False
                        print('\n===============================\n'+str(E).upper())
                        print('Error Description:\n', i)
                        break
                    newList.append(coordinate(val))                    
                else:
                    newList.append(i)
        except ValueError:
            return None, False
        return newList, check
        
    @staticmethod
    def evaluateVariable(varList, tfs):
        if varList[0] == 'distance':
            key1 = key.retrieveKey(varList[1]).ind.edit(tfs)
            key2 = key.retrieveKey(varList[2]).ind.edit(tfs)
            if 'point' in key2.type and 'lineSegment' in key1.type:
                key1, key2 = key2, key1
            if key1.type == key2.type:
                if 'point' in key1.type:
                    return key1.value.subtract(key2.value).length()
                elif 'lineSegment' in key1.type:
                    if key1.parallel(key2):
                        return key1.root.subtract(key2.projection(key1.root)).length()
            else:
                return key1.value.subtract(key2.projection(key1.value)).length()
        if varList[0] == 'angle':
            key1 = key.retrieveKey(varList[1]).ind.edit(tfs)
            key2 = key.retrieveKey(varList[2]).ind.edit(tfs)
            if 'lineSegment' not in key1.type or 'lineSegment' not in key2.type:
                return 0.01
            v0 = key1.direction
            v1 = key2.direction
            nom = v0.x*v1.x + v0.y*v1.y
            dem = ((v0.x*v0.x + v0.y*v0.y)*(v1.x*v1.x + v1.y*v1.y)).sqrt()
            if dem == c0:
                raise ZeroDivisionError('One of the vectors are (0,0) --> division by zero')
            cosTheta = nom / dem
            return coordinate(acos(cosTheta.value))
        return 0
    
    @staticmethod
    def evaluateBounds(bounds, tfSet):
        if len(bounds) == 4:
            if isinstance(bounds[-1], list):
                b1 = graphMap.evaluateVariable(bounds[-1], tfSet)
            else:
                b1 = bounds[-1]
            if isinstance(bounds[-2], list):
                b0 = graphMap.evaluateVariable(bounds[-2], tfSet)
            else:
                b0 = bounds[-2]
            rand = randint(0, 100000)
            boundVal = (b1-b0)*rand/100000 + b0
        elif len(bounds) == 3:
            if isinstance(bounds[-1], list):
                boundVal = graphMap.evaluateVariable(bounds[-1], tfSet)
            else:
                boundVal = bounds[-1]
        return boundVal
        
    @staticmethod
    def solvePointOnLine(graphTransf, obj, bounds, adaptive = False):        
        i = 0
        while i < len(graphTransf.unresolved):
            u = graphTransf.unresolved[i]
            startP = obj.start
            endP = obj.end
            if obj.containsPoint(u) and (not startP == u) and (not endP == u):
                if adaptive:
                    for j in graphTransf.vects:
                        if obj.containsPoint(j[0]):
                            if j[0] < u:
                                if j[0] > startP:
                                    startP = j[0]
                            else:
                                if j[0] < endP:
                                    endP = j[0]
                
                graphTransf.unresolved.pop(i)
                nStart = graphTransf.apply(startP)
                nEnd = graphTransf.apply(endP)
                if obj.reverse:
                    nStart, nEnd = nEnd, nStart
                rand = randint(0, 100000)
                ratio = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
                p = nStart.add(nEnd.subtract(nStart).scale(ratio))
                graphTransf.add(u, p)
            else:
                i += 1


    def solveAngleEnd(self, comparand, intList, pos, reverse):
        if reverse:
            pos = -1 if pos == 0 else 0
        intersect = intList.pop(pos)
        while comparand == intersect:
            intersect = intList.pop(pos)
        return intersect

    
    def solveAngle(self, tfSet, bounds, lhs):
        # getting angle from the directive range
        angle = self.evaluateBounds(bounds, tfSet)
        while angle > pi:
            angle -= 2*pi
        """
        while angle > 180:
            angle -= 360
        angle = angle * pi / 180 
        """

        base = bounds[0].ind
        new = bounds[1].ind
        
        graphTransf = self.getTransform(tfSet)
        
        intsn = base.intersection(new)
        if intsn == base.end:
            angle = angle - pi
        
        anchorVect = None
        containedPoints = []
        for i in graphTransf.vects:
            if i[0].subtract(new.root).parallel(new.direction):
                containedPoints.append(i[0])
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
            if anchorVect is not None and len(containedPoints) > 0:
                break
        if len(containedPoints) == 0:
            raise ValueError('line with tag '+bounds[1].name+' does not pass through any point/line intersection point/line end point on the lhs')
        
        intStart = None
        intEnd = None
        intersections = {}
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if ints == new.start:
                    intStart = (ints, i.forms[0])
                if ints == new.end:
                    intEnd = (ints, i.forms[0])
                if new.containsPoint(ints):
                    intersections[ints] = i.forms[0]
                                    
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
                
        if nBaseS == nBaseE:
            return []
            
        reverse = (nBaseS > nBaseE) 
        d1 = nBaseE.subtract(nBaseS)
            
        cp = containedPoints[0]
        nCP = graphTransf.apply(cp)
                
        
        if anchorVect is not None and self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            angle = -angle
            
        ca = cos(angle.value)
        sa = sin(angle.value)
        d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
        
        pM = []
        
        
        if graphTransf.contains(new.start) and graphTransf.contains(new.end):
            nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
            cStart = False
            cEnd = False
            for i in graphTransf.directiveResults:
                if i[0][0] == new.start or i[1][0] == new.start:
                    cStart = True
                elif i[0][0] == new.end or i[1][0] == new.end:
                    cEnd = True
                if cStart and cEnd:
                    if new.start == intsn:
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        nStart = nBase.intersection(pLine)
                        nEnd = nStart.add(d2.scale(nLen))
                    elif new.end == intsn:
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        nEnd = nBase.intersection(pLine)
                        nStart = nEnd.add(d2.scale(nLen))
                    else:
                        nLen = nLen / 2
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        intersection = pLine.intersection(nBase)
                        nStart = intersection.subtract(d2.scale(nLen))
                        nEnd = intersection.add(d2.scale(nLen))
                elif cStart:
                    nEnd = graphTransf.apply(new.end)
                    nStart = nEnd.add(d2.scale(nLen))
                elif cEnd:
                    nStart = graphTransf.apply(new.start)
                    nEnd = nStart.add(d2.scale(nLen))
                break
            pM.insert(0,(new.end, nEnd))
            pM.insert(0,(new.start, nStart))
            newLine = new.__class__(new.sort, nStart, nEnd)
            
            for rhsl in self.rhsLines:
                if graphTransf.contains(rhsl.start) and graphTransf.contains(rhsl.end) and rhsl.intersect(new):
                    rhsl1 = rhsl.edit(tfSet)
                    if not newLine.intersect(rhsl1):
                        continue
                    pM.append((new.intersection(rhsl), rhsl1.intersection(newLine)))
                    
            posMatch = []
            posMatch.append(pM)
            return posMatch
            
        newLine = line2D(new.sort, nCP, nCP.add(d2))
        nBase = base.edit(tfSet)
        nIntsn = newLine.intersection(nBase)
        
        if intStart is not None and intEnd is not None:
            nSLine = intStart[1].edit(tfSet)
            if nSLine.intersect(newLine):
                nStart = nSLine.intersection(newLine)
            else:
                return []
            nELine = intEnd[1].edit(tfSet)
            if nELine.intersect(newLine):
                nEnd = nELine.intersection(newLine)
            else:
                return []
        else:
            newints = []
            for intersection0 in intersections:
                if graphTransf.contains(intersection0):
                    intersection = graphTransf.apply(intersection0)
                    newints.append(intersection)
                else:
                    nIntLine = intersections[intersection0].edit(tfSet) 
                    if not nIntLine.intersect(newLine):
                        continue
                    intersection = nIntLine.intersection(newLine)
                    newints.append(intersection)
                    if intersection0 == new.start or intersection0 == new.end:
                        continue
                    pM.append((intersection0,intersection))
            newints = sorted(newints)
            dirCheck = d2 > vz
            if intsn <= new.start or intsn >= new.end:
                index = 0
                while index < len(newints):
                    if newints[index] == nIntsn or (newints[index] > nIntsn) == dirCheck:
                        index += 1
                    else:
                        newints.pop(index)
            if len(newints) >= 2:
                if intStart is not None:
                    nSLine = intStart[1].edit(tfSet)
                    if nSLine.intersect(newLine):
                        nStart = nSLine.intersection(newLine)
                    else:
                        return []
                    if not graphTransf.contains(new.end):
                        nEnd = self.solveAngleEnd(nStart, newints, -1, reverse)
                elif intEnd is not None:
                    nELine = intEnd[1].edit(tfSet)
                    if nELine.intersect(newLine):
                        nEnd = nELine.intersection(newLine)
                    else:
                        return []
                    if not graphTransf.contains(new.start):
                        nStart = self.solveAngleEnd(nEnd, newints, 0, reverse)
                elif graphTransf.contains(new.start):
                    nStart = graphTransf.apply(new.start)
                    nEnd = self.solveAngleEnd(nStart, newints, -1, reverse)
                elif graphTransf.contains(new.end):
                    nEnd = graphTransf.apply(new.end)
                    nStart = self.solveAngleEnd(nEnd, newints, 0, reverse)
                else:
                    nStart, nEnd = newints[0], newints.pop(-1)
                    while nEnd == nStart:
                        nEnd = newints.pop(-1)
                    if reverse:
                        nStart, nEnd = nEnd , nStart
            else:
                nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
                if graphTransf.contains(new.start):
                    nStart = graphTransf.apply(new.start)
                    nEnd = nStart.add(d2.scale(nLen))
                elif graphTransf.contains(new.end):
                    nEnd = graphTransf.apply(new.end)
                    nStart = nEnd.add(d2.scale(nLen))
                elif new.start == intsn:
                    pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                    nBase = base.edit(tfSet)
                    nStart = nBase.intersection(pLine)
                    nEnd = nStart.add(d2.scale(nLen))
                elif new.end == intsn:
                    pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                    nBase = base.edit(tfSet)
                    nEnd = nBase.intersection(pLine)
                    nStart = nEnd.add(d2.scale(nLen))
                else:
                    nLen = nLen / 2
                    pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                    nBase = base.edit(tfSet)
                    intersection = pLine.intersection(nBase)
                    nStart = intersection.subtract(d2.scale(nLen))
                    nEnd = intersection.add(d2.scale(nLen))
        
        pM.insert(0,(new.end, nEnd))
        pM.insert(0,(new.start, nStart))
        
        for rhsl in self.rhsLines:
            if graphTransf.contains(rhsl.start) and graphTransf.contains(rhsl.end) and rhsl.intersect(new):
                rhsl1 = rhsl.edit(tfSet)
                if not newLine.intersect(rhsl1):
                    continue
                pM.append((new.intersection(rhsl), rhsl1.intersection(newLine)))
                
        posMatch = []
        posMatch.append(pM)
        
        
        return posMatch
        
        
        
        
    """
    def solveAngle(self, tfSet, bounds, lhs):
        if len(bounds) == 4:
            rand = randint(0, 100000)
            angle = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
        elif len(bounds) == 3:        
            angle = bounds[-1]
        while angle > 180:
            angle -= 360
        angle = angle * pi / 180 
        base = bounds[0].ind
        new = bounds[1].ind

        graphTransf = self.getTransform(tfSet)
    
        anchorVect = None
        containedPoints = []
        for i in graphTransf.vects:
            if i[0].subtract(new.root).parallel(new.direction):
                containedPoints.append(i[0])
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
        if len(containedPoints) == 0:
            raise ValueError('line with tag '+bounds[1].name+' does not pass through any point/line intersection point/line end point on the lhs')

        intStart = None
        intEnd = None
        intersections = []
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if ints == new.start:
                    intStart = (ints, i.forms[0])
                if ints == new.end:
                    intEnd = (ints, i.forms[0])
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
                    
        intsn = base.intersection(new)
        delayBase = False
        if base.start >= intsn:
            baseDir = True
            p0 = base.end
        elif base.end <= intsn:
            baseDir = False
            p0 = base.start
        else:
            delayBase = True
        
        if new.start >= intsn:
            p1 = new.end
        elif new.end <= intsn:
            p1 = new.start
        else:
            if delayBase:
                p0 = base.start
                delayBase = False
            endProj = base.projection(new.end)
            if (abs(angle)*2 < pi) == ((endProj > intsn) == (p0 > intsn)):
                p1 = new.end
                baseDir = True
            else:
                p1 = new.start
                baseDir = False
                
        if delayBase:
            p1Proj = base.projection(p1)
            if (abs(angle)*2 < pi) == ((p1Proj > intsn) == (base.start > intsn)):
                p0 = base.start
                baseDir = False
            else:
                p0 = base.end
                baseDir = True
        
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
        
        '''
        if angle > 0 and self.ccw(p0, intsn, p1) :
            angle = -angle
        elif angle < 0 and not self.ccw(p0, intsn, p1) :
            angle = -angle

        if anchorVect is not None and self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            angle = -angle
        
            
        if nBaseS > nBaseE:
            baseDir = not baseDir
            angle = pi-angle
            angle = -angle
        '''

        if baseDir:
            d1 = nBaseE.subtract(nBaseS)
        else:
            d1 = nBaseS.subtract(nBaseE)
            
        cp = containedPoints[0]
        nCP = graphTransf.apply(cp)
        
                
        ca = cos(angle)
        sa = sin(angle)
        d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
        
        
        nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
        if graphTransf.contains(new.start) and graphTransf.contains(new.end):
            cStart = False
            cEnd = False
            for i in graphTransf.directiveResults:
                if i[0][0] == new.start or i[1][0] == new.start:
                    cStart = True
                elif i[0][0] == new.end or i[1][0] == new.end:
                    cEnd = True
                if cStart and cEnd:
                    if new.start == intsn:
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        nStart = nBase.intersection(pLine)
                        nEnd = nStart.add(d2.scale(nLen))
                    elif new.end == intsn:
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        nEnd = nBase.intersection(pLine)
                        nStart = nEnd.add(d2.scale(nLen))
                    else:
                        nLen = nLen / 2
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        intersection = pLine.intersection(nBase)
                        nStart = intersection.subtract(d2.scale(nLen))
                        nEnd = intersection.add(d2.scale(nLen))
                elif cStart:
                    nEnd = graphTransf.apply(new.end)
                    nStart = nEnd.add(d2.scale(nLen))
                elif cEnd:
                    nStart = graphTransf.apply(new.start)
                    nEnd = nStart.add(d2.scale(nLen))
        elif graphTransf.contains(new.start):
            nStart = graphTransf.apply(new.start)
            nEnd = nStart.add(d2.scale(nLen))
        elif graphTransf.contains(new.end):
            nEnd = graphTransf.apply(new.end)
            nStart = nEnd.add(d2.scale(nLen))
        elif new.start == intsn:
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            nStart = nBase.intersection(pLine)
            nEnd = nStart.add(d2.scale(nLen))
        elif new.end == intsn:
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            nEnd = nBase.intersection(pLine)
            nStart = nEnd.add(d2.scale(nLen))
        else:
            nLen = nLen / 2
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            intersection = pLine.intersection(nBase)
            nStart = intersection.subtract(d2.scale(nLen))
            nEnd = intersection.add(d2.scale(nLen))
            
            
        nLine = new.__class__(new.sort, nStart, nEnd)

        check = True
        pM = []
        for intersectionSet in intersections:
            intersection0 = intersectionSet[0]
            if graphTransf.contains(intersection0):
                intersection = graphTransf.apply(intersection0)
            else:
                nIntLine = intersectionSet[1].edit(tfSet) 
                if not nIntLine.intersect(nLine):
                    continue
                intersection = nIntLine.intersection(nLine)
                
            if not intersection.subtract(nLine.root).parallel(nLine.direction):
                check = False
                break
            if intersection < nLine.start:
                nLine.start = intersection
            elif intersection > nLine.end:
                nLine.end = intersection
        if not check:
            return []
        
        
        if not nLine.reverse:
            nStart = nLine.start
            nEnd = nLine.end
        else:
            nStart = nLine.end
            nEnd = nLine.start
            
        if intStart is not None and intEnd is not None:
            nSLine = intStart[1].edit(tfSet)
            nELine = intEnd[1].edit(tfSet)
            if nSLine.intersect(nLine) and nELine.intersect(nLine):
                nStart = nSLine.intersection(nLine)
                nEnd = nELine.intersection(nLine)

        pM.append((new.start, nStart))
        pM.append((new.end, nEnd))
        for intersectionSet in intersections:
            intersection0 = intersectionSet[0]
            if graphTransf.contains(intersection0) or intersection0 == new.start or intersection0 == new.end:
                continue
            else:
                nIntLine = intersectionSet[1].edit(tfSet) 
                if not nIntLine.intersect(nLine):
                    continue
                intersection = nIntLine.intersection(nLine)
                pM.append((intersection0,intersection))
        posMatch = []
        posMatch.append(pM)
        if anchorVect is None:
            sa = -sa
            d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
            
            nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
            if graphTransf.contains(new.start):
                nStart = graphTransf.apply(new.start)
                nEnd = nStart.add(d2.scale(nLen))
            elif graphTransf.contains(new.end):
                nEnd = graphTransf.apply(new.end)
                nStart = nEnd.subtract(d2.scale(nLen))
            elif new.start == intsn:
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                nStart = nBase.intersection(pLine)
                nEnd = nStart.add(d2.scale(nLen))
            elif new.end == intsn:
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                nEnd = nBase.intersection(pLine)
                nStart = nEnd.subtract(d2.scale(nLen))
            else:
                nLen = nLen / 2
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                intersection = pLine.intersection(nBase)
                nStart = intersection.subtract(d2.scale(nLen))
                nEnd = intersection.add(d2.scale(nLen))
            
            nLine = new.__class__(new.sort, nStart, nEnd)
            if intStart is not None and intEnd is not None:
                nSLine = intStart[1].edit(tfSet)
                nELine = intEnd[1].edit(tfSet)
                if nSLine.intersect(nLine) and nELine.intersect(nLine):
                    nStart = nSLine.intersection(nLine)
                    nEnd = nELine.intersection(nLine)
                    
            pM = []
            
            for intersectionSet in intersections:
                intersection0 = intersectionSet[0]
                if graphTransf.contains(intersection0):
                    continue
                else:
                    nIntLine = intersectionSet[1].edit(tfSet) 
                    if not nIntLine.intersect(nLine):
                        continue
                    
            if not nLine.reverse:
                nStart = nLine.start
                nEnd = nLine.end
            else:
                nStart = nLine.end
                nEnd = nLine.start
            pM.append((new.start, nStart))
            pM.append((new.end, nEnd))
            posMatch.append(pM)
        return posMatch
    """


    def solveLineLineDistance(self, tfSet, bounds):
        distance = self.evaluateBounds(bounds, tfSet)
        base = bounds[0].ind
        new = bounds[1].ind

        if (new.start.y > base.projection(new.start).y) != (distance > 0):
            distance = -distance

        graphTransf = self.getTransform(tfSet)
        
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
                break

        intersections = []
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
                    
        for i in self.rhsLines:
            if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.intersect(new):
                ints = i.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i])
                
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
            
        if anchorVect is not None and self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            distance = -distance
            
        if nBaseS > nBaseE:
            d1 = nBaseS.subtract(nBaseE)
            distance = -distance
        else:
            d1 = nBaseE.subtract(nBaseS)
        
        pDir = d1.__class__(-d1.y, d1.x)
        
        sc = ((distance*distance)/(pDir.dotproduct(pDir))).sqrt()
            
        if distance < 0:
            sc = -sc
        
        nStart = nBaseS.add(pDir.scale(sc))
        nEnd = nBaseE.add(pDir.scale(sc))
        
        nLine = new.__class__(new.sort, nStart, nEnd)
        check = True
        
        intSet = []
        for intersectionSet in intersections:
            intersection = intersectionSet[0]
            if graphTransf.contains(intersection):
                intersection = graphTransf.apply(intersection)
            else:
                intersection = intersectionSet[1].edit(tfSet).intersection(nLine)
            if not intersection.subtract(nLine.root).parallel(nLine.direction):
                check = False
                break
            intSet.append(intersection)
        if not check:
            return []
        
        intSet = sorted(intSet)
        if len(intSet) == 1:
            if intSet[0] < nLine.start:
                nLine.start = intSet[0]
            else:
                nLine.end = intSet[0]
        elif len(intSet) >= 2:
            nLine.start, nLine.end = intSet[0], intSet[-1]
        if not nLine.reverse:
            nStart, nEnd = nLine.start, nLine.end
        else:
            nStart, nEnd = nLine.end, nLine.start
            
        posMatch = []
        posMatch.append([[new.start, nStart],[new.end, nEnd]])
        
        if anchorVect is None:
            sc = -sc
            
            nStart = nBaseS.add(pDir.scale(sc))
            nEnd = nBaseE.add(pDir.scale(sc))
            
            nLine = new.__class__(new.sort, nStart, nEnd)
            
            if not nLine.reverse:
                nStart, nEnd = nLine.start, nLine.end
            else:
                nStart, nEnd = nLine.end, nLine.start
            posMatch.append([[new.start, nStart],[new.end, nEnd]])
        return posMatch
    
    

    def solveLinePointDistance(self, tfSet, bounds):
        distance = self.evaluateBounds(bounds, tfSet)

        base = bounds[0].ind
        new = bounds[1].ind
        graphTransf = self.getTransform(tfSet)
                    
        if (new.value.y > base.projection(new.value).y) != (distance > 0):
            distance = -distance

        nCline = None
        for i in self.nodeList:
            if i.nodeElement.containsPoint(new.value):
                nCline = i.forms[0].edit(tfSet)
                break
        if nCline is None:
            for i in self.rhsLines:
                if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.containsPoint(new.value):
                    cline = i
                    nCline = cline.edit(tfSet)
                    break            
            if nCline is None:
                return []
        
            
        
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
                break
            
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
            
        if self.ccw(base.end, base.start, anchorVect[0]) != self.ccw(nBaseE, nBaseS, anchorVect[1]):
            distance = -distance
            
        if nBaseS > nBaseE:
            d1 = nBaseS.subtract(nBaseE)
            distance = -distance
        else:
            d1 = nBaseE.subtract(nBaseS)
            
        pDir = d1.__class__(-d1.y, d1.x)
        
        sc = ((distance*distance)/(pDir.dotproduct(pDir))).sqrt()
            
        if distance < 0:
            sc = -sc
        
        nStart = nBaseS.add(pDir.scale(sc))
        nLine = base.__class__(base.sort, nStart, nStart.add(d1))
        try:            
            nPoint = nLine.intersection(nCline)
        except ValueError:
            return []
        return [[[new.value, nPoint]]]
        
    def solvePointLineDistance(self, tfSet, bounds):
        distance = self.evaluateBounds(bounds, tfSet)

        base = bounds[0].ind
        new = bounds[1].ind

        graphTransf = self.getTransform(tfSet)
        
        cPoint = []
        lineN = new.parametricNode()
        
        for i in graphTransf.vects:
            if lineN.containsPoint(i[0]):
                cPoint.append(i[0])
                cPoint.append(i[1])
                if new.containsPoint(i[0]): cPoint.append(True)
                else: cPoint.append(False)
                break
        if len(cPoint) == 0:
            raise ValueError
            
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.value, cPoint[0]):
                anchorVect = i
                break

        intersections = []
        for i in self.nodeList:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])

        for i in self.rhsLines:
            if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.intersect(new):
                ints = i.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i])

        nBase = graphTransf.apply(base.value)
        
        d1 = nBase.subtract(cPoint[1])
        d1len = d1.length()
        sa = distance / d1len
        
        try:
            ca = (d1len*d1len - distance*distance).sqrt() / d1len
        except ValueError:
            return []
        
        sPoint = new.start
        if sPoint == cPoint[0]:
            sPoint = new.end
        if self.ccw(base.value, cPoint[0], sPoint):
            sa = -sa
        
        if anchorVect is not None:
            ccwC = self.ccw(base.value, cPoint[0], anchorVect[0])
            if ccwC != self.ccw(nBase, cPoint[1], anchorVect[1]):
                sa = -sa
        
        d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
        nLine = line2D(new.sort, cPoint[1], cPoint[1].add(d2))
        
        newInts = []
        for i in intersections:
            l = i[1].edit(tfSet)
            newInts.append(l.intersection(nLine))
        posMatch = []
        if cPoint[-1] is True:
            if cPoint[1] not in newInts:
                newInts.append(cPoint[1])
            if len(newInts) < 2:
                newInts.append(nLine.projection(nBase))
            newInts = sorted(newInts)
            if new.start == cPoint[0]:
                if cPoint[1] == newInts[-1]:
                    posMatch.append([[new.end, newInts[0]]])
                else:
                    posMatch.append([[new.end, newInts[-1]]])
            elif new.end == cPoint[0]:
                if cPoint[1] == newInts[0]:
                    posMatch.append([[new.start, newInts[-1]]])
                else:
                    posMatch.append([[new.start, newInts[0]]])
            else:
                posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
        else:
            newInts = sorted(newInts)
            if len(newInts) < 2:
                newInts.append(nLine.projection(nBase))
                if len(newInts) < 2:
                    newInts.append(cPoint[1])
            posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
            
        if anchorVect is None:
            sa = -sa
            d2 = d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)
            nLine = line2D(new.sort, cPoint[1], cPoint[1].add(d2))
            if new.start == cPoint[0]:
                posMatch.append([[new.end, nLine.projection(nBase)]])
            elif new.end == cPoint[0]:
                posMatch.append([[new.start, nLine.projection(nBase)]])
            else:
                posMatch.append([[new.start, cPoint[1]],[new.end, nLine.projection(nBase)]])
        return posMatch
        
    def solvePointPointDistance(self, tfSet, bounds):

        distance = self.evaluateBounds(bounds, tfSet)
        base = bounds[0].ind
        new = bounds[1].ind
        graphTransf = self.getTransform(tfSet)
        
        nCline = None
        for i in self.nodeList:
            if i.nodeElement.containsPoint(new.value):
                cline = i.forms[0]
                nCline = cline.edit(tfSet)
                break
        if nCline is None:
            for i in self.rhsLines:
                if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.containsPoint(new.value):
                    cline = i
                    nCline = cline.edit(tfSet)
                    break            
            if nCline is None:
                return []
        
        proj = cline.projection(base.value)
        if proj > new.value:
            distance = -distance
        if nCline.reverse:
            distance = -distance
        
        nBase = graphTransf.apply(base.value)
        nProj = nCline.projection(nBase)
        nProsBaseDist = nProj.subtract(nBase)
        
        dProjP = distance**2-(nProsBaseDist.dotproduct(nProsBaseDist))
        if dProjP < 0:
            return []
        elif dProjP == 0:
            dProjP = c0
        dProjP = dProjP.sqrt()
        
        if distance < 0:
            dProjP = -dProjP
            
        nPoint = nProj.add(nCline.unitDir.scale(dProjP))
        return [[[new.value, nPoint]]]
        
    def solveDoubleLinePointDistance(self, tfSet, d1, d2, line1, line2, pointNew):
        nl1 = line1.edit(tfSet)
        nl2 = line2.edit(tfSet)
        if nl1.parallel(nl2):
            return []
        n1 = vector2D(nl1.unitDir.y, -nl1.unitDir.x)
        r1_1 = nl1.root.add(n1.scale(d1))
        r1_2 = nl1.root.subtract(n1.scale(d1))
        nl1_1 = nl1.__class__(nl1.sort, r1_1, r1_1.add(nl1.unitDir))
        nl1_2 = nl1.__class__(nl1.sort, r1_2, r1_2.add(nl1.unitDir))
        n2 = vector2D(nl2.unitDir.y, -nl2.unitDir.x)
        r2_1 = nl2.root.add(n2.scale(d2))
        r2_2 = nl2.root.subtract(n2.scale(d2))
        nl2_1 = nl2.__class__(nl2.sort, r2_1, r2_1.add(nl2.unitDir))
        nl2_2 = nl2.__class__(nl2.sort, r2_2, r2_2.add(nl2.unitDir))
        i1 = nl1_1.intersection(nl2_1)
        i2 = nl1_2.intersection(nl2_1)
        i3 = nl1_1.intersection(nl2_2)
        i4 = nl1_2.intersection(nl2_2)
        return [[[pointNew.value, i1]],[[pointNew.value, i2]],[[pointNew.value, i3]],[[pointNew.value, i4]]]        
        
    def solvePointPointLinePointDistance(self, tfSet, d1, d2, point1, line1, pointNew):
        np1 = point1.edit(tfSet)
        nl1 = line1.edit(tfSet)
        nProj = nl1.projection(np1.value)
        d3 = nProj.subtract(np1.value).length()
        if (d1 + d2) < d3:
            return []
        elif (d1 + d2) == d3:
            nPoint = nProj.subtract(np1.value).scale(d1/d3).add(np1.value)
            return [[[pointNew.value, nPoint]]]
        elif abs(d1-d2) == d3:
            if d1 > d2:
                nPoint = nProj.subtract(np1.value).scale(d1/d3).add(np1.value)
            else:
                nPoint = np1.value.subtract(nProj).scale(d2/d3).add(nProj)
            return [[[pointNew.value, nPoint]]]
        posMatch = []
        
        if d1 > (d2 + d3):
            a = (d1*d1 - (d3+d2)**2).sqrt()
            v = nProj.subtract(np1.value)
            nPt1 = nProj.add(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
            nPt2 = nProj.subtract(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
            posMatch.append([[pointNew.value, nPt1]])
            posMatch.append([[pointNew.value, nPt2]])
        a = (d1*d1 - (d3-d2)**2).sqrt()
        v = np1.value.subtract(nProj)
        nPt1 = nProj.add(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
        nPt2 = nProj.subtract(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
        posMatch.append([[pointNew.value, nPt1]])
        posMatch.append([[pointNew.value, nPt2]])
        return posMatch

    def solveDoublePointLineDistance(self, tfSet, d1, d2, point1, point2, lineNew):
        np1 = point1.edit(tfSet)
        np2 = point2.edit(tfSet)
        u = np2.value.subtract(np1.value)
        d3 = u.length()
        posMatch = []
        if d1+d2 < d3:
            h1 = d3*d1/(d1+d2)
            intPt = u.scale(h1/d3).add(np1.value)
            alpha = acos(d1/h1)
            startPoint1 = vector2D(u.x*cos(alpha)-u.y*sin(alpha), u.x*sin(alpha)+u.y*cos(alpha)).scale(d1/d3).add(np1.value)
            startPoint2 = vector2D(u.x*cos(alpha)+u.y*sin(alpha), u.y*cos(alpha)-u.x*sin(alpha)).scale(d1/d3).add(np1.value)
            endPoint1 = intPt.subtract(startPoint1).scale(d3/d1).add(startPoint1)
            endPoint2 = intPt.subtract(startPoint2).scale(d3/d1).add(startPoint2)
            posMatch.append([[lineNew.start, startPoint1], [lineNew.end, endPoint1]])
            posMatch.append([[lineNew.start, startPoint2], [lineNew.end, endPoint2]])
        elif d1+d2 == d3:
            h1 = d3*d1/(d1+d2)
            intPt = u.scale(h1/d3).add(np1.value)
            v = vector2D(u.y, -u.x)
            pt1 = intPt.add(v)
            pt2 = intPt.subtract(v)
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
        if d1 == d2:
            v = vector2D(u.y, -u.x).scale(d1/d3)
            spt = np1.value.add(v)
            ept = np2.value.add(v)
            posMatch.append([[lineNew.start, spt], [lineNew.end, ept]])
        elif abs(d1-d2) == d3:
            v = vector2D(u.y, -u.x)
            if d1>d2:
                intPt = u.scale(d1/d3).add(np1.value)
            else:
                intPt = np1.value.subtract(u.scale(d1/d3))
            pt1 = intPt.add(v)
            pt2 = intPt.subtract(v)
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
        elif abs(d1-d2) < d3:
            alpha = acos((d1-d2)/d3)
            v1 = vector2D(u.x*cos(alpha)-u.y*sin(alpha), u.x*sin(alpha)+u.y*cos(alpha)).scale(1/d3)
            pt1 = np1.value.add(v1.scale(d1))
            pt2 = np2.value.add(v1.scale(d2))
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
            v2 = vector2D(u.x*cos(alpha)+u.y*sin(alpha), u.y*cos(alpha)-u.x*sin(alpha)).scale(1/d3)
            pt1 = np1.value.add(v2.scale(d1))
            pt2 = np2.value.add(v2.scale(d2))
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
        return posMatch

    def solveDoublePointPointDistance(self, tfSet, d1, d2, point1, point2, pointNew):
        np1 = point1.edit(tfSet)
        np2 = point2.edit(tfSet)
        d3 = np2.value.subtract(np1.value).length()
        if (d1 + d2) < d3 or abs(d1 - d2) > d3:
            return []
        elif (d1 + d2) == d3:
            nPoint = np2.value.subtract(np1.value).scale(d1/d3).add(np1.value)
            return [[[pointNew.value, nPoint]]]
        elif abs(d1-d2) == d3:
            if d1 > d2:
                nPoint = np2.value.subtract(np1.value).scale(d1/d3).add(np1.value)
            else:
                nPoint = np1.value.subtract(np2.value).scale(d2/d3).add(np2.value)
            return [[[pointNew.value, nPoint]]]
        u = np2.value.subtract(np1.value)
        alpha = acos((d1*d1 + d3*d3 - d2*d2) / (2*d1*d3))
        nPoint1 = vector2D(u.x*cos(alpha)-u.y*sin(alpha), u.x*sin(alpha)+u.y*cos(alpha)).scale(d1/d3).add(np1.value)
        nPoint2 = vector2D(u.x*cos(alpha)+u.y*sin(alpha), u.y*cos(alpha)-u.x*sin(alpha)).scale(d1/d3).add(np1.value)
        return [[[pointNew.value, nPoint1]],[[pointNew.value, nPoint2]]]
        
    def readjustLength(self, tfSet, bounds, lhs, rhs):
        if len(bounds) == 3:
            rand = randint(0, 100000)
            length = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
        elif len(bounds) == 2:
            length = bounds[-1]
        else:
            raise ValueError()
        
        # line = line to be adjusted
        line = bounds[0].ind
        graphTransf = self.getTransform(tfSet)
        
        cS = graphTransf.contains(line.start)
        cE = graphTransf.contains(line.end)
        if cS and cE:
            rhsCheckE = [False]
            rhsCheckS = [False]
            
            # check if there are any point/line containing line.end or line.start
            rhs.accept(self, direction = ['containsCheck', line.start, rhsCheckS, line])
            rhs.accept(self, direction = ['containsCheck', line.end, rhsCheckE, line])

            if rhsCheckE[0] == rhsCheckS[0] == False: # if no other line/point containing both line.start and line.end:
                adjust = []
                for i in graphTransf.vects:
                    if i[0] == line.start:
                        adjust.insert(0,line.start)
                    elif i[0] == line.end:
                        adjust.insert(0,line.end)
            elif rhsCheckE[0] is False: # if there are no lines/points containing line.end then pass --> adjust is None
                adjust = [line.end, line.start]
            elif rhsCheckS[0] is False: # if 
                adjust = [line.start, line.end]
            else:
                lhsCheckE = [False]
                lhs.accept(self, direction = ['containsCheck', line.end, lhsCheckE, line])
                lhsCheckS = [False]
                lhs.accept(self, direction = ['containsCheck', line.start, lhsCheckS, line])
                if lhsCheckE[0] is False and lhsCheckS[0] is False:
                    for i in self.directives:
                        lhsCheckE = False
                        lhsCheckS = False
                        if i == 'length' or i == 'point_on_line':
                            continue
                        for j in i[1]:
                            ind = j[1]
                            if 'lineSegment' in ind.type:
                                if line.end == ind.start or line.end == ind.end:
                                    lhsCheckE = True
                                if line.start == ind.start or line.start == ind.end:
                                    lhsCheckS = True
                            elif 'point' in ind.type:
                                if line.end == ind.value:
                                    lhsCheckE = True
                                if line.start == ind.value:
                                    lhsCheckS = True
                    if lhsCheckE and lhsCheckS:
                        raise ValueError('cannot apply directive "Length" onto line '+str(line))
                    elif lhsCheckE is False:
                        adjust = [line.end, line.start]
                    else:
                        adjust = [line.start, line.end]
                elif lhsCheckE[0] is False:
                    adjust = [line.end, line.start]
                elif lhsCheckS[0] is False:
                    adjust = [line.start, line.end]
                else:
                    raise ValueError('cannot apply directive "Length" onto line '+str(line))
            checkVect = True
            for i in adjust:
                if not graphTransf.contains(i):
                    checkVect = False
                    break
            if checkVect:
                new0 =  graphTransf.apply(adjust[0])
                new1 =  graphTransf.apply(adjust[1])
                d0 = new0.subtract(new1)
                newPoint = new1.add(d0.scale(length/d0.length()))
                #print(adjust[0], new0, new1)
                return [[[adjust[0], newPoint]]]
            else:
                cVects = []
                for i in graphTransf.vects:
                    if line.containsPoint(i[0]):
                        cVects.append(i[1])
                cVects = sorted(cVects)
                if len(cVects) < 2:
                    return [[]]
                elif len(cVects) > 2:
                    for i in range(len(cVects)-2):
                        for j in range(i+1,len(cVects)-1):
                            for k in range(i+2,len(cVects)):
                                if not cVects[i].colinear(cVects[j],cVects[k]):
                                    return [[]]
                start = cVects[0]
                end = cVects[-1]
                direction = end.subtract(start)
                ratio = length / (direction.length() * 2)
                newStart = start.subtract(direction.scale(ratio))
                newEnd = end.add(direction.scale(ratio))
                for i in adjust:
                    if i in graphTransf.unresolved:
                        graphTransf.unresolved.remove(i)
                return [[[adjust[0], newStart],[adjust[1], newEnd]]]
        elif cS:
            contained0 = []
            contained1 = []
            start = graphTransf.apply(line.start)
            for i in graphTransf.vects:
                if i[0] == line.start:
                    continue
                if line2D.containsPoint(line, i[0]):
                    contained0.append(i[0])
                    contained1.append(i[1])
            if len(contained1) == 0:
                return []
            elif len(contained1) == 1:
                end = contained1[0]
            else:
                if (contained0[0] <= contained0[-1]) == (contained1[0] <= contained1[-1]):
                    end = max(contained1[1])
                else:
                    end = min(contained1[1])
            if start == end:
                return []
            nDir = end.subtract(start)
            l = length - nDir.length()
            nDir = nDir.unit()
            nEnd = end.add(nDir.scale(l))
            nLine = line.__class__(line.sort, start, nEnd)
            posMatch = [[line.end, nEnd]]
            for i in self.nodeList.lines:
                if line.intersect(i.nodeElement):
                    intsn = i.nodeElement.intersection(line)
                    if intsn == line.start or intsn == line.end or graphTransf.contains(intsn):
                        continue
                    nIL = i.forms[0].edit(tfSet)
                    if not line2D.intersect(nLine, nIL):
                        continue
                    nIntsn = line2D.intersection(nLine, nIL)
                    posMatch.append([intsn, nIntsn])
            return [posMatch]
        elif cE:
            contained0 = []
            contained1 = []
            end = graphTransf.apply(line.end)
            for i in graphTransf.vects:
                if i[0] == line.end:
                    continue
                if line2D.containsPoint(line, i[0]):
                    contained0.append(i[0])
                    contained1.append(i[1])
            if len(contained1) == 0:
                return []
            elif len(contained1) == 1:
                start = contained1[0]
            else:
                if (contained0[0] <= contained0[-1]) == (contained1[0] <= contained1[-1]):
                    start = max(contained1[1])
                else:
                    start = min(contained1[1])
            if start == end:
                return []
            nDir = end.subtract(start)
            l = length - nDir.length()
            nDir = nDir.unit()
            nStart = start.subtract(nDir.scale(l))
            nLine = line.__class__(line.sort, nStart, end)
            posMatch = [[line.start, nStart]]
            for i in self.nodeList.lines:
                if line.intersect(i.nodeElement):
                    intsn = i.nodeElement.intersection(line)
                    if intsn == line.start or intsn == line.end or graphTransf.contains(intsn):
                        continue
                    nIL = i.forms[0].edit(tfSet)
                    if not line2D.intersect(nLine, nIL):
                        continue
                    nIntsn = line2D.intersection(nLine, nIL)
                    posMatch.append([intsn, nIntsn])
            return [posMatch]
        else:
            contained = []
            for i in graphTransf.vects:
                if line2D.containsPoint(line, i[0]):
                    contained.append(i[1])
            if len(contained) < 2:
                return []
            start = min(contained)
            end = max(contained)
            if start == end:
                return []
            nDir = end.subtract(start)
            l = nDir.length()
            extendedlength = (length - l) / 2
            nDir = nDir.unit()
            nStart = start.subtract(nDir.scale(extendedlength))
            nEnd = end.add(nDir.scale(extendedlength))
            nLine = line.__class__(line.sort, nStart, nEnd)
            posMatch = [[line.start, nStart],[line.end, nEnd]]
            for i in self.nodeList.lines:
                if line.intersect(i.nodeElement):
                    intsn = i.nodeElement.intersection(line)
                    if intsn == line.start or intsn == line.end or graphTransf.contains(intsn):
                        continue
                    nIL = i.forms[0].edit(tfSet)
                    if not line2D.intersect(nLine, nIL):
                        continue
                    nIntsn = line2D.intersection(nLine, nIL)
                    posMatch.append([intsn, nIntsn])
            return [posMatch]

