# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 12:37:59 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import round
from builtins import range
"""
from future import standard_library
standard_library.install_aliases()
"""
from .mapping import mapping
from .distinguishable import distinguishable
from ..struct.coordinate import coordinate
from ..transform.transformSet import transformSet
from ..transform.affineTransform import affineTransform2D
from ..transform.transform import transform

from ..ind.point import point2D
from ..ind.circle import circle2D
from ..ind.ellipse import ellipse2D
from ..ind.line import line2D
from enum import Enum

p = coordinate.p
class similarityMap2D(mapping):
    
    type = 'map-distinguishable-similarityMap2D'
    
        
    class casus(Enum):
        CASE_2PT_NOT_COINCIDENT = 1, """DONE"""
        CASE_1PT_1ARC_NOT_COINCIDENT = 2,
        CASE_1PT_1LN_NOT_COLINEAR = 3, """DONE"""
        
        CASE_2ARC_NOT_COINCIDENT = 4,
        CASE_1ARC_1LN_NOT_COLINEAR = 5,
        
        #CASE_1D_2LN_PARALLEL = 4, """**DONE**"""
        CASE_3LN_NO_TWO_PARALLEL = 6, """DONE"""
        CASE_3LN_FIRST_TWO_PARALLEL = 7, """DONE"""
        CASE_3LN_LAST_TWO_PARALLEL = 8, """DONE"""
        REDUNDANT = 0
        
    class distinguishes(Enum):
        INSUFFICIENT = 1, 
        UNSUITABLE = 2, 
        SUCCESS = 3

    def __init__(self, *args, **kwargs):
        super(similarityMap2D, self).__init__()     
        self.distinguishables = []
        self.transforms = []
        self.fromSet = [None,None,None]
        #self.checkTree = None
        self.distinguishableSuccess = False
        self.reflection = 0
        self.transformset = transformSet()
        self.casus = similarityMap2D.casus.REDUNDANT
        self.subCase = []
        self.cosAngle = None
        self.ratio = None
        return
        
    @staticmethod
    def getInstance():
        return similarityMap2D()
        
    def isValid(self):
        if self.distinguishableSuccess:
            return True
        if len(self.distinguishables) == 1:
            if (self.distinguishables[0].element.individual.type == 'ind-point2D' 
            or self.distinguishables[0].element.individual.type == 'ind-line2D'):
                return True
        return False
    
    def clear(self):
        for i in self.distinguishables:
            i.clear()
        self.transforms = []
        return
    
    def map(self, fromInd, toInd):
        for i in self.distinguishables:
            elmInd = i.element.individual
            if elmInd == fromInd:
                if elmInd.hasAttribute():
                    if fromInd.hasAttribute() and elmInd.attribute.checkPartOf(toInd.attribute):
                        i.addCounterpart(toInd)
                        return True
                else:
                    i.addCounterpart(toInd)
                    return True
        return False
    
    def distinguish(self, ind):
        element = ind.codescriptor()
        if ind.hasAttribute():
            element.addAttribute(ind.attribute)
        dist = None
        for i in self.distinguishables:
            if element == i.element.individual:
                dist = i
                dist.addInstance(ind)
                return dist
        if self.distinguishableSuccess:
            return None
        dist = distinguishable(element)
        
        #cell =  [element, dist]    
        self.distinguishables.append(dist)
        dist.addInstance(ind)
        size = len(self.distinguishables)
        last = element
        first = None
        second = None
        #third = None

        if size > 1:
            first = self.distinguishables[0].element.individual
            if size > 2:
                second = self.distinguishables[1].element.individual
                #if size > 3:
                    #third = next(iterDistinguishables)
                
        success = self.assessDistinguishables(size, last, first, second)#, third)
        if success == similarityMap2D.distinguishes.UNSUITABLE:
            del self.distinguishables[-1]
        if success == similarityMap2D.distinguishes.SUCCESS:
            self.distinguishableSuccess = True
        return dist

    def assessDistinguishables(self, size, last, first, second):#, second, third):
        if size == 1:
            # 
            if last.__class__ in (point2D, circle2D, ellipse2D, line2D):
                return similarityMap2D.distinguishes.INSUFFICIENT
                
        if size == 2:
            # Case 6,7,8: 3 lines cases
            if last.__class__ == line2D and first.__class__ == line2D:
                return similarityMap2D.distinguishes.INSUFFICIENT

        if size >= 2: #and first.__class__ == point2D:
        
            # Case 1: 2 points not coincident
            if last.__class__ == point2D and first.__class__ == point2D:
                if last == first:
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[1] = last.value
                self.fromSet[0] = first.value                
                self.casus = similarityMap2D.casus.CASE_2PT_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 2: 1 point + 1 arc not coincident (point and center of arc)
            elif first.__class__ == point2D and last.__class__ in (circle2D, ellipse2D):
                if first.value == last.center:
                    if second is not None:
                        if second.value == last.center:
                            return similarityMap2D.distinguishes.UNSUITABLE
                        self.fromSet[1] = last.center
                        self.fromSet[0] = second.value     
                    else:
                        return similarityMap2D.distinguishes.UNSUITABLE
                else:
                    self.fromSet[1] = last.center
                    self.fromSet[0] = first.value         
                
                self.casus = similarityMap2D.casus.CASE_1PT_1ARC_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 3: 1 point + 1 line not colinear
            elif first.__class__ == point2D and last.__class__ == line2D:
                self.fromSet[1] = first.value
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                self.fromSet[0] = last.root.add(u.scale(v))
                
                if self.fromSet[1] == self.fromSet[0]:
                    if second is not None:
                        self.fromSet[1] = second.value
                        u = last.direction
                        v = self.fromSet[1].subtract(last.root)
                        self.fromSet[0] = last.root.add(u.scale(v))
                        if self.fromSet[1] == self.fromSet[0]:
                            return similarityMap2D.distinguishes.UNSUITABLE
                        self.distinguishables.pop(0)
                    else:
                        return similarityMap2D.distinguishes.UNSUITABLE
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                self.casus = similarityMap2D.casus.CASE_1PT_1LN_NOT_COLINEAR
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 4: 2 arcs with centers not coincident
            elif first.__class__ in (circle2D, ellipse2D) and last.__class__ in (circle2D, ellipse2D):
                if first.center == last.center:
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[1] = last.center
                self.fromSet[0] = first.center                
                self.casus = similarityMap2D.casus.CASE_2ARC_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 5: 1 arc + 1 line not colinear
            elif first.__class__ in (circle2D, ellipse2D) and last.__class__ == line2D:
                self.fromSet[1] = first.center
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                if u.parallel(v):
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[0] = last.root.add(u.scale(v))     
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                self.casus = similarityMap2D.casus.CASE_1ARC_1LN_NOT_COLINEAR
                return similarityMap2D.distinguishes.SUCCESS
                
            elif last.__class__ == line2D and first.__class__ == line2D and second.__class__ == line2D:
                # case 6: 3 lines no two parallel
                fpl = last.parallel(first)
                fps = second.parallel(first)
                spl = second.parallel(last)
                if not (fpl or fps or spl):
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = first.intersection(second)
                    self.fromSet[2] = second.intersection(last)
                    
                    #if self.fromSet[0] == self.fromSet[1] or self.fromSet[0] == self.fromSet[2] or self.fromSet[2] == self.fromSet[1]:
                    if self.fromSet[0] == self.fromSet[1] == self.fromSet[2]:
                        return similarityMap2D.distinguishes.UNSUITABLE
                        
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[2]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))

                    u = self.fromSet[1].subtract(self.fromSet[0])
                    v = self.fromSet[2].subtract(self.fromSet[1])
                    w = self.fromSet[2].subtract(self.fromSet[0])
                    self.ratio = [u.squareLength(),v.squareLength(),w.squareLength()]
                    
                    #self.checkTree = True                    
                    
                    self.casus = similarityMap2D.casus.CASE_3LN_NO_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
                    
                # unsuccessful case: 3 lines all parallel with one another
                if fpl and fps and spl:
                    return similarityMap2D.distinguishes.UNSUITABLE
                    
                # if 1st parallel last (highly unlikely) --> swap 2nd and last so that first two lines parallel
                if fpl:
                    self.distinguishables[1], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[1]
                    second, last = last, second
                    fps = second.parallel(first)
                    
                # if 2nd parallel last --> swap 1st and last so that first two lines parallel
                if spl:
                    self.distinguishables[0], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[0]
                    first, last = last, first
                    fps = True
                    
                # case 7: 3 lines first two parallel
                if fps:
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = second.intersection(last)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    
                    u = first.unitDir
                    v = last.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    #self.checkTree = True                  
                    
                    self.casus = similarityMap2D.casus.CASE_3LN_FIRST_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
                    
        return similarityMap2D.distinguishes.UNSUITABLE

    def assessTargetSet(self, target):
        if self.casus == similarityMap2D.casus.CASE_3LN_NO_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        if self.casus == similarityMap2D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if not first.parallel(second):
                return False
            if second.compareValue(first) == 0:
                return False
            return True
        elif self.casus == similarityMap2D.casus.CASE_3LN_LAST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            intersectingPoint = second.intersection(first)
            if not(second.containsPoint(intersectingPoint) and first.containsPoint(intersectingPoint)):
                return False
            return True
        return True
        
    
    def potentialTransforms(self): ########################
        #print('CASUS:', self.casus)
        if len(self.transforms) != 0:
            return self.transforms

        if self.casus == similarityMap2D.casus.REDUNDANT:
            return self.specialPotentialTransforms()
        size = len(self.distinguishables)
        source = []
        counterparts = [] 
                        
        for elm in self.distinguishables:
            source.append(elm.element)
            counterparts.append(elm.getCounterparts())
        target = [None]*size
        """
        if self.checkTree and size == 2:
            for i in range(2):
                if len(counterparts[i]) == 0:
                    return self.transforms
            if (counterparts[0][0].individual.sort.name == 
                counterparts[1][0].individual.sort.name):
                self.checkTree = [[True for i in range(len(counterparts[1]))] 
                                        for j in range(len(counterparts[0]))]
            else:
                self.checkTree = True
                                 
        elif self.checkTree and size == 3:
            for i in range(3):
                if len(counterparts[i]) == 0:
                    return self.transforms
            print(len(counterparts[0]),len(counterparts[1]),len(counterparts[2]))
            if (counterparts[0][0].individual.sort.name == 
                counterparts[1][0].individual.sort.name == 
                counterparts[2][0].individual.sort.name):
                self.checkTree = [[[True for i in range(len(counterparts[2]))] 
                                         for j in range(len(counterparts[1]))] 
                                         for k in range(len(counterparts[0]))]
            else:
                self.checkTree = True            
        """                    
        for e0 in range(len(counterparts[0])):
            target[0] = counterparts[0][e0]
            if size > 1:
                for e1 in range(len(counterparts[1])):
                    if counterparts[1][e1].individual == target[0].individual:
                        continue
                    target[1] = counterparts[1][e1]
                    if not self.assessTargetSet(target):
                        continue
                    if size > 2:
                        for e2 in range(len(counterparts[2])):
                                """
                            if isinstance(self.checkTree,list):
                                if self.checkTree[e0][e1][e2]:
                                    self.checkTree[e0][e1][e2] = False
                                    self.checkTree[e0][e2][e1] = False
                                    self.checkTree[e1][e0][e2] = False
                                    self.checkTree[e1][e2][e0] = False
                                    self.checkTree[e2][e0][e1] = False
                                    self.checkTree[e2][e1][e0] = False
                                    target[2] = counterparts[2][e2]
                                    if (counterparts[2][e2].individual == target[1].individual or 
                                        counterparts[2][e2].individual == target[0].individual):
                                        continue
                                    self.detectTransforms(source, target)
                            else:
                                """
                                target[2] = counterparts[2][e2]
                                self.detectTransforms(source, target)
                    else:
                            """
                        if isinstance(self.checkTree,list):
                            if self.checkTree[e0][e1]:
                                self.checkTree[e0][e1] = False
                                self.checkTree[e1][e0] = False
                                self.detectTransforms(source, target)
                        else:
                            """
                            self.detectTransforms(source, target)
        return self.transforms
        
        
    
    def detectTransforms(self, source, target):
        size = len(source)
        #leftLast = source[-1].codescriptor()
        last = target[-1].individual
        if size > 1:
            #leftFirst = source[0].codescriptor()
            first = target[0].individual
            if size > 2:
                second = target[1].individual
        #type = last.__class__
        toSet = [None, None, None]
        #########################################################
        # case 1: 2 points not coincident
        if self.casus == similarityMap2D.casus.CASE_2PT_NOT_COINCIDENT:
            if last == first:
                return False
            toSet[1] = last.value
            toSet[0] = first.value
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
            
        # Case 2: 1 point + 1 arc not coincident (point and center of arc)
        if self.casus == similarityMap2D.casus.CASE_1PT_1ARC_NOT_COINCIDENT:
            if last.center == first:
                return False
            toSet[1] = last.center
            toSet[0] = first.value
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
            
        #########################################################
        # Case 3: 1 point + 1 line not colinear
        elif self.casus == similarityMap2D.casus.CASE_1PT_1LN_NOT_COLINEAR:
            if last.type == 'ind-point2D':
                temp = first
                first = last
                last = temp
            toSet[1] = first.value
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return False                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength()==0:
                return False
                
            for i in target[1].forms[target[1].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[0] 
                and (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check = True
                    break
                check = False
            if not check:
                return
                
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
            
        #########################################################
        # Case 4: 2 arcs with centers not coincident
        if self.casus == similarityMap2D.casus.CASE_2ARC_NOT_COINCIDENT:
            if last.center == first.center:
                return False
            toSet[1] = last.center
            toSet[0] = first.center
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True

        #########################################################
        # Case 5: 1 arc + 1 line not colinear
        elif self.casus == similarityMap2D.casus.CASE_1ARC_1LN_NOT_COLINEAR:
            toSet[1] = first.center
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return False                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength()==0:
                return False
                
            for i in target[1].forms[target[1].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[0] 
                and (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check = True
                    break
                check = False
            if not check:
                return
                
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
                    
        #########################################################
        # case 6: 3 lines no two parallel
        elif self.casus == similarityMap2D.casus.CASE_3LN_NO_TWO_PARALLEL:
            fpl = last.parallel(first)
            fps = second.parallel(first)
            spl = second.parallel(last)
            if (fpl or fps or spl):
                return False
            is1 = first.intersection(last)
            is2 = first.intersection(second)
            is3 = second.intersection(last)
            
            
            if is1 == is2 or is1 == is3 or is2 == is3:
                return False
            
            u = is2.subtract(is1).squareLength()
            v = is3.subtract(is2).squareLength()
            w = is3.subtract(is1).squareLength()
            
            if u/self.ratio[0] == v/self.ratio[1] and u/self.ratio[0] == w/self.ratio[2]:
                toSet = [is1,is2,is3,target[0],target[1],target[2]]
                for i in toSet[3].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    return False
                for i in toSet[4].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    return False
                for i in toSet[5].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    return False
                    
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
            
            
            """
            u0= self.ratio[0]
            v0= self.ratio[1]
            w0= self.ratio[2]
            ts = []
            if u/u0 == v/v0 and u/u0 == w/w0:
                ts.append([is1,is2,is3,target[0],target[1],target[2]])
            if u/u0 == w/v0 and u/u0 == v/w0:
                ts.append([is2,is1,is3,target[0],target[2],target[1]])
            if v/u0 == u/v0 and v/u0 == w/w0:
                ts.append([is3,is2,is1,target[1],target[0],target[2]])
            if v/u0 == w/v0 and v/u0 == u/w0:
                ts.append([is2,is3,is1,target[1],target[2],target[0]])
            if w/u0 == u/v0 and w/u0 == v/w0:
                ts.append([is3,is1,is2,target[2],target[0],target[1]])
            if w/u0 == v/v0 and w/u0 == u/w0:
                ts.append([is1,is3,is2,target[2],target[1],target[0]])
            for toSet in ts:
                for i in toSet[3].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                for i in toSet[4].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                for i in toSet[5].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    continue
                    
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
            """
                        
        #########################################################
        # case 7: 3 lines first two parallel
        elif self.casus == similarityMap2D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            if last.parallel(first) or second.parallel(last) or not second.parallel(first):
                return False
                
            u = first.unitDir
            v = last.unitDir
            cosA = abs(u.dotproduct(v))
            if not self.cosAngle ==  cosA:
                return False
            
            toSet[0] = first.intersection(last)
            toSet[1] = second.intersection(last)
            
            ###########
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[0]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[1]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[2] and
                    i.containsPoint(toSet[1]) >= self.subCase[3]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:               
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
            #------------------------------------------------------------    
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[1]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[0]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[3] and
                    i.containsPoint(toSet[1]) >= self.subCase[2]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:              
                tfs = affineTransform2D.similarityTransform(self.fromSet, [toSet[1],toSet[0],None], self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
            #------------------------------------------------------------    

        return False
        
    def specialPotentialTransforms(self):
        if len(self.distinguishables) == 1:
            dist = self.distinguishables[0]
            if dist.element.individual.type == 'ind-point2D':
                fromVect = dist.element.individual.value
                for i in dist.counterparts:
                    toVect = i.individual.value
                    self.transforms.append(affineTransform2D.translation(toVect.subtract(fromVect)))
                    
            if dist.element.individual.type == 'ind-line2D':
                fromVects = [dist.element.individual.assoc.start, dist.element.individual.assoc.end]
                for i in dist.counterparts:
                    for line in i.forms[dist.element.individual.assoc.sort.name]:

                        toVects = [line.start, line.end]
                        tfs = affineTransform2D.similarityTransform(fromVects, toVects, 1)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        
                        toVects = [line.end, line.start]
                        tfs = affineTransform2D.similarityTransform(fromVects, toVects, 1)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
        if len(self.distinguishables) == 2:
            dist0 = self.distinguishables[0]
            dist1 = self.distinguishables[1]
            if not dist0.element.individual.type == dist1.element.individual.type == 'ind-line2D':
                return self.transforms
            if dist1.element.individual.parallel(dist0.element.individual):
                fromVects = []
                fromVects.append(dist0.element.forms[dist0.element.individual.assoc.sort.name][0].start)
                fromVects.append(dist1.element.forms[dist1.element.individual.assoc.sort.name][0].start)
                for i in dist0.counterparts:
                    for j in dist1.counterparts:
                        if i.individual == j.individual:
                            continue
                        if not i.individual.parallel(j.individual):
                            continue
                        for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                            for line1 in j.forms[dist1.element.individual.assoc.sort.name]:
                                toVects = [line0.start,line1.start]
                                tfs = affineTransform2D.similarityTransform(fromVects, toVects, 1)
                                for tf in tfs:
                                    if tf is not None:
                                        self.transforms.append(tf)
                                toVects = [line0.end,line1.end]
                                tfs = affineTransform2D.similarityTransform(fromVects, toVects, 1)
                                for tf in tfs:
                                    if tf is not None:
                                        self.transforms.append(tf)
                                        
            else:
                fromVects = [dist0.element.individual.intersection(dist1.element.individual)]
                checkContains = False
                for i in dist0.element.forms[dist0.element.individual.assoc.sort.name]:
                    if i.containsPoint(fromVects[0]):
                        checkContains = True
                        pt = i.start if not i.start == fromVects[0] else i.end
                        fromVects.append(pt)
                if not checkContains:
                    l = dist0.element.forms[dist0.element.individual.assoc.sort.name][0]
                    pt = l.start if not l.start == fromVects[0] else l.end
                    fromVects.append(pt)
                for i in dist0.counterparts:
                    for j in dist1.counterparts:
                        if i.individual.parallel(j.individual):
                            continue
                        toVects = [i.individual.intersection(j.individual), None]
                        for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                            if not toVects[0] == line0.start:
                                toVects[1] = line0.start
                                tfs = affineTransform2D.similarityTransform(fromVects, toVects, 1)
                                for tf in tfs:
                                    if tf is not None:
                                        self.transforms.append(tf)
                            if not toVects[0] == line0.end:
                                toVects[1] = line0.end
                                tfs = affineTransform2D.similarityTransform(fromVects, toVects, 1)
                                for tf in tfs:
                                    if tf is not None:
                                        self.transforms.append(tf)
                                        
                
        return self.transforms
        
    def checkTransform(self):
        return
     