# -*- coding: utf-8 -*-
"""
Created on Tue Aug 30 14:19:32 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .sort import sort
from ..form.metaForm import metaForm

"""
 * A Disjunctive Sort specifies a disjunctive composition of
 * {Sort}s under the sum operation. Under the sum operation,
 * a form of the resulting sort may contain a form of every operand sort.
 * If any component sort is an unnamed disjunctive sort, instead,
 * its components become part of the disjunctive composition.
 * If any component sort is part of another, unnamed component sort, the former
 * component sort is not included in the composition.
 * A disjunctive sort may have a name assigned.
 * 
 * The DisjunctiveSort class represents a disjunctive sort additionally
 * as an ordered set of component sorts. 
 * In the definition of a composite sort, the operation of sum is denoted by '+'.
 * In its canonical version, component sorts cannot themselves be disjunctive sorts
 * and XXX.
 * (a + b) + (a + c) -> a + b + c.
"""

class disjunctiveSort(sort):
    separator = ' + '
    type = 'sort-disjunctive'
    
    """
    initialize the sort, a disjunctive sort should not contain any other disjunctive sort.
    needs to go through all the component and compile all the subordinate sorts
    e.g. (a + b) + (a + c) -> a + b + c
    then make the description for the sort
    """
    def __init__(self, name, component = [], *args, **kwargs):
        if self.checkSort(name) == True:
            raise ValueError('A sort with name '+name+' already exists')
        super(disjunctiveSort, self).__init__(name)
        
        if 'absoluteComp' in kwargs and kwargs['absoluteComp'] is True:
            self.component = component
        else:
            #compiling all sorts from the component
            uniqueComp = disjunctiveSort.manageComp(component)
        
            #assign to self.component
            self.component = uniqueComp

        self.dimensional = None
        for i in self.component:
            dim = i.baseSort().characteristic.type[-2:]
            if dim != '2D' and dim != '3D':
                continue
            if self.dimensional is None:
                self.isParametric = True if i.baseSort().characteristic.type[-3] == 'P' else False
                self.dimensional = dim
            else:
                if self.isParametric != (i.baseSort().characteristic.type[-3] == 'P'):
                    print(i.baseSort().characteristic.type)
                    raise ValueError('Cannot have both Parametric and Non-Parametric objects under the same disjunctive sort '+name)
                if self.dimensional != dim:
                    print(i.baseSort().characteristic.type)
                    raise ValueError('Cannot have both 2D and 3D objects under the same disjunctive sort '+name)

        self.description = self.description(self.component)
        return
        
    @classmethod
    def create(cls, name, component = []):
        uniqueComp = disjunctiveSort.manageComp(component)
        if sort.checkSort(name) == True: 
            s = sort.getSort(name)
            if s.type != 'sort-disjunctive' or uniqueComp != s.component:
                for i in uniqueComp:
                    print(i)
                print()
                for i in s.component:
                    print(i)
                raise ValueError('A sort with name '+name+' already exists')
            return s
        return disjunctiveSort(name, uniqueComp, absoluteComp = True)
        
    @staticmethod
    def manageComp(component):
        #compiling all sorts from the component
        uniqueComp = []
        for i in component:
            if i.type == 'sort-disjunctive':
                for j in i.component:
                    if j not in uniqueComp:
                        uniqueComp.append(j)
            elif i not in uniqueComp:
                uniqueComp.append(i)
                
        #combining attribute sorts with the same base sort
        iIndex = 0
        while iIndex < len(uniqueComp):
            i = uniqueComp[iIndex]
            jIndex = iIndex + 1
            while jIndex < len(uniqueComp):
                j = uniqueComp[jIndex]
                if i.type == 'sort-attribute' and j.type == 'sort-attribute' and i.base.name == j.base.name:
                    nameList = []
                    if i.weight.type == 'sort-disjunctive':
                        for comp in i.weight.component:
                            if comp.name not in nameList:
                                nameList.append(comp.name)
                    else:
                        nameList.append(i.weight.name)
                        
                    if j.weight.type == 'sort-disjunctive':
                        for comp in j.weight.component:
                            if comp.name not in nameList:
                                nameList.append(comp.name)
                    else:
                        nameList.append(j.weight.name)
                        
                    
                    newDisj = disjunctiveSort(disjunctiveSort.separator.join(nameList), [i.weight, j.weight])
                    i.weight = newDisj
                    i.name= i.base.name+'_'+'('+i.weight.name.replace(" ", "")+')'
                    uniqueComp.remove(j)
                else:
                    jIndex += 1
            iIndex += 1
            
        #sort the list
        uniqueComp.sort() 
        return uniqueComp
        
    def __eq__(self,other):
        if self.__class__ != other.__class__:
            return False
        for i in self.component:
            check = False
            for j in other.component:
                if i == j:
                    check = True
            if check == False:
                return False
        return True
    
    """
    add a component sort to this sort
    """
    def addComponent(self, sort):
        for i in self.component:
            if sort.name == i.name:
                return
        self.component.append(sort)

    @staticmethod
    def sameComponent(component):
        uniqueComp = disjunctiveSort.manageComp(component)
        for name in sort.register:
            s = sort.register[name]
            if s.type != 'sort-disjunctive':
                continue
            if uniqueComp == s.component:
                return s
        return False
    
    """
    creating the description for this sort
    """
    def description(self, component):
        if len(component) <= 1:
            desc = ''
            for i in component:
                desc += i.description
        else:
            desc = '('
            for i in component:
                if desc == '(':
                    desc += i.description
                else:
                    desc += disjunctiveSort.separator + i.description
            desc += ')'
        return desc
        
    """
    checking if this sort contain other sort
    """
    def contains(self, other):
        if other.type == 'sort-disjunctive':
            for i in other.component:
                Check = False
                for j in self.component:
                    if j.contains(i):
                        Check = True
                if Check == False:
                    return False
            return True
        else:
            for i in self.component:
                if i.contains(other):
                    return True
        return False
        
    """
    find the result when subtracting this sort by another sort
    """
    def complementSort(self, other):
        if not self.contains(other):
            raise ValueError("Disjunctive sort "+self.name+" does not contain the sort named "+other.name)
        comp = []
        if other.type != 'sort-disjunctive':
            for i in self.component:
                if i == other:
                    continue
                comp.append(i)
        else:
            for i in self.component:
                check = False
                for j in other.component:
                    if i == j:
                        check = True
                        break
                if check:
                    continue
                comp.append(i)
        nName = newName = self.name+'_'+other.name+'_complement'
        count = 0
        while self.checkSort(nName):
            nName = newName+str(count)
            count+=1
        return disjunctiveSort(nName, comp)
        
    """
    dimension of this sort is 0
    """
    def dimension(self):
        return 0
        
    """
    creating a new metaform of this sort
    """    
    def newForm(self):
        try:
            form = metaForm(self)
            return form
        except:
            print('exception')
    
    