# -*- coding: utf-8 -*-
"""
Created on Fri Aug 12 14:33:20 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from __future__ import with_statement

from builtins import int
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from builtins import object
from mpmath import sqrt, pi, sin, cos, tan, asin, acos, atan, atan2
from mpmath import isint, mp, nstr, almosteq

"""
 * Coordinate is the representational unit for geometric
 * Coordinate is saved and represented as symbolic mathematics
 * allowing the exact representation of real number
 * e.g. sin(2*pi/3) will be represented as sqrt(3)/2 
"""
class coordinate(object):
    """
    setting the precision of the floating point conversion to 50
    i.e. the value of the coordinate is rounded to 
         50 decimal places when converted to float value
    """
    mp.dps = 40
    p = 9
    defaultP = 9
    cp = mp.mpf('1e-'+str(p))
    c0 = mp.mpf(0)
    c1 = mp.mpf(1)
    type = 'struct-coordinate'
    
    def __init__(self, num):
        if isinstance(num, mp.mpf):
            self.value = num
        elif isinstance(num, coordinate):
            # raise ValueError('initiating an already existing coordinate object')
            self.value = num.value
        elif isinstance(num, int) or isinstance(num, str):
            self.value = mp.mpf(num)
        else:
            try:
                self.value = mp.mpf(num)
            except:
                print('Cannot create a coordinate with '+str(num))
    
    def __float__(self):
        #raise NotImplementedError
        return float(self.value)
        
    def __int__(self):
        return int(self.value)
        
    def __repr__(self):
        return nstr(self.value,3)
        
    def __str__(self):
        return nstr(self.value,16)
        return str(self.value)
        
    @staticmethod
    def precision():
        return coordinate.p
        
    @staticmethod
    def setPrecision(prec = None):
        if prec is None:
            coordinate.p = coordinate.defaultP
        else:
            coordinate.p = prec
        
    @staticmethod
    def setDefaultPrecision():
        coordinate.p = coordinate.defaultP
        
    def __round__(self, n = None):
        if n is None:
            return coordinate(round(self.value))
        return coordinate(round(self.value,n))
                
        
    """
    finding the sign of the coordinate: 1 if positive, -1 if negative, 0 if 0
    """
    def sign(self):
        if self == 0:
            return 0
        if self > 0:
            return 1
        return -1
        
    """
    check if the coordinate is equal to 0
    """
    def isZero(self):
        return almosteq(self.value, 0, self.cp)
        
    """
    check if the coordinate is equal to 1
    """
    def isOne(self):
        return round(self.value, self.p) == 1
    
    """
    finding square root of the coordinate
    """
    def sqrt(self):
        return coordinate(sqrt(self.value))
        
    """
    addition
    """
    def __add__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value + other.value)   

    """
    subtraction
    """
    def __sub__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value - other.value)

    """
    multiplication
    """
    def __mul__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value * other.value)

    """
    division
    """
    def __truediv__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value / other.value)

    """
    floor dividsion (only taking the integer part of the result)
    e.g. 23 // 2 = 11
    """
    def __floordiv__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value // other.value)  

    """
    power
    """
    def __pow__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value ** other.value)

    """
    modular
    """
    def __mod__(self, other):
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value % other.value)   
        
    """
    addition
    """
    def __radd__(self, other):
        other = coordinate(other)
        return coordinate(other.value + self.value)   

    """
    subtraction
    """
    def __rsub__(self, other):
        other = coordinate(other)
        return coordinate(other.value - self.value)

    """
    multiplication
    """
    def __rmul__(self, other):
        other = coordinate(other)
        return coordinate(other.value * self.value)

    """
    division
    """
    def __rtruediv__(self, other):
        other = coordinate(other)
        return coordinate(other.value / self.value)

    """
    floor dividsion (only taking the integer part of the result)
    e.g. 23 // 2 = 11
    """
    def __rfloordiv__(self, other):
        other = coordinate(other)
        return coordinate(other.value // self.value)  

    """
    power
    """
    def __rpow__(self, other):
        raise ValueError('POWER: this is not coordinate')
        other = coordinate(other)
        return coordinate(other.value ** self.value) 

    """
    modular
    """
    def __rmod__(self, other):
        other = coordinate(other)
        return coordinate(other.value % self.value)   

    """
    negation
    """
    def __neg__ (self):
        return coordinate(-self.value)   

    """
    """
    def __pos__ (self):
        return  coordinate(self.value)

    """
    absolute value
    """
    def __abs__(self):
        return coordinate(abs(self.value))

    """
    inverting
    """
    def invert (self):
        return coordinate(mp.mpf('1') / self.value)   

    """
    check equality
    """
    #@workdps(30)
    
    def __eq__(self, other):
        if isinstance(other, coordinate):
            return almosteq(self.value, other.value, self.cp)
        elif isinstance(other, mp.mpf):
            return almosteq(self.value, other, self.cp)
        elif isinstance(other, int) or isinstance(other, str):
            return almosteq(self.value, mp.mpf(other), self.cp)
        raise TypeError('not supported type')
        
        """
        if isinstance(other, coordinate):
            return round(self.value, self.p) == round(other.value, self.p)
        elif other is 0:
            t = self.c0
        elif other is 1:
            t = self.c1
        return round(self.value, self.p) == round(t, self.p)
        
        if isinstance(other, coordinate):
            if str(self.value) == str(other.value):
                return True
            return round(self.value, self.p) == round(other.value, self.p)
        if str(self.value) == str(other):
            return True
        return round(self.value, self.p) == round(mp.mpf(other), self.p)
        """
        
    def __ne__(self, other):
        return not self == other

    """
    <
    """
    
    def __lt__(self,other):
        """
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return False
            return self.value < other.value
        if almosteq(self.value, mp.mpf(other), self.cp):
            return False
        return self.value < mp.mpf(other)
        
        if isinstance(other, coordinate):
            if self.value < other.value and not almosteq(self.value, other.value, self.cp):
                return True
            return False
        elif other is 0:
            t = self.c0
        elif other is 1:
            t = self.c1
        if self.value < t and not almosteq(self.value, t, self.cp):
            return True
        return False
        """
        t = None
        if isinstance(other, coordinate):
            if self.value < other.value and not almosteq(self.value, other.value, self.cp):
                return True
            return False
        elif other is 0:
            t = self.c0
        elif other is 1:
            t = self.c1
        else:
            t = mp.mpf(other)
        if self.value < t and not almosteq(self.value, t, self.cp):
            return True
        return False
        

    """
    >
    """
    
    def __gt__(self,other):
        """
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return False
            return self.value > other.value
        if almosteq(self.value, mp.mpf(other), self.cp):
            return False
        return self.value > mp.mpf(other)
        """
        t = None
        if isinstance(other, coordinate):
            if self.value > other.value and not almosteq(self.value, other.value, self.cp):
                return True
            return False
        elif other is 0:
            t = self.c0
        elif other is 1:
            t = self.c1
        else:
            t = mp.mpf(other)
        if self.value > t and not almosteq(self.value, t, self.cp):
            return True
        return False

    """
    <=
    """
    def __le__(self,other):
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return True
            return self.value <= other.value
        if almosteq(self.value, mp.mpf(other), self.cp):
            return True
        return self.value <= mp.mpf(other)

    """
    >=
    """
    def __ge__(self,other):
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return True
            return self.value >= other.value
        if almosteq(self.value, mp.mpf(other), self.cp):
            return True
        return self.value >= mp.mpf(other)
    """
    >>
    """
    def __rshift__(self, other):
        if isinstance(other, coordinate):
            return coordinate(int(self.value) >> int(other.value))
        return coordinate(int(self.value) >> int(other))
        

    """
    comparing values
    """
    def compareValue(self,other):
        if isinstance(other, coordinate):
            """
            if self.value > other.value:
                return 1
            elif self.value < other.value:
                return -1
            return 0
            """
            if almosteq(self.value, other.value, self.cp):
                return 0
            elif self.value > other.value:
                return 1
            return -1
        """
        if self.value > mp.mpf(other):
            return 1
        elif self.value < mp.mpf(other):
            return -1
        return 0
        """
        if almosteq(self.value, mp.mpf(other), self.cp):
            return True
        elif self.value > mp.mpf(other):
            return 1
        return -1

c0 = coordinate(0)
c1 = coordinate(1)
c2 = coordinate(2)
cn1 = coordinate(-1)
c0p5 = coordinate('0.5')
