# -*- coding: utf-8 -*-
"""
Created on Tue Sep  6 12:40:57 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import
from builtins import str
from builtins import range
from builtins import round
"""
from future import standard_library
standard_library.install_aliases()
"""
from builtins import object
from ..struct.coordinate import coordinate, c0, c1, c2, cn1, c0p5, sin, cos
#from decimal import Decimal



"""
transformation for geometric shape grammar

each affineTransform object consists of a transformation matrix that can be applied
onto vectors by matrix multiplication
"""
class affineTransform2D(object):
    SIZE = 3 #4
    type = 'transform-affineTransform2D'
    
    """
    Affine Transformation object in 2D: Consisting of a 2x3 matrix
    
    """
    def __init__(self, reflection = None):
        self.matrix = [[c1,c0,c0],
                       [c0,c1,c0]]
        if reflection is not None:
            self.matrix[reflection][reflection] = cn1
        self.id = None
        return
    
    def __repr__(self):
        return 'affineTransf'
        
    def __str__(self):
        return str(self.matrix[0])+'\n'+str(self.matrix[1])+'\n'
        
    def printTransf(self):
        print('matrix:')
        for i in self.matrix:
            #print('|',i[0],i[1],i[2])
            print('|',round(i[0],3),round(i[1],3),round(i[2],3))
        print()
            
    def duplicate(self):
        newTr = affineTransform2D()
        for i in range(len(self.matrix)):
            for j in range(len(self.matrix[i])):
                newTr.matrix[i][j] = self.matrix[i][j]
        return newTr
        
    
    def scalar(self, n, m):
        if n < 0 or n >= affineTransform2D.SIZE or m < 0 or m >= affineTransform2D.SIZE:
            raise IndexError
        return self.matrix[n][m]
            
    @staticmethod
    def translation(vector):
        result = affineTransform2D()
        result.matrix = [[c1, c0, vector.x],
                         [c0, c1, vector.y]]
        return result
        
    @staticmethod
    def scale(factor):
        result = affineTransform2D()
        f = coordinate(factor)
        result.matrix = [[f, c0, c0],
                         [c0, f, c0]]
        return result
        
    @staticmethod
    def rotation(theta):
        result = affineTransform2D()
        ct = coordinate(cos(theta))
        result.matrix = [[ct                    , coordinate(-sin(theta)), c0],
                         [coordinate(sin(theta)), ct                     , c0]]
        return result
        
    @staticmethod
    def scaledRotation(theta, lengthFrom, lengthTo):
        result = affineTransform2D()
        l = lengthTo / lengthFrom
        ct = (coordinate(cos(theta))*l)
        result.matrix = [[ct                        , (coordinate(-sin(theta))*l), c0],
                         [(coordinate(sin(theta))*l), ct                         , c0]]
        return result
        
    @staticmethod
    def reflectionX():
        result = affineTransform2D()
        result.matrix = [[c1, c0,  c0],
                         [c0, cn1, c0]]
        return result
        
    @staticmethod
    def reflectionY():
        result = affineTransform2D()
        result.matrix = [[cn1, c0, c0],
                         [c0,  c1, c0]]
        return result
        
    @staticmethod
    def reflectionXY():
        result = affineTransform2D()
        result.matrix = [[c0, c1, c0],
                         [c1, c0, c0]]
        return result
                
    def combine(self, other):
        result = affineTransform2D()
        a = self.matrix
        b = other.matrix
        for i in range(affineTransform2D.SIZE-1):
            for j in range(affineTransform2D.SIZE-1):
                result.matrix[i][j]= (a[i][0]*b[0][j]+a[i][1]*b[1][j])
        result.matrix[0][2] = a[0][0]*b[0][2]+a[0][1]*b[1][2]+a[0][2]
        result.matrix[1][2] = a[1][0]*b[0][2]+a[1][1]*b[1][2]+a[1][2]
        return result
    
    @staticmethod    
    def similarityTransform(fromE, toE, index):
        dirFrom = fromE[1].subtract(fromE[0])
        dirTo = toE[1].subtract(toE[0])
        
        fromLength = dirFrom.length()
        toLength = dirTo.length()

        transformation= None
        transformationRefl = None
        
        baseRotation = affineTransform2D.scaledRotation(dirFrom.angle(dirTo), fromLength, toLength)
        transformation = affineTransform2D.translation(toE[0].subtract(fromE[0].transform(baseRotation))).combine(baseRotation)
        if index == 1:
            reflRotation = affineTransform2D.scaledRotation(dirFrom.transform(affineTransform2D.reflectionX()).angle(dirTo), fromLength, toLength).combine(affineTransform2D.reflectionX())
            transformationRefl = affineTransform2D.translation(toE[0].subtract(fromE[0].transform(reflRotation))).combine(reflRotation)
        result = [transformation, transformationRefl]
        return result


class affineTransform3D(object):
    SIZE = 4
    
    rotX = 0
    rotY = 1
    rotZ = 2
    
    type = 'transform-affineTransform3D'
    
    
    def __init__(self, reflection = None):
        self.matrix = [[c1,c0,c0,c0],
                       [c0,c1,c0,c0],
                       [c0,c0,c1,c0]]
        if reflection is not None:
            self.matrix[reflection][reflection] = cn1
        self.id = None
        return
        
    def __str__(self):
        return str(self.matrix[0])+'\n'+str(self.matrix[1])+'\n'+str(self.matrix[2])+'\n'
        
    def printTransf(self):
        for i in self.matrix:
            try:
                print('|\t',round(i[0],3),'\t',round(i[1],3),'\t',round(i[2],3),'\t',round(i[3],3))
            except TypeError:
                print('|',i[0],'\t',i[1],'\t',i[2],'\t',i[3])
        print('')
        
    def duplicate(self):
        newTr = affineTransform3D()
        for i in range(len(self.matrix)):
            for j in range(len(self.matrix[i])):
                newTr.matrix[i][j] = self.matrix[i][j]
        return newTr
    
    def scalar(self, n, m):
        if n < 0 or n >= affineTransform3D.SIZE or m < 0 or m >= affineTransform3D.SIZE:
            raise IndexError
        return self.matrix[n][m]
    
    @staticmethod
    def tensorProduct(vector)  :     
        result = affineTransform3D()
        coords = [vector.x,vector.y,vector.z]
        for i in range(len(result.matrix)):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j] = coords[i] * coords[j]
        return result;

    @staticmethod
    def crossProduct(vector):
        result = affineTransform3D()
        result.matrix = [[c0       , -vector.z, vector.y , c0],
                         [vector.z , c0       , -vector.x, c0],
                         [-vector.y, vector.x , c0       , c0]]
        return result
        
    def add(self, other):
        result = affineTransform3D()
        for i in range(len(result.matrix)):
            for j in range(len(result.matrix[i])):
                result.matrix[i][j] = self.matrix[i][j]+other.matrix[i][j]
        return result
        
    def tScale(self,factor):
        result = affineTransform3D()
        for i in range(len(result.matrix)):
            for j in range(affineTransform3D.SIZE):
                result.matrix[i][j] = factor * self.matrix[i][j]
        return result
        
    def transpose(self):
        result = affineTransform3D()
        for i in range(affineTransform3D.SIZE-1):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j] = self.matrix[j][i]
        result.matrix[0][3] = c0
        result.matrix[1][3] = c0
        result.matrix[2][3] = c0
        return result



    @staticmethod
    def translation(vector):
        result = affineTransform3D()
        result.matrix = [[c1, c0, c0, vector.x],
                         [c0, c1, c0, vector.y],
                         [c0, c0, c1, vector.z]]
        return result
        
    @staticmethod
    def scale(factor):
        result = affineTransform3D()
        f = coordinate(factor)
        result.matrix = [[f, c0, c0, c0],
                         [c0, f, c0, c0],
                         [c0, c0, f, c0]]
        return result
        
    @staticmethod
    def rotation(normal, sinSq, cos):
        squareLength = normal.dotproduct(normal)
        result = affineTransform3D().tScale(cos)
        tf = affineTransform3D.tensorProduct(normal).tScale(c1 / squareLength).tScale(c1-cos)
        result = result.add(tf)
        tf = affineTransform3D.crossProduct(normal).tScale((sinSq / squareLength) ** c0p5)
        result = result.add(tf)
        return result
                
    @staticmethod
    def scaledRotation(normal, sinSq, cosSq, cosSign, factorSq):
        squareLength = normal.dotproduct(normal)
        fcos = (cosSq * factorSq) ** c0p5
        if cosSign < 0:
            fcos = -fcos
        result = affineTransform3D().tScale(fcos)
        tf = affineTransform3D.tensorProduct(normal).tScale(c1/squareLength).tScale(factorSq**c0p5-fcos)
        result = result.add(tf)
        fsinSq = factorSq * (sinSq/squareLength)
        tf = affineTransform3D.crossProduct(normal).tScale(fsinSq**c0p5)
        result = result.add(tf)
        return result
                                
    def combine(self, other):
        result = affineTransform3D()
        a = self.matrix
        b = other.matrix
        for i in range(affineTransform3D.SIZE-1):
            for j in range(affineTransform3D.SIZE-1):
                result.matrix[i][j]= (a[i][0]*b[0][j]+a[i][1]*b[1][j]+a[i][2]*b[2][j])
        result.matrix[0][3] = (a[0][0]*b[0][3]+a[0][1]*b[1][3]+a[0][2]*b[2][3]+a[0][3])
        result.matrix[1][3] = (a[1][0]*b[0][3]+a[1][1]*b[1][3]+a[1][2]*b[2][3]+a[1][3])
        result.matrix[2][3] = (a[2][0]*b[0][3]+a[2][1]*b[1][3]+a[2][2]*b[2][3]+a[2][3])
        return result
                
    @staticmethod
    def reflectionX():
        result = affineTransform3D()
        result.matrix = [[c1, c0,  c0, c0],
                         [c0, cn1, c0, c0],
                         [c0, c0,  cn1, c0]]
        return result
        
    @staticmethod
    def reflectionY():
        result = affineTransform3D()
        result.matrix = [[cn1, c0, c0,  c0],
                         [c0,  c1, c0,  c0],
                         [c0,  c0, cn1, c0]]
        return result
        
    @staticmethod
    def reflectionZ():
        result = affineTransform3D()
        result.matrix = [[cn1, c0,  c0, c0],
                         [c0,  cn1, c0, c0],
                         [c0,  c0,  c1, c0]]
        return result
        
    @staticmethod
    def reflectionO():
        result = affineTransform3D()
        result.matrix = [[cn1, c0,  c0,  c0],
                         [c0,  cn1, c0,  c0],
                         [c0,  c0,  cn1, c0]]
        return result
        
    @staticmethod    
    def similarityTransform(fromE, toE, index, pypass = False):   
        uf = fromE[1].subtract(fromE[0])
        vf = fromE[2].subtract(fromE[0])
        ut = toE[1].subtract(toE[0])
        vt = toE[2].subtract(toE[0])
        

        #normal vectors of from set and to set
        normalfrom = uf.product(vf)
        normalto = ut.product(vt)

        #mapping the normal vector of the from set and to set together
        step1 = affineTransform3D()
        axis = normalfrom.product(normalto)
        
        if not axis.isZero():
            sqLengths = normalfrom.dotproduct(normalfrom) * normalto.dotproduct(normalto)
            sinSq = axis.dotproduct(axis) / sqLengths
            cos = normalfrom.dotproduct(normalto) / sqLengths.sqrt()
            step1 = affineTransform3D.rotation(axis, sinSq, cos)
            
        #transforming the vectors of from set, then find the direction vectors of the 2 sets
        directionfrom = fromE[2].transform(step1).subtract(fromE[1].transform(step1))
        directionto = toE[2].subtract(toE[1])
        #finding the scaled rotation to map directionfrom to directionto (no translation)
        axis = directionfrom.product(directionto)
        
        sqLengths = directionfrom.dotproduct(directionfrom) * directionto.dotproduct(directionto)
        cosSq = directionfrom.dotproduct(directionto) ** c2 / sqLengths
        scalarSq = directionto.dotproduct(directionto) / directionfrom.dotproduct(directionfrom)
        
        if not axis.isZero():
            sinSq = axis.dotproduct(axis) / sqLengths
            step2 = affineTransform3D.scaledRotation(axis, sinSq, cosSq, directionfrom.dotproduct(directionto).sign(), scalarSq)
            if index == 1:
                axisNeg = axis.scale(-1)
                step2Neg = affineTransform3D.scaledRotation(axisNeg, sinSq, cosSq, directionfrom.dotproduct(directionto).sign()*cn1, scalarSq)
        else:
            step2 = affineTransform3D.scaledRotation(normalto, c0, cosSq, directionfrom.dotproduct(directionto).sign(), scalarSq)
            if index == 1:
                step2Neg = affineTransform3D.scaledRotation(normalto, c0, cosSq, directionfrom.dotproduct(directionto).sign()*cn1, scalarSq)
        
        """
        axis = normalto.product(vector3D(0,0,1))
        rotZ = affineTransform3D()
        
        if not axis.isZero():
            sqLengths = normalto.dotproduct(normalto)
            sinSq = axis.dotproduct(axis) / sqLengths
            cos = normalto.z / (sqLengths ** c0p5)
            rotZ = affineTransform3D.rotation(axis, sinSq, cos)
        """
    
        baseRots = step2.combine(step1)
        tf = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(baseRots))).combine(baseRots)

        if index == 1:
            baseRotsNeg = step2Neg.combine(step1).tScale(cn1)

        transformation= None
        #transformationRefl = None
        
        """ 
        ###########################################
        ### For Testing:
        ###############
        print('\n================================')
        print('fromE', fromE)
        print('toE', toE)
        print('base rotation:')
        baseRots.printTransf()
        print('tf:')
        tf.printTransf()
        """
        
        result = []
        if fromE[1].transform(tf) == toE[1] and fromE[2].transform(tf) == toE[2]:
            transformation = tf
            result.append(transformation)
            if index == 1:
                """
                reflRots = baseRots.combine(rotZ.transpose()).combine(affineTransform3D(2)).combine(rotZ)
                transformationRefl = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(reflRots))).combine(reflRots)
                result.append(transformationRefl)
                """
                tf = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(baseRotsNeg))).combine(baseRotsNeg)
                result.append(tf)
                
            if not pypass:
                return result


        #180 degree rotation matrix
        rot180 = affineTransform3D.rotation(directionto, c0, cn1)
        
        tf = rot180.combine(baseRots)
        tf = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(tf))).combine(tf)

        
        """
        ###########################################
        ### For Testing:
        ###############
        print('----------------')
        print('tf-180:')
        tf.printTransf()
        print('.:.__',fromE[0].transform(tf), toE[0], str(fromE[0].transform(tf)) == str(toE[0]))
        print('.:.__',fromE[1].transform(tf), toE[1], str(fromE[1].transform(tf)) == str(toE[1]))
        print('.:.__',fromE[2].transform(tf), toE[2], str(fromE[2].transform(tf)) == str(toE[2]))
        """


        if fromE[1].transform(tf) == toE[1] and fromE[2].transform(tf) == toE[2]:
            transformation = tf
            result.append(transformation)
            
            if index == 1:
                """
                reflRots = baseRots.combine(rotZ.transpose()).combine(affineTransform3D(2)).combine(rotZ)
                tf = rot180.combine(reflRots)
                transformationRefl = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(tf))).combine(tf)
                result.append(transformationRefl)
                """
            
                tf = rot180.combine(baseRotsNeg)
                tf = affineTransform3D.translation(toE[0].subtract(fromE[0].transform(tf))).combine(tf)
                result.append(tf)

        
        return result

